/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Implements generic_service.h */

#include <iostream>
#include <signal.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h>

#include "generic_service.h"
#include "../util/Util.h"
#include "../udp-lib/UDPLib.h"
#include "../serv-comp/ServCompIFPkt.h"
#include "../serv-comp/ServComp.h"

bool generic_service::should_listen = true;

generic_service::generic_service(generic_client_info_t *arg_cl_list,
				 int arg_max_cl, Debug_t *arg_dbg) {
  cl_list = arg_cl_list; max_cl = arg_max_cl;
  generic_client_info::dbg = dbg = arg_dbg;
} // End generic_service::generic_service()

JString generic_service::myName() {
  return "generic_service";
} // End generic_service::myName(0

void generic_service::usage(const char *progname) {
  fprintf(stderr, "Usage: %s <port-base>\n", progname);
} // End generic_service::usage()

void generic_service::intr_catcher(int) {
  should_listen = false;
} // End generic_service::intr_catcher()

int generic_service::main(int argc, char *argv[]) {
  util_init();
  if(argc != 2) { usage(argv[0]); return 1; }

  sc_listen_port = atoi(argv[1]);
  data_port_base = sc_listen_port+100;
  generic_client_info::myIP = myIP = nslookup(getenv("HOSTNAME"));
  strcpy(generic_client_info::name, myName().cstr());

  dbg->print() << myName() << " listen-start sc_listen_port="
	       << sc_listen_port << " data_port_base=" << data_port_base
	       << endl;

  // prepare the UDP socket for listening
  udp_sock = sock_and_bind(sc_listen_port);
  if(udp_sock < 0) { perror("socket"); return 3; }

  // init the udp library
  UDPLib::init();

  // get the request response block for sends
  generic_client_info::rrBlock = UDPLib::getReqRespBlock();
  if(generic_client_info::rrBlock == NULL) {
    cerr << "rrBlock is NULL" << endl; return 23;
  }

  cerr << myName() << " service running on port " << sc_listen_port
       << "\a" << endl;

  // Activate interrupt catcher
  signal(SIGINT, intr_catcher);
  signal(SIGTERM, intr_catcher);

  // forever,
  // (i) listen for SC pkts
  // (ii) listen for data/data-ack pkts
  // (iii) listen for ctl/ctl-ack pkts
  // (iii) check for pending pkts to be sent

  should_listen = true;
  while(should_listen) {
    /*---- Check to receive packets ----*/
    fd_set rfds; FD_ZERO(&rfds);
    FD_SET(udp_sock, &rfds);
    int max_fd = udp_sock;

    for(int i = 0; i < max_cl; i++) {
      generic_client_info_t &cl = get_cl(i);
      if(cl.state == IDLE_CL) { continue; }
      FD_SET(cl.data_sock_fd, &rfds);
      if(max_fd < cl.data_sock_fd) { max_fd = cl.data_sock_fd; }
      FD_SET(cl.ctl_sock_fd, &rfds);
      if(max_fd < cl.ctl_sock_fd) { max_fd = cl.ctl_sock_fd; }
      FD_SET(cl.data_ack_sock_fd, &rfds);
      if(max_fd < cl.data_ack_sock_fd) { max_fd = cl.data_ack_sock_fd; }
      FD_SET(cl.ctl_ack_sock_fd, &rfds);
      if(max_fd < cl.ctl_ack_sock_fd) { max_fd = cl.ctl_ack_sock_fd; }
      FD_SET(cl.state_sock_fd, &rfds);
      if(max_fd < cl.state_sock_fd) { max_fd = cl.state_sock_fd; }
    }

    struct timeval timeout = { 0, 20*1000 }; // 20ms
    int retval = select(max_fd+1, &rfds, NULL, NULL, &timeout);
    if(retval < 0) { perror("gs.select"); continue; }

    for(int i = 0; i < max_cl; i++) {
      generic_client_info_t &cl = get_cl(i);
      if(cl.state == IDLE_CL) { continue; }
      if(FD_ISSET(cl.data_sock_fd, &rfds)) { cl.handleData(); }
      if(cl.state == IDLE_CL) { continue; }
      if(FD_ISSET(cl.ctl_sock_fd, &rfds)) { cl.handleCtl(); }
      if(cl.state == IDLE_CL) { continue; }
      if(FD_ISSET(cl.data_ack_sock_fd, &rfds)) { cl.handleDataAck(); }
      if(cl.state == IDLE_CL) { continue; }
      if(FD_ISSET(cl.ctl_ack_sock_fd, &rfds)) { cl.handleCtlAck(); }
      if(cl.state == IDLE_CL) { continue; }
      if(FD_ISSET(cl.state_sock_fd, &rfds)) { cl.handleSoftState(); }
    }
    // pkt from SC
    if(FD_ISSET(udp_sock, &rfds)) { handleSC(); }

    /*---- Check for packets to be (re)sent ----*/
    for(int i = 0; i < max_cl; i++) {
      generic_client_info_t &cl = get_cl(i);
      if(cl.state == IDLE_CL) { continue; }
      cl.checkDataSend();	if(cl.state == IDLE_CL) { continue; }
      cl.checkCtlSend();	if(cl.state == IDLE_CL) { continue; }
      cl.checkSoftStateSend();	if(cl.state == IDLE_CL) { continue; }
      cl.checkHBSend();		if(cl.state == IDLE_CL) { continue; }
      cl.checkPathTimeout();
    }

  } // End for(;;)

  for(int i = 0; i < max_cl; i++) {
    generic_client_info_t &cl = get_cl(i);
    if(cl.state != IDLE_CL) { cl.endSession(); }
  }

  dbg->close();
  cerr << myName() << " service terminated on port " << sc_listen_port
       << "\a" << endl;

  return 0;
} // End generic_service::main()

int generic_service::findCl(ServCompIFPkt_t *sPkt) {
  for(int i = 0; i < max_cl; i++) {
    generic_client_info_t &cl = get_cl(i);
    if((cl.state != IDLE_CL) && (cl.pathID == sPkt->fixed.pathID) &&
       (cl.path_dest_scid == sPkt->fixed.path_dest_scid) &&
       (cl.version == sPkt->fixed.version) &&
       (cl.serv_my_scid == sPkt->fixed.serv_my_scid) &&
       (cl.serv_dest_scid == sPkt->fixed.serv_dest_scid)) {
      return i;
    }
  }
  return -1;
} // End generic_service::findCl()

void generic_service::handleSC() {
  // recv the pkt
  UDPLibPkt_t uPkt; ServCompIFPkt_t sPkt;
  int recv_len; struct sockaddr_in from_addr;
  unsigned long rcv_count = 0; bool pkt_ready = true;
  int res = UDPLib::blockingRecv(udp_sock, &uPkt, &recv_len,
				 &from_addr, NULL, &rcv_count, pkt_ready);
  if(res != 1) { cerr << "res = " << res << endl; exit(4); }
  // parse it
  int parseRes = sPkt.parsePkt(uPkt.data, recv_len);
  if(parseRes != 1) {
    cerr << "parseRes=" << parseRes << " recv_len=" << recv_len << endl;
    exit(5);
  }

  ServCompIFPkt_t resp_sPkt;
  if(sPkt.fixed.type == APP_CREATE_SERV_INST) {
    // if create instance request, create a path
    handleCreate(&sPkt, &resp_sPkt);

  } else if(sPkt.fixed.type == APP_CHANGE_SERV_INST) {
    // if change service instance request, change destn
    handleCSI(&sPkt, &resp_sPkt);

  } else if(sPkt.fixed.type == APP_KILL_SERV_INST) {
    // if kill service instance, kill path
    handleKill(&sPkt, &resp_sPkt);

  } else {
    // return error
    cerr << "gs: unknown pkt type ["
	 << ServCompIFPkt::type2str((ServCompPktType_t)sPkt.fixed.type)
	 << "]\n";
    resp_sPkt.fixed.success = 0;

  }

  int send_len = resp_sPkt.toStr(uPkt.data, sizeof(uPkt.data));
  int send_fd = udp_sock; // can use same sock to send pkts too
  bool use_tunnel = false; // should not send pkt via the delaying tunnel
  unsigned long count = 0;
  res = UDPLib::sendResp(send_fd, &uPkt, send_len,
			 from_addr.sin_addr, SERV_COMP_SERVER_PORT,
			 use_tunnel, &count);
  if(res != 1) {
    cerr << "gs.handleSC: res = " << res << endl; return;
  }

} // End generic_service::handleSC()

bool generic_service::checkDup(ServCompIFPkt_t *sPkt, ServCompIFPkt_t *resp_sPkt) {
  for(int i = 0; i < max_cl; i++) {
    generic_client_info_t &cl = get_cl(i);
    if(cl.state == IDLE_CL) { continue; }
    if((cl.pathID == sPkt->fixed.pathID) &&
       (cl.path_dest_scid == sPkt->fixed.path_dest_scid) &&
       (cl.version == sPkt->fixed.version) &&
       (cl.serv_my_scid == sPkt->fixed.serv_my_scid) &&
       (cl.serv_dest_scid == sPkt->fixed.serv_dest_scid)) {
      dbg->print() << myName() << " dup-req p_id=" << sPkt->fixed.pathID
		   << ':' << sPkt->fixed.path_dest_scid
		   << " path_ver=" << sPkt->fixed.version
		   << " serv_my_scid=" << sPkt->fixed.serv_my_scid
		   << " serv_dest_scid=" << sPkt->fixed.serv_dest_scid
		   << endl;
      resp_sPkt->fixed.success = 1;
      JString my_dest_spec =
	ip2str(myIP) + ":" + long2str(cl.data_listen_port);
      resp_sPkt->addPair(PREV_DEST_SPEC_F, my_dest_spec.cstr());
      return true;
    }
  }
  return false;
} // End generic_service::checkDup()

int generic_service::getFreeCl() {
  for(int i = 0; i < max_cl; i++) {
    generic_client_info_t &cl = get_cl(i);
    if(cl.state == IDLE_CL) { return i; }
  }
  return -1;
} // End generic_service::getFreeCl()

void generic_service::handleCreate(ServCompIFPkt_t *sPkt,
				   ServCompIFPkt_t *resp_sPkt,
				   enum cl_state_t next_state) {
  bool to_ret = checkDup(sPkt, resp_sPkt);
  if(to_ret) { return; }

  int cl_index = getFreeCl();
  if(cl_index < 0) {
    dbg->print() << myName() << " no-free-cl p_id=" << sPkt->fixed.pathID
		 << ':' << sPkt->fixed.path_dest_scid << " version="
		 << sPkt->fixed.version << " serv_dest_scid="
		 << sPkt->fixed.serv_dest_scid << " serv_my_scid="
		 << sPkt->fixed.serv_my_scid << endl;
    resp_sPkt->fixed.success = 0; return;
  }
  generic_client_info_t &cl = get_cl(cl_index);
  int data_listen_port = data_port_base + 10*cl_index;
  // get params from the SC pkt
  cl.init_from_sc(sPkt, resp_sPkt, data_listen_port);
  cl.state = next_state; // change state

  dbg->print() << myName() << " create-serv p_id=" << cl.pathID
	       << ':' << cl.path_dest_scid << " path_ver="
	       << cl.version << " data_listen_port="
	       << cl.data_listen_port << " dest=" << ip2str(cl.destIP)
	       << ':' << cl.destPort << " serv_dest_scid="
	       << cl.serv_dest_scid << " serv_my_scid="
	       << cl.serv_my_scid << endl;

  // add in the response info
  resp_sPkt->fixed.success = 1;
  JString my_dest_spec = ip2str(myIP) + ":" + long2str(cl.data_listen_port);
  resp_sPkt->addPair(PREV_DEST_SPEC_F, my_dest_spec.cstr());
} // End generic_service::handleCreate()

void generic_service::handleCSI(ServCompIFPkt_t *sPkt, ServCompIFPkt_t *resp_sPkt) {
  // find the matching client
  int cl_index = findCl(sPkt);
  if(cl_index < 0) {
    cerr << "gs.CSI: no match for p_id=" << sPkt->fixed.pathID
	 << ':' << sPkt->fixed.path_dest_scid << " path_ver="
	 << sPkt->fixed.version << endl;
    resp_sPkt->fixed.success = 0;
    return;
  }

  // sanity check
  generic_client_info_t &cl = get_cl(cl_index);
  if(sPkt->fixed.serv_my_scid != cl.serv_my_scid) {
    cerr << "cl.handleCSI: cl.serv_my_scid=" << cl.serv_my_scid
	 << " sPkt->fixed.serv_my_scid=" << sPkt->fixed.serv_my_scid << endl;
    resp_sPkt->fixed.success = 0;
    return;
  }

  // update destn info for that client
  if(!cl.parseDestSpec(sPkt->getValue(PREV_DEST_SPEC_F))) {
    cerr << "parseDestSpec error\n";
    resp_sPkt->fixed.success = 0;
    return;
  }
  cl.serv_dest_scid = sPkt->fixed.serv_dest_scid;

  dbg->print() << myName() << " change-serv p_id=" << cl.pathID
	       << ':' << cl.path_dest_scid << " path_ver=" << cl.version
	       << " dest=" << ip2str(cl.destIP) << ':' << cl.destPort << endl;

  cl.sendUpstreamInfo(); // send upstream information, to the elt downstream

  // return success
  resp_sPkt->fixed.success = 1;
} // End generic_service::handleCSI()

void generic_service::handleKill(ServCompIFPkt_t *sPkt, ServCompIFPkt_t *resp_sPkt) {
  // find the matching client
  int cl_index = findCl(sPkt);
  if(cl_index < 0) {
    cerr << "gs.Kill: no match for p_id=" << sPkt->fixed.pathID
	 << ':' << sPkt->fixed.path_dest_scid << " path_ver="
	 << sPkt->fixed.version << endl;
    resp_sPkt->fixed.success = 0;
    return;
  }

  // end session for this particular client
  generic_client_info_t &cl = get_cl(cl_index);
  cl.endSession();

  dbg->print() << myName() << " kill-serv p_id=" << cl.pathID
	       << ':' << cl.path_dest_scid << " path_ver="
	       << cl.version << endl;

  resp_sPkt->fixed.success = 1;
} // End generic_service::handleKill()
