/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Implements the routines in LinkState.h */

#include "LinkState.h"
#include "NetworkGraph.h"
#include "../util/Util.h"
#include <netdb.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

pthread_t LinkState::sendThr;
pthread_t LinkState::listenThr;

int LinkState::listen_fd;
int LinkState::s_ack_listen_fd;
int LinkState::f_ack_listen_fd;

reqRespBlock_t *LinkState::send_rrBlock;
UDPPktBlock_t *LinkState::send_pktBlock;
reqRespBlock_t *LinkState::fwd_rrBlock;
UDPPktBlock_t *LinkState::fwd_pktBlock;

const unsigned short LinkState::MAX_SEQNO = 0x7fff;

unsigned short LinkState::maxThreshold = 40; // 40ms
unsigned short LinkState::softStatePeriod = 60; // 60sec
unsigned short LinkState::rand_period = 60; // 60sec
struct timeval LinkState::lastSendTime;

bool LinkState::run_send_thread = true;
bool LinkState::run_listen_thread = true;
bool LinkState::first_send = true;
unsigned short LinkState::send_seqNo = 0;

Uniform LinkState::U;

bool LinkState::start_send = false;
bool LinkState::start_listen = false;
bool LinkState::stop_send = false;
bool LinkState::stop_listen = false;
Debug_t *LinkState::sDbg;
Debug_t *LinkState::lDbg;

unsigned long LinkState::numSent = 0;
unsigned long LinkState::numRecd = 0;

int LinkState::init(Debug_t *arg_s_dbg, Debug_t *arg_l_dbg) {
  stop_send = stop_listen = false;
  numRecd = numSent = 0;

  sDbg = arg_s_dbg;
  lDbg = arg_l_dbg;

  // socket for listening
  listen_fd = sock_and_bind(LINK_STATE_PORT);
  if(listen_fd < 0) { perror("LinkState::listenThread.socket"); return 0; }

  s_ack_listen_fd = sock_and_bind(LINK_STATE_SEND_ACK_PORT);
  if(s_ack_listen_fd < 0) { perror("LinkState::init.socket"); return 0; }

  f_ack_listen_fd = sock_and_bind(LINK_STATE_FWD_ACK_PORT);
  if(f_ack_listen_fd < 0) { perror("LinkState::init.socket"); return 0; }

  // get the request response block for sends
  send_rrBlock = UDPLib::getReqRespBlock();
  if(send_rrBlock == NULL) {
    cerr << "LinkState::init() send_rrBlock is NULL" << endl;
    return 0;
  }

  // get a block of packet space for sends
  send_pktBlock = UDPPktBlock::getUDPPktBlock();
  if(send_pktBlock == NULL) {
    cerr << "LinkState::init() send_pktBlock is NULL" << endl;
    return 0;
  }

  // get the request response block for forwards
  fwd_rrBlock = UDPLib::getReqRespBlock();
  if(fwd_rrBlock == NULL) {
    cerr << "LinkState::init() fwd_rrBlock is NULL" << endl;
    return 0;
  }

  // get a block of packet space for forwards
  fwd_pktBlock = UDPPktBlock::getUDPPktBlock();
  if(fwd_pktBlock == NULL) {
    cerr << "LinkState::init() fwd_pktBlock is NULL" << endl;
    return 0;
  }

  // set the seed for the random number generator
  // set random seed
  struct timeval tv;
  gettimeofday(&tv, NULL);
  unsigned long seedLong = tv.tv_usec%211;
  float seedFloat = ((seedLong+1)*1.0)/212;
  Uniform::Set(seedFloat);

  first_send = true;
  send_seqNo = 0;

  return 1;
} // End LinkState::init()

void LinkState::startSendThread(bool arg_run_send_thread) {
  run_send_thread = arg_run_send_thread;
  start_send = true;
  if(!run_send_thread) { return; }
  // Spawn the thread
  pthread_create(&sendThr, NULL, LinkState::sendThread, 0);
} // End LinkState::startSendThread()

void LinkState::startListenThread(bool arg_run_listen_thread) {
  run_listen_thread = arg_run_listen_thread;
  start_listen = true;
  if(!run_listen_thread) { return; }
  // Spawn the thread
  pthread_create(&listenThr, NULL, LinkState::listenThread, 0);
} // End LinkState::startListenThread()

void LinkState::sendToPeers(int send_fd, UDPLibPkt_t &send_pkt,
			    int send_len, vector<Peer_t> &peer_list,
			    int numPeers, unsigned short *flags,
			    enum ls_send_or_fwd_t send_or_fwd) {
  UDPPktBlock_t *pktBlock;
  reqRespBlock_t *rrBlock;
  int ack_listen_fd;
  if(send_or_fwd == LS_SEND) {
    pktBlock = send_pktBlock;
    rrBlock = send_rrBlock;
    ack_listen_fd = s_ack_listen_fd;
  } else {
    pktBlock = fwd_pktBlock;
    rrBlock = fwd_rrBlock;
    ack_listen_fd = f_ack_listen_fd;
  }

  for(int i = 0; i < numPeers; i++) {
    if(!flags[i]) { continue; }
    if((peer_list[i].hb->getPeerStatus() == DEAD) ||
       (peer_list[i].hb->getMyStatusPeerView() == DEAD)) {
      continue;
    }

    UDPLibPkt_t *s_pkt = pktBlock->getPkt();
    if(s_pkt == NULL) {
      fprintf(stderr, "LinkState::sendToPeers: No space for s pkt\n");
      exit(33);
    }

    memcpy((void*)(s_pkt->data), (void*)(send_pkt.data), send_len);

    UDPLibPkt_t *ack_pkt = pktBlock->getPkt();
    if(ack_pkt == NULL) {
      fprintf(stderr, "LinkState::sendToPeers: No space for ack pkt\n");
      exit(33);
    }

    struct timeval timeout = peer_list[i].getTimeout();
    bool use_tunnel = true;
    int reqNo = rrBlock->insertRequest
      (send_fd, s_pkt, send_len, peer_list[i].ip_addr,
       LINK_STATE_PORT, ack_listen_fd, ack_pkt, timeout, use_tunnel,
       peer_list[i].scid, NetworkGraph::mySCID);
    if(reqNo == -1) {
      fprintf(stderr, "LinkState: no space for request\n");
      pktBlock->freePkt(ack_pkt);
      pktBlock->freePkt(s_pkt);
      continue;
    } else {
      if(send_or_fwd == LS_SEND) { Peer::send_addReq(reqNo); }
      else { Peer::fwd_addReq(reqNo); }
    }

    /*
    cerr << "[S" << (int)send_or_fwd << "] reqNo=" << reqNo
	 << " ts=" << ntohl(s_pkt->ts) << " nonce="
	 << ntohl(s_pkt->nonce) << " peerSCID=" << peer_list[i].scid
	 << '(' << ip2str(peer_list[i].ip_addr) << ')' << endl;
    */

  }

} // End LinkState::sendToPeers()

void* LinkState::sendThread(void*) {
  if(!run_send_thread) {
    fprintf(stderr, "LS::sendThread running without run_send_thread set\n");
    exit(96);
  }

  sDbg->setCurrFn(SENDTHREAD_FN);

  // send out a flood right after starting up
  sendHelper();

  // forever
  for(;;) {
    if(stop_send) { break; }
    /* check if any space in the send_pktBlock has to be freed */
    checkAcks(LS_SEND, run_send_thread);
    // Don't spin loop.  sleep for 30ms -- negligible compared to
    // other delays in failure detection
    usleep(30*1000);
    // call sendHelper
    sendHelper();
  } // End for(;;)

  return 0;
} // End LinkState::sendThread()

void LinkState::sendHelper() {
  if(!start_send || stop_send) { return; }

  sDbg->setCurrFn(SENDTHREAD_FN);

  // can use same socket for sending
  int send_fd = s_ack_listen_fd;

  // get list of peers
  unsigned short numPeers;
  vector<Peer_t> &peer_list = NetworkGraph::getPeerList(&numPeers);

  // send latency info "of all peers", "to all peers" -- check if any
  // latency has changed significantly simultaneously, form the packet
  // to be sent if so, send "to all peers"

  // number of peers for which latency has changed significantly
  int numChanges = 0; 

  UDPLibPkt_t send_pkt;
  int pkt_len = 0;
  ls_pkt_t *ls_p = (ls_pkt_t*)send_pkt.data;

  entry_t *entries = &ls_p->first_entry;
  unsigned short flags[MAX_PEERS];
  bool toSendAll = false;

  // am i sending for the first time?
  if(first_send) { toSendAll = true; }

  if(!first_send && (now(lastSendTime) > rand_period*1000)) {
    // send everything every 'softStatePeriod' -- simple timeout to
    // implement soft-state
    toSendAll = true;
  }

  for(int i = 0; i < numPeers; i++) {
    flags[i] = 1;

    bool toSend = toSendAll;

    // check if new latency/status has to be sent
    short peer_status = peer_list[i].hb->getPeerStatus();
    short my_status_peer_view = peer_list[i].hb->getMyStatusPeerView();
    short link_status = LIVE;
    if((peer_status == DEAD) || (my_status_peer_view == DEAD)) {
      link_status = DEAD;
    }
    short currRTT = (short)peer_list[i].latMeas->getSRTT();

    if(!first_send) {
      if((link_status == DEAD) && (peer_list[i].lastRTTSent >= 0)) {
	// link was up earlier, is dead now
	if(sDbg->use_stream) {
	  sDbg->print() << "peer-link-dead peerSCID=" << peer_list[i].scid << endl;
	} else {
	  LogLine_t ll; ll.desc = peer_link_dead_desc;
	  ll.peerSCID = peer_list[i].scid; sDbg->addLL(ll);
	}
	toSend = true;
      } // end case status change from live to dead

      if(link_status == LIVE) {
	if(peer_list[i].lastRTTSent == -1) {
	  // was dead earlier, is live now
	  if(sDbg->use_stream) {
	    sDbg->print() << "peer-link-live peerSCID=" << peer_list[i].scid << endl;
	  } else {
	    LogLine_t ll; ll.desc = peer_link_live_desc;
	    ll.peerSCID = peer_list[i].scid; sDbg->addLL(ll);
	  }
	  toSend = true;
	} else {
	  if(abs(currRTT - peer_list[i].lastRTTSent) >= maxThreshold) {
	    if(sDbg->use_stream) {
	      sDbg->print() << "peer-cost-change peerSCID="
			    << peer_list[i].scid << " curr=" << currRTT
			    << " last=" << peer_list[i].lastRTTSent << endl;
	    } else {
	      LogLine_t ll; ll.desc = peer_cost_change_desc;
	      ll.peerSCID = peer_list[i].scid;
	      ll.utype = LS_U; ll.u.ls.init(); ll.u.ls.currRTT = currRTT;
	      ll.u.ls.lastRTT = peer_list[i].lastRTTSent; sDbg->addLL(ll);
	    }
	    toSend = true;
	  }
	}
      } // end case link LIVE

      // end if(!first_send)
    } else {
      if(link_status == DEAD) {
	if(sDbg->use_stream) {
	  sDbg->print() << "start-dead peerSCID=" << peer_list[i].scid << endl;
	} else {
	  LogLine_t ll; ll.desc = start_dead_desc;
	  ll.peerSCID = peer_list[i].scid; sDbg->addLL(ll);
	}
      } else {
	if(sDbg->use_stream) {
	  sDbg->print() << "start-live peerSCID=" << peer_list[i].scid << endl;
	} else {
	  LogLine_t ll; ll.desc = start_live_desc;
	  ll.peerSCID = peer_list[i].scid; sDbg->addLL(ll);
	}
      }

      // end else of if(!first_send)
    }

    if(!toSend) { continue; }

    entries[numChanges].origin_scid = htons(peer_list[i].scid);
    short oneWayDelay;
    if(link_status == DEAD) {
      // -ve value indicates that the path is dead
      peer_list[i].lastRTTSent = -1;
      oneWayDelay = -1;
    } else {
      // path to peer is live
      peer_list[i].lastRTTSent = currRTT;
      oneWayDelay = peer_list[i].lastRTTSent/2; // assume symmetry
    }
    entries[numChanges].oneWayDelay = htons(oneWayDelay);

    numChanges++;
    pkt_len += sizeof(entry_t);

    // add entry to graph
    NetworkGraph::addEdge(NetworkGraph::mySCID, peer_list[i].scid,
			  oneWayDelay);
    if(first_send || (oneWayDelay == -1)) {
      // first_send -- add reverse edge also to begin with
      // no point in having just one edge for graph computations!
      // so mark the other direction also as INFINITE cost (-1)
      NetworkGraph::addEdge(peer_list[i].scid, NetworkGraph::mySCID,
			    oneWayDelay);
    }

  } // End for(i)

  if(numChanges > 0) {
    ls_p->dest_scid = htons(NetworkGraph::mySCID);
    pkt_len += sizeof(ls_p->dest_scid);
    ls_p->seq_no = htons(send_seqNo);
    pkt_len += sizeof(ls_p->seq_no);
    send_seqNo++;
    if(send_seqNo > MAX_SEQNO) { send_seqNo = 0; }
    ls_p->num_entries = htons(numChanges);
    pkt_len += sizeof(ls_p->num_entries);

    //cerr << "Sending: num_entries = " << numChanges
    //<< " seqNo = " << send_seqNo << endl;

    // send the packet "to all peers"
    gettimeofday(&lastSendTime, NULL);
    sendToPeers(send_fd, send_pkt, pkt_len, peer_list, numPeers,
		flags, LS_SEND);
    // reset soft-state period
    rand_period = (unsigned short)((0.75+U.Next()*0.25)*softStatePeriod);
    // reset first_send
    first_send = false;

    if(gdbg) {
      fprintf(stderr, "[%d|%d]", (int)NetworkGraph::mySCID,
	      (int)rand_period);
    }
  } // end if(numChanges > 0)

} // End LinkState::sendHelper()

void* LinkState::listenThread(void*) {
  if(!run_listen_thread) {
    fprintf(stderr, "LS::listenThread running without run_listen_thread set\n");
    exit(96);
  }

  lDbg->setCurrFn(LISTENTHREAD_FN);

  // listen for packets forever, deal as appropriate
  for(;;) {
    if(stop_listen) { break; }

    /* check if any space in the fwd_pktBlock has to be freed */
    checkAcks(LS_FWD, run_listen_thread);
    // if sending thread has stopped, then we're in clean up mode
    if(run_send_thread && stop_send) {
      checkAcks(LS_SEND, run_send_thread);
    }

    bool pkt_ready = false;
    listenHelper(pkt_ready);

  } // End for(;;)

  return 0;
} // End LinkState::listenThread()

void LinkState::listenHelper(bool pkt_ready) {
  lDbg->setCurrFn(LISTENTHREAD_FN);

  struct sockaddr_in from_addr;
  UDPLibPkt_t p;
  int recv_len;

  // receive packet (blocking)
  struct timeval timeout = { 0, 120*1000 }; // 120ms
  int res = UDPLib::blockingRecv(listen_fd, &p, &recv_len,
				 &from_addr, &timeout, &numRecd, pkt_ready);
  if(res == -1) { return; }
  if(stop_send) { return; }

  Peer_t *peer = NetworkGraph::getPeerWithIP(from_addr.sin_addr);
  if(peer == NULL) {
    fprintf(stderr, "%d: LS from unknown peer %s\n",
	    NetworkGraph::mySCID, inet_ntoa(from_addr.sin_addr));
    exit(86);
  }

  ls_pkt_t *ls_p = (ls_pkt_t*)p.data;
  unsigned short dest_scid = ntohs(ls_p->dest_scid);
  unsigned short recv_seqNo = ntohs(ls_p->seq_no);
  int ack_port = LINK_STATE_FWD_ACK_PORT;
  enum ls_send_or_fwd_t s_o_f = LS_FWD;
  if(peer->scid == dest_scid) {
    ack_port = LINK_STATE_SEND_ACK_PORT;
    s_o_f = LS_SEND;
  }

  /*
  cerr << "[R" << (int)s_o_f << "] reqNo=" << ntohl(p.reqNo)
       << " ts=" << ntohl(p.ts) << " nonce=" << ntohl(p.nonce)
       << " peerSCID=" << peer->scid << '(' << ip2str(peer->ip_addr)
       << ')' << endl;
  */

  // send UDPLib response
  int send_fd = listen_fd; // can use the same socket for sending
  int send_len = 0; // this is just an ack
  bool use_tunnel = true;
  res = UDPLib::sendResp
    (send_fd, &p, send_len, from_addr.sin_addr, ack_port,
     use_tunnel, &numSent);
  if(res != 1) { fprintf(stderr, "LS Warning: error sending response\n"); }

  // check if duplicate
  //cerr << "dest_scid = " << dest_scid
  //<< " recv_seqNo = " << recv_seqNo << endl;
  if(dest_scid == NetworkGraph::mySCID) {
    // what goes around, comes around
    if(gdbg) { fprintf(stderr, "[g]"); }
    return;
  }
  if(NetworkGraph::isDuplicate(dest_scid, recv_seqNo)) {
    if(gdbg) { fprintf(stderr, "[d]"); }
    return;
  }

  // note the latency changes in graph
  unsigned short num_entries = ntohs(ls_p->num_entries);
  //cerr << "num_entries = " << num_entries << endl;
  entry_t *entries = &ls_p->first_entry;

  for(int i = 0; i < num_entries; i++) {
    unsigned short origin_scid = ntohs(entries[i].origin_scid);
    short oneWayDelay = ntohs(entries[i].oneWayDelay);
    //cerr << "origin_scid = " << origin_scid
    //<< " oneWayDelay = " << oneWayDelay << endl;
    if(origin_scid == NetworkGraph::mySCID) {
      Peer_t *peer = NetworkGraph::getPeerWithSCID(dest_scid);
      if(peer == NULL) {
	if(lDbg->use_stream) {
	  lDbg->print() << "ls-update-from-unknown-peer peerSCID="
			<< dest_scid << endl;
	} else {
	  LogLine_t ll; ll.desc = ls_update_from_unknown_peer_desc;
	  ll.peerSCID = dest_scid; lDbg->addLL(ll);
	}
	lDbg->close(); exit(129);
      }
      if((peer->hb->getMyStatusPeerView() == DEAD) ||
	 (peer->hb->getPeerStatus() == DEAD)) {
	oneWayDelay = -1;
      }
    }

    unsigned short oldCost =
      NetworkGraph::addEdge(dest_scid, origin_scid, oneWayDelay);
    if((oldCost == PriorityQueue<unsigned short>::INFINITY) &&
       (oneWayDelay != -1)) {
      if(lDbg->use_stream) {
	lDbg->print() << "graph-link-live dest_scid=" << dest_scid
		      << " origin_scid=" << origin_scid << endl;
      } else {
	LogLine_t ll; ll.desc = graph_link_live_desc;
	ll.dest_scid = dest_scid; ll.origin_scid = origin_scid;
	lDbg->addLL(ll);
      }
    }

    if(oneWayDelay == -1) {
      // no point in having just one edge for graph computations!
      // so mark the other direction also as INFINITE cost (-1)
      NetworkGraph::addEdge(origin_scid, dest_scid, oneWayDelay);
      if(lDbg->use_stream) {
	lDbg->print() << "graph-link-dead dest_scid=" << dest_scid
		      << " origin_scid=" << origin_scid << endl;
      } else {
	LogLine_t ll; ll.desc = graph_link_dead_desc;
	ll.dest_scid = dest_scid; ll.origin_scid = origin_scid;
	lDbg->addLL(ll);
      }
    }

  } // end for(i) over the entries in the recd pkt

  // forward message along all directions (using UDPLib)
  unsigned short numPeers;
  vector<Peer_t> &peer_list = NetworkGraph::getPeerList(&numPeers);

  int checkNumForwards = 0; // just a check
  unsigned short flags[MAX_PEERS];
  for(int i = 0; i < numPeers; i++) {
    if(peer_list[i].ip_addr.s_addr == from_addr.sin_addr.s_addr) {
      // no need to flood along link on which pkt was received
      flags[i] = 0;
    } else {
      checkNumForwards++;
      flags[i] = 1;
    }
  } // End for(i)

  if(checkNumForwards != (numPeers-1)) {
    fprintf(stderr, "LS Err: checkNumForwards=%d, numPeers-1=%d\n",
	    checkNumForwards, numPeers-1);
    exit(32);
  }

  if(numPeers > 1) {
    // forward to peers
    sendToPeers(send_fd, p, recv_len, peer_list, numPeers, flags, LS_FWD);
  }
} // End LinkState::listenHelper()

void LinkState::checkAcks(enum ls_send_or_fwd_t send_or_fwd, bool should_listen) {

  UDPPktBlock_t *pktBlock; reqRespBlock_t *rrBlock;
  int ack_listen_fd;
  if(send_or_fwd == LS_SEND) {
    pktBlock = send_pktBlock; rrBlock = send_rrBlock;
    ack_listen_fd = s_ack_listen_fd;
  } else {
    pktBlock = fwd_pktBlock; rrBlock = fwd_rrBlock;
    ack_listen_fd = f_ack_listen_fd;
  }

  if(should_listen) {
    int num_recd_now = rrBlock->check_all(ack_listen_fd);
    if(num_recd_now > 100) { fprintf(stderr, "[r=%d]", num_recd_now); }
  }

  //const int MAX_REQ_NOS = 32;
  //int reqNos[MAX_REQ_NOS];
  vector<reqInfo_t> &reqs = rrBlock->reqs;
  vector<int> &reqNos = (send_or_fwd == LS_SEND) ?
    Peer::send_vec : Peer::fwd_vec;
  vector<bool> &reqNo_vec = (send_or_fwd == LS_SEND) ?
    Peer::send_reqNo_vec : Peer::fwd_reqNo_vec;

  int numReqs;
  // get the request numbers to check
  if(send_or_fwd == LS_SEND) {
    //numReqs = send_popReqNos(reqNos, MAX_REQ_NOS);
    numReqs = Peer::send_vec_pop();
  } else {
    //numReqs = Peer::fwd_popReqNos(reqNos, MAX_REQ_NOS);
    numReqs = Peer::fwd_vec_pop();
  }

  for(int j = 0; j < numReqs; j++) {
    int reqNo = reqNos[j];
    if((reqNo < 0) || (reqNo >= reqs.size())) {
      cerr << NetworkGraph::mySCID << ": reqNos[" << j << "]=" << reqNos[j]
	   << ", reqs.size()=" << reqs.size() << endl;
      exit(45);
    }

    int ackRecd = 0;
    if(should_listen && reqs[reqNo].recd_now) { ackRecd = 1; }
    else { ackRecd = rrBlock->check_t_o_resend(reqNo); }
    if(!should_listen && (ackRecd == 1)) {
      cerr << NetworkGraph::mySCID << ": ls-ack-err reqNo=" << reqNo
	   << " active=" << reqNo_vec[reqNo]
	   << " s_o_f=" << send_or_fwd << endl;
      Debug_t *dbg = (send_or_fwd == LS_SEND) ? sDbg : lDbg;
      dbg->close(); exit(44);
    }
    struct sockaddr_in ack_from = reqs[reqNo].recv_from;
    //struct sockaddr_in ack_from;
    //int ackRecd = rrBlock->checkResp(reqNo, NULL, &ack_from);
    if(ackRecd != 0) {
      checkAckHelper(send_or_fwd, ackRecd, reqNo);
    } else {
      // push back the reqNo onto the queue for later examination
      if(send_or_fwd == LS_SEND) { Peer::send_addReq(reqNo); }
      else { Peer::fwd_addReq(reqNo); }
    }
  } // End for(j)

} // End LinkState::checkAcks()

void LinkState::recvAck(enum ls_send_or_fwd_t send_or_fwd) {
  // assume that the ack is ready, go ahead and receive it, and
  // process it
  UDPPktBlock_t *pktBlock; reqRespBlock_t *rrBlock;
  int ack_listen_fd;
  if(send_or_fwd == LS_SEND) {
    pktBlock = send_pktBlock; rrBlock = send_rrBlock;
    ack_listen_fd = s_ack_listen_fd;
  } else {
    pktBlock = fwd_pktBlock; rrBlock = fwd_rrBlock;
    ack_listen_fd = f_ack_listen_fd;
  }
  vector<bool> &reqNo_vec = (send_or_fwd == LS_SEND) ?
    Peer::send_reqNo_vec : Peer::fwd_reqNo_vec;

  // get the ack pkt
  int reqNo = rrBlock->check_one(ack_listen_fd);
  if(reqNo < 0) {
    // must be an old ack -- discard
    cerr << "[l:r=-1]"; return;
  }

  // sanity checks
  if(reqNo >= reqNo_vec.size()) {
    cerr << NetworkGraph::mySCID << ": LS::recvAck reqNo=" << reqNo
	 << " s_or_f=" << (int)send_or_fwd << " reqNo_vec.size="
	 << reqNo_vec.size() << endl;
    exit(43);
  }
  if(!reqNo_vec[reqNo]) {
    cerr << NetworkGraph::mySCID << ": LS::recvAck inactive reqNo="
	 << reqNo << " s_or_f=" << (int)send_or_fwd << " active="
	 << reqNo_vec[reqNo] << endl;
    exit(45);
  }

  int ackRecd = 1;
  checkAckHelper(send_or_fwd, ackRecd, reqNo);
  reqNo_vec[reqNo] = false;
} // End LinkState::recvAck()

void LinkState::checkAckHelper(enum ls_send_or_fwd_t send_or_fwd, int ackRecd,
			       int reqNo) {
  if(ackRecd == -1) {
    fprintf(stderr, "[T%d]", (int)send_or_fwd);
    //cerr << " reqNo=" << reqNo << endl;
  }
  if(ackRecd == 1) {
    //fprintf(stderr, "[A%d]", (int)send_or_fwd);
    //cerr << " reqNo=" << reqNo << endl;
  }
  UDPPktBlock_t *pktBlock; reqRespBlock_t *rrBlock;
  if(send_or_fwd == LS_SEND) {
    pktBlock = send_pktBlock; rrBlock = send_rrBlock;
  } else {
    pktBlock = fwd_pktBlock; rrBlock = fwd_rrBlock;
  }
  // ack has been received, or timeout has happened
  // free the send and receive packets
  pktBlock->freePkt(rrBlock->reqs[reqNo].send_pkt);
  pktBlock->freePkt(rrBlock->reqs[reqNo].recv_pkt);
  // free the udp-lib entry
  rrBlock->freeEntry(reqNo);
} // End LinkState::checkAckHelper()

void LinkState::stopSendThread() {
  stop_send = true;
  if(!run_send_thread) { return; }
  pthread_join(sendThr, NULL);
} // End LinkState::stopSendThread()

void LinkState::stopListenThread() {
  stop_listen = true;
  if(!run_listen_thread) { return; }
  pthread_join(listenThr, NULL);
} // End LinkState::stopListenThread()

unsigned long LinkState::getNumSent() {
  return numSent + send_rrBlock->numSent + fwd_rrBlock->numSent;
} // End LinkState::getNumSent()

unsigned long LinkState::getNumRecd() {
  return numRecd + send_rrBlock->numRecd + fwd_rrBlock->numRecd;
} // End LinkState::getNumRecd()
