/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
#ifndef LINK_STATE_H
#define LINK_STATE_H

#include <pthread.h>
#include <vector>

#include "Peer.h"
#include "../udp-lib/UDPLib.h"
#include "../udp-lib/UDPPktBlock.h"
#include "../util/newran02/include.h"
#include "../util/newran02/newran.h"
#include "../util/Debug.h"

#define LINK_STATE_PORT 3500
#define LINK_STATE_SEND_ACK_PORT 3502
#define LINK_STATE_FWD_ACK_PORT 3504

struct entry {
  /** service cluster id of the origin */
  unsigned short origin_scid;
  /** the one-way delay, negative value indicates failure of link */
  short oneWayDelay;
}; // End struct entry

typedef struct entry entry_t;

/** The packet sent for link-state updates.  NOTE: this is of variable
    size */
struct ls_pkt {
  /** service cluster id of the destination */
  unsigned short dest_scid;
  /** the sequence number of the update */
  unsigned short seq_no;
  /** number of entries */
  unsigned short num_entries;
  /** the entries -- can't list all entries (variable number) -- take
      pointer to first_entry and go from there */
  entry_t first_entry;
}; // End struct ls_pkt

typedef struct ls_pkt ls_pkt_t;

enum ls_send_or_fwd_t { LS_SEND=1, LS_FWD=2 };

/** The class that defines the routines for implementing the
    link-state info exchange layer for service composition */
class LinkState {
protected:
  /** The debug log file for the send thread */
  static Debug_t *sDbg;
  /** The debug log file for the listen thread */
  static Debug_t *lDbg;

  /** See 'startSendThread' */
  static pthread_t sendThr;
  /** See 'startListenThread' */
  static pthread_t listenThr;
  /** The number of packets sent in sendResp calls */
  static unsigned long numSent;
  /** The number of packets sent in blockingRecv calls */
  static unsigned long numRecd;

  /** The request response block for sending packets */
  static reqRespBlock_t *send_rrBlock;
  /** The block of packet -- for async sends */
  static UDPPktBlock_t *send_pktBlock;

  /** The request response block for forwarding packets */
  static reqRespBlock_t *fwd_rrBlock;
  /** The block of packet -- for async forwards */
  static UDPPktBlock_t *fwd_pktBlock;

  /** Implements 'sendThr' */
  static void* sendThread(void*);
  /** Implements 'listenThr' */
  static void* listenThread(void*);

  /** The maximum threshold -- will wait for latency to change at
      least this much before sending LS update for a peer (in
      addition, there is a soft-state LS update sent periodically) */
  static unsigned short maxThreshold;
  /** The period for the soft-state LS updates (in seconds) */
  static unsigned short softStatePeriod;
  /** The random soft-state period (to avoid synchronization) (in sec) */
  static unsigned short rand_period;
  /** The last time an LS update was sent */
  static struct timeval lastSendTime;
  /** The uniform random number generator (for varying maxThreshold) */
  static Uniform U;

  /** Help routine -- to send a given packet to all peers */
  static void sendToPeers(int send_fd, UDPLibPkt_t &send_pkt,
			  int send_len, vector<Peer_t> &peer_list,
			  int numPeers, unsigned short *flags,
			  enum ls_send_or_fwd_t send_or_fwd);

  /** Helper for checkAcks */
  static void checkAckHelper(enum ls_send_or_fwd_t s_or_f, int ackRecd,
			     int reqNo);

public:
  /** Should I run a send thread? */
  static bool run_send_thread;
  /** Should I run a listen thread? */
  static bool run_listen_thread;
  /** Is this the first time I'm sending? */
  static bool first_send;
  /** The sending sequence number */
  static unsigned short send_seqNo;

  /** the maximum sequence number of updates sent -- beyond this, the
      sequence number wraps around -- note that the number of unique
      sequence numbers is (MAX_SEQNO+1) -- including '0' */
  static const unsigned short MAX_SEQNO;

  /** Should I start sending? */
  static bool start_send;
  /** Should I start listening? */
  static bool start_listen;

  /** Should I stop sending? */
  static bool stop_send;
  /** Should I stop listening? */
  static bool stop_listen;

  /** The socket for listening for LS packets */
  static int listen_fd;
  /** The socket for listening for link-state acks send/fwd */
  static int s_ack_listen_fd;
  static int f_ack_listen_fd;

  /** Init the class variables -- socket for listening for acks,
      ReqRespBlock for UDPLib, etc.  Return 1 on success, 0 on failure */
  static int init(Debug_t *arg_s_dbg, Debug_t *arg_l_dbg);

  /** Start the thread that implements 'sendThr' -- this periodically
      floods the network with the link-state info of peers.  The
      argument tells whether a send thread should be run. */
  static void startSendThread(bool arg_run_send_thread);

  /** Start the thread that implements 'listenThr' -- this listens for
      incoming link-state updates and processes/forwards them as
      appropriate.  The argument tells whether a listen thread should
      be run. */
  static void startListenThread(bool arg_run_listen_thread);

  /** Stop the send thread */
  static void stopSendThread();
  /** Stop the listen thread */
  static void stopListenThread();

  /** Helper for the send thread -- called from outside when send
      thread is not running */
  static void sendHelper();
  /** Helper for the listen thread -- called from outside when send
      thread is not running.  The argument indicates if the pkt is
      ready to be received. */
  static void listenHelper(bool pkt_ready);

  /** To check for acks -- called from outside when either thread is
      not running.  The 'should_listen' flag tells if this routine
      should listen for acks.  If it is 'false', then it is assumed
      that someone else is listening for acks, and only resends are
      handled. */
  static void checkAcks(enum ls_send_or_fwd_t s_or_f, bool should_listen);
  /** To be called from outside when an ack pkt is ready */
  static void recvAck(enum ls_send_or_fwd_t s_or_f);

  /** Return the number of pkts sent by all the threads */
  static unsigned long getNumSent();
  /** Return the number of packets received */
  static unsigned long getNumRecd();

}; // End class LinkState

#endif
