/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Implements the class NetworkGraph */

#include <iostream>
#include <fstream>
#include <netdb.h>

#include "NetworkGraph.h"
#include "../util/JString.h"

const unsigned short NetworkGraph::INVALID_SCID = 0;
const short NetworkGraph::INVALID_NODE = -1;
unsigned short NetworkGraph::mySCID;

vector<GraphNode_t> NetworkGraph::nodes;
JHashtable<unsigned short, short> NetworkGraph::scidHash;
vector<Peer_t> NetworkGraph::peers;
JHashtable<unsigned short, short> NetworkGraph::peer_scidHash;
JHashtable<unsigned long, short> NetworkGraph::peer_ipHash;

void NetworkGraph::init(const char *scid_mapping, const char *graph,
			unsigned short arg_mySCID, bool read_all) {
  mySCID = arg_mySCID; // note my SCID
  //cerr << "mySCID is " << mySCID << endl;

  // set the hash functions
  scidHash.setHashFunction(ushortHash);
  peer_scidHash.setHashFunction(ushortHash);
  peer_ipHash.setHashFunction(ulongHash);

  // read the 'graph' file to get your adjacency list
  readGraph(graph, read_all);

  // read the file for the list of peers
  readSCIDMapping(scid_mapping);

} // End NetworkGraph::init()

void NetworkGraph::readGraph(const char *graph, bool read_all) {
  //cerr << "reading graph file " << graph << endl;
  ifstream g_fin(graph);
  for(;;) {
    // read each line
    if(g_fin.eof()) break;
    char line[8192*2];
    g_fin.getline(line, sizeof(line)); // read a line
    if(g_fin.eof()) break;
    if(g_fin.fail()) {
      cerr << "Error reading " << graph << endl;
      exit(99);
    }

    // line has format:
    // "<dest-scid> <origin-scid1> <cost1> <origin-scid2> <cost2>..."
    vector<JString> sVec;
    int numFields = split(line, sVec);
    if(numFields <= 0) {
      cerr << "Format of " << graph << " is wrong" << endl;
      exit(97);
    }
    unsigned short dest_scid = JString_atoi(sVec[0]);
    if(dest_scid == 0) {
      cerr << "dest_scid has to be > 0" << endl;
      exit(99);
    }

    if(!read_all && (dest_scid != mySCID)) { continue; }

    if(((numFields-1)%2) != 0) {
      cerr << "odd number of fields for scid " << dest_scid
	   << " in " << graph << endl;
      exit(98);
    }
    int numEdges = (numFields-1)/2;

    for(int i = 0; i < numEdges; i++) {
      unsigned short origin_scid = JString_atoi(sVec[1+i*2]);
      if(origin_scid == 0) {
	cerr << "origin_scid has to be > 0" << endl;
	exit(99);
      }
      unsigned short cost = JString_atoi(sVec[1+i*2+1]);
      addEdge(dest_scid, origin_scid, cost);

      if(dest_scid == mySCID) {
	Peer_t p;
	p.scid = origin_scid;
	// insert the peer structure in the database
	peer_scidHash.put(p.scid, peers.size());
	peers.push_back(p);
	if(!read_all) {
	  addEdge(p.scid, mySCID, cost); // add the back edge also
	}
      }

    } // End for(i) reading all edges  

  } // End for(;;) reading all lines in 'graph'
  g_fin.close();
} // End NetworkGraph::readGraph()

void NetworkGraph::readSCIDMapping(const char *scid_mapping) {
  vector<sc_node_t> sc1;
  util_readSCIDMapping(scid_mapping, sc1);
  for(unsigned int i = 0; i < sc1.size(); i++) {
    Peer_t *p = getPeerWithSCID(sc1[i].scid);
    if(p != NULL) { p->ip_addr = sc1[i].ipAddr; }
  }
  // prepare the peer_ipHash mapping
  for(unsigned int i = 0; i < peers.size(); i++) {
    peer_ipHash.put(peers[i].ip_addr.s_addr, i);
  }
} // End NetworkGraph::readSCIDMapping()

int NetworkGraph::isDupInfo(unsigned short recv_seqNo, DupInfo_t *d,
			    unsigned short MAX_SEQNO,
			    unsigned short recycleTime) {
  if(!d->recdLSPkts) {
    // this is the first packet from this node
    d->recdLSPkts = 1;
    gettimeofday(&d->lastLSUpdate, NULL);
    d->lastLSSeqNo = recv_seqNo;
    return 0;

  } else {
    // this is not the first packet

    float timeSinceLastUpdate = now(d->lastLSUpdate);
    if(timeSinceLastUpdate >= recycleTime*1000) {
      // recycleTime over, take any packet
      gettimeofday(&d->lastLSUpdate, NULL);
      d->lastLSSeqNo = recv_seqNo;
      return 0;

    } else {
      // recycleTime not over since last LS update

      if(recv_seqNo > d->lastLSSeqNo) {
	if((recv_seqNo - d->lastLSSeqNo) >= ((MAX_SEQNO+1)/2)) {
	  // this is NOT a valid forward sequence number
	  return 1; // is a duplicate
	} else {
	  gettimeofday(&d->lastLSUpdate, NULL);
	  d->lastLSSeqNo = recv_seqNo;
	  return 0; // not a duplicate
	}
      } else {
	if((d->lastLSSeqNo - recv_seqNo) < ((MAX_SEQNO+1)/2)) {
	  // this is NOT a valid forward sequence number
	  return 1; // is a duplicate
	} else {
	  gettimeofday(&d->lastLSUpdate, NULL);
	  d->lastLSSeqNo = recv_seqNo;
	  return 0; // not a duplicate
	}
      }

    }
  }

} // End NetworkGraph::isDupInfo()

int NetworkGraph::isDuplicate(unsigned short origin_scid,
			      unsigned short recv_seqNo) {
  GraphNode_t *p = getNodeWithSCID(origin_scid);
  if(p == NULL) {
    //cerr << "isDuplicate: No node with SCID " << origin_scid << endl;
    return 0;
  }
  // 100 seconds for link-state update sequence numbers to turn around
  // half-way
  unsigned short recycleTime = 100;
  return isDupInfo(recv_seqNo, &p->lsDupInfo, LinkState::MAX_SEQNO,
		   recycleTime);
} // End NetworkGraph::isDuplicate()

int NetworkGraph::isDuplicateServInfo(unsigned short origin_scid,
				      unsigned short recv_seqNo) {
  GraphNode_t *p = getNodeWithSCID(origin_scid);
  if(p == NULL) {
    cerr << "isDuplicate: No node with SCID " << origin_scid << endl;
    return 0;
  }
  // 100 seconds for link-state update sequence numbers to turn around
  // half-way
  unsigned short recycleTime = 100;
  return isDupInfo(recv_seqNo, &p->servDupInfo,
		   (unsigned short)LinkState::MAX_SEQNO, recycleTime);
} // End NetworkGraph::isDuplicateServInfo()

unsigned short NetworkGraph::addEdge(unsigned short dest_scid,
				     unsigned short origin_scid,
				     short srtt) {
  // see if the nodes are present
  if(!scidHash.containsKey(origin_scid)) {
    scidHash.put(origin_scid, nodes.size());
    GraphNode_t gNode;
    gNode.scid = origin_scid;
    gNode.lsDupInfo.recdLSPkts = false;
    gNode.servDupInfo.recdLSPkts = false;
    nodes.push_back(gNode);
  }
  if(!scidHash.containsKey(dest_scid)) {
    scidHash.put(dest_scid, nodes.size());
    GraphNode_t gNode;
    gNode.scid = dest_scid;
    nodes.push_back(gNode);
  }

  // add the information under dest_scid
  short destIndex = scidHash.get(dest_scid);
  short srcIndex = scidHash.get(origin_scid);
  vector<GraphEdge_t> &incoming = nodes[destIndex].incoming;
  int done = 0;
  unsigned short realCost = (srtt == -1) ?
    PriorityQueue<unsigned short>::INFINITY : srtt;
  unsigned short oldCost = PriorityQueue<unsigned short>::INFINITY;
  for(unsigned int i = 0; i < incoming.size(); i++) {
    if(incoming[i].srcVertex == srcIndex) {
      oldCost = incoming[i].arcCost;
      incoming[i].arcCost = realCost; done = 1; break;
    }
  }
  if(!done) {
    GraphEdge_t gEdge = { srcIndex, realCost };
    incoming.push_back(gEdge);
  }
  return oldCost;

} // End NetworkGraph::addEdge()

unsigned short NetworkGraph::getEdgeCost(unsigned short dest_scid,
					 unsigned short origin_scid) {
  if(!scidHash.containsKey(dest_scid)) {
    return PriorityQueue<unsigned short>::INFINITY;
  }

  short destIndex = scidHash.get(dest_scid);
  short srcIndex = scidHash.get(origin_scid);
  vector<GraphEdge_t> &incoming = nodes[destIndex].incoming;
  for(unsigned int i = 0; i < incoming.size(); i++) {
    if(incoming[i].srcVertex == srcIndex) {
      return incoming[i].arcCost;
    }
  }

  return PriorityQueue<unsigned short>::INFINITY;
} // End NetworkGraph::getEdgeCost()

vector<Peer_t> &NetworkGraph::getPeerList(unsigned short *numPeers) {
  *numPeers = peers.size();
  return peers;
} // End NetworkGraph::getPeerList()

Peer_t *NetworkGraph::getPeerWithSCID(unsigned short scid) {
  if(peer_scidHash.containsKey(scid)) {
    short nIndex = peer_scidHash.get(scid);
    return &peers[nIndex];
  } else {
    return NULL;
  }
} // End NetworkGraph::getPeerWithSCID()

Peer_t *NetworkGraph::getPeerWithIP(struct in_addr ip) {
  if(peer_ipHash.containsKey(ip.s_addr)) {
    short nIndex = peer_ipHash.get(ip.s_addr);
    return &peers[nIndex];
  } else {
    return NULL;
  }
} // End NetworkGraph::getPeerWithIP()

short NetworkGraph::getNumNodes() {
  return nodes.size();
} // End NetworkGraph::getNodes()

void NetworkGraph::getNodes(TreeNode_t *tVec, short numNodes) {
  for(short i = 0; i < numNodes; i++) {
    tVec[i].scid = nodes[i].scid;
  }
} // End NetworkGraph::getNodes()

GraphNode_t *NetworkGraph::getNodeWithSCID(unsigned short scid) {
  if(scidHash.containsKey(scid)) {
    short nIndex = scidHash.get(scid);
    return &nodes[nIndex];
  } else {
    return NULL;
  }
} // End NetworkGraph::getNodeWithSCID()

const vector<GraphEdge_t> *NetworkGraph::getIncomingEdges(unsigned short scid) {
  if(scidHash.containsKey(scid)) {
    short nIndex = scidHash.get(scid);
    return &nodes[nIndex].incoming;
  } else {
    return NULL;
  }
} // End NetworkGraph::getIncomingEdges()

void NetworkGraph::printGraph() {
  cerr << "-------- Begin Graph --------" << endl;
  int numNodes = nodes.size();
  for(int i = 0; i < numNodes; i++) {
    vector<GraphEdge_t> &eVec = nodes[i].incoming;
    int numEdges = eVec.size();
    for(int j = 0; j < numEdges; j++) {
      cerr << nodes[i].scid << " <-- " << nodes[eVec[j].srcVertex].scid
	   << " (" << eVec[j].arcCost << ")" << endl;
    }
  }
  cerr << "-------- End Graph --------" << endl;
} // End NetworkGraph::printGraph()
