/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
#ifndef NETWORK_GRAPH_H
#define NETWORK_GRAPH_H

#include "LinkState.h"
#include "../serv-comp/TreeInfo.h"
#include "../util/JString.h"

/** Represents a graph edge -- is stored along with a node, which
    represents the destination vertex of the edge */
class GraphEdge {
public:
  short srcVertex;
  unsigned short arcCost;
}; // End class GraphEdge

typedef class GraphEdge GraphEdge_t;

/** Structure with info to detect duplicate messages during flood (of
    service information or link-state updates) */
class DupInfo {
public:
  /** last sequence number received from this peer */
  unsigned short lastLSSeqNo;
  /** not yet received any packets from this peer */
  bool recdLSPkts;
  /** last time a link-state packet was received */
  struct timeval lastLSUpdate;
}; // End class DupInfo

typedef class DupInfo DupInfo_t;

class GraphNode {
public:
  vector<GraphEdge_t> incoming;
  unsigned short scid;
  DupInfo_t lsDupInfo; // to keep track of last link-state update
  DupInfo_t servDupInfo; // to keep track of last service info update
}; // End GraphNode

typedef class GraphNode GraphNode_t;

/** This is the class that maintains the network graph */
class NetworkGraph {
protected:
  /** Read the graph file.  See init() for more details. */
  static void readGraph(const char *graph, bool read_all);

  /** Read the scid_mapping file.  See init() for more details. */
  static void readSCIDMapping(const char *scid_mapping);

public:
  /** The invalid service cluster ID */
  const static unsigned short INVALID_SCID;
  /** The invalid node index */
  const static short INVALID_NODE;
  /** My service cluster ID */
  static unsigned short mySCID;

  /** The nodes in the graph */
  static vector<GraphNode_t> nodes;

  /** The hash of scid to index in 'nodes' array */
  static JHashtable<unsigned short, short> scidHash;

  /** The peers of this node */
  static vector<Peer_t> peers;

  /** The hash of scid to 'peers' index */
  static JHashtable<unsigned short, short> peer_scidHash;
  /** The hash of ip address to 'peers' index */
  static JHashtable<unsigned long, short> peer_ipHash;

  /** This init function has to be called before everything will work.
      The file 'scid_mapping' has info about the location of managers
      in each service cluster (only those of the peers are needed by
      each node).  The format of 'scid_mapping' is, in each line:
      "<scid> <host-name-of-manager>".

      The file 'graph' gives network graph (only the adjacency list of
      self is required).  Format of 'graph' is, in each line:
      "<dest-scid> <origin-scid1> <cost1> <origin-scid2> <cost2>...".
      (The list of incoming edges and costs are given).

      The SCID for this node is also given as argument.

      The flag 'read_all' is set to true during debugging, when the
      entire graph has to be read in to begin with, rather than built
      using the link-state algorithm. */
  static void init(const char *scid_mapping, const char *graph,
		   unsigned short arg_mySCID, bool read_all);

  /** Check if the given sequence number is a repeat, given the info
      in the DupInfo structure, the maximum sequence number, and the
      recycleTime.  The 'recycleTime' is a lower bound on the time
      taken for the sequence number to wrap around half way.  It is
      also a lower bound on the time for a machine to reboot. */
  static int isDupInfo(unsigned short recv_seqNo, DupInfo_t *d,
		       unsigned short MAX_SEQNO,
		       unsigned short recycleTime);

  /** Checks if the given link-state update packet is a duplicate --
      checks for timeout and sequence number.  Return 1 if duplicate,
      0 otherwise. */
  static int isDuplicate(unsigned short origin_scid,
			 unsigned short recv_seqNo);

  /** Checks if the given service information packet is a duplicate --
      checks for timeout and sequence number.  Return 1 if duplicate,
      0 otherwise. */
  static int isDuplicateServInfo(unsigned short origin_scid,
				 unsigned short recv_seqNo);

  /** Add the given edge to the graph (update if edge already
      present).  If the cost is negative, then the link is down.
      NOTE: only one thread should call this function (this not
      enforced though).  Returns the old cost of the edge (INFINITY if
      edge did not exists earlier). */
  static unsigned short addEdge(unsigned short origin_scid,
				unsigned short dest_scid,
				short srtt);

  /** Return the edge cost (could be INFINITY) */
  static unsigned short getEdgeCost(unsigned short dest_scid,
				    unsigned short origin_scid);

  /** Return list of peers as a vector.  Returns the number of peers
      in '*numPeers' -- this number has to be used rather than the
      vector length -- the vector length could increase if the list of
      peers has grown (it cannot decrease though) */
  static vector<Peer_t> &getPeerList(unsigned short *numPeers);

  /** Return the peer with the given SCID.  Returns NULL if there is
      no such peer. */
  static Peer_t *getPeerWithSCID(unsigned short scid);

  /** Return the peer with the given ip address.  Returns NULL if
      there is no such peer. */
  static Peer_t *getPeerWithIP(struct in_addr ip);

  /** Return the number of nodes in the graph currently.  NOTE: this
      number could increase as more information comes in (but this
      number is guaranteed not to decrease) */
  static short getNumNodes();

  /** Return the list of scid information of nodes in the given
      TreeInfo array.  Need to return 'numNodes' elements (this is
      assumed to be from an earlier call to getNumNodes() */
  static void getNodes(TreeNode_t *tVec, short numNodes);

  /** Return the node with the given SCID.  Return NULL on failure. */
  static GraphNode_t *getNodeWithSCID(unsigned short scid);

  /** Return the set of incoming edges on a given SCID.  Return NULL
      on failure. */
  static const vector<GraphEdge_t> *getIncomingEdges(unsigned short scid);

  /** Print the graph onto stderr */
  static void printGraph();

}; // End class NetworkGraph

#endif
