/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
#ifndef PEER_H
#define PEER_H

#include <queue>
#include <list>
#include <netinet/in.h>
#include "../perf-meas/LatMeas.h"
#include "../perf-meas/HB.h"
#include "../udp-lib/ReqRespBlock.h"

struct ackInfo {
  int reqNo;
  UDPLibPkt_t *uPkt;
  struct in_addr toReplyAddr;
  int toReplyPort;
  int pathID;
  unsigned int path_dest_scid;
  // 'version' is used only while communicating with a local service,
  // not while communicating with a peer cluster manager
  unsigned short version;
  bool to_app; // is this ack going back to the app?
  ackInfo() { version = 0; to_app = false; }
}; // End struct ackInfo
typedef struct ackInfo ackInfo_t;

typedef struct active_ackInfo {
  bool active; // is entry active?
  ackInfo_t aInfo;
  active_ackInfo() { active = false; }
} active_ackInfo_t; // End struct active_ackInfo

struct repair_ackInfo {
  int reqNo;
  UDPLibPkt_t *uPkt;
  struct in_addr toReplyAddr;
  int toReplyPort;
  int pathID;
  unsigned int path_dest_scid;
  unsigned short fail_dest_scid;
  unsigned short fail_origin_scid;
  short posn;
  bool local; // indicate if this is a local repair
}; // End struct repair_ackInfo
typedef struct repair_ackInfo repair_ackInfo_t;

typedef struct active_repair_ackInfo {
  bool active; // is entry active?
  repair_ackInfo_t rep_aInfo;
  active_repair_ackInfo() { active = false; }
} active_repair_ackInfo_t;

typedef struct {
  int pathID;
  unsigned short path_dest_scid;
} path_id_t;

/** Class for managing info about a peer */
class Peer {
public:
  /** Should I use a queue of request numbers, or a vector indexed on
      the reqNo? */
  static bool use_queue;

  /** The queue of request numbers for sends */
  static queue<int, list<int> > send_reqNoQueue;
  static vector<bool> send_reqNo_vec;
  static unsigned int send_lastRet;
  static vector<int> send_vec;
  /** The queue of request numbers for forwards */
  static queue<int, list<int> > fwd_reqNoQueue;
  static vector<bool> fwd_reqNo_vec;
  static unsigned int fwd_lastRet;
  static vector<int> fwd_vec;

  /** The queue of request numbers for serv-comp path heart-beats */
  static queue<int, list<int> > sc_hb_reqNoQueue;
  /** Similar to sc_serv_reqNo_vec */
  static vector<bool> sc_hb_reqNo_vec;
  /** Similar to sc_serv_lastRet */
  static unsigned int sc_hb_lastRet;
  /** Similar to sc_serv_vec */
  static vector<int> sc_hb_vec;

  /** The queue of request numbers for service information "sends" */
  queue<int, list<int> > send_reqNoQueueServInfo;
  vector<bool> send_reqNoServInfo_vec;
  unsigned int send_ServInfo_lastRet;
  vector<int> send_ServInfo_vec;
  /** The queue of request numbers for service information "forwards" */
  queue<int, list<int> > fwd_reqNoQueueServInfo;
  vector<bool> fwd_reqNoServInfo_vec;
  unsigned int fwd_ServInfo_lastRet;
  vector<int> fwd_ServInfo_vec;

  /** The queue of request numbers for service composition messages.
      Also have to remember the udp-lib packet, and the address/port
      to which ack has to be sent */
  static queue<ackInfo_t, list<ackInfo_t> > sc_reqNoQueue;
  static vector<active_ackInfo_t> sc_reqNo_vec;
  static unsigned int sc_lastRet;
  static vector<ackInfo_t> sc_vec;

  /** The queue of request numbers for service composition messages
      related to patching paths.  Also have to remember the udp-lib
      packet, and the address/port to which ack has to be sent */
  static queue<repair_ackInfo_t, list<repair_ackInfo_t> > sc_repair_reqNoQueue;
  static vector<active_repair_ackInfo_t> sc_repair_reqNo_vec;
  static unsigned int sc_repair_lastRet;
  static vector<repair_ackInfo_t> sc_repair_vec;

  /** The queue of request numbers for service composition messages
      related to end-2-end recovery.  Also have to remember the
      udp-lib packet, and the address/port to which ack has to be sent */
  static queue<ackInfo_t, list<ackInfo_t> > sc_alt_reqNoQueue;
  static vector<active_ackInfo_t> sc_alt_reqNo_vec;
  static unsigned int sc_alt_lastRet;
  static vector<ackInfo_t> sc_alt_vec;

  /** The queue of request numbers for service composition messages
      related to communication with a local service. */
  static queue<ackInfo_t, list<ackInfo_t> > sc_serv_reqNoQueue;
  /** This is an alternate structure for maintaining the set of active
      request numbers -- this is used IFF sc_serv_reqNoQueue is not
      used.  This vector is indexed on the reqNo itself. */
  static vector<active_ackInfo_t> sc_serv_reqNo_vec;
  /** To maintain a queue-like structure to the sc_serv_reqNo_vec
      vector.  This is the position in the vector, of the last
      returned reqNo, in the call to popReqNos, or vec_pop */
  static unsigned int sc_serv_lastRet;
  /** This is for returning a bunch of active ackInfo structures */
  static vector<ackInfo_t> sc_serv_vec;

  /** The list of paths that pass through this peer */
  vector<path_id_t> path_list;

  /** For handling 'send' or 'fwd' addReq */
  static void addReq(int reqNo, queue<int, list<int> > *reqNoQueue,
		     vector<bool> *reqNo_vec);
  /** For handling 'send' or 'fwd' popReqNos */
  static int popReqNos(int *reqNos, int maxPop,
		       queue<int, list<int> > *reqNoQueue,
		       vector<bool> *reqNo_vec, unsigned int *lastRet);
  /** Similar to 'sc_vec_pop' */
  static int vec_pop(queue<int, list<int> > *reqNoQueue,
		     vector<bool> *reqNo_vec, unsigned int *lastRet,
		     vector<int> *vec);

  /** For handling operations related to ackInfo_t */
  static void a_addReq(ackInfo_t &aInfo,
		       queue<ackInfo_t, list<ackInfo_t> > *reqNoQueue,
		       vector<active_ackInfo_t> *reqNo_vec);
  static int a_popReqNos(ackInfo_t *aVec, int maxPop,
			 queue<ackInfo_t, list<ackInfo_t> > *reqNoQueue,
			 vector<active_ackInfo_t> *reqNo_vec,
			 unsigned int *lastRet);
  static int a_vec_pop(queue<ackInfo_t, list<ackInfo_t> > *reqNoQueue,
		       vector<active_ackInfo_t> *reqNo_vec,
		       unsigned int *lastRet, vector<ackInfo_t> *vec);

  /** For handling operations related to repair_ackInfo_t */
  static void a_rep_addReq(repair_ackInfo_t &rep_aInfo,
			   queue<repair_ackInfo_t, list<repair_ackInfo_t> > *reqNoQueue,
			   vector<active_repair_ackInfo_t> *reqNo_vec);
  static int a_rep_popReqNos(repair_ackInfo_t *aVec, int maxPop,
			     queue<repair_ackInfo_t, list<repair_ackInfo_t> > *reqNoQueue,
			     vector<active_repair_ackInfo_t> *reqNo_vec,
			     unsigned int *lastRet);
  static int a_rep_vec_pop(queue<repair_ackInfo_t, list<repair_ackInfo_t> > *reqNoQueue,
			   vector<active_repair_ackInfo_t> *reqNo_vec,
			   unsigned int *lastRet, vector<repair_ackInfo_t> *vec);

public:
  /** The queue of paths to be patched */
  queue<path_id_t, list<path_id_t> > patchQ;
  /** The last time paths from this peer were patched */
  struct timeval last_patch;

  /** The ip address of the peer */
  struct in_addr ip_addr;
  /** The service cluster ID of the peer */
  unsigned short scid;

  /** The latency measurement object associated with this peer */
  LatMeas_t *latMeas;
  /** The heart-beat object associated with this peer */
  HB_t *hb;

  /** The time the last measurement "of" the path to this peer was
      sent over the network */
  // ARPANET-algo struct timeval lastSendTime;
  /** The last RTT that was sent over the network */
  short lastRTTSent;
  /** The current 'threshold' for the latency of this path -- this is
      from the original ARPANET algorithm */
  // ARPANET-algo unsigned short threshold;

  /** The last 'status' that was noted by the SC layer */
  short lastStatus;
  /** The last 'my_status_peer_view' that was noted by the SC layer */
  short lastMyStatusPeerView;

  Peer() {
    lastMyStatusPeerView = lastStatus = LIVE;
    gettimeofday(&last_patch, NULL);
    send_ServInfo_lastRet = fwd_ServInfo_lastRet = 0;
  } // End Peer

  /** Return an appropriate timeout value for responses from this
      particular peer */
  struct timeval getTimeout() {
    struct timeval toRet = { 6, 0 }; // 6 seconds
    return toRet;
  } // End Peer::getTimeout()

  /** Return an appropriate timeout value for responses from this
      particular peer - for service information sends/forwards */
  struct timeval getServInfoTimeout() {
    struct timeval toRet = { 10, 0 }; // 10 seconds
    return toRet;
  } // End Peer::getServInfoTimeout()

  /** Return an appropriate timeout value for responses from this
      particular peer - for service composition related messages */
  struct timeval getServCompTimeout() {
    struct timeval toRet = { 6, 0 }; // 6 seconds
    return toRet;
  } // End Peer::getServCompTimeout()

  //---- Routines related to LS 'send' ----//
  /** Add this particular request number to the set of 'send' requests
      that has been sent to this peer, for which an ack is awaited.
      NOTE: the ReqRespBlock for which this reqNo is valid is to be
      maintained by the caller */
  static void send_addReq(int reqNo);
  /** Pop some 'send' request numbers added earlier -- note, this is
      "pop", meaning the request numbers will be removed from the
      stored list.  Return the number of request numbers popped.  The
      max number of reqNos that can be popped is given as argument. */
  static int send_popReqNos(int *reqNos, int maxPop);
  /** Similar to sc_vec_pop() */
  static int send_vec_pop();

  //---- Routines related to LS 'fwd' ----//
  /** Same as 'send_addReq' but for forwards */
  static void fwd_addReq(int reqNo);
  /** Same as 'send_popReqNos' but for forwards */
  static int fwd_popReqNos(int *reqNos, int maxPop);
  /** Similar to sc_vec_pop() */
  static int fwd_vec_pop();

  //---- Routines related to SC heart-beats ----//
  /** Same as 'send_addReq' but for serv-comp path heart-beats */
  static void sc_hb_addReq(int reqNo);
  /** Same as 'send_popReqNos' but for serv-comp path heart-beats */
  static int sc_hb_popReqNos(int *reqNos, int maxPop);
  /** Similar to sc_vec_pop() */
  static int sc_hb_vec_pop();

  //---- Routines related to service-locn 'send' ----//
  /** Same as 'send_addReq' but for service information "sends" */
  void send_addServInfoReq(int reqNo);
  /** Same as 'send_popReqNos' but for service info "sends" */
  int send_popReqNosServInfo(int *reqNos, int maxPop);

  //---- Routines related to service-locn 'fwd' ----//
  /** Same as 'send_addServInfoReq' but for service information "forwards" */
  void fwd_addServInfoReq(int reqNo);
  /** Same as 'send_popReqNosServInfo' but for service info "forwards" */
  int fwd_popReqNosServInfo(int *reqNos, int maxPop);

  //---- Routines related to SC messages ----//
  /** Same as 'send_addReq', but for service composition messages */
  static void sc_addReq(int reqNo, UDPLibPkt_t *uPkt,
			struct in_addr toReplyAddr, int toReplyPort,
			int pathID, unsigned int path_dest_scid,
			bool to_app = false);
  /** Same as 'send_popReqNos' but for service composition messages */
  static int sc_popReqNos(ackInfo_t *aVec, int maxPop);
  /** Alternate method of accessing the queue of request numbers --
      put it into the Peer::appropriate_vector -- this handles an
      arbitrary number of request numbers */
  static int sc_vec_pop();

  //---- Routines related to SC repair messages ----//
  /** Same as 'sc_addReq', but for service composition messages
      related to patching paths */
  static void sc_repair_addReq(int reqNo, UDPLibPkt_t *uPkt,
			       struct in_addr toReplyAddr,
			       int toReplyPort, int pathID,
			       unsigned int path_dest_scid,
			       unsigned short fail_dest_scid,
			       unsigned short fail_origin_scid,
			       short posn, bool local);
  /** Same as 'sc_popReqNos' but for service composition messages
      related to patching paths */
  static int sc_repair_popReqNos(repair_ackInfo_t *aVec, int maxPop);
  /** Similar to sc_vec_pop() */
  static int sc_repair_vec_pop();

  //---- Routines related to SC alt-path messages ----//
  /** Same as 'sc_addReq', but for service composition messages
      related to end-2-end recovery */
  static void sc_alt_addReq(int reqNo, UDPLibPkt_t *uPkt,
			    struct in_addr toReplyAddr, int toReplyPort,
			    int pathID, unsigned int path_dest_scid);
  /** Same as 'sc_popReqNos' but for service composition messages
      related to end-2-end recovery */
  static int sc_alt_popReqNos(ackInfo_t *aVec, int maxPop);
  /** Similar to sc_vec_pop() */
  static int sc_alt_vec_pop();

  /*---- Routines related to SC comm with service instances.  NOTE:
    these are not associated with any particular peer -- these
    routines are here since all async request queues are here anyway.
    ----*/
  /** Same as 'sc_addReq', but for service composition messages
      related to communication with local service. */
  static void sc_serv_addReq(int reqNo, UDPLibPkt_t *uPkt,
			     struct in_addr toReplyAddr, int toReplyPort,
			     int pathID, unsigned int path_dest_scid,
			     unsigned short version, bool to_app);
  /** Same as 'sc_popReqNos' but for service composition messages
      related to communication with local service */
  static int sc_serv_popReqNos(ackInfo_t *aVec, int maxPop);
  /** Similar to sc_vec_pop() */
  static int sc_serv_vec_pop();

  /** Add to the list of paths that pass through this edge */
  void addPath(int pathID, unsigned short path_dest_scid);
  /** Delete from the list of paths that pass through this edge */
  void delPath(int pathID, unsigned short path_dest_scid);

  /** Return the list of paths that pass through this edge */
  vector<path_id_t> *getPathList() { return &path_list; }

}; // End class Peer

typedef class Peer Peer_t;

#endif
