/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Program to test the link-state algorithm */

#include <unistd.h>
#include <stdio.h>

#include "NetworkGraph.h"
#include "LinkState.h"
#include "../perf-meas/HB.h"
#include "../perf-meas/LatMeas.h"
#include "../udp-lib/timeSync.h"

void usage(const char *progname) {
  cerr << "Usage: " << progname
       << " <scid-mapping-file> <graph-file> <mySCID>" << endl;
} // End usage()

int main(int argc, char *argv[]) {
  util_init();
  if(argc != 4) { usage(argv[0]); exit(1); }

  // Init the udp-lib
  UDPLib::init();

  // Init the NetworkGraph
  int mySCID = atoi(argv[3]);
  bool read_all = false;
  NetworkGraph::init(argv[1], argv[2], (unsigned short)mySCID, read_all);

  // Init the debug object
  struct timeval startTime;
  struct in_addr ipAddr;
  ipAddr.s_addr = inet_addr(timeServer);
  bool res = getStartTimeAtHost(&startTime, ipAddr);
  if(!res) {
    cerr << "Can't get time from server " << timeServer << endl;
    exit(11);
  }
  char ls_send_logFile[128];
  sprintf(ls_send_logFile, "ls-send-log.%d", mySCID);
  Debug_t ls_send_dbg(ls_send_logFile, mySCID, startTime, cfg_to_file,
		      cfg_use_stream);

  char ls_listen_logFile[128];
  sprintf(ls_listen_logFile, "ls-listen-log.%d", mySCID);
  Debug_t ls_listen_dbg(ls_listen_logFile, mySCID, startTime,
			cfg_to_file, cfg_use_stream);

  // Start the listen threads for HB and LatMeas
  bool run_listen_thread = true, run_send_thread = true;
  HB::startListenThread(run_listen_thread);
  LatMeas::startListenThread(run_listen_thread);

  // Allocate the HB and LatMeas threads for each peer
  // Also start the send threads on each of those
  unsigned short numPeers;
  vector<Peer_t> &peer_list = NetworkGraph::getPeerList(&numPeers);
  for(int i = 0; i < numPeers; i++) {
    peer_list[i].hb = new HB(peer_list[i].ip_addr, peer_list[i].scid);
    peer_list[i].latMeas =
      new LatMeas(peer_list[i].ip_addr, peer_list[i].scid);
  }
  HB::start_all_sends(run_send_thread, mySCID);
  LatMeas::start_all_sends(run_send_thread, mySCID);

  // LinkState stuff
  bool run_ls_listen_thread = true, run_ls_send_thread = true;
  if(!LinkState::init(&ls_send_dbg, &ls_listen_dbg)) {
    cerr << "LinkState::init() failed" << endl;
    exit(100);
  }
  LinkState::startListenThread(run_ls_listen_thread);

  /** Wait for peers to start up before starting to send LS updates */
  int sleep_time = 6;
  fprintf(stderr, "Sleeping for %d sec before LS send\n", sleep_time);
  sleep(sleep_time);

  LinkState::startSendThread(run_ls_send_thread);

  struct timeval last_print_time;
  gettimeofday(&last_print_time, NULL);

  // loop forever now
  for(;;) {

    //---- Listen stuff ----//
    fd_set rfds; FD_ZERO(&rfds);
    int max_fd = -1;

    if(!HB::run_listen_thread) {
      FD_SET(HB::listen_fd, &rfds);
      if(HB::listen_fd > max_fd) { max_fd = HB::listen_fd; }
    }

    if(!LatMeas::run_listen_thread) {
      FD_SET(LatMeas::listen_fd, &rfds);
      if(LatMeas::listen_fd > max_fd) { max_fd = LatMeas::listen_fd; }
    }

    if(!LinkState::run_listen_thread) {
      FD_SET(LinkState::listen_fd, &rfds);
      if(LinkState::listen_fd > max_fd) { max_fd = LinkState::listen_fd; }
      FD_SET(LinkState::f_ack_listen_fd, &rfds);
      if(LinkState::f_ack_listen_fd > max_fd) { max_fd = LinkState::f_ack_listen_fd; }
    }

    if(!LinkState::run_send_thread || LinkState::stop_send) {
      FD_SET(LinkState::s_ack_listen_fd, &rfds);
      if(LinkState::s_ack_listen_fd > max_fd) { max_fd = LinkState::s_ack_listen_fd; }
    }

    if(max_fd != -1) {
      struct timeval timeout = { 0, 50*1000 }; // 100ms
      select(max_fd+1, &rfds, NULL, NULL, &timeout);

      // call appropriate routines based on which fd recd packet
      if(!HB::run_listen_thread && FD_ISSET(HB::listen_fd, &rfds)) {
	HB::recvHelper();
      }
      if(!LatMeas::run_listen_thread && FD_ISSET(LatMeas::listen_fd, &rfds)) {
	LatMeas::recvHelper();
      }
      if(!LinkState::run_listen_thread &&
	 FD_ISSET(LinkState::listen_fd, &rfds)) {
	bool pkt_ready = true; LinkState::listenHelper(pkt_ready);
      }
      if(!LinkState::run_listen_thread &&
	 FD_ISSET(LinkState::f_ack_listen_fd, &rfds)) {
	LinkState::recvAck(LS_FWD);
      }
      if((!LinkState::run_send_thread || LinkState::stop_send) &&
	 FD_ISSET(LinkState::s_ack_listen_fd, &rfds)) {
	LinkState::recvAck(LS_SEND);
      }

      // end case when i have to listen for something
    } else {

      usleep(50*1000);
      // end case when i don't have to listen to anything
    }

    //---- Send stuff ----//
    if(!HB::run_send_thread) { HB::all_checkSend(); }
    if(!LatMeas::run_send_thread) { LatMeas::all_checkSend(); }
    if(!LinkState::run_send_thread) {
      LinkState::sendHelper();
      // this checks for retransmits
      LinkState::checkAcks(LS_SEND, LinkState::run_send_thread);
    }
    if(!LinkState::run_listen_thread) {
      // this checks for retransmits
      LinkState::checkAcks(LS_FWD, LinkState::run_listen_thread);
    }

    // print the graph
    if(now(last_print_time) >= 5000) {
      NetworkGraph::printGraph();
      gettimeofday(&last_print_time, NULL);
    }
  }

} // End main()
