/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Implements PktDrop.h */

#include <iostream>
#include <fstream>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#include "PktDrop.h"
#include "../util/Util.h"

PktDrop::PktDrop(const char *distFileName, unsigned short a_dest_scid,
		 unsigned short a_origin_scid, Debug_t *arg_dbg) {
  dbg = arg_dbg;
  dest_scid = a_dest_scid; origin_scid = a_origin_scid;

  // first count the number of lines
  ifstream fin(distFileName);
  numEntries = 0;
  for(;;) {
    // read each line
    if(fin.eof()) break;
    char line[1024];
    fin.getline(line, sizeof(line)); // read a line
    if(fin.eof()) break;
    if(fin.fail()) {
      cerr << "Error reading " << distFileName << endl;
      exit(99);
    }
    numEntries++;
  } // End for(;;) reading lines of file first time
  fin.close();

  // now allocate for the CDF storage
  CDF = new Entry_t[numEntries];
  if(CDF == NULL) {
    cerr << "Cannot alloc for CDF" << endl;
    exit(100);
  }

  // now read the file again
  ifstream fin1(distFileName);
  float cumProb = 0;
  int entryNum = 0;

  for(;;) {
    // read each line
    if(fin1.eof()) break;
    char line[1024];
    fin1.getline(line, sizeof(line)); // read a line
    if(fin1.eof()) break;
    if(fin1.fail()) {
      cerr << "Error reading " << distFileName << endl;
      exit(99);
    }

    // read in the params
    unsigned long t;
    float freq, p;
    int numRead = sscanf(line, "%ld %f %f", &t, &freq, &p);
    if(numRead != 3) {
      cerr << "Format error in " << distFileName << endl;
      exit(101);
    }

    // make entry in CDF
    cumProb += p;
    CDF[entryNum].t = t;
    // hack for testing - reduce by a factor of 10!
    // CDF[entryNum].t /= 10;
    CDF[entryNum].F = cumProb;
    entryNum++;
  } // End for(;;) reading lines of file second time
  fin1.close();

  // ensure that the last entry is correct
  if(fabs(CDF[numEntries-1].F - 1) >= 0.00005) {
    cerr << "Error: last CDF value differs from 1: "
	 << CDF[numEntries-1].F << endl;
    exit(127);
  }
  CDF[numEntries-1].F = 1;

  // set random seed
  struct timeval tv;
  gettimeofday(&tv, NULL);
  unsigned long seedLong = tv.tv_usec%211;
  float seedFloat = ((seedLong+1)*1.0)/212;
  fprintf(stderr, "setting random seed done with %ld(%f)\n",
	  seedLong, seedFloat);
  Uniform::Set(seedFloat);

  // init the current gap, and its start time
  currGap = 0;
  gettimeofday(&currGapStart, NULL);

  // log
  //char valStr[1024]; char *lastPos = valStr; *lastPos = '['; lastPos++;
  //lastPos = my_strcpy(lastPos, distFileName); *lastPos = ']'; lastPos++;
  //*lastPos = 0;
  JString valStr = "[" + JString(distFileName) + "]";
  if(dbg->use_stream) {
    dbg->print() << "pkt-drop dest_scid=" << dest_scid
		 << " origin_scid=" << origin_scid << " file="
		 << valStr << " seed=" << seedLong << endl;
  } else {
    dbg->setCurrFn(PKTDROP_FN);
    LogLine_t ll; ll.desc = pkt_drop_desc;
    ll.dest_scid = dest_scid; ll.origin_scid = origin_scid;
    ll.addPair(FILE_LK, valStr); ll.utype = EMUL_U; ll.u.emul.init();
    ll.u.emul.seedLong = seedLong;
    dbg->addLL(ll);
  }

} // End PktDrop::PktDrop()

PktDrop::~PktDrop() {
  delete[] CDF;
} // End PktDrop::~PktDrop()

void PktDrop::print() {
  cerr << dest_scid << "-" << origin_scid << " currGap=" << currGap
       << " currGapStart=[" << currGapStart.tv_sec << ":"
       << currGapStart.tv_usec << "]" << endl;
  for(int i = 0; i < numEntries; i++) {
    cerr << "\tt=" << CDF[i].t << "\tF=" << CDF[i].F << endl;
  }

} // End PktDrop::print()

unsigned long PktDrop::CDF_inv(float y) {
  for(int i = 0; i < numEntries; i++) {
    if(CDF[i].F >= y) { return CDF[i].t; }
  }
  return CDF[numEntries-1].t;
} // End PktDrop::CDF_inv()

bool PktDrop::dropNow() {
  // this routine is called when a pkt arrives, to decide whether or
  // not to drop it

  // check if within current gap period
  float timeSinceCurrGapStart = now(currGapStart);
  if(timeSinceCurrGapStart < currGap) {
    return true;
  }

  // have to generate new gap period
  float unifRand = U.Next();
  unsigned long randGap = CDF_inv(unifRand);

  // hack! have to account for UDP/Ping distributions separately
  unsigned long realRandGap;
  unsigned long distQuanta = CDF[1].t - CDF[0].t; // should be 300ms
  // for UDP distributions, and 1000ms for ping distributions -- using
  // distQuanta, instead of hardcoded 300ms, since that would allow us
  // to 'scale' time and test this program (only have to change
  // CDF[i].t when reading it in
  if(randGap <= 2*distQuanta) {
    realRandGap = 0;
  } else if((randGap%distQuanta) == 0) {
    // this is a UDP distribution
    // should implement 300ms loss period for gaps seen above 600, but
    // below 900, 600ms loss period for 900 <= gap < 1200, 900ms for
    // 1200 <= gap 1500, and so on.
    realRandGap = randGap-2*distQuanta;
  } else {
    // this is a ping distribution
    // should implement 1000ms loss period for single loss run, 2000ms
    // loss period for loss run of length two, etc.
    realRandGap = randGap;
  }

  if(realRandGap == 0) {
    currGap = 0;
    gettimeofday(&currGapStart, NULL);
    // can let pkt to go through
    return false;
  } else {
    // need to decide 'where' to place the gap window w.r.t. the pkt
    // arrival time (which is 'now')
    unsigned long gapInPast = (unsigned long)(U.Next()*realRandGap);
    unsigned long gapInFuture = realRandGap-gapInPast;
    // ensure that we don't overlap with the previous gap period
    if(timeSinceCurrGapStart < (currGap + gapInPast)) {
      // we do overlap ==> adjust gapInPast and gapInFuture
      gapInPast = (unsigned long)(timeSinceCurrGapStart)-currGap;
      gapInFuture = realRandGap-gapInPast;
    }
    currGap = realRandGap;
    currGapStart = past(gapInPast);
    dbg->setCurrFn(PKTDROP_FN);
    if(dbg->use_stream) {
      dbg->print() << "drop-start dest_scid=" << dest_scid
		   << " origin_scid=" << origin_scid << " gap="
		   << gapInFuture << endl;
    } else {
      LogLine_t ll; ll.desc = drop_start_desc;
      ll.dest_scid = dest_scid; ll.origin_scid = origin_scid;
      ll.utype = EMUL_U; ll.u.emul.init();
      ll.u.emul.gapInFuture = gapInFuture; dbg->addLL(ll);
    }
    // now drop packet
    return true;
  }

} // End PktDrop::dropNow()
