/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Program to forward Raw IP packets along an IP tunnel */

#ifdef IP_TUNNEL

#include <iostream>
#include <fstream>
#include <vector>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>

#include "fwdRaw.h"
#include "../udp-lib/UDPLibPkt.h"
#include "../udp-lib/timeSync.h"
#include "../util/newran02/include.h"
#include "../util/newran02/newran.h"
#include "../util/Util.h"
#include "../util/Debug.h"

#ifdef MEAS_FWDRAW
#include "measFwdRaw.h"
#endif

void usage(char *progname) {
#ifdef MEAS_FWDRAW
  fprintf(stderr, "Usage: %s <scid-mapping-file> <graph-config-file> "
	  "<output-file>\n", progname);
#else
  fprintf(stderr, "Usage: %s <scid-mapping-file> <graph-config-file>\n",
	  progname);
#endif
} // End usage()

//-------------------- File reading routines --------------------//

Emulator_t *emul;
graph_t *gr;
unsigned long numRecd = 0;
unsigned long numSent = 0;
unsigned long numDropped = 0;

unsigned long pktDist[16];

void printPktDist() {
  ofstream fout("/tmp/pkt-dist.out");
  if(fout.fail()) {
    cerr << "Could not write pkt distribution to /tmp/pkt-dist.out\n";
    return;
  }

  unsigned long cumDist[16];
  cumDist[0] = pktDist[0];
  for(int i = 1; i < 16; i++) {
    cumDist[i] = cumDist[i-1] + pktDist[i];
  }

  fout << "0\t0\t0" << endl;
  for(int i = 0; i < 16; i++) {
    float perc = (cumDist[15] > 0) ? 100.0*cumDist[i]/cumDist[15] : 100;
    fout << 100*(i+1)-1 << '\t' << perc << '\t' << cumDist[i] << endl;
  }

  fout.close();

  cerr << "Pkt distribution written to /tmp/pkt-dist.out\n";
} // End printPktDist()

void readDet(bool firstTime) {
  const char detFile[] = "det-fail.cfg";

  fprintf(stderr, "\nReading file %s\n", detFile);

  for(int i = 0; i < gr->numEdges; i++) {
    emul->links[i].det->startAgain();
  }

  cerr << endl;
  cerr << "numRecd=" << numRecd << ", numSent=" << numSent
       << ", numDropped=" << numDropped << endl;
  cerr << "numSent+numDropped=" << numSent+numDropped << endl;
  cerr << endl;

  ifstream p_fin(detFile);
  if(p_fin.fail()) {
    if(!firstTime) {
      cerr << "Can't open detFile: " << detFile << " -- exiting" << endl;
      printPktDist();
      emul->dbg->close();
      exit(0);
    } else {
      cerr << "Can't open detFile: " << detFile << " -- ignoring" << endl;
      return;
    }
  }

  for(;;) {

    // read each line
    if(p_fin.eof()) break;
    char line[512];
    p_fin.getline(line, sizeof(line)); // read a line
    if(p_fin.eof()) break;
    if(p_fin.fail()) {
      cerr << "Error reading " << detFile << endl;
      exit(99);
    }

    // line has the format: "dest_scid src_scid gap_start gap_duration"
    vector<JString> sVec;
    int numFields = split(line, sVec);
    if(numFields != 4) {
      cerr << "Format of " << detFile << " is wrong" << endl;
      exit(97);
    }

    unsigned short dest_scid = JString_atoi(sVec[0]);
    unsigned short src_scid = JString_atoi(sVec[1]);
    unsigned long gap_start = JString_atoi(sVec[2]);
    unsigned long gap_duration = JString_atoi(sVec[3]);

    // check for validity of entry
    if(!gr->scidHash.containsKey(dest_scid)) {
      cerr << "Unknown SCID " << dest_scid << " in " << detFile << endl;
      continue;
    }
    if(!gr->scidHash.containsKey(src_scid)) {
      cerr << "Unknown SCID " << dest_scid << " in " << detFile << endl;
      continue;
    }

    //JString destStr = ip2str(gr->scidHash.get(dest_scid));
    //JString srcStr = ip2str(gr->scidHash.get(src_scid));
    //JString key = srcStr + ":" + destStr;
    unsigned long key = 100*1000*src_scid + dest_scid;

    if(!emul->linkHash->containsKey(key)) {
      cerr << "no rule for " << key << endl;
      continue;
    }
    int lcIndex = emul->linkHash->get(key);

    gap_t g1 = { gap_start, gap_duration };
    emul->links[lcIndex].det->addRule(g1);

  } // End for(;;) reading file

  p_fin.close();

} // End readDet()

void readGraph(const char *graphFile, graph_t *g) {
  cerr << "reading graph file " << graphFile << endl;
  ifstream g_fin(graphFile);
  g->maxCost = 0;
  g->numEdges = 0;
  for(;;) {
    // read each line
    if(g_fin.eof()) break;
    char line[8192*2];
    g_fin.getline(line, sizeof(line)); // read a line
    if(g_fin.eof()) break;
    if(g_fin.fail()) {
      cerr << "Error reading " << graphFile << endl;
      cerr << line << endl;
      exit(99);
    }

    // line has format:
    // "<dest-scid> <origin-scid1> <cost1> <origin-scid2> <cost2>..."
    vector<JString> sVec;
    int numFields = split(line, sVec);
    if(numFields <= 0) {
      cerr << "Format of " << graphFile << " is wrong" << endl;
      exit(97);
    }
    unsigned short dest_scid = JString_atoi(sVec[0]);
    if(dest_scid == 0) {
      cerr << "dest_scid=" << dest_scid << " has to be > 0" << endl;
      cerr << "sVec[0]=[" << sVec[0] << "]" << endl;
      exit(99);
    }

    if(((numFields-1)%2) != 0) {
      cerr << "odd number of fields for scid " << dest_scid
	   << " in " << graphFile << endl;
      exit(98);
    }
    int numEdges = (numFields-1)/2;

    node_t n;
    n.scid = dest_scid;

    for(int i = 0; i < numEdges; i++) {
      unsigned short origin_scid = JString_atoi(sVec[1+i*2]);
      if(origin_scid == 0) {
	cerr << "origin_scid=" << origin_scid << " has to be > 0" << endl;
	cerr << "scid-str=[" << sVec[1+i*2] << ']' << endl;
	exit(99);
      }
      short cost = JString_atoi(sVec[1+i*2+1]);
      edge_t e;
      e.origin_scid = origin_scid;
      e.cost = cost;
      n.incoming.push_back(e);

      // update graph's maxCost
      if(g->maxCost < cost) { g->maxCost = cost; }
    } // End for(i) reading all edges  

    // update graph's numEdges
    g->numEdges += numEdges;

    // add the node to the graph
    g->nHash.put(n.scid, g->nodes.size());
    g->nodes.push_back(n);

  } // End for(;;) reading all lines in 'graphFile'
  g_fin.close();
} // End readGraph()

void readSCIDMapping(const char *scid_mapping,
		     JHashtable<unsigned short, struct in_addr> *scidHash) {
  vector<sc_node_t> sc1;
  util_readSCIDMapping(scid_mapping, sc1);
  for(unsigned int i = 0; i < sc1.size(); i++) {
    // put the mapping onto the hashtable
    scidHash->put(sc1[i].scid, sc1[i].ipAddr);
  }
} // End readSCIDMapping()

//-------------------- Emulator class routines --------------------//

Emulator::Emulator(graph_t *g, Debug_t *arg_dbg) {
  dbg = arg_dbg;

  // allocate for set of link characteristics
  links = new LinkCharacteristics_t[g->numEdges];
  if(links == NULL) {
    cerr << "Emulator::Emulator: can't allocate for links" << endl;
    exit(103);
  }
  linkHash = new JHashtable<unsigned long, int>(g->numEdges);
  if(linkHash == NULL) {
    cerr << "Emulator::Emulator: can't alloc for linkHash" << endl;
    exit(103);
  }
  //linkHash->setHashFunction(JStringHash);
  linkHash->setHashFunction(ulongHash);

  // init the packet buffer
  numPkts = g->numEdges*5; // 5 pkts per overlay link
#ifdef MEAS_FWDRAW
  numPkts = 40000;
#else
  numPkts = 40000; // use that much buffer space anyway
#endif
  pktBuf = new EmulPktBuf_t[numPkts];
  if(pktBuf == NULL) {
    cerr << "Emulator::Emulator: can't alloc for pktBuf" << endl;
    exit(103);
  }

  // init the priority queue
  pktQ = new PriorityQueue<unsigned long>(numPkts);
  if(pktQ == NULL) {
    cerr << "Emulator::Emulator: can't alloc for pktQ" << endl;
    exit(103);
  }
  gettimeofday(&startTime, NULL);

  // init the free lists
  startFree = 0;
  for(int i = 0; i < numPkts-1; i++) {
    pktBuf[i].nextFreeIndex = i+1;
    pktBuf[i].allocLen = 0;
  }
  pktBuf[numPkts-1].nextFreeIndex = -1; // signal end of list
  Step1 = 100; Step2 = 200; // these are the "steps" for storage of pkts
  startSubStep1 = startSubStep2 = -1; // both are empty lists to start with

  // init the uniform random number generator for generating random
  // run lengths (for GenLat), and also for generating a random
  // density file (for PktDrop)
  Uniform U;
  // set random seed
  struct timeval tv;
  gettimeofday(&tv, NULL);
  unsigned long seedLong = tv.tv_usec%211;
  float seedFloat = ((seedLong+1)*1.0)/212;
  fprintf(stderr, "setting random seed done with %ld(%f)\n",
	  seedLong, seedFloat);
  Uniform::Set(seedFloat);

  // 10% of the links have 15min run lengths for the latency, 80% have
  // 52min run lengths, and 10% have 100min run lengths

  // this is the list of density files to be used for PktDrop
  const int numDist = 1;
  char densityFiles[numDist][256] = {
    //"udp-distributions/test.density",
    "udp-distributions/no-failures.density",

    /*
    "udp-distributions/aruna-jake.density",
    "udp-distributions/aruna-pandora.density",
    "udp-distributions/beetle-jake.density",
    "udp-distributions/beetle-plastique.density",
    "udp-distributions/jake-aruna.density",
    "udp-distributions/jake-beetle.density",
    "udp-distributions/jake-kofte.density",
    "udp-distributions/jake-pandora.density",
    "udp-distributions/jake-plastique-2.density",
    "udp-distributions/jake-plastique.density",
    "udp-distributions/kofte-jake.density",
    "udp-distributions/kofte-plastique.density",
    "udp-distributions/pandora-aruna.density",
    "udp-distributions/pandora-jake.density",
    "udp-distributions/plastique-beetle.density",
    "udp-distributions/plastique-jake-2.density",
    "udp-distributions/plastique-jake.density",
    "udp-distributions/plastique-kofte.density",
    */

    /*
    "ping-distributions/cgi.shellnet.co.uk-hnets.uia.ac.be-dat.density",
    "ping-distributions/cgi.shellnet.co.uk-sites.inka.de-dat.density",
    "ping-distributions/d18183f47.rochester.rr.com-www.his.com-dat.density",
    "ping-distributions/hnets.uia.ac.be-sites.inka.de-dat.density",
    "ping-distributions/www.csu.net-www.atmos.albany.edu-dat.density",
    "ping-distributions/www.csu.net-zeus.lyceum.com-dat.density",
    "ping-distributions/www.hit.net-d18183f47.rochester.rr.com-dat.density",
    "ping-distributions/www.hit.net-www.his.com-dat.density",
    "ping-distributions/www.interlog.com-inoc.iserv.net-dat.density",
    "ping-distributions/www.interworld.net-inoc.iserv.net-dat.density",
    "ping-distributions/www.interworld.net-www.interlog.com-dat.density",
    "ping-distributions/zeus.lyceum.com-www.atmos.albany.edu-dat.density"
    */
  };

  // 100ms is the normalized max cost (one way) of overlay links
  const short normalizedMaxCost = 100;

  // init the link characteristics from the graph
  int edgeNum = 0;
  for(unsigned int i = 0; i < g->nodes.size(); i++) {
    node_t &n = g->nodes[i];
    //JString destStr = ip2str(g->scidHash.get(n.scid));
    unsigned short dest_scid = n.scid;

    for(unsigned int j = 0; j < n.incoming.size(); j++) {
      //JString srcStr = ip2str(g->scidHash.get(n.incoming[j].origin_scid));
      unsigned short src_scid = n.incoming[j].origin_scid;
      unsigned short normalizedCost =
	(n.incoming[j].cost*normalizedMaxCost)/g->maxCost;

      float unifRand = U.Next();
      unsigned long runLength;
      if(unifRand <= 0.1) {
	runLength = 15*60*1000; // 15min, in ms
      } else if(unifRand <= 0.9) {
	runLength = 52*60*1000; // 52min, in ms
      } else {
	runLength = 100*60*1000; // 100min, in ms
      }

      // init the lat/loss structures
      float unifRand1 = U.Next();
      int densityIndex = (int)(unifRand1*numDist);
      densityIndex = edgeNum%numDist; // use deterministically
      links[edgeNum].lat = new GenLat(normalizedCost, runLength, arg_dbg);
      links[edgeNum].loss = new PktDrop
	(densityFiles[densityIndex], n.scid,
	 n.incoming[j].origin_scid, arg_dbg);
      links[edgeNum].det = new DetFail(arg_dbg);
      if((links[edgeNum].lat == NULL) || (links[edgeNum].loss == NULL) ||
	 (links[edgeNum].det == NULL)) {
	cerr << "Emulator::Emulator: can't alloc for lat/loss/det" << endl;
	exit(103);
      }

      // put the index in the hash of links
      //JString key = srcStr+":"+destStr;
      unsigned long key = 100*1000*src_scid + dest_scid;
      linkHash->put(key, edgeNum);
      edgeNum++;

      cerr << key << endl
	   << "\tGenLat(" << normalizedCost << "ms,"
	   << (runLength/(60*1000)) << "min); " << endl
	   << "\tPktDrop(" << densityFiles[densityIndex] << ')' << endl;

    } // End for(j) over edges

  } // End for(i) over nodes

  // init socket stuff
  // prepare raw socket
  raw_sock = socket(AF_INET, SOCK_RAW, IPPROTO_UDP);
  if(raw_sock < 0) {
    perror("socket");
    exit(1);
  }
  /* do a IP_HDRINCL call, to make sure that the kernel knows the
     header is included in the data */
  int one = 1;
  const int *val = &one;
  int res = setsockopt(raw_sock, IPPROTO_IP, IP_HDRINCL, val, sizeof(one));
  if(res < 0) {
    perror("setsockopt IP_HDRINCL");
    exit(2);
  }

  // prepare the UDP socket for listening
  udp_sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if(udp_sock < 0) {
    perror("socket");
    exit(3);
  }
  // bind to TUN_PORT
  struct sockaddr_in listen_addr;
  bzero((char *)&listen_addr, sizeof(listen_addr));
  listen_addr.sin_family =  AF_INET;
  listen_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  listen_addr.sin_port = htons(TUN_PORT);
  if(bind(udp_sock, (struct sockaddr *)&listen_addr,
	  sizeof(listen_addr)) < 0) {
    fprintf(stderr, "Can't bind to port %d\n", TUN_PORT);
    perror("bind failed");
    close(udp_sock);
    exit(4);
  }

} // End Emulator::Emulator()

int Emulator::alloc(int len) {
  // handle 3 cases -- pkt size less than 'Step1', between 'Step1' and
  // 'Step2', and above 'Step2'
  if(len <= Step1) {
    if(startSubStep1 != -1) {
      int toRet = startSubStep1;
      startSubStep1 = pktBuf[startSubStep1].nextFreeIndex;
      return toRet;

    } else if(startSubStep2 != -1) {
      int toRet = startSubStep2;
      startSubStep2 = pktBuf[startSubStep2].nextFreeIndex;
      return toRet;

    } else if (startFree != -1) {
      int toRet = startFree;
      startFree = pktBuf[startFree].nextFreeIndex;
      if(pktBuf[toRet].data == NULL) {
	pktBuf[toRet].data = new char[Step1];
	assert(pktBuf[toRet].data != NULL);
	pktBuf[toRet].allocLen = Step1;
      }
      return toRet;

    } else {
      return -1;
    }

    // End case if(len <= Step1)
  } else if(len <= Step2) {

    if(startSubStep2 != -1) {
      int toRet = startSubStep2;
      startSubStep2 = pktBuf[startSubStep2].nextFreeIndex;
      return toRet;

    } else if (startFree != -1) {
      int toRet = startFree;
      startFree = pktBuf[startFree].nextFreeIndex;
      if(pktBuf[toRet].data == NULL) {
	pktBuf[toRet].data = new char[Step2];
	assert(pktBuf[toRet].data != NULL);
	pktBuf[toRet].allocLen = Step2;
      }
      return toRet;

    } else if(startSubStep1 != -1) {
      int toRet = startSubStep1;
      startSubStep1 = pktBuf[startSubStep1].nextFreeIndex;
      assert(pktBuf[toRet].data != NULL);
      delete[] pktBuf[toRet].data;
      pktBuf[toRet].data = new char[Step2];
      assert(pktBuf[toRet].data != NULL);
      pktBuf[toRet].allocLen = Step2;
      return toRet;

    } else {
      return -1;
    }

    // end case (Step1 < len <= Step2)
  } else {
    if(startFree != -1) {
      int toRet = startFree;
      startFree = pktBuf[startFree].nextFreeIndex;
      if((pktBuf[toRet].data != NULL) && (len > pktBuf[toRet].allocLen)) {
	delete[] pktBuf[toRet].data;
	pktBuf[toRet].data = NULL;
      }
      if(pktBuf[toRet].data == NULL) {
	pktBuf[toRet].data = new char[len];
	assert(pktBuf[toRet].data != NULL);
	pktBuf[toRet].allocLen = len;
      }
      return toRet;

    } else if(startSubStep1 != -1) {
      int toRet = startSubStep1;
      startSubStep1 = pktBuf[startSubStep1].nextFreeIndex;
      assert(pktBuf[toRet].data != NULL);
      delete[] pktBuf[toRet].data;
      pktBuf[toRet].data = new char[len];
      assert(pktBuf[toRet].data != NULL);
      pktBuf[toRet].allocLen = len;
      return toRet;

    } else if(startSubStep2 != -1) {
      int toRet = startSubStep2;
      startSubStep2 = pktBuf[startSubStep2].nextFreeIndex;
      assert(pktBuf[toRet].data != NULL);
      delete[] pktBuf[toRet].data;
      pktBuf[toRet].data = new char[len];
      assert(pktBuf[toRet].data != NULL);
      pktBuf[toRet].allocLen = len;
      return toRet;

    } else {
      return -1;
    }

    // end case (Step2 < len)
  }

} // End Emulator::alloc()

void Emulator::dealloc(int index) {
  // should return index to the right 'free' list
  if(pktBuf[index].allocLen == Step1) {
    pktBuf[index].nextFreeIndex = startSubStep1;
    startSubStep1 = index;
  } else if(pktBuf[index].allocLen == Step2) {
    pktBuf[index].nextFreeIndex = startSubStep2;
    startSubStep2 = index;
  } else {
    pktBuf[index].nextFreeIndex = startFree;
    startFree = index;
  }
} // End Emulator::dealloc()

//-------------------- The interrupt handler --------------------//

float max_rate = 0;

#ifdef MEAS_FWDRAW
meas_t M[MAX_MEAS];
char *outFile;

void intr_catcher(int) {
  // open file for output
  ofstream fout(outFile);
  if(fout.fail()) {
    cerr << "Can't open output file " << outFile << endl;
    exit(2);
  }
  // write onto results file
  for(int i = 0; i < MAX_MEAS; i++) {
    fout << i << " \t" << M[i].seqNo << " \t" << M[i].recd << " \t"
	 << M[i].oneWayDelay << endl;
  }
  fout.close();
  cerr << endl << "Output written to " << outFile << endl;
  cerr << "max_rate=" << max_rate << " pkts/sec" << endl;
  printPktDist();
  exit(0);
} // End intr_catcher()

#else
void intr_catcher(int) {
  bool firstTime = false;
  cerr << "max_rate=" << max_rate << " pkts/sec" << endl;
  readDet(firstTime); // read the deterministic failures file
} // End intr_catcher()

#endif

//-------------------- The main function --------------------//

int main(int argc, char *argv[]) {
  util_init();

#ifdef MEAS_FWDRAW
  int argExpected = 4;
#else
  int argExpected = 3;
#endif

  if(argc != argExpected) { usage(argv[0]); exit(1); }

  const char *scidMappingFile = argv[1];
  const char *graphConfigFile = argv[2];

#ifdef MEAS_FWDRAW
  outFile = argv[3];
  Uniform U; // for deciding the random graph edge to look at
#endif

  // Activate interrupt catcher
  signal(SIGINT, intr_catcher);

  // read configuration
  graph_t g;
  readGraph(graphConfigFile, &g);
  readSCIDMapping(scidMappingFile, &g.scidHash);

  // Init the debug object
  struct timeval startTime;
  struct in_addr ipAddr;
  ipAddr.s_addr = inet_addr(timeServer);
  bool res1 = getStartTimeAtHost(&startTime, ipAddr);
  if(!res1) {
    cerr << "Can't get time from server " << timeServer << endl;
    exit(11);
  }
  char logFile[128];
  unsigned short mySCID = 0;
  sprintf(logFile, "/tmp/fwd-log.%d", mySCID);
  Debug_t dbg(logFile, mySCID, startTime, cfg_to_file, cfg_use_stream);

  // init the emulator
  Emulator_t E(&g, &dbg);

  gr = &g; // for use by readDet();
  emul = &E; // for use by readDet()
  bool firstTime = true;
  readDet(firstTime);

  numDropped = numSent = numRecd = 0;

  float last_mark_time = 0;
  unsigned long count_since_last_mark = 0;
  const float mark_period = 100; // 100ms

  // forever,
  // (1) see if there are any packets in the priority queue
  // (2) check for incoming packets, and put them in priority queue
  for(;;) {

    // (1) see if there are any packets in the priority queue
    unsigned long whenToSend;
    int pktIndex = E.pktQ->minVertex(&whenToSend);
    float timeNow = now(E.startTime);
    while((pktIndex >= 0) && (whenToSend <= timeNow)) {
      // there is a packet that has to be sent now
      // remove it from the packet queue (actually, reuse it)
      E.pktQ->reuseMin();

      // send the packet
      char *data = E.pktBuf[pktIndex].data;
      struct ip *iph = (struct ip*)data;
      struct udphdr *udph = (struct udphdr*)(data + sizeof(struct ip));

      struct sockaddr_in to_addr;
      bzero((char *)&to_addr, sizeof(to_addr));
      to_addr.sin_family = AF_INET;
      to_addr.sin_addr.s_addr = iph->ip_dst.s_addr;
      to_addr.sin_port = udph->dest;
      int res = sendto(E.raw_sock, data, ntohs(iph->ip_len), 0,
		       (struct sockaddr *)&to_addr, sizeof(to_addr));
      numSent++;
      if(res < 0) {
	perror("sendto");
	exit(3);
      } else {
	if(gdbg) { fputc('.', stderr); }
      }

      // deallocate the packet (put it into a free list)
      E.dealloc(pktIndex);
      // see if there are more packets to send at this time
      pktIndex = E.pktQ->minVertex(&whenToSend);
    }

    // (2) check for incoming packets, and put them in priority queue
    fd_set rfds; struct timeval tv; int retval;
    FD_ZERO(&rfds); FD_SET(E.udp_sock, &rfds);
    long toLong; // timeout in ms
    if(pktIndex < 0) {
      toLong = 100; /* no pkts to send now */
    } else {
      /* can sleep as long as we have time to send the next packet */
      toLong = (long)(whenToSend-timeNow);
    }
    tv.tv_sec = toLong/1000; tv.tv_usec = 1000*(toLong%1000);
    retval = select(E.udp_sock+1, &rfds, NULL, NULL, &tv);

    if(FD_ISSET(E.udp_sock, &rfds)) {
      // receive the packet
      char data[4096];
      struct ip *iph = (struct ip*)data;
      struct udphdr *udph = (struct udphdr*)(data + sizeof(struct ip));
      char *payload = data + sizeof(struct ip) + sizeof(struct udphdr);
      UDPLibPkt_t *upkt = (UDPLibPkt_t*)payload;

#ifdef MEAS_FWDRAW
      meas_pkt_t *mp = (meas_pkt_t*)(upkt->data);
#endif

      struct sockaddr_in from_addr;
      unsigned int from_len = sizeof(from_addr);
      int plen = recvfrom(E.udp_sock, payload,
			  sizeof(data)-sizeof(struct ip)-sizeof(struct udphdr),
			  0, (struct sockaddr*)&from_addr, &from_len);

      // check on the pkt rate
      if(numRecd == 0) {
	last_mark_time = timeNow;
	count_since_last_mark = 0;
      } else {
	float time_since_last_mark = timeNow-last_mark_time;
	if(time_since_last_mark >= mark_period) {
	  float rate = count_since_last_mark*1000/mark_period;
	  if(rate > 20000) { fputc('R', stderr); }
	  else if(rate > 15000) { fputc('r', stderr); }
	  if(rate > max_rate) { max_rate = rate; }
	  last_mark_time = timeNow;
	  count_since_last_mark = 0;
	} else {
	  count_since_last_mark++;
	}
      }

      numRecd++;

      if(plen < 0) {
	perror("recvfrom");
      }

      int dist_ind = plen/100;
      if(dist_ind > 15) { dist_ind = 15; }
      pktDist[dist_ind]++;

      // set the udp header
      udph->source = from_addr.sin_port;
      udph->dest = upkt->dest_port;
      udph->len = htons(sizeof(struct udphdr) + plen);
      udph->check = 0; // how to compute checksum?

      // set the IP fields
      iph->ip_hl = 5; // header length = 20 bytes = 5 X 32 bits
      iph->ip_v = 4; // IP version
      iph->ip_tos = 0; // type of service
      iph->ip_len =
	htons(sizeof(struct ip) + sizeof(struct udphdr) + plen);
      iph->ip_id = htonl(0); // fragment id (value does not matter here)
      iph->ip_off = 0; // fragment offset
      iph->ip_ttl = 255; // time to live
      iph->ip_p = IPPROTO_UDP; // protocol
      iph->ip_sum = 0; // will be set by kernel
      iph->ip_src.s_addr = from_addr.sin_addr.s_addr;
      iph->ip_dst.s_addr = upkt->dest_addr.s_addr;

      // see if this matches any rule
#ifdef MEAS_FWDRAW
      // need to make it look as if its coming from here
      iph->ip_src.s_addr = tun_ip.s_addr;
      udph->source = TUN_PORT;
      // generate random edge index
      int lcIndex = (int)(g.numEdges*U.Next());
#else
      // see if src and destn IP are same
      if(iph->ip_src.s_addr == iph->ip_dst.s_addr) {
	iph->ip_src.s_addr = tun_ip.s_addr;
	upkt->src_dest_same = htons(1);
      } else {
	upkt->src_dest_same = htons(0);
      }
      // get edge index from src:dest
      unsigned short src_scid = ntohs(upkt->my_scid);
      unsigned short dest_scid = ntohs(upkt->dest_scid);
      //JString key = ip2str(from_addr.sin_addr)+":"+ip2str(upkt->dest_addr);
      unsigned long key = 100*1000*src_scid + dest_scid;
      if(!E.linkHash->containsKey(key)) {
	cerr << "no rule for " << key << " src=" << ip2str(from_addr.sin_addr)
	     << ":" << ntohs(from_addr.sin_port) << " dst="
	     << ip2str(upkt->dest_addr) << ":"
	     << ntohs(upkt->dest_port) << endl;
	continue;
      }
      int lcIndex = E.linkHash->get(key);
#endif

      bool dropNow = E.links[lcIndex].loss->dropNow();
      bool detDropNow = E.links[lcIndex].det->dropNow();
      if(dropNow || detDropNow) {
#ifdef MEAS_FWDRAW
	// check for flagged pkt, and note measurement
	unsigned short flag = ntohs(mp->flag);
	if(flag) {
	  unsigned long measIndex = ntohl(mp->measIndex);
	  unsigned long seqNo = ntohl(mp->seqNo);
	  M[measIndex].seqNo = seqNo;
	}
#endif
	// pkt has to be dropped now
	if(gdbg) {
	  if(detDropNow) fputc('D', stderr);
	  else fputc('#', stderr);
	}
	numDropped++;
	continue;
      }

      // get space for packet
      int buflen = ntohs(iph->ip_len);
      int pktIndex = E.alloc(buflen);
      if(pktIndex < 0) {
	cerr << "running out of buffer space" << endl;
	continue;
      }
      // copy the pkt
      memcpy(E.pktBuf[pktIndex].data, data, buflen);
      // stash away the pkt to be forwarded later
      float timeNow = now(E.startTime);
      unsigned short pktDelay = E.links[lcIndex].lat->currLat();
      unsigned long delay = pktDelay + (unsigned long) timeNow;
      E.pktQ->decreaseCost(pktIndex, delay);
#ifdef MEAS_FWDRAW
      // check for flagged pkt, and note measurement
      unsigned short flag = ntohs(mp->flag);
      if(flag) {
	unsigned long measIndex = ntohl(mp->measIndex);
	unsigned long seqNo = ntohl(mp->seqNo);
	M[measIndex].seqNo = seqNo;
	M[measIndex].recd = true;
	M[measIndex].oneWayDelay = pktDelay;
      }
#endif

    } // End if(FD_ISSET)

  } // End for(;;)

} // End main()

#else

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

int main(int argc, char *argv[]) {
  fprintf(stderr, "IP_TUNNEL option not set while compiling\n");
  exit(0);
} // End main()

#endif
