/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
#ifndef FWDRAW_H
#define FWDRAW_H

#include <vector>

#include "../util/Util.h"
#include "../util/Debug.h"
#include "../util/JHashtable.h"
#include "../util/PriorityQueue.h"
#include "GenLat.h"
#include "PktDrop.h"
#include "DetFail.h"

typedef struct edge {
  unsigned short origin_scid; // origin of the arc
  short cost;
} edge_t; // End struct edge

typedef struct node {
  unsigned short scid;
  vector<edge_t> incoming;
} node_t; // End struct node

typedef struct graph {
  vector<node_t> nodes; // vector of nodes
  // hash of node scid to index in 'nodes' array
  JHashtable<unsigned short, int> nHash;
  // hash of node scid to ip_addr
  JHashtable<unsigned short, struct in_addr> scidHash;
  // the maximum edge cost in the graph (to scale down later)
  short maxCost;
  int numEdges; // the number of edges (to alloc the JHashtable)

  graph() {
    nHash.setHashFunction(ushortHash);
    scidHash.setHashFunction(ushortHash);
    maxCost = 0;
    numEdges = 0;
  }
} graph_t; // End struct graph

/** Represents the delay/loss patterns of a single overlay link (a
    wide-area path) */
struct LinkCharacteristics {
  GenLat_t *lat; // the latency generator
  PktDrop_t *loss; // the loss gap generator
  DetFail_t *det; // the deterministic losses
}; // End class LinkCharacteristics
typedef struct LinkCharacteristics LinkCharacteristics_t;

/** Represents the collection of pkts used as buffers */
struct EmulPktBuf {
  // The data this buffer is holding.  Invariant: if this is not
  // allocated, it is set of NULL
  char *data;
  int allocLen; // the allocated length of 'data'
  int len; // the actual length of the 'data'
  /** If this pkt is part of the free list, this gives the index of
      the next free pkt */
  int nextFreeIndex;
  EmulPktBuf() {
    data = NULL;
    len = 0;
  }
};
typedef struct EmulPktBuf EmulPktBuf_t;

/** Represents the emulator -- set of data structures for emulating
    the wide-area overlay network */
struct Emulator {
  // the debug structure
  Debug_t *dbg;

  LinkCharacteristics_t *links; // the array of links
  // the hashtable of src*100,000+dest JString to the index in 'links'
  JHashtable<unsigned long, int> *linkHash;

  // the collection of packets to use as buffer
  EmulPktBuf_t *pktBuf;
  // the number of packets in the buffer
  int numPkts;

  // the priority queue of packets to send -- priority queue on the
  // time when the packet has to be sent
  PriorityQueue<unsigned long> *pktQ;
  // this is time w.r.t. which the 'keys' of the priority queue are
  // determined
  struct timeval startTime;

  /** The start of the list of free packets that are unallocated (-1
      if nothing list is empty) */
  int startFree;
  /** See 'startSubStep1' */
  int Step1;
  /** The start of the list of free packets that are allocated, and
      less than 'Step1' bytes (-1 if list is empty) */
  int startSubStep1;
  /** See 'startSubStep2' */
  int Step2;
  /** The start of the list of free packets that are allocated, and
      over 'Step1' bytes, but less than 'Step2' bytes (-1 if list is
      empty) */
  int startSubStep2;

  /*---- socket stuff ----*/
  /** The raw IP socket, for sending */
  int raw_sock;
  /** The UDP packet, for receiving */
  int udp_sock;

  /** the constructor */
  Emulator(graph_t *g, Debug_t *arg_dbg);

  /** Allocate a packet of at least given size.  The allocation is
      done, and the index to the 'pktBuf' array is returned.  -1 is
      returned on error. */
  int alloc(int len);

  /** Deallocate the pktBuf structure with the given index.  The
      memory may not be deleted -- the index could just be pushed to a
      free list.  It is assumed that the 'index' is currently
      allocated -- this is not checked. */
  void dealloc(int index);

};
typedef struct Emulator Emulator_t;

#endif
