/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Program to measure overhead of fwdRaw */

#include <iostream>
#include <fstream>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <netdb.h>

#include "measFwdRaw.h"
#include "../udp-lib/UDPLibPkt.h"
#include "../util/Util.h"
#include "../util/newran02/include.h"
#include "../util/newran02/newran.h"

void usage(const char *progname) {
  fprintf(stderr, "Usage:\t %s <packet-rate> <pkt-size> <meas-output-file>\n",
	  progname);
  fprintf(stderr, "\t packet rate should be given in num-pkts/sec\n");
} // End usage()

int main(int argc, char *argv[]) {
  util_init();
  if(argc != 4) { usage(argv[0]); exit(1); }
  int pktRate = atoi(argv[1]);
  int PKT_LEN = atoi(argv[2]);
  const char *outFile = argv[3];
  const char *localAddr = getenv("HOSTNAME");
  struct in_addr localIP = nslookup(localAddr); // DNS lookup

  // set random seed
  Uniform U;
  struct timeval tv;
  gettimeofday(&tv, NULL);
  unsigned long seedLong = tv.tv_usec%211;
  float seedFloat = ((seedLong+1)*1.0)/212;
  fprintf(stderr, "setting random seed done with %ld(%f)\n",
	  seedLong, seedFloat);
  Uniform::Set(seedFloat);

  // prepare the UDP socket for listening
  int udp_sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if(udp_sock < 0) {
    perror("socket");
    exit(3);
  }
  // bind to localIP:MEAS_PORT
  struct sockaddr_in listen_addr;
  bzero((char *)&listen_addr, sizeof(listen_addr));
  listen_addr.sin_family =  AF_INET;
  listen_addr.sin_addr.s_addr = localIP.s_addr;
  //listen_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  listen_addr.sin_port = htons(MEAS_PORT);
  if(bind(udp_sock, (struct sockaddr *)&listen_addr,
	  sizeof(listen_addr)) < 0) {
    fprintf(stderr, "Can't bind to port %s:%d\n", localAddr, MEAS_PORT);
    perror("bind failed");
    close(udp_sock);
    exit(4);
  }
  cerr << "Sending and listening on " << localAddr << ":" << MEAS_PORT
       << endl;

  // open results file
  ofstream fout(outFile);
  if(fout.fail()) {
    cerr << "Can't open output file " << outFile << endl;
    exit(5);
  }

  struct timeval startTime;
  gettimeofday(&startTime, NULL);
  int numPktsSent = 0;

  int numMeas = 0;
  meas_t M[MAX_MEAS];
  const float flagProb = 0.01;

  // the last time a packet was received
  struct timeval lastPktRecd;
  gettimeofday(&lastPktRecd, NULL);

  for(;;) {

    // check if time to send
    float timeSinceStart = now(startTime); // time in ms

    if((numMeas < MAX_MEAS) && (numPktsSent < timeSinceStart*pktRate/1000.0)) {

      // generate random pkt size
      int len = (int)(PKT_LEN*U.Next());
      len = PKT_LEN;
      len += sizeof(meas_pkt_t)+sizeof(UDPLibPkt_t)-MAXDATASZ;
      UDPLibPkt_t uPkt;
      meas_pkt_t *mp = (meas_pkt_t*)uPkt.data;
      mp->seqNo = htonl(numPktsSent);
      mp->measIndex = htonl(numMeas);
      mp->flag = htons(0);

      // decide at random, whether or not to flag packet
      float unifRand = U.Next();
      if(unifRand <= flagProb) {
	mp->flag = htons(1);
	gettimeofday(&M[numMeas].start, NULL);
	M[numMeas].seqNo = numPktsSent;
	if((numMeas % 100) == 0) {
	  cerr << "numMeas=" << numMeas << ", MAX_MEAS=" << MAX_MEAS << endl;
	}
	numMeas++;
	//fprintf(stderr, "+");
      } else {
	//fprintf(stderr, "_");
      }
      bool use_tunnel = true;
      unsigned long count = 0;
      unsigned short dest_scid = 0, my_scid = 0;
      int res = lib_sendto(udp_sock, &uPkt, len, 0, &listen_addr,
			   use_tunnel, &count, dest_scid, my_scid);
      if(res == -1) { perror("lib_sendto"); }
      numPktsSent++;

    }

    // check socket for packet receipt
    fd_set rfds; struct timeval tv; int retval;
    FD_ZERO(&rfds); FD_SET(udp_sock, &rfds);
    tv.tv_sec = 0; tv.tv_usec = 0; // don't block
    retval = select(udp_sock+1, &rfds, NULL, NULL, &tv);

    if(FD_ISSET(udp_sock, &rfds)) {
      // receive the packet
      UDPLibPkt_t recvPkt;
      struct sockaddr_in from_addr;
      unsigned int from_len = sizeof(from_addr);
      int plen = recvfrom(udp_sock, &recvPkt, sizeof(UDPLibPkt_t),
			  0, (struct sockaddr*)&from_addr, &from_len);
      if(plen < 0) {
	perror("recvfrom");
      }
      // mark the measure, if it had been flagged
      meas_pkt_t *mp = (meas_pkt_t*)(recvPkt.data);
      unsigned short flag = ntohs(mp->flag);
      if(flag) {
	unsigned long measIndex = ntohl(mp->measIndex);
	M[measIndex].recd = true;
	M[measIndex].oneWayDelay =
	  (unsigned short)(now(M[measIndex].start));
	//fprintf(stderr, "#");
      } else {
	//fprintf(stderr, "~");
      }
      // reset the lastPktRecd time
      gettimeofday(&lastPktRecd, NULL);
    }

    // stop if 'MAX_MEAS' packets flagged totally, and no packets recd
    // in the last 5 seconds
    if((numMeas == MAX_MEAS) && (now(lastPktRecd) > 5*1000)) {
      break;
    }
  }

  // write onto results file
  for(int i = 0; i < MAX_MEAS; i++) {
    fout << i << " \t" << M[i].seqNo << " \t" << M[i].recd << " \t"
	 << M[i].oneWayDelay << endl;
  }
  fout.close();
  cerr << endl << "Output written to " << outFile << endl;

  for(int i = 0; i < 5; i++) {
    cerr << '\a';
    sleep(1);
  }

} // End main()
