/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Implements the routines in HB.h */

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

#include <iostream>

#include "HB.h"
#include "../udp-lib/UDPLibPkt.h"
#include "../util/Util.h"

pthread_t HB::listenThr;
int HB::numHBThreads;
class HB* HB::theThreads[MAX_PEERS];
bool HB::start_send = false;
bool HB::start_listen = false;
bool HB::stop_send = false;
bool HB::stop_listen = false;
unsigned long HB::numRecd = 0;

int HB::send_fd;
int HB::listen_fd;

bool HB::run_send_thread = true;
bool HB::run_listen_thread = true;
struct timeval HB::lastSent;
long HB::HB_PERIOD = 300; // 300ms
unsigned short HB::mySCID = 0;

pthread_t HB::all_sendThr;

extern int print_ctl;

HB::HB(const struct in_addr &arg_peer_ipaddr, unsigned short arg_peer_scid) {
  // Init remote address
  peer_ipaddr = arg_peer_ipaddr;
  peer_scid = arg_peer_scid;

  // Update the 'static' list of HBThreads
  // Note: this update is not protected by a mutex -- assume that all
  // HB threads are init by the same main thread
  if(numHBThreads < MAX_PEERS) {
    theThreads[numHBThreads] = this;
    numHBThreads++;
  } else {
    fprintf(stderr, "numHBThreads (%d) exceeded MAX_PEERS\n", numHBThreads);
  }

  // Init status and peer_view_status
  peer_status = my_status_peer_view = DEAD;

  // the number of packets sent
  numSent = 0;
} // End HB::HB()

void HB::startListenThread(bool arg_run_listen_thread) {
  run_listen_thread = arg_run_listen_thread;

  start_listen = true;
  stop_send = stop_listen = false;
  // the number of packets recd
  numRecd = 0;
  // socket for listening
  listen_fd = sock_and_bind(HB_PORT);
  if(listen_fd < 0) { perror("HB::listenThread.socket"); return; }

  if(!run_listen_thread) { return; }

  // Spawn the thread
  pthread_create(&listenThr, NULL, HB::listenThread, 0);
} // End HB::startListenThread()

void* HB::listenThread(void *) {
  if(!run_listen_thread) {
    fprintf(stderr, "listenThread running without run_listen_thread set\n");
    exit(43);
  }

  // listen for packets forever, deal as appropriate
  for(;;) {
    if(stop_listen) { break; }

    struct timeval timeout = { 2, 0 }; // 2sec
    fd_set rfds; FD_ZERO(&rfds); FD_SET(listen_fd, &rfds);
    select(listen_fd+1, &rfds, NULL, NULL, &timeout);
    if(!FD_ISSET(listen_fd, &rfds)) { continue; }

    recvHelper();

  } // End for(;;)

  return 0;
} // End HB::listenThread()

void HB::recvHelper() {
  struct sockaddr_in from_addr;
  UDPLibPkt_t upkt;
  hb_pkt_t *p = (hb_pkt_t*)upkt.data;
  int expected_len = sizeof(hb_pkt_t) + sizeof(upkt) - MAXDATASZ;

  // receive packet (blocking)
  // NOTE: called only when something is ready to be read
  int res = lib_recvfrom(listen_fd, &upkt, expected_len, &from_addr);
  numRecd++;
  if(res == -1) { perror("HB::listenThread.recvfrom"); return; }
  if(res != expected_len) {
    fprintf(stderr, "res=%d, expected_len=%d\n", res, expected_len);
    return;
  }

  // Find the right thread, and call processPkt on it
  for(int i = 0; i < numHBThreads; i++) {
    if(theThreads[i]->peer_ipaddr.s_addr == from_addr.sin_addr.s_addr) {
      theThreads[i]->processPkt(p);
    }
  }
} // End HB::recvHelper()

void HB::processPkt(hb_pkt_t* p) {
  my_status_peer_view = ntohs(p->status);
  peer_status = LIVE;
  gettimeofday(&lastMeas, NULL);
} // End HB::processPkt()

void HB::startSendThread() {
  // Spawn the thread
  pthread_create(&sendThr, NULL, HB::sendThread, this);
} // End HB::startSendThread()

void HB::start_all_sends(bool arg_run_send_thread, unsigned short arg_mySCID) {
  run_send_thread = arg_run_send_thread;
  mySCID = arg_mySCID;
  start_send = true;
  stop_send = false;

  // socket
  send_fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if(send_fd < 0) { perror("HB::sendThread.socket"); return; }

  // init the last time a packet was sent
  gettimeofday(&lastSent, NULL);

  if(!run_send_thread) { return; }

#ifdef HB_ALL_SEND_IN_ONE_THR
  // Spawn the all_sendThr
  pthread_create(&all_sendThr, NULL, HB::all_sendThread, 0);
#else
  for(int i = 0; i < numHBThreads; i++) {
    theThreads[i]->startSendThread();
  }
#endif

} // End HB::start_all_sends()

void* HB::sendThread(void *arg_thisPtr) {

#ifdef HB_ALL_SEND_IN_ONE_THR
  fprintf(stderr, "sendThread: HB_ALL_SEND_IN_ONE_THR def\n");
  exit(33);
#endif

  if(!run_send_thread) {
    fprintf(stderr, "sendThread running without run_send_thread set\n");
    exit(43);
  }

  HB_t* thisPtr = (HB_t *)arg_thisPtr;

  // forever, send packet, check for timeout, and sleep
  for(;;) {
    if(HB::stop_send) { break; }
    thisPtr->sendHelper();
    // sleep for HB_PERIOD ms
    usleep(HB_PERIOD*1000);
  }

  return 0;
} // End HB::sendThread()

void* HB::all_sendThread(void*) {

#ifndef HB_ALL_SEND_IN_ONE_THR
  fprintf(stderr, "all_sendThread: HB_ALL_SEND_IN_ONE_THR not def\n");
  exit(33);
#endif

  if(!run_send_thread) {
    fprintf(stderr, "all_sendThread running without run_send_thread set\n");
    exit(43);
  }

  for(;;) {
    if(HB::stop_send) { break; }
    for(int i = 0; i < numHBThreads; i++) {
      theThreads[i]->sendHelper();
    }
    // sleep for HB_PERIOD ms
    usleep(HB_PERIOD*1000);
  }

  return 0;
} // End HB::all_sendThread()

void HB::all_checkSend() {
  if(run_send_thread) {
    fprintf(stderr, "all_checkSend called with run_send_thread set\n");
    exit(98);
  }

  if(!start_send || stop_send) { return; }

  float timeSinceLastSend = now(lastSent);
  if(timeSinceLastSend < HB_PERIOD) { return; }

  for(int i = 0; i < numHBThreads; i++) {
    theThreads[i]->sendHelper();
  }

  gettimeofday(&lastSent, NULL);
} // End HB::all_checkSend()

void HB::sendHelper() {
  UDPLibPkt_t upkt;
  hb_pkt_t *p = (hb_pkt_t*)upkt.data;
  int expected_len = sizeof(hb_pkt_t) + sizeof(upkt) - MAXDATASZ;
  struct sockaddr_in to_addr;
  bzero((char *)&to_addr, sizeof(to_addr));
  to_addr.sin_family = PF_INET;
  to_addr.sin_addr = this->peer_ipaddr;
  to_addr.sin_port = htons(HB_PORT);

  p->status = htons(this->peer_status);
  bool use_tunnel = true;
  int res = lib_sendto(send_fd, &upkt, expected_len, 0, &to_addr,
		       use_tunnel, &this->numSent, this->peer_scid, mySCID);
  if(res < 0) {
    perror("HB::sendThread.lib_sendto");
  }
  if(res != expected_len) {
    fprintf(stderr, "res=%d, expected_len=%d\n", res, expected_len);
  }

  // check for timeout
  float timeout = 1800; // 1800ms = 1.8sec
  if(now(this->lastMeas) >= timeout) {
    this->peer_status = DEAD;
  }
} // End HB::sendHelper()

void HB::stopSendThreads() {
  stop_send = true;
  if(!run_send_thread) { return; }
#ifdef HB_ALL_SEND_IN_ONE_THR
  pthread_join(all_sendThr, NULL);
#else
  for(int i = 0; i < numHBThreads; i++) {
    pthread_join(theThreads[i]->sendThr, NULL);
  }
#endif
} // End HB::stopSendThreads()

void HB::stopListenThread() {
  stop_listen = true;
  if(!run_listen_thread) { return; }
  pthread_join(listenThr, NULL);
} // End HB::stopListenThread()

unsigned long HB::getNumSent() {
  unsigned long totSent = 0;
  for(int i = 0; i < numHBThreads; i++) {
    totSent += theThreads[i]->numSent;
  }
  return totSent;
} // End HB::getNumSent()

unsigned long HB::getNumRecd() {
  return numRecd;
} // End HB::getNumRecd()
