/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
#ifndef HB_H
#define HB_H

/** Class to implement the thread that sends heart-beats to a given
    host.  It assumes that a similar thread is running on the other
    end.  The heart-beat is done using UDP packets.  This class also
    implements a static thread to receive heart-beat packets.  */

#include <pthread.h>
#include <sys/time.h>
#include <netinet/in.h>

// This means that all the send threads should be a single thread in
// implementation
#define HB_ALL_SEND_IN_ONE_THR 1

#define MAX_PEERS 64

#define HB_PORT 3400

struct hb_pkt {
  short status;
};

typedef struct hb_pkt hb_pkt_t;

#define DEAD 0
#define LIVE 1

class HB {
protected:
  /** my SCID */
  static unsigned short mySCID;

  /** The HB period, in ms */
  static long HB_PERIOD;
  /** The last time HB was sent (used only when run_threads is not set) */
  static struct timeval lastSent;

  /** Should I start sending? */
  static bool start_send;
  /** Should I start listening? */
  static bool start_listen;

  /** Should I stop sending? */
  static bool stop_send;
  /** Should I stop listening? */
  static bool stop_listen;

  /** The last time HB was received */
  struct timeval lastMeas;
  /** The peer's status: dead/live */
  short peer_status;
  /** The peer's view of my status: dead/live */
  short my_status_peer_view;
  /** The IP address of the remote peer host */
  struct in_addr peer_ipaddr;
  /** The SCID of the remote peer host */
  unsigned short peer_scid;

  /** The thread to listen for incoming heart-beats from peer(s) */
  static pthread_t listenThr;
  /** Implements listenThr */
  static void* listenThread(void*);

  /** The number of HB threads so far */
  static int numHBThreads;
  /** The array of HB threads so far */
  static class HB* theThreads[MAX_PEERS];

  /** Process the incoming pkt.  This just updates
      'my_status_peer_view' and 'lastMeas'.  'peer_status' is turned
      from LIVE to DEAD by 'sendThread', and DEAD to LIVE or LIVE to
      LIVE by 'processPkt'  */
  void processPkt(hb_pkt_t* p);

  /** The thread that sends packets to the peer periodically */
  pthread_t sendThr;
  /** The total number of packets sent */
  unsigned long numSent;
  /** The total number of packets sent */
  static unsigned long numRecd;
  /** Implements sendThr */
  static void* sendThread(void* thisPtr);
  /** Start the sendThread for this peer */
  void startSendThread();

  /** The single thread that implements all the send threads */
  static pthread_t all_sendThr;
  /** The single send thread for all peers */
  static void* all_sendThread(void*);

  /** Helper for sendThread/all_sendThread */
  void sendHelper();

public:
  /** The socket descriptors for sending and listening */
  static int send_fd;
  static int listen_fd;

  /** Should run listen thread or not?  NOTE: this is supposed to be
      set by the startListenThread routine */
  static bool run_listen_thread;
  /** Should run send thread(s) or not?  NOTE: this is supposed to be
      set by the start_all_sends routine */
  static bool run_send_thread;

  /** Start the listenThread to listen on incoming packets, or prepare
      for listening if arg_run_listen_thread is false */
  static void startListenThread(bool arg_run_listen_thread);
  /** Start all the sendThreads, or the single all_sendThread, or
      prepare to send if arg_run_send_thread is false */
  static void start_all_sends(bool arg_run_send_thread,
			      unsigned short arg_mySCID);

  /** Check for all sends -- should be called only when run_threads is
      false */
  static void all_checkSend();

  /** Receive a HB pkt and deal as appropriate -- to be called only
      when there is a pkt ready to be received */
  static void recvHelper();

  /** The constructor -- the host to which heart-beats should be sent
      is given */
  HB(const struct in_addr &arg_peer_ipaddr, unsigned short arg_peer_scid);
  /** Return the status of the peer: DEAD/LIVE */
  short getPeerStatus() { return peer_status; }
  /** Return my status as seen by the peer: DEAD/LIVE */
  short getMyStatusPeerView() { return my_status_peer_view; }
  /** When was the last heart-beat received?  The return value has no
      meaning if the no HB has been recd so far (in this case,
      peer_status and my_status_peer_view would be 0). */
  struct timeval getLastMeas() { return lastMeas; }

  /** Stop the send threads */
  static void stopSendThreads();
  /** Stop the listen thread */
  static void stopListenThread();

  /** Return the number of pkts sent by all the threads */
  static unsigned long getNumSent();
  /** Return the number of packets received */
  static unsigned long getNumRecd();

}; // End class HB

typedef class HB HB_t;

#endif
