/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Implements the routines in LatMeas.h */

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

#include "LatMeas.h"
#include "../udp-lib/UDPLibPkt.h"
#include "../util/Util.h"

pthread_t LatMeas::listenThr;
int LatMeas::numLatMeasThreads;
class LatMeas* LatMeas::theThreads[MAX_PEERS];
bool LatMeas::start_send = false;
bool LatMeas::start_listen = false;
bool LatMeas::stop_send = false;
bool LatMeas::stop_listen = false;
unsigned long LatMeas::numRespSent = 0;
unsigned long LatMeas::numRecd = 0;

int LatMeas::send_fd;
int LatMeas::listen_fd;

bool LatMeas::run_send_thread = true;
bool LatMeas::run_listen_thread = true;
struct timeval LatMeas::lastSent;
long LatMeas::LM_PERIOD = 2000; // 2sec
unsigned short LatMeas::mySCID = 0;

pthread_t LatMeas::all_sendThr;

extern int print_ctl;

LatMeas::LatMeas(const struct in_addr &arg_peer_ipaddr,
		 unsigned short arg_peer_scid) {
  // Init remote address
  peer_ipaddr = arg_peer_ipaddr;
  peer_scid = arg_peer_scid;

  // Update the 'static' list of LatMeasThreads
  // Note: this update is not protected by a mutex -- assume that all
  // LatMeas threads are init by the same main thread
  if(numLatMeasThreads < MAX_PEERS) {
    theThreads[numLatMeasThreads] = this;
    numLatMeasThreads++;
  } else {
    fprintf(stderr, "numLatMeasThreads exceeded MAX_PEERS\n");
  }

  // Init srtt and rtt
  rtt = srtt = 0;

  // The number of packets sent by this threads
  numSent = 0;
} // End LatMeas::LatMeas()

void LatMeas::startListenThread(bool arg_run_listen_thread) {
  run_listen_thread = arg_run_listen_thread;

  start_listen = true;
  stop_send = stop_listen = false;
  numRecd = numRespSent = 0;
  // socket for listening
  listen_fd = sock_and_bind(LAT_MEAS_PORT);
  if(listen_fd < 0) { perror("LatMeas::listenThread.socket"); return; }

  if(!run_listen_thread) { return; }

  // Spawn the thread
  pthread_create(&listenThr, NULL, LatMeas::listenThread, 0);
} // End LatMeas::startListenThread()

void* LatMeas::listenThread(void *) {
  // listen for packets forever, deal as appropriate
  for(;;) {
    if(stop_listen) { break; }

    struct timeval timeout = { 2, 0 }; // 2sec
    fd_set rfds; FD_ZERO(&rfds); FD_SET(listen_fd, &rfds);
    select(listen_fd+1, &rfds, NULL, NULL, &timeout);
    if(!FD_ISSET(listen_fd, &rfds)) { continue; }

    recvHelper();
  } // End for(;;)

  return 0;
} // End LatMeas::listenThread()

void LatMeas::recvHelper() {

  struct sockaddr_in from_addr;
  UDPLibPkt_t upkt;
  pkt_t *p = (pkt_t*)upkt.data;
  int expected_len = sizeof(pkt_t) + sizeof(upkt) - MAXDATASZ;

  // receive packet (blocking)
  int res = lib_recvfrom(listen_fd, &upkt, expected_len, &from_addr);
  numRecd++;
  if(res == -1) {
    perror("LatMeas::listenThread.recvfrom"); return;
  }
  if(res != expected_len) {
    fprintf(stderr, "res=%d, expected_len=%d\n", res, expected_len);
    return;
  }

  // if pkt type is REQ_TYPE, send a response
  if(ntohs(p->pkt_type) == REQ_TYPE) {
    p->pkt_type = htons(RESP_TYPE);
    struct sockaddr_in to_addr;
    bzero((char *)&to_addr, sizeof(to_addr));
    to_addr.sin_family = PF_INET;
    to_addr.sin_addr = from_addr.sin_addr;
    to_addr.sin_port = htons(LAT_MEAS_PORT);

    bool use_tunnel = true;
    unsigned short pkt_dest_scid = 0;
#ifdef IP_TUNNEL
    pkt_dest_scid = ntohs(upkt.my_scid);
#endif
    int numSent1 = lib_sendto(listen_fd, &upkt, expected_len, 0,
			      &to_addr, use_tunnel, &numRespSent,
			      pkt_dest_scid, mySCID);
    if(numSent1 < 0) {
      perror("LatMeas::listenThread.lib_sendto");
      return;
    }
    if(numSent1 != expected_len) {
      fprintf(stderr, "numSent1=%d, expected_len=%d\n", numSent1, expected_len);
      return;
    }

    return;
  }

  if(ntohs(p->pkt_type) != RESP_TYPE) {
    fprintf(stderr, "Warning: p->pkt_type is %d\n", ntohs(p->pkt_type));
    return;
  }

  // It is a response packet -- find the right thread, and call
  // processPkt on it
  for(int i = 0; i < numLatMeasThreads; i++) {
    if(theThreads[i]->peer_ipaddr.s_addr == from_addr.sin_addr.s_addr) {
      theThreads[i]->processPkt(p);
    }
  }
} // End LatMeas::recvHelper()

void LatMeas::processPkt(pkt_t* p) {
  // Implement the TCP algorithm for smoothed RTT
  // Also compute the rttvar -- not used as of now (11 Aug 2001)
  rtt = now(p->ts);
  if(srtt == 0) {
    srtt = rtt;
    rttvar = rtt/2;

  } else {
    float alpha = 1.0/8.0;
    float beta = 1.0/4.0;
    srtt = (1-alpha)*srtt + alpha*rtt;
    rttvar = (1-beta)*rttvar + beta*fabs(srtt-rtt);

  }
  gettimeofday(&lastMeas, NULL);
  //fprintf(stderr, "rtt=%f, srtt=%f, rttvar=%f\n", rtt, srtt, rttvar);
} // End LatMeas::processPkt()

void LatMeas::startSendThread() {
  // Spawn the thread
  pthread_create(&sendThr, NULL, LatMeas::sendThread, this);
} // End LatMeas::startSendThread()

void LatMeas::start_all_sends(bool arg_run_send_thread,
			      unsigned short arg_mySCID) {
  run_send_thread = arg_run_send_thread;
  mySCID = arg_mySCID;
  start_send = true;
  stop_send = false;

  // socket
  send_fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if(send_fd < 0) { perror("LatMeas::start_all_sends.socket"); return; }

  // init the last time a packet was sent
  gettimeofday(&lastSent, NULL);

  if(!run_send_thread) { return; }

#ifdef LM_ALL_SEND_IN_ONE_THR
  // Spawn the all_sendThr
  pthread_create(&all_sendThr, NULL, LatMeas::all_sendThread, 0);
#else
  for(int i = 0; i < numLatMeasThreads; i++) {
    theThreads[i]->startSendThread();
  }
#endif

} // End LatMeas::start_all_sends()

void* LatMeas::sendThread(void *arg_thisPtr) {

#ifdef LM_ALL_SEND_IN_ONE_THR
  fprintf(stderr, "sendThread: LM_ALL_SEND_IN_ONE_THR def\n");
  exit(33);
#endif

  class LatMeas* thisPtr = (class LatMeas *)arg_thisPtr;

  // forever, send packet and sleep
  for(;;) {
    if(LatMeas::stop_send) { break; }
    thisPtr->sendHelper();
    usleep(LM_PERIOD*1000);
  }

  return 0;
} // End LatMeas::sendThread()

void* LatMeas::all_sendThread(void*) {

#ifndef LM_ALL_SEND_IN_ONE_THR
  fprintf(stderr, "all_sendThread: LM_ALL_SEND_IN_ONE_THR not def\n");
  exit(33);
#endif

  // socket
  int send_fd = socket(PF_INET, SOCK_DGRAM, getprotobyname("udp")->p_proto);
  if(send_fd < 0) {
    perror("LatMeas::sendThread.socket");
    return 0;
  }

  // forever, send packet and sleep
  for(;;) {
    if(LatMeas::stop_send) { break; }
    for(int i = 0; i < numLatMeasThreads; i++) {
      theThreads[i]->sendHelper();
    }
    usleep(LM_PERIOD*1000);
  }

  return 0;
} // End LatMeas::all_sendThread()

void LatMeas::all_checkSend() {
  if(run_send_thread) {
    fprintf(stderr, "all_checkSend called with run_send_thread set\n");
    exit(98);
  }

  if(!start_send || stop_send) { return; }

  float timeSinceLastSend = now(lastSent);
  if(timeSinceLastSend < LM_PERIOD) { return; }

  for(int i = 0; i < numLatMeasThreads; i++) {
    theThreads[i]->sendHelper();
  }

  gettimeofday(&lastSent, NULL);
} // End LatMeas::all_checkSend()

void LatMeas::sendHelper() {
  UDPLibPkt_t upkt;
  pkt_t *p = (pkt_t*)upkt.data;
  struct sockaddr_in to_addr;
  bzero((char *)&to_addr, sizeof(to_addr));
  to_addr.sin_family = PF_INET;
  to_addr.sin_addr = this->peer_ipaddr;
  to_addr.sin_port = htons(LAT_MEAS_PORT);
  p->pkt_type = htons(REQ_TYPE);
  int expected_len = sizeof(pkt_t) + sizeof(upkt) - MAXDATASZ;

  gettimeofday(&p->ts, NULL);
  bool use_tunnel = true;
  int res = lib_sendto(send_fd, &upkt, expected_len, 0, &to_addr,
		       use_tunnel, &this->numSent, this->peer_scid, mySCID);
  if(res < 0) { perror("LatMeas::sendThread.lib_sendto"); }
  if(res != expected_len) {
    fprintf(stderr, "res=%d, expected_len=%d\n", res, expected_len);
  }
} // End LatMeas::sendHelper()

void LatMeas::stopSendThreads() {
  stop_send = true;
  if(!run_send_thread) { return; }
#ifdef HB_ALL_SEND_IN_ONE_THR
  pthread_join(all_sendThr, NULL);
#else
  for(int i = 0; i < numLatMeasThreads; i++) {
    pthread_join(theThreads[i]->sendThr, NULL);
  }
#endif
} // End LatMeas::stopSendThreads()

void LatMeas::stopListenThread() {
  stop_listen = true;
  if(!run_listen_thread) { return; }
  pthread_join(listenThr, NULL);
} // End LatMeas::stopListenThread()

unsigned long LatMeas::getNumSent() {
  unsigned long totSent = numRespSent;
  for(int i = 0; i < numLatMeasThreads; i++) {
    totSent += theThreads[i]->numSent;
  }
  return totSent;
} // End LatMeas::getNumSent()

unsigned long LatMeas::getNumRecd() {
  return numRecd;
} // End LatMeas::getNumRecd()
