/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
#ifndef LATMEAS_H
#define LATMEAS_H

/** Class to implement the thread that measures the latency to a given
    host.  It assumes that a similar thread is running on the other
    end.  The measurement is done using UDP packets, and RTT averaging
    is done like in TCP.  */

#include <pthread.h>
#include <sys/time.h>
#include <netinet/in.h>

// This means that all the send threads should be a single thread in
// implementation
#define LM_ALL_SEND_IN_ONE_THR 1

#define MAX_PEERS 64

#define LAT_MEAS_PORT 3200

#define REQ_TYPE 1
#define RESP_TYPE 2

struct pkt {
  short pkt_type; // One of REQ_TYPE/RESP_TYPE
  struct timeval ts; // The timestamp while sending
};

typedef struct pkt pkt_t;

class LatMeas {
protected:
  /** my SCID */
  static unsigned short mySCID;

  /** The measurement period, in ms */
  static long LM_PERIOD;
  /** The last time a measurement was sent (used only when run_threads
      is not set) */
  static struct timeval lastSent;

  /** Should I start sending? */
  static bool start_send;
  /** Should I start listening? */
  static bool start_listen;

  /** Should I stop sending? */
  static bool stop_send;
  /** Should I stop listening? */
  static bool stop_listen;

  /** The last measured RTT */
  float rtt;
  /** The smoothed RTT */
  float srtt;
  /** The running RTT variance */
  float rttvar;
  /** The last time RTT was measured */
  struct timeval lastMeas;
  /** The IP address of the remote peer host */
  struct in_addr peer_ipaddr;
  /** The SCID of the remote peer host */
  unsigned short peer_scid;

  /** The thread to listen for incoming packets from peer(s) */
  static pthread_t listenThr;
  /** The number of responses sent by listenThread */
  static unsigned long numRespSent;
  /** The number of packets received */
  static unsigned long numRecd;
  /** Implements listenThr */
  static void* listenThread(void*);

  /** The number of LatMeas threads so far */
  static int numLatMeasThreads;
  /** The array of LatMeas threads so far */
  static class LatMeas* theThreads[MAX_PEERS];

  /** Process the incoming pkt */
  void processPkt(pkt_t* p);

  /** The thread that sends packets to the peer periodically */
  pthread_t sendThr;
  /** The number of packets sent by this thread */
  unsigned long numSent;
  /** Implements sendThr */
  static void* sendThread(void* thisPtr);
  /** Start the thread: the one for periodically sending UDP
      packets. */
  void startSendThread();

  /** The single thread that implements all the send threads */
  static pthread_t all_sendThr;
  /** The single send thread for all peers */
  static void* all_sendThread(void*);

  /** Helper for sendThread/all_sendThread */
  void sendHelper();

public:
  /** Should run listen thread or not?  NOTE: this is supposed to be
      set by the startListenThread routine */
  static bool run_listen_thread;
  /** Should run send thread(s) or not?  NOTE: this is supposed to be
      set by the start_all_sends routine */
  static bool run_send_thread;

  /** The socket descriptors for sending and listening */
  static int send_fd;
  static int listen_fd;

  /** Start the listenThread to listen on incoming packets, or prepare
      for listening if arg_run_listen_thread is false */
  static void startListenThread(bool arg_run_listen_thread);
  /** Start all the sendThreads, or the single all_sendThread, or
      prepare for sending if arg_run_send_thread is false */
  static void start_all_sends(bool arg_run_send_thread,
			      unsigned short arg_mySCID);

  /** Check for all sends -- should be called only when run_threads is
      false */
  static void all_checkSend();

  /** Receive a LatMeas pkt and deal as appropriate -- to be called
      only when there is a pkt ready to be received */
  static void recvHelper();

  /** The constructor -- the host to which latency measurements should
      be made is given */
  LatMeas(const struct in_addr &arg_peer_ipaddr, unsigned short arg_peer_scid);
  /** Return the current smoothed round trip time in 'ms'.  NOTE: this
      does not account for long lapses in measurement -- the failure
      detection mechanism is supposed to handle that.  */
  float getSRTT() { return srtt; }
  /** Return the last RTT measurement */
  float getRTT() { return rtt; }
  /** When was the last RTT measurement taken?  The return value has
      no meaning if the no measurement has been made so far (in this
      case, rtt and srtt would be 0). */
  struct timeval getLastMeas() { return lastMeas; }

  /** Stop the send threads */
  static void stopSendThreads();
  /** Stop the listen thread */
  static void stopListenThread();

  /** Return the number of pkts sent by all the threads */
  static unsigned long getNumSent();
  /** Return the number of packets received */
  static unsigned long getNumRecd();

}; // End class LatMeas

typedef class LatMeas LatMeas_t;

#endif
