/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Program to print out components of a (disconnected) graph.
    Outputs a new connected super graph. */

#include <stdio.h>

#include <iostream>
#include <fstream>
#include <vector>

#include "../util/PriorityQueue.h"
#include "../util/Util.h"
#include "../util/JHashtable.h"

typedef struct edge {
  unsigned short origin_scid; // origin of the arc
  short cost;
} edge_t;

typedef struct node {
  unsigned short scid;
  vector<edge_t> incoming;
} node_t;

typedef struct graph {
  vector<node_t> nodes;
  JHashtable<unsigned short, int> nHash;

  graph() {
    nHash.setHashFunction(ushortHash);
  }
} graph_t;

void usage(const char *progname) {
  fprintf(stderr, "Usage: %s <old-graph-file> <new-graph-file>\n",
	  progname);
} // End usage()

void readGraph(const char *graphFile, graph_t *g) {
  cerr << "reading graph file " << graphFile << endl;
  ifstream g_fin(graphFile);
  for(;;) {
    // read each line
    if(g_fin.eof()) break;
    char line[8192*2];
    g_fin.getline(line, sizeof(line)); // read a line
    if(g_fin.eof()) break;
    if(g_fin.fail()) {
      cerr << "Error reading " << graphFile << endl;
      cerr << line << endl;
      exit(99);
    }

    // line has format:
    // "<dest-scid> <origin-scid1> <cost1> <origin-scid2> <cost2>..."
    vector<JString> sVec;
    int numFields = split(line, sVec);
    if(numFields <= 0) {
      cerr << "Format of " << graphFile << " is wrong" << endl;
      exit(97);
    }
    unsigned short dest_scid = JString_atoi(sVec[0]);
    if(dest_scid == 0) {
      cerr << "dest_scid has to be > 0" << endl;
      exit(99);
    }

    if(((numFields-1)%2) != 0) {
      cerr << "odd number of fields for scid " << dest_scid
	   << " in " << graphFile << endl;
      exit(98);
    }
    int numEdges = (numFields-1)/2;

    node_t n;
    n.scid = dest_scid;

    for(int i = 0; i < numEdges; i++) {
      unsigned short origin_scid = JString_atoi(sVec[1+i*2]);
      if(origin_scid == 0) {
	cerr << "origin_scid has to be > 0" << endl;
	exit(99);
      }
      short cost = JString_atoi(sVec[1+i*2+1]);
      edge_t e;
      e.origin_scid = origin_scid;
      e.cost = cost;
      n.incoming.push_back(e);
    } // End for(i) reading all edges  

    // add the node to the graph
    g->nHash.put(n.scid, g->nodes.size());
    g->nodes.push_back(n);

  } // End for(;;) reading all lines in 'graphFile'
  g_fin.close();
} // End readGraph()

void writeGraph(const char *graphFile, graph_t *g) {
  ofstream fout(graphFile);
  for(unsigned int i = 0; i < g->nodes.size(); i++) {
    fout << g->nodes[i].scid << " ";
    for(unsigned int j = 0; j < g->nodes[i].incoming.size(); j++) {
      fout << g->nodes[i].incoming[j].origin_scid << " "
	   << g->nodes[i].incoming[j].cost << " ";
    }
    fout << '\n';
  }
  fout.close();
} // End writeGraph()

int main(int argc, char *argv[]) {
  util_init();
  if(argc != 3) { usage(argv[0]); exit(1); }

  const char *oldGraphFile = argv[1];
  const char *newGraphFile = argv[2];

  // read in the graph
  graph g;
  readGraph(oldGraphFile, &g);

  // do a dijkstra's until all nodes are covered (graph connected),
  // or until the min node has an infinite cost (graph disconnected)
  class PriorityQueue<unsigned short> pq(g.nodes.size());
  pq.decreaseCost(0, 0); // start with the 0th node

  vector<int> seenNodes; // vector of seen nodes
  unsigned short maxEdgeCost = 0; // keep track of the max edge cost

  vector<unsigned short> costs; // the min costs
  for(unsigned int i = 0; i < g.nodes.size(); i++) {
    costs.push_back(PriorityQueue<unsigned short>::INFINITY);
  }

  // set random seed
  struct timeval tv;
  gettimeofday(&tv, NULL);
  srandom(tv.tv_usec);
  fprintf(stderr, "set random seed done with %ld\n", tv.tv_usec);

  int numComponents = 1; // the number of components in the graph

  for(;;) {

    unsigned short minCost;
    int minIndex = pq.minVertex(&minCost);

    if((minIndex >= 0) && (minCost < PriorityQueue<unsigned short>::INFINITY)) {
      // expand the min node
      for(unsigned int j = 0; j < g.nodes[minIndex].incoming.size(); j++) {
	int originIndex =
	  g.nHash.get(g.nodes[minIndex].incoming[j].origin_scid);
	pq.decreaseCost(originIndex,
			minCost+g.nodes[minIndex].incoming[j].cost);
	if(maxEdgeCost < g.nodes[minIndex].incoming[j].cost) {
	  maxEdgeCost = g.nodes[minIndex].incoming[j].cost;
	}
      }

      // add min node to the list of seen nodes
      seenNodes.push_back(minIndex);

      // note the cost of the min node
      costs[minIndex] = minCost;

      // remove the min node
      pq.removeMin();

      // update the minIndex and minCost
      minIndex = pq.minVertex(&minCost);
    }

    if(minIndex < 0) {
      // priority queue is empty ==> all nodes are reachable ==> graph
      // is connected
      break;
    }

    if(minCost >= PriorityQueue<unsigned short>::INFINITY) {
      // minVertex is not in the same component as the set of seen
      // nodes

      // pick a random vertex among the seen nodes and attach up the
      // minVertex
      int randIndex = random()%(seenNodes.size());
      edge_t e1, e2;
      e1.origin_scid = g.nodes[randIndex].scid;
      e2.origin_scid = g.nodes[minIndex].scid;
      e1.cost = e2.cost = (random()%maxEdgeCost) + 1;
      g.nodes[minIndex].incoming.push_back(e1);
      g.nodes[randIndex].incoming.push_back(e2);

      cerr << "added edge " << g.nodes[randIndex].scid << " <--> "
	   << g.nodes[minIndex].scid << " (" << e1.cost << ")" << endl;

      // decrease the cost in the priority queue and move on
      pq.decreaseCost(minIndex, costs[randIndex]+e1.cost);

      // increase the count of number of components in the graph
      numComponents++;
    }

  } // End for(;;) connecting up the graph

  // write the final graph
  writeGraph(newGraphFile, &g);

  // print the number of components
  cerr << "numComponents = " << numComponents << endl;

} // End main()
