/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Implements Paths.h */

#include <stdio.h>
#include <iostream>

#include "../link-state/NetworkGraph.h"
#include "Paths.h"

const bool Paths::only_e2e = true;
const bool Paths::bunched = true;
const bool Paths::rate_control = true;
const bool Paths::joint_listen = true;

const long path_hb::hb_period = 10*1000; // 10sec
const long path_hb::hb_period_var = 1*1000; // 1sec
Uniform path_hb::U;
struct timeval path_hb::startTime;

const int Paths::MAX_PATH_ID = 0x7fff;
int Paths::lastPathID;
JHashtable<unsigned long, PathInfo_t> Paths::pathHash;
JHashtable<JString, PathInfo_t> Paths::repairHash;
JHashtable<unsigned long, PathInfo_t> Paths::altHash;
JHashtable<unsigned long, PathInfo_t> Paths::readPathHash;
Debug_t *Paths::dbg;

void Paths::init(Debug_t *arg_dbg) {
  pathHash.setHashFunction(ulongHash);
  repairHash.setHashFunction(JStringHash);
  altHash.setHashFunction(ulongHash);
  readPathHash.setHashFunction(ulongHash);
  lastPathID = 0;
  dbg = arg_dbg;
} // End Paths::init()

int Paths::getPathID() {
  for(;;) {
    // wrap around and increment lastPathID by 1
    if(lastPathID == MAX_PATH_ID) { lastPathID = 0; }
    lastPathID++;
    // check if this pathID has been taken up
    unsigned long key = lastPathID*10000+NetworkGraph::mySCID;
    //char keyStr[64];
    //sprintf(keyStr, "%d:%u", lastPathID, NetworkGraph::mySCID);
    if(pathHash.containsKey(key)) { continue; }
    else { return lastPathID; }
  }
} // End Paths::getPathID()

void Paths::recyclePathID(int pathID, unsigned short path_dest_scid) {
  unsigned long key = pathID*10000+path_dest_scid;
  //char keyStr[64];
  //sprintf(keyStr, "%d:%u", pathID, path_dest_scid);
  if(pathHash.containsKey(key)) {
    pathHash.remove(key);
  } else {
    cerr << "Trying to remove non-existent path " << key << endl;
  }
} // End Paths::recyclePathID()

void Paths::recycleAltPath(int pathID, unsigned short path_dest_scid) {
  unsigned long key = pathID*10000+path_dest_scid;
  //char keyStr[64];
  //sprintf(keyStr, "%d:%u", pathID, path_dest_scid);
  if(altHash.containsKey(key)) {
    altHash.remove(key);
  } else {
    cerr << "Trying to remove non-existent alt-path " << key << endl;
  }
} // End Paths::recycleAltPath()

void Paths::recycleRepair(int pathID, unsigned short path_dest_scid,
			  unsigned short fail_dest_scid,
			  unsigned short fail_origin_scid, short posn) {
  char keyStr[128];
  sprintf(keyStr, "%d:%u:%u:%d:%d", pathID, path_dest_scid,
	  fail_dest_scid, fail_origin_scid, (int)posn);
  if(repairHash.containsKey(keyStr)) {
    repairHash.remove(keyStr);
  } else {
    cerr << "Trying to remove non-existent repair-path " << keyStr << endl;
  }
} // End Paths::recycleRepair()

void Paths::setPathInfo(int pathID, unsigned short path_dest_scid,
			const PathInfo_t &path) {
  unsigned long key = pathID*10000+path_dest_scid;
  //char keyStr[64];
  //sprintf(keyStr, "%d:%u", pathID, path_dest_scid);
  pathHash.put(key, path);
} // End Paths::setPathInfo()

void Paths::setRepairInfo(int pathID, unsigned short path_dest_scid,
			  unsigned short fail_dest_scid,
			  unsigned short fail_origin_scid,
			  short posn, const PathInfo_t &sub_p) {
  char keyStr[128];
  sprintf(keyStr, "%d:%u:%u:%d:%d", pathID, path_dest_scid,
	  fail_dest_scid, fail_origin_scid, (int)posn);
  repairHash.put(keyStr, sub_p);
} // End Paths::setRepairInfo()

void Paths::setAltPathInfo(int pathID, unsigned short path_dest_scid,
			   const PathInfo_t &path) {
  unsigned long key = pathID*10000+path_dest_scid;
  //char keyStr[64];
  //sprintf(keyStr, "%d:%u", pathID, path_dest_scid);
  altHash.put(key, path);
} // End Paths::setAltPathInfo()

void Paths::setReadPathInfo(int pathID, unsigned short path_dest_scid,
			    const PathInfo_t &path) {
  unsigned long key = pathID*10000+path_dest_scid;
  //char keyStr[64];
  //sprintf(keyStr, "%d:%u", pathID, path_dest_scid);
  readPathHash.put(key, path);
} // End Paths::setReadPathInfo()

bool Paths::pathExists(int pathID, unsigned short path_dest_scid) {
  //char keyStr[64];
  //sprintf(keyStr, "%d:%u", pathID, path_dest_scid);
  unsigned long key = pathID*10000+path_dest_scid;
  bool entryExists = pathHash.containsKey(key);
  if(entryExists) {
    PathInfo_t &p = pathHash.get(key);
    float nowTime = now(path_hb::startTime);

    if(nowTime >= (p.lastHBRecd + 5*path_hb::hb_period)) {
      // remove any sub-path, if any
      if(p.repairExists) {
	char keyStr1[128];
	sprintf(keyStr1, "%d:%u:%u:%d:%d", pathID, path_dest_scid,
		p.fail_dest_scid, p.fail_origin_scid, (int)p.failPosn);
	dbg->setCurrFn(PATHEXISTS_FN);
	if(dbg->use_stream) {
	  dbg->print() << "timing-out-sub-path p_id=" << pathID << ':'
		       << path_dest_scid << " fail_dest_scid="
		       << p.fail_dest_scid << " fail_origin_scid="
		       << p.fail_origin_scid << " failPosn="
		       << p.failPosn << endl;
	} else {
	  LogLine_t ll; ll.desc = timing_out_sub_path_desc;
	  ll.pathID = pathID; ll.path_dest_scid = path_dest_scid;
	  ll.fail_dest_scid = p.fail_dest_scid;
	  ll.fail_origin_scid = p.fail_origin_scid;
	  ll.failPosn = p.failPosn; dbg->addLL(ll);
	}
	repairHash.remove(keyStr1);
      }

      if(dbg->use_stream) {
	dbg->print() << "timing-out-path p_id=" << pathID << ':'
		     << path_dest_scid << " path_ver=" << p.version << endl;
      } else {
	LogLine_t ll; ll.desc = timing_out_path_desc; ll.pathID = pathID;
	ll.path_dest_scid = path_dest_scid; dbg->addLL(ll);
      }
      pathHash.remove(key);
      return false;

    } else {
      return true;
    }
  } else {
    return false;
  }
} // End Paths::pathExists()

bool Paths::repairExists(int pathID, unsigned short path_dest_scid,
			 unsigned short fail_dest_scid,
			 unsigned short fail_origin_scid, short posn) {
  char keyStr[128];
  sprintf(keyStr, "%d:%u:%u:%d:%d", pathID, path_dest_scid,
	  fail_dest_scid, fail_origin_scid, (int)posn);
  return repairHash.containsKey(keyStr);
} // End Paths::repairExists()

bool Paths::altPathExists(int pathID, unsigned short path_dest_scid) {
  unsigned long key = pathID*10000+path_dest_scid;
  //char keyStr[64];
  //sprintf(keyStr, "%d:%u", pathID, path_dest_scid);
  return altHash.containsKey(key);
} // End Paths::altPathExists()

bool Paths::readPathExists(int pathID, unsigned short path_dest_scid) {
  unsigned long key = pathID*10000+path_dest_scid;
  //char keyStr[64];
  //sprintf(keyStr, "%d:%u", pathID, path_dest_scid);
  return readPathHash.containsKey(key);
} // End Paths::readPathExists()

PathInfo_t &Paths::getPathInfo(int pathID,
			       unsigned short path_dest_scid) {
  unsigned long key = pathID*10000+path_dest_scid;
  //char keyStr[64];
  //sprintf(keyStr, "%d:%u", pathID, path_dest_scid);
  return pathHash.get(key);
} // End Paths::getPathInfo()

PathInfo_t &Paths::getRepairInfo(int pathID, unsigned short path_dest_scid,
				 unsigned short fail_dest_scid,
				 unsigned short fail_origin_scid,
				 short posn) {
  char keyStr[128];
  sprintf(keyStr, "%d:%u:%u:%d:%d", pathID, path_dest_scid,
	  fail_dest_scid, fail_origin_scid, (int)posn);
  return repairHash.get(keyStr);
} // End Paths::getRepairInfo()

PathInfo_t &Paths::getAltPathInfo(int pathID,
				  unsigned short path_dest_scid) {
  unsigned long key = pathID*10000+path_dest_scid;
  //char keyStr[64];
  //sprintf(keyStr, "%d:%u", pathID, path_dest_scid);
  return altHash.get(key);
} // End Paths::getAltPathInfo()

PathInfo_t &Paths::getReadPathInfo(int pathID,
				   unsigned short path_dest_scid) {
  unsigned long key = pathID*10000+path_dest_scid;
  //char keyStr[64];
  //sprintf(keyStr, "%d:%u", pathID, path_dest_scid);
  return readPathHash.get(key);
} // End Paths::getReadPathInfo()
