/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
#ifndef PATHS_H
#define PATHS_H

#include <queue>
#include <vector>

#include "../util/JHashtable.h"
#include "../util/JString.h"
#include "../util/Util.h"
#include "../util/Debug.h"
#include "../service-locn/AppPkt.h"

struct path_hb {
  /** This is the HB period, in ms */
  const static long hb_period;
  /** This is the variation in the HB period, to avoid synchronization */
  const static long hb_period_var;
  /** To randomize the HB send time */
  static Uniform U;
  /** This is when the program started */
  static struct timeval startTime;

  static void init();

  /** The pathID for which HB has to be sent */
  int pathID;
  /** The version for which HB has to be sent */
  unsigned short version;
  /** The time at which HB has to be sent, w.r.t startTime */
  float whenToSend;
};
typedef struct path_hb path_hb_t;

struct ServInst {
  GlobalServInfo_t gInfo;
  JString sName; // the service name
  JString dest_spec; // the destn. spec. of this service instance
}; // End struct ServInst
typedef struct ServInst ServInst_t;

struct PathInfo {
  // the physical path -- downstream to upstream
  vector<ServInst_t> physPath;
  // the number of elements of the physPath that have been created
  // numCreated is 'n' ==> the first 'n' downstream instances of the
  // path have been created
  unsigned short numCreated;

  /** This is the time when the last HB was received - relative to
      'ServComp::startTime' (in ms) */
  unsigned long lastHBRecd;
  /** This is the 'posn' in the path from where the last heart-beat
      was sent OR forwarded -- used for detecting duplicates */
  unsigned short lastHBPosn;

  /* variables for local path recovery */
  // does the path have any repaired part passing through this node?
  bool repairExists;
  // if 'repairExists' is true, this is the sub-path for which repair exists
  unsigned short fail_dest_scid;
  unsigned short fail_origin_scid;
  short failPosn;

  /* 'path_origin_scid', 'dest_spec', and 'services' are required for
     e2e path recovery (need to be set only at the 'path_dest_scid' */
  // this is the path_origin_scid (could be INVALID_SCID if the app
  // did not request for a specific origin)
  unsigned short path_origin_scid;
  // this is the destination specified by the application
  JString dest_spec;
  // this is the vector of services requested for creation by the app
  vector<JString> services;
  // this is the path version
  unsigned short version;
  PathInfo() {
    lastHBPosn = 0;
    numCreated = 0;
    repairExists = false;
    version = 100;
  }
};
typedef struct PathInfo PathInfo_t;

/** Class to maintain path information for all paths running through
    the service cluster */
class Paths {
protected:
  /** The debug object */
  static Debug_t *dbg;

  /** The maximum value of pathID */
  const static int MAX_PATH_ID;
  /** The last pathID that was given out */
  static int lastPathID;

  /** The hash of pathID to path information.  The 'key' for this
      hashtable is the JString "pathID:path_dest_scid". */
  static JHashtable<unsigned long, PathInfo_t> pathHash;
  /** The information for paths being repaired currently.  The 'key'
      for this hashtable is the JString
      "pathID:path_dest_scid:fail_dest_scid:fail_origin_scid:posn".
      'posn' here corresponds to the position of the fail_dest_scid in
      the original path. */
  static JHashtable<JString, PathInfo_t> repairHash;
  /** The hash of pathID to path information -- for alternate paths
      that are being created now.  The 'key' for this hashtable is the
      JString "pathID:path_dest_scid". */
  static JHashtable<unsigned long, PathInfo_t> altHash;
  /** Paths read in from a file */
  static JHashtable<unsigned long, PathInfo_t> readPathHash;

public:
  /** Should use only end-to-end recovery? */
  const static bool only_e2e;
  /** Should I do bunched recovery? */
  const static bool bunched;
  /** Should I do rate control in recovery? */
  const static bool rate_control;
  /** Should I do a joint listen? */
  const static bool joint_listen;

  /** This should be called before any of the other routines will work
      (not enforced). */
  static void init(Debug_t *arg_dbg);

  /** Return a new pathID for assigning to mySCID */
  static int getPathID();

  /** Can reuse the given pathID/path_dest_scid pair */
  static void recyclePathID(int pathID, unsigned short path_dest_scid);
  /** The alt path does not exist anymore */
  static void recycleAltPath(int pathID, unsigned short path_dest_scid);
  /** Recycle the repair info */
  static void recycleRepair(int pathID, unsigned short path_dest_scid,
			    unsigned short fail_dest_scid,
			    unsigned short fail_origin_scid, short posn);

  /** Store the given path information vector under the given
      'pathID', and the given 'path_dest_scid' */
  static void setPathInfo(int pathID, unsigned short path_dest_scid,
			  const PathInfo_t &path);
  /** Store the given path repair information */
  static void setRepairInfo(int pathID, unsigned short path_dest_scid,
			    unsigned short fail_dest_scid,
			    unsigned short fail_origin_scid,
			    short posn, const PathInfo_t &sub_p);
  /** Store the given alternate path information vector under the
      given 'pathID', and the given 'path_dest_scid' */
  static void setAltPathInfo(int pathID, unsigned short path_dest_scid,
			     const PathInfo_t &path);
  /** Store the given path info in the set of paths read in from a
      file */
  static void setReadPathInfo(int pathID, unsigned short path_dest_scid,
			      const PathInfo_t &path);

  /** Return true if the given path exists, false otherwise */
  static bool pathExists(int pathID, unsigned short path_dest_scid);
  /** Return true if the given repair path exists, false otherwise */
  static bool repairExists(int pathID, unsigned short path_dest_scid,
			   unsigned short fail_dest_scid,
			   unsigned short fail_origin_scid, short posn);
  /** Return true if the given alternate path exists, false otherwise */
  static bool altPathExists(int pathID, unsigned short path_dest_scid);
  /** Returns true if the given path exists in the list of read-in
      paths */
  static bool readPathExists(int pathID, unsigned short path_dest_scid);

  /** Return the stored path information for the given 'pathID' and
      the given 'path_dest_scid' */
  static PathInfo_t &getPathInfo(int pathID,
				 unsigned short path_dest_scid);
  /** Return the stored path information for the given tuple */
  static PathInfo_t &getRepairInfo(int pathID,
				   unsigned short path_dest_scid,
				   unsigned short fail_dest_scid,
				   unsigned short fail_origin_scid,
				   short posn);
  /** Return the stored alternate path information for the given
      'pathID' and the given 'path_dest_scid' */
  static PathInfo_t &getAltPathInfo(int pathID,
				    unsigned short path_dest_scid);
  /** Return the read-in path info */
  static PathInfo_t &getReadPathInfo(int pathID,
				     unsigned short path_dest_scid);

}; // End class Paths

#endif
