/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Implements the service composition recovery algorithms */

#include <netdb.h>
#include <stdio.h>
#include <unistd.h>
#include <stdio.h>
#include <iostream>

#include "ServComp.h"
#include "TreeInfo.h"
#include "Paths.h"
#include "Bunch.h"
#include "../link-state/NetworkGraph.h"
#include "../service-locn/ServiceLocn.h"

void ServComp::checkFailures() {
  dbg->setCurrFn(CHECKFAILURES_FN);

  // get list of peers
  unsigned short numPeers;
  vector<Peer_t> &peer_list = NetworkGraph::getPeerList(&numPeers);

  for(unsigned short i = 0; i < numPeers; i++) {

    // Note: need to read status off the network graph (LS layer), not
    // directly from the HB layer
    unsigned short currStatus = LIVE;
    if(NetworkGraph::getEdgeCost(NetworkGraph::mySCID, peer_list[i].scid) ==
       PriorityQueue<unsigned short>::INFINITY) {
      currStatus = DEAD;
    }
    unsigned short currMyStatusPeerView = LIVE;
    if(NetworkGraph::getEdgeCost(peer_list[i].scid, NetworkGraph::mySCID) ==
       PriorityQueue<unsigned short>::INFINITY) {
      currMyStatusPeerView = DEAD;
    }

    if((peer_list[i].lastStatus == currStatus) &&
       (peer_list[i].lastMyStatusPeerView == currMyStatusPeerView)) {
      // nothing has changed
      continue;
    }

    if((peer_list[i].lastStatus == LIVE) && (currStatus == DEAD)) {
      if(dbg->use_stream) {
	dbg->print() << "incoming-dead peerSCID=" << peer_list[i].scid
		     << endl;
      } else {
	LogLine_t ll; ll.desc = incoming_dead_desc;
	ll.peerSCID = peer_list[i].scid; dbg->addLL(ll);
      }
      if(Paths::rate_control) { patchPathsFromPeerRC(&peer_list[i]); }
      else { patchPathsFromPeer(&peer_list[i]); }
      peer_list[i].lastStatus = currStatus;
      peer_list[i].lastMyStatusPeerView = currMyStatusPeerView;
      continue;
    }

    if((peer_list[i].lastMyStatusPeerView == LIVE) &&
       (currMyStatusPeerView == DEAD)) {
      if(dbg->use_stream) {
	dbg->print() << "outgoing-dead peerSCID=" << peer_list[i].scid
		     << endl;
      } else {
	LogLine_t ll; ll.desc = outgoing_dead_desc;
	ll.peerSCID = peer_list[i].scid; dbg->addLL(ll);
      }
      if(Paths::rate_control) { patchPathsFromPeerRC(&peer_list[i]); }
      else { patchPathsFromPeer(&peer_list[i]); }
      peer_list[i].lastStatus = currStatus;
      peer_list[i].lastMyStatusPeerView = currMyStatusPeerView;
      continue;
    }

    if((peer_list[i].lastStatus == DEAD) && (currStatus == LIVE)) {
      if(dbg->use_stream) {
	dbg->print() << "incoming-restored peerSCID=" << peer_list[i].scid
		     << endl;
      } else {
	LogLine_t ll; ll.desc = incoming_restored_desc;
	ll.peerSCID = peer_list[i].scid; dbg->addLL(ll);
      }
    }

    if((peer_list[i].lastMyStatusPeerView == DEAD) &&
       (currMyStatusPeerView == LIVE)) {
      if(dbg->use_stream) {
	dbg->print() << "outgoing-restored peerSCID=" << peer_list[i].scid
		     << endl;
      } else {
	LogLine_t ll; ll.desc = incoming_restored_desc;
	ll.peerSCID = peer_list[i].scid; dbg->addLL(ll);
      }
    }

    peer_list[i].lastStatus = currStatus;
    peer_list[i].lastMyStatusPeerView = currMyStatusPeerView;
  }

} // End ServComp::checkFailures()

void ServComp::patchPathsFromPeer(Peer_t *peer) {
  dbg->setCurrFn(PATCHPATHSFROMPEER_FN);

  vector<path_id_t> *toPatch = peer->getPathList();
  if(toPatch == NULL) {
    // nothing to patch
    if(dbg->use_stream) {
      dbg->print() << "nothing-to-patch peerSCID=" << peer->scid << endl;
    } else {
      LogLine_t ll; ll.desc = nothing_to_patch_desc;
      ll.peerSCID = peer->scid; dbg->addLL(ll);
    }
    return;
  }

  // call patchPath on each
  Bunch_t b;
  b.fail_dest_scid = NetworkGraph::mySCID;
  b.fail_origin_scid = peer->scid;
  Bunch_t *bPtr = (Paths::bunched) ? &b : NULL;
  for(unsigned int i = 0; i < toPatch->size(); i++) {
    patchPath(peer, &(*toPatch)[i], bPtr);
  }

  if(Paths::bunched) {
    sendBunch(&b);
  }

} // End ServComp::patchPathsFromPeer()

void ServComp::patchPathsFromPeerRC(Peer_t *peer) {
  dbg->setCurrFn(PATCHPATHSFROMPEERRC_FN);

  float timeDiff = now(peer->last_patch);
  if(timeDiff < 5000) { return; }

  vector<path_id_t> *toPatch = peer->getPathList();
  if(toPatch == NULL) {
    // nothing to patch
    if(dbg->use_stream) {
      dbg->print() << "nothing-to-patch peerSCID=" << peer->scid << endl;
    } else {
      LogLine_t ll; ll.desc = nothing_to_patch_desc;
      ll.peerSCID = peer->scid; dbg->addLL(ll);
    }
    return;
  }

  if(dbg->use_stream) {
    dbg->print() << "detect-time fail_origin_scid=" << peer->scid
		 << " num_to_patch=" << toPatch->size() << endl;
  } else {
    LogLine_t ll; ll.desc = detect_time_desc; ll.fail_origin_scid = peer->scid;
    ll.utype = PATCH_U; ll.u.patch.init();
    ll.u.patch.num_to_patch = toPatch->size(); dbg->addLL(ll);
  }

  for(unsigned int i = 0; i < toPatch->size(); i++) {
    peer->patchQ.push((*toPatch)[i]);
  }
} // End ServComp::patchPathsFromPeerRC()

void ServComp::checkPatchQueues() {
  // get list of peers
  unsigned short numPeers;
  vector<Peer_t> &peer_list = NetworkGraph::getPeerList(&numPeers);
  for(unsigned int i = 0; i < numPeers; i++) {
    checkPatchQ(&peer_list[i]);
  }
} // End ServComp::checkPatchQueues()

void ServComp::checkPatchQ(Peer_t *peer) {
  dbg->setCurrFn(CHECKPATCHQ_FN);

  if(peer->patchQ.empty()) { return; }

  float timeDiff = now(peer->last_patch);
  if(timeDiff < min_patch_interval) { return; }

  unsigned short cost =
    NetworkGraph::getEdgeCost(NetworkGraph::mySCID, peer->scid);
  if(cost < PriorityQueue<unsigned short>::INFINITY) {
    if(dbg->use_stream) {
      dbg->print() << "stale-to-patch fail_origin_scid=" << peer->scid
		   << endl;
    } else {
      LogLine_t ll; ll.desc = stale_to_patch_desc;
      ll.fail_origin_scid = peer->scid; dbg->addLL(ll);
    }
    return;
  }

  Bunch_t b;
  b.fail_dest_scid = NetworkGraph::mySCID;
  b.fail_origin_scid = peer->scid;
  unsigned int num_patched_now = 0;

  for(;;) {
    if(peer->patchQ.empty()) { break; }
    path_id_t tp = peer->patchQ.front();
    peer->patchQ.pop();

    // call patchPath on each
    Bunch_t *bPtr = (Paths::bunched) ? &b : NULL;
    patchPath(peer, &tp, bPtr);
    num_patched_now++;
    if(num_patched_now == max_to_patch_now) { break; }
  }

  if(Paths::bunched) {
    sendBunch(&b);
  }

  gettimeofday(&peer->last_patch, NULL);
} // End ServComp::checkPatchQ()

void ServComp::patchPath(Peer_t *peer, path_id_t *path, Bunch_t *b) {
  dbg->setCurrFn(PATCHPATH_FN);

  bool pathExists = Paths::pathExists(path->pathID, path->path_dest_scid);
  if(!pathExists) {
    if(dbg->use_stream) {
      dbg->print() << "del-non-existent-path p_id=" << path->pathID
		   << ':' << path->path_dest_scid << endl;
    } else {
      LogLine_t ll; ll.desc = del_non_existent_path_desc;
      ll.pathID = path->pathID; ll.path_dest_scid = path->path_dest_scid;
      dbg->addLL(ll);
    }
    peer->delPath(path->pathID, path->path_dest_scid);
    return;
  }

  PathInfo_t &pInfo = Paths::getPathInfo(path->pathID, path->path_dest_scid);
  /* SUPRESS
  if(dbg->use_stream) {
    dbg->print() << "try-patch p_id=" << path->pathID << ':'
		 << path->path_dest_scid << " path_ver="
		 << pInfo.version << endl;
  } else {
    LogLine_t ll; ll.desc = try_patch_desc; ll.pathID = path->pathID;
    ll.path_dest_scid = path->path_dest_scid; ll.version = pInfo.version;
    dbg->addLL(ll);
  }
  */
  if(pInfo.numCreated != pInfo.physPath.size()) {
    // path not fully created
    // just let things timeout and fail -- path will (should) be
    // re-created by the 'path_dest_scid' node
    if(dbg->use_stream) {
      dbg->print() << "incomplete-path" << endl;
    } else {
      LogLine_t ll; ll.desc = incomplete_path_desc; dbg->addLL(ll);
    }
    return;
  }

  if(pInfo.repairExists) {
    // This path has already been repaired locally.  In fact, mySCID
    // could be part of the patched sub-path.  This situation has to
    // be handled end-2-end since local recovery would be too
    // complicated.
    if(dbg->use_stream) {
      dbg->print() << "repair-exists-try-e2e" << endl;
    } else {
      LogLine_t ll; ll.desc = repair_exists_try_e2e_desc; dbg->addLL(ll);
    }
    patchEdgeE2E(peer, path, pInfo, b);
    return;
  }

  // find the (first) position of this edge in the path
  int posn = -1;
  int rev_posn = -1; // first position of the reverse edge
  bool dup = false; // to see if this edge occurs twice in the path
  for(unsigned int i = 0; i < pInfo.physPath.size()-1; i++) {
    if((pInfo.physPath[i].gInfo.scid == NetworkGraph::mySCID) &&
       (pInfo.physPath[i+1].gInfo.scid == peer->scid)) {
      // this is a failed edge
      if(posn != -1) { dup = true; }
      else { posn = i; }
    }
    if((pInfo.physPath[i+1].gInfo.scid == NetworkGraph::mySCID) &&
       (pInfo.physPath[i].gInfo.scid == peer->scid)) {
      // this is the back-edge of the failed edge
      if(posn != -1) { dup = true; }
      if(rev_posn == -1) { rev_posn = i; }
    }
  }

  // check if upstream edge also occurs downstream ==> will be fixed
  // by the peer, and i can ignore this failure
  if((posn != -1) && (rev_posn != -1) && (rev_posn < posn)) {
    if(dbg->use_stream) {
      dbg->print() << "leave-for-downstream" << endl;
    } else {
      LogLine_t ll; ll.desc = leave_for_downstream_desc; dbg->addLL(ll);
    }
    return;
  }

  if(dup) {
    // this edge occurs more than once in the path, then local recovery
    // is too complicated
    if(dbg->use_stream) {
      dbg->print() << "edge-repeat-try-e2e" << endl;
    } else {
      LogLine_t ll; ll.desc = edge_repeat_try_e2e_desc; dbg->addLL(ll);
    }
    patchEdgeE2E(peer, path, pInfo, b);
    return;
  }

  if(posn == -1) {
    // this is an error condition -- can actually happen because paths
    // are never cleaned up
    if(dbg->use_stream) {
      dbg->print() << "non-existent-edge-del-path" << endl;
    } else {
      LogLine_t ll; ll.desc = non_existent_edge_del_path_desc; dbg->addLL(ll);
    }
    peer->delPath(path->pathID, path->path_dest_scid);
    return;
  }

  // this edge occurs only once in the path, and no other repair has
  // been done on the path so far -- try a local repair
  if(Paths::only_e2e) {
    patchEdgeE2E(peer, path, pInfo, b);
  } else {
    patchEdgeLocal(peer, path, pInfo, posn, b);
  }

} // End ServComp::patchPath()

void ServComp::patchEdgeLocal(Peer_t *peer, path_id_t *path,
			      PathInfo_t &pInfo, unsigned int posn,
			      Bunch_t *b) {
  dbg->setCurrFn(PATCHEDGELOCAL_FN);

  unsigned short fail_origin_scid = peer->scid;
  unsigned short fail_dest_scid = NetworkGraph::mySCID;

  if(dbg->use_stream) {
    dbg->print() << "try-local fail_origin_scid=" << fail_origin_scid
		 << " p_id=" << path->pathID << ':' << path->path_dest_scid
		 << " path_ver=" << pInfo.version << " failPosn="
		 << posn << endl;
  } else {
    LogLine_t ll; ll.desc = try_local_desc; ll.pathID = path->pathID;
    ll.path_dest_scid = path->path_dest_scid;
    ll.fail_origin_scid = fail_origin_scid;
    ll.version = pInfo.version; ll.failPosn = posn;
    dbg->addLL(ll);
  }

  // call findPath on the empty path, starting from peer->scid
  vector<JString> emptyPath;
  PathInfo_t sub_p; // the new sub path to replace the failed edge
  vector<ServInst_t> &newSubPath = sub_p.physPath;

  bool forceRebuild = false;
  int pathLen = findPath(peer->scid, emptyPath, newSubPath,
			 forceRebuild, fail_dest_scid, fail_origin_scid);
  if((pathLen != -1) && !checkPath(newSubPath)) {
    if(dbg->use_stream) {
      dbg->print() << "graph-state-stale" << endl;
    } else {
      LogLine_t ll; ll.desc = graph_state_stale_desc; dbg->addLL(ll);
    }
    // discard the old physPath
    while(newSubPath.size() > 0) { newSubPath.pop_back(); }
    // try to find new one with current graph state
    forceRebuild = true;
    pathLen = findPath(peer->scid, emptyPath, newSubPath,
		       forceRebuild, fail_dest_scid, fail_origin_scid);
  }

  if(pathLen == -1) {
    if(dbg->use_stream) {
      dbg->print() << "no-replacement-try-e2e fail_origin_scid="
		   << peer->scid << endl;
    } else {
      LogLine_t ll; ll.desc = no_replacement_try_e2e_desc;
      ll.fail_origin_scid = peer->scid; dbg->addLL(ll);
    }
    patchEdgeE2E(peer, path, pInfo, b);
    return;
  }

  // check that this new path is of length at least 3, and other
  // sanity checks
  if(newSubPath.size() < 3) {
    if(dbg->use_stream) {
      ostream &fout = dbg->print();
      fout << "new-sub-path-weird-try-e2e size=" << newSubPath.size()
	   << " peers=[";
      unsigned short numPeers;
      vector<Peer_t> &peer_list = NetworkGraph::getPeerList(&numPeers);
      for(unsigned int i = 0; i < numPeers; i++) {
	short peer_status = LIVE;
	if(NetworkGraph::getEdgeCost(NetworkGraph::mySCID, peer_list[i].scid) ==
	   PriorityQueue<unsigned short>::INFINITY) {
	  peer_status = DEAD;
	}
	short my_status_peer_view = LIVE;
	if(NetworkGraph::getEdgeCost(peer_list[i].scid, NetworkGraph::mySCID) ==
	   PriorityQueue<unsigned short>::INFINITY) {
	  peer_status = DEAD;
	}
	fout << peer_list[i].scid << ':'
	     << ((peer_status == DEAD) ? "DEAD" : "LIVE") << ':'
	     << ((my_status_peer_view == DEAD) ? "DEAD" : "LIVE") << ' ';
      }
      fout << ']' << endl;
    } else {
      LogLine_t ll; ll.desc = new_sub_path_weird_try_e2e_desc;
      dbg->addLL(ll);
    }
    patchEdgeE2E(peer, path, pInfo, b);
    return;
  }

  if((newSubPath[0].gInfo.scid != NetworkGraph::mySCID) ||
     (newSubPath[newSubPath.size()-1].gInfo.scid != peer->scid)) {
    if(dbg->use_stream) {
      dbg->print() << "new-sub-path-weird destSCID=" << newSubPath[0].gInfo.scid
		   << " originSCID="
		   << newSubPath[newSubPath.size()-1].gInfo.scid
		   << " fail_origin_scid=" << peer->scid << endl;
    } else {
      LogLine_t ll; ll.desc = new_sub_path_weird_desc;
      ll.dest_scid = newSubPath[0].gInfo.scid;
      ll.origin_scid = newSubPath[newSubPath.size()-1].gInfo.scid;
      ll.fail_origin_scid = peer->scid; dbg->addLL(ll);
    }
    dbg->close(); exit(113);
    return;
  }

  // check that this sub-path + old path do not have any repeat nodes
  if(checkRepeat(pInfo.physPath, newSubPath)) {
    if(dbg->use_stream) {
      dbg->print() << "node-repeat-try-e2e" << endl;
    } else {
      LogLine_t ll; ll.desc = node_repeat_try_e2e_desc; dbg->addLL(ll);
    }
    patchEdgeE2E(peer, path, pInfo, b);
    return;
  }

  char valStr[1024]; char *lastPos = valStr; *lastPos = '['; lastPos++;
  for(unsigned int i = 0; i < newSubPath.size(); i++) {
    lastPos = longPrint(lastPos, newSubPath[i].gInfo.scid);
    *lastPos = ':'; lastPos++;
    lastPos = my_strcpy(lastPos, newSubPath[i].sName.cstr());
    *lastPos = ' '; lastPos++;
  }
  *lastPos = ']'; lastPos++; *lastPos = 0;
  if(dbg->use_stream) {
    dbg->print() << "replacement-found newSubPath=" << valStr << endl;
  } else {
    LogLine_t ll; ll.desc = replacement_found_desc;
    ll.addPair(NEWSUBPATH_LK, valStr); dbg->addLL(ll);
  }

  /*---- send request upstream ----*/

  // form the packet to send
  ServCompIFPkt_t scPkt;
  UDPLibPkt_t *uPkt = listen_pktBlock->getPkt();
  if(uPkt == NULL) {
    fprintf(stderr, "patchEdgeLocal: can't alloc for uPkt\n");
    exit(77);
  }
  scPkt.fixed.type = PEER_PATCH_PATH;
  scPkt.fixed.repType = LOCAL_R;
  scPkt.fixed.path_dest_scid = path->path_dest_scid;
  scPkt.fixed.pathID = path->pathID;
  scPkt.fixed.fail_origin_scid = fail_origin_scid;
  scPkt.fixed.fail_dest_scid = fail_dest_scid;
  scPkt.addPair(PREV_DEST_SPEC_F, pInfo.physPath[posn].dest_spec.cstr());
  scPkt.fixed.version = pInfo.version;
  char oldPath[1024]; char *lastOld = oldPath;
  for(unsigned int i = 0; i < pInfo.physPath.size(); i++) {
    lastOld = longPrint(lastOld, pInfo.physPath[i].gInfo.scid);
    *lastOld = ' '; *lastOld++;
    lastOld = my_strcpy(lastOld, pInfo.physPath[i].sName.cstr());
    *lastOld = ' '; lastOld++;
  }
  *lastOld = 0;
  char newSubPathStr[1024]; char *lastNew = newSubPathStr;
  // Note: NOT stripping off the two ends of the newSubPath (they will
  // be the ends of the 'edge' being repaired now)
  for(unsigned int j = 0; j < newSubPath.size(); j++) {
    lastNew = longPrint(lastNew, newSubPath[j].gInfo.scid);
    *lastNew = ' '; lastNew++;
    lastNew = my_strcpy(lastNew, newSubPath[j].sName.cstr());
    *lastNew = ' '; lastNew++;
  }
  *lastNew = 0;
  scPkt.addPair(OLD_PATH_F, oldPath);
  scPkt.addPair(NEW_SUB_PATH_F, newSubPathStr);
  scPkt.fixed.failPosn = posn;
  int uPktLen = scPkt.toStr(uPkt->data, sizeof(uPkt->data));

  // get the address to send
  unsigned short nextSCID = newSubPath[1].gInfo.scid;
  Peer_t *peer1 = NetworkGraph::getPeerWithSCID(nextSCID);

  if(dbg->use_stream) {
    dbg->print() << "pass-req-upstream nextSCID=" << nextSCID << endl;
  } else {
    LogLine_t ll; ll.desc = pass_req_upstream_desc;
    ll.nextSCID = nextSCID; dbg->addLL(ll);
  }

  // send the packet
  UDPLibPkt_t *u_ackPkt = listen_pktBlock->getPkt();
  if(u_ackPkt == NULL) {
    fprintf(stderr, "patchEdgeLocal: can't alloc for u_ackPkt\n");
    exit(78);
  }
  struct timeval timeout = peer1->getServCompTimeout();
  int send_fd = listen_fd; // can use same socket for sending too
  bool use_tunnel = true;
  int reqNo = listen_rrBlock->insertRequest
    (send_fd, uPkt, uPktLen, peer1->ip_addr, SERV_COMP_PORT,
     ack_listen_fd, u_ackPkt, timeout, use_tunnel, peer1->scid,
     NetworkGraph::mySCID);
  if(reqNo == -1) {
    fprintf(stderr, "patchEdgeLocal: no space for request\n");
    exit(79);
  }

  // add the 'being patched' path info to database
  sub_p.physPath[0] = pInfo.physPath[posn];
  Paths::setRepairInfo(path->pathID, path->path_dest_scid,
		       fail_dest_scid, fail_origin_scid, posn, sub_p);
  pInfo.repairExists = true;
  pInfo.fail_dest_scid = fail_dest_scid;
  pInfo.fail_origin_scid = fail_origin_scid;
  pInfo.failPosn = posn;

  // add the reqNo to the queue, also remember the pkt for the
  // ack, and the peer from whom the original request was recd
  UDPLibPkt_t *reqRaw = NULL; // no request recd from anyone
  struct in_addr ackIPAddr; // to whom ack must be sent
  ackIPAddr.s_addr = inet_addr("0.0.0.0");
  bool local = true;
  Peer::sc_repair_addReq(reqNo, reqRaw, ackIPAddr, 0, path->pathID,
			 path->path_dest_scid, fail_dest_scid,
			 fail_origin_scid, posn, local);
  // ack should trigger replacement of old path info with new path info

} // End ServComp::patchEdgeLocal()

void ServComp::patchEdgeE2E(Peer_t *peer, path_id_t *path,
			    PathInfo_t &pInfo, Bunch_t *b) {

  unsigned short fail_origin_scid = peer->scid;
  unsigned short fail_dest_scid = NetworkGraph::mySCID;

  patchEdgeE2EHelper(peer, path, pInfo, fail_dest_scid,
		     fail_origin_scid, pInfo.version, b);

} // End ServComp::patchEdgeE2E()

void ServComp::patchEdgeE2EHelper(Peer_t *peer, path_id_t *path,
				  PathInfo_t &pInfo,
				  unsigned short fail_dest_scid,
				  unsigned short fail_origin_scid,
				  unsigned short fail_path_ver,
				  Bunch_t *b) {

  dbg->setCurrFn(PATCHEDGEE2EHELPER_FN);

  unsigned short path_dest_scid = path->path_dest_scid;
  int pathID = path->pathID;

  if(dbg->use_stream) {
    dbg->print() << "e2e-patch p_id=" << path->pathID << ':'
		 << path->path_dest_scid << " fail_dest_scid="
		 << fail_dest_scid << " fail_origin_scid=" << fail_origin_scid
		 << " fail_path_ver=" << fail_path_ver
		 << " peerSCID=" << peer->scid << endl;
  } else {
    LogLine_t ll; ll.desc = e2e_patch_desc; ll.pathID = path->pathID;
    ll.path_dest_scid = path->path_dest_scid;
    ll.fail_dest_scid = fail_dest_scid;
    ll.fail_origin_scid = fail_origin_scid; ll.fail_version = fail_path_ver;
    ll.peerSCID = peer->scid; dbg->addLL(ll);
  }

  unsigned short nextSCID = NetworkGraph::INVALID_SCID;

  // find the broken edge in the path
  int failPosn = -1;
  int failSubPosn = -1;
  for(unsigned int i = 0; i < pInfo.physPath.size()-1; i++) {
    if((pInfo.physPath[i].gInfo.scid == fail_dest_scid) &&
       (pInfo.physPath[i+1].gInfo.scid == fail_origin_scid)) {
      failPosn = i; break;
    }
  }
  if((failPosn == -1) && (pInfo.repairExists)) {
    PathInfo_t &repInfo = Paths::getRepairInfo
      (pathID, path_dest_scid, pInfo.fail_dest_scid,
       pInfo.fail_origin_scid, pInfo.failPosn);
    for(unsigned int i = 0; i < repInfo.physPath.size()-1; i++) {
      if((repInfo.physPath[i].gInfo.scid == fail_dest_scid) &&
	 (repInfo.physPath[i+1].gInfo.scid == fail_origin_scid)) {
	failSubPosn = i; break;
      }
    }
  }

  // find my position in the path
  int myPosn = -1;
  for(unsigned int i = 0; i < pInfo.physPath.size(); i++) {
    if(pInfo.physPath[i].gInfo.scid == NetworkGraph::mySCID) {
      myPosn = i; break;
    }
  }

  // check path version
  if(fail_path_ver != pInfo.version) {
    if(dbg->use_stream) {
      dbg->print() << "version-mismatch fail_path_ver=" << fail_path_ver
		   << " path_ver=" << pInfo.version << endl;
    } else {
      LogLine_t ll; ll.desc = version_mismatch_desc;
      ll.fail_version = fail_path_ver; ll.version = pInfo.version;
      dbg->addLL(ll);
    }
    //if(fail_path_ver > pInfo.version) { exit(125); }
    return;
  }

  if(path->path_dest_scid == NetworkGraph::mySCID) {
    if((failPosn == -1) && (failSubPosn == -1)) {
      // edge is non-existent -- must be an old path
      if(dbg->use_stream) {
	dbg->print() << "non-existent-edge-ignoring" << endl;
      } else {
	LogLine_t ll; ll.desc = non_existent_edge_ignoring_desc; dbg->addLL(ll);
      }
    }
    // i have to create the new path
    if(dbg->use_stream) {
      dbg->print() << "create-alt-path" << endl;
    } else {
      LogLine_t ll; ll.desc = create_alt_path_desc; dbg->addLL(ll);
    }
    createAltPath(peer, path, pInfo, fail_dest_scid, fail_origin_scid);
    return;
  }

  if(myPosn != -1) {
    // i am part of the original path
    nextSCID = pInfo.physPath[myPosn-1].gInfo.scid;

    // check the case when the downstream edge was already broken and
    // was fixed earlier
    if((pInfo.repairExists) &&
       (pInfo.fail_dest_scid == nextSCID) &&
       (pInfo.fail_origin_scid == NetworkGraph::mySCID)) {
      PathInfo_t &repInfo =
	Paths::getRepairInfo(pathID, path_dest_scid,
			     pInfo.fail_dest_scid,
			     pInfo.fail_origin_scid, pInfo.failPosn);
      unsigned int subPathSize = repInfo.physPath.size();
      nextSCID = repInfo.physPath[subPathSize-2].gInfo.scid;
      if(dbg->use_stream) {
	dbg->print() << "am-upstream-of-sub-path nextSCID=" << nextSCID << endl;
      } else {
	LogLine_t ll; ll.desc = am_upstream_of_sub_path_desc;
	ll.nextSCID = nextSCID; dbg->addLL(ll);
      }
    } else {
      if(dbg->use_stream) {
	dbg->print() << "am-part-of-orig-path nextSCID=" << nextSCID << endl;
      } else {
	LogLine_t ll; ll.desc = am_part_of_orig_path_desc;
	ll.nextSCID = nextSCID; dbg->addLL(ll);
      }
    }

  } else {
    // i am part of a repaired sub path
    int mySubPosn = -1;

    if(pInfo.repairExists) {

      PathInfo_t &repInfo = Paths::getRepairInfo
	(pathID, path_dest_scid, pInfo.fail_dest_scid,
	 pInfo.fail_origin_scid, pInfo.failPosn);

      for(unsigned int i = 1; i < repInfo.physPath.size()-1; i++) {
	if(repInfo.physPath[i].gInfo.scid == NetworkGraph::mySCID) {
	  mySubPosn = i; break;
	}
      }

      nextSCID = repInfo.physPath[mySubPosn-1].gInfo.scid;
      if(dbg->use_stream) {
	dbg->print() << "am-in-repaired-sub-path nextSCID=" << nextSCID << endl;
      } else {
	LogLine_t ll; ll.desc = am_in_repaired_sub_path_desc;
	ll.nextSCID = nextSCID; dbg->addLL(ll);
      }
    }

    if(mySubPosn <= 0) {
      if(dbg->use_stream) {
	dbg->print() << "no-recognized-sub-path" << endl;
      } else {
	LogLine_t ll; ll.desc = no_recognized_sub_path_desc; dbg->addLL(ll);
      }
      return;
    }

  }

  // If bunched info has to be sent, fill info in the argument 'b'
  if(Paths::bunched) {
    b->addInfo(nextSCID, path->pathID, path->path_dest_scid, pInfo.version);
    return;
  }

  // send patch-path downstream (from the closest occurrence of
  // 'mySCID' to 'path_dest_scid')

  // form the packet to send
  ServCompIFPkt_t scPkt;
  UDPLibPkt_t *uPkt = listen_pktBlock->getPkt();
  if(uPkt == NULL) {
    fprintf(stderr, "patchEdgeE2EHelper: can't alloc for uPkt\n");
    exit(77);
  }
  scPkt.fixed.type = PEER_PATCH_PATH;
  scPkt.fixed.repType = E2E_R;
  scPkt.fixed.path_dest_scid = path->path_dest_scid;
  scPkt.fixed.pathID = path->pathID;
  scPkt.fixed.fail_origin_scid = fail_origin_scid;
  scPkt.fixed.fail_dest_scid = fail_dest_scid;
  scPkt.fixed.version = pInfo.version;
  int uPktLen = scPkt.toStr(uPkt->data, sizeof(uPkt->data));

  // get the address to send
  Peer_t *peer1 = NetworkGraph::getPeerWithSCID(nextSCID);
  if(peer1 == NULL) {
    if(dbg->use_stream) {
      dbg->print() << "unknown-peer peerSCID=" << nextSCID << endl;
    } else {
      LogLine_t ll; ll.desc = unknown_peer_desc;
      ll.nextSCID = nextSCID; dbg->addLL(ll);
    }
    dbg->close(); exit(128);
  }

  // send the packet
  UDPLibPkt_t *u_ackPkt = listen_pktBlock->getPkt();
  if(u_ackPkt == NULL) {
    fprintf(stderr, "patchEdgeE2EHelper: can't alloc for u_ackPkt\n");
    exit(78);
  }
  struct timeval timeout = peer1->getServCompTimeout();
  int send_fd = listen_fd; // can use same socket for sending too
  bool use_tunnel = true;
  int reqNo = listen_rrBlock->insertRequest
    (send_fd, uPkt, uPktLen, peer1->ip_addr, SERV_COMP_PORT,
     ack_listen_fd, u_ackPkt, timeout, use_tunnel, peer1->scid,
     NetworkGraph::mySCID);
  if(reqNo == -1) {
    fprintf(stderr, "patchEdgeE2EHelper: no space for request\n");
    exit(79);
  }

  // add the reqNo to the queue, also remember the pkt for the
  // ack, and the peer from whom the original request was recd
  UDPLibPkt_t *reqRaw = NULL; // no request recd from anyone
  struct in_addr ackIPAddr; // to whom ack must be sent
  ackIPAddr.s_addr = inet_addr("0.0.0.0");
  bool local = false;
  Peer::sc_repair_addReq(reqNo, reqRaw, ackIPAddr, 0, pathID,
			 path_dest_scid, fail_dest_scid,
			 fail_origin_scid, failPosn, local);

  // don't kill path yet, let it timeout, (or may be it won't timeout
  // if the 'path_dest_scid' sends in a renew command for the path)

} // End ServComp::patchEdgeE2EHelper()

void ServComp::peerPatchPath(UDPLibPkt_t *reqPktRaw,
			     ServCompIFPkt_t *reqIFPkt,
			     ServCompIFPkt_t *replyIFPkt,
			     struct in_addr from_ip_addr,
			     int *replyPort, int *toReply) {
  // call peerPatchPathLocal() or peerPatchPathE2E()
  int16_t repType = reqIFPkt->fixed.repType;
  if(repType == LOCAL_R) {
    peerPatchPathLocal(reqPktRaw, reqIFPkt, replyIFPkt,
		       from_ip_addr, replyPort, toReply);
  } else if(repType == E2E_R) {
    peerPatchPathE2E(reqPktRaw, reqIFPkt, replyIFPkt,
		     from_ip_addr, replyPort, toReply);
  } else {
    cerr << "peerPatchPath: unknown repairType [" << repType << "]"
	 << endl;
    exit(111);
  }
} // End ServComp::peerPatchPath()

void ServComp::peerPatchPathLocal(UDPLibPkt_t *reqPktRaw,
				  ServCompIFPkt_t *reqIFPkt,
				  ServCompIFPkt_t *replyIFPkt,
				  struct in_addr from_ip_addr,
				  int *replyPort, int *toReply) {

  dbg->setCurrFn(PEERPATCHPATHLOCAL_FN);

  *replyPort = SERV_COMP_ACK_PORT;

  // get parameters
  unsigned short path_dest_scid = reqIFPkt->fixed.path_dest_scid;
  int pathID = reqIFPkt->fixed.pathID;
  unsigned short fail_origin_scid = reqIFPkt->fixed.fail_origin_scid;
  unsigned short fail_dest_scid = reqIFPkt->fixed.fail_dest_scid;
  const char *prev_dest_spec = reqIFPkt->getValue(PREV_DEST_SPEC_F);
  unsigned int posn = reqIFPkt->fixed.failPosn;
  const char *oldPath = reqIFPkt->getValue(OLD_PATH_F);
  const char *newSubPathStr = reqIFPkt->getValue(NEW_SUB_PATH_F);
  unsigned short fail_path_ver = reqIFPkt->fixed.version;

  if(dbg->use_stream) {
    dbg->print() << "req p_id=" << pathID << ':' << path_dest_scid
		 << " fail_path_ver=" << fail_path_ver
		 << " fail_dest_scid=" << fail_dest_scid
		 << " fail_origin_scid=" << fail_origin_scid
		 << " failPosn=" << posn << endl;
  } else {
    LogLine_t ll; ll.desc = req_desc; ll.pathID = pathID;
    ll.path_dest_scid = path_dest_scid; ll.fail_version = fail_path_ver;
    ll.fail_dest_scid = fail_dest_scid;
    ll.fail_origin_scid = fail_origin_scid; ll.failPosn = posn;
    dbg->addLL(ll);
  }

  // get new sub-path vector
  vector<JString> newSubPathVec;
  int numFields1 = split(newSubPathStr, newSubPathVec);
  int numInNewSubPath = numFields1/2;
  if(((numFields1 % 2) != 0) || (numInNewSubPath == 0)) {
    cerr << "peerPatchPathLocal: numFields1 (" << numFields1 << ") is not even" << endl;
    exit(88);
  }

  // get old path vector
  vector<JString> oldPathVec;
  int numFields2 = split(oldPath, oldPathVec);
  if((numFields2 % 2) != 0) {
    cerr << "peerPatchPathLocal: numFields2 (" << numFields2 << ") is not even" << endl;
    exit(88);
  }
  int numInOldPath = numFields2/2;

  if(fail_origin_scid == NetworkGraph::mySCID) {

    if(dbg->use_stream) {
      dbg->print() << "am-last-in-sub-path" << endl;
    } else {
      LogLine_t ll; ll.desc = am_last_in_sub_path_desc; dbg->addLL(ll);
    }

    // I am the final node in the new path
    bool pathExists = Paths::pathExists(pathID, path_dest_scid);
    if(!pathExists) {
      if(dbg->use_stream) {
	dbg->print() << "non-existent-path" << endl;
      } else {
	LogLine_t ll; ll.desc = non_existent_path_desc; dbg->addLL(ll);
      }
      replyIFPkt->fixed.success = 0;
      *toReply = 1;
      return;
    }

    PathInfo_t &pInfo = Paths::getPathInfo(pathID, path_dest_scid);
    if(pInfo.repairExists) {
      // repair has already been done for this path
      // have to distinguish between duplicate request, and some other
      // node sending request
      bool repairExists = Paths::repairExists(pathID, path_dest_scid,
					      fail_dest_scid,
					      fail_origin_scid, posn);
      if(repairExists) {
	// this must be a duplicate request -- may the ack got lost
	// earlier
	if(dbg->use_stream) { dbg->print() << "dup-req" << endl; }
	else { LogLine_t ll; ll.desc = dup_req_desc; dbg->addLL(ll); }
	replyIFPkt->fixed.success = 1;
      } else {
	// this is a request from some other node -- cannot handle
	// using local recovery -- return failure
	if(dbg->use_stream) {
	  dbg->print() << "local-exists-cant-handle" << endl;
	} else {
	  LogLine_t ll; ll.desc = local_exists_cant_handle_desc; dbg->addLL(ll);
	}
	replyIFPkt->fixed.success = 0;
      }
      *toReply = 1;
      return;
    }

    if(fail_path_ver != pInfo.version) {
      // this does not match path version
      if(dbg->use_stream) {
	dbg->print() << "version-mismatch path_ver=" << pInfo.version << endl;
      } else {
	LogLine_t ll; ll.desc = version_mismatch_desc;
	ll.version = pInfo.version; dbg->addLL(ll);
      }
      //if(fail_path_ver > pInfo.version) { exit(125); }
      replyIFPkt->fixed.success = 0;
      *toReply = 1;
      return;
    }

    if(!pathsEqual(pInfo.physPath, oldPathVec)) {
      // it does not match oldPath -- this is a sanity check
      if(dbg->use_stream) { dbg->print() << "paths-unequal" << endl; }
      else { LogLine_t ll; ll.desc = paths_unequal_desc; dbg->addLL(ll); }
      replyIFPkt->fixed.success = 0;
      *toReply = 1;
      return;
    }

    if(pInfo.numCreated != pInfo.physPath.size()) {
      // the old path has not been fully created, why am i already
      // getting a patch path request?
      if(dbg->use_stream) { dbg->print() << "incomplete-path" << endl; }
      else { LogLine_t ll; ll.desc = incomplete_path_desc; dbg->addLL(ll); }
      replyIFPkt->fixed.success = 0;
      *toReply = 1;
      return;
    }

    // note down info in database
    pInfo.repairExists = true;
    pInfo.fail_dest_scid = fail_dest_scid;
    pInfo.fail_origin_scid = fail_origin_scid;
    pInfo.failPosn = posn;
    PathInfo_t sub_p;
    sub_p.physPath.push_back(pInfo.physPath[posn]);
    for(int i = 1; i < numInNewSubPath-1; i++) {
      ServInst_t s;
      s.gInfo.scid = JString_atoi(newSubPathVec[2*i]);
      s.sName = newSubPathVec[2*i+1];
      if(s.sName != "noop") {
	cerr << "peerPatchPathLocal: sub-path has unknown service "
	     << s.sName << endl;
	exit(114);
      }
      sub_p.physPath.push_back(s);
    }
    sub_p.physPath.push_back(pInfo.physPath[posn+1]);
    Paths::setRepairInfo(pathID, path_dest_scid, fail_dest_scid,
			 fail_origin_scid, posn, sub_p);

    // Send request to service instance
    unsigned short serv_dest_scid =
      sub_p.physPath[numInNewSubPath-2].gInfo.scid;
    int succ = changeServInst(pathID, path_dest_scid, pInfo.version,
			      &pInfo.physPath[posn+1].gInfo,
			      prev_dest_spec, serv_dest_scid);

    // Send ack back
    replyIFPkt->fixed.success = succ;
    *toReply = 1;
    return;

  } else {
    // I am NOT the final node in the new sub-path

    bool pathExists = Paths::pathExists(pathID, path_dest_scid);
    if(pathExists) {
      // this path exists ==> it could be part of the old path, or the
      // newly patched sub-path
      PathInfo_t &p = Paths::getPathInfo(pathID, path_dest_scid);
      if(p.repairExists) {
	// have to distinguish between duplicate requests, and
	// requests from other nodes for patching the same path
	bool repairExists =
	  Paths::repairExists(pathID, path_dest_scid, fail_dest_scid,
			      fail_origin_scid, posn);
	if(repairExists) {

	  // this is a duplicate request
	  PathInfo_t repInfo =
	    Paths::getRepairInfo(pathID, path_dest_scid,
				 fail_dest_scid, fail_origin_scid, posn);
	  if(repInfo.numCreated == repInfo.physPath.size()) {
	    // must be a lost ack case
	    if(dbg->use_stream) {
	      dbg->print() << "lost-ack" << endl;
	    } else {
	      LogLine_t ll; ll.desc = lost_ack_desc; dbg->addLL(ll);
	    }
	    *toReply = 1;
	    replyIFPkt->fixed.success = 1;
	    return;
	  } else {
	    // getting a duplicate before sub-path construction is
	    // complete ==> ignore packet
	    if(dbg->use_stream) { dbg->print() << "dup-req" << endl; }
	    else { LogLine_t ll; ll.desc = dup_req_desc; dbg->addLL(ll); }
	    *toReply = 0;
	    listen_pktBlock->freePkt(reqPktRaw);
	    return;
	  }

	} else {
	  // this is a case where another node is trying to setup a
	  // sub-path patch through me --> cannot handle (has to be
	  // handled by end-2-end recovery)
	  if(dbg->use_stream) {
	    dbg->print() << "repair-exists-cant-handle" << endl;
	  } else {
	    LogLine_t ll; ll.desc = repair_exists_cant_handle_desc;
	    dbg->addLL(ll);
	  }
	  *toReply = 1;
	  replyIFPkt->fixed.success = 0;
	  return;
	}

      }

      if(fail_path_ver != p.version) {
	// this does not match path version
	if(dbg->use_stream) {
	  dbg->print() << "version-mismatch path_ver=" << p.version << endl;
	} else {
	  LogLine_t ll; ll.desc = version_mismatch_desc;
	  ll.version = p.version; dbg->addLL(ll);
	}
	//if(fail_path_ver > p.version) { exit(125); }
	replyIFPkt->fixed.success = 0;
	*toReply = 1;
	return;
      }

      // other sanity checks
      if(!pathsEqual(p.physPath, oldPathVec)) {
	// this is an error condition
	if(dbg->use_stream) {
	  dbg->print() << "unequal-paths" << endl;
	} else {
	  LogLine_t ll; ll.desc = unequal_paths_desc; dbg->addLL(ll);
	}
	*toReply = 1;
	replyIFPkt->fixed.success = 0;
	return;
      }

      if(dbg->use_stream) {
	dbg->print() << "path-exists-here" << endl;
      } else {
	LogLine_t ll; ll.desc = path_exists_here_desc; dbg->addLL(ll);
      }

      // end case if(pathExists)
    } else {
      if(dbg->use_stream) {
	dbg->print() << "creating-entry-for-path" << endl;
      } else {
	LogLine_t ll; ll.desc = creating_entry_for_path_desc; dbg->addLL(ll);
      }

      // create the path info
      PathInfo_t pInfo;
      for(int i = 0; i < numInOldPath; i++) {
	ServInst_t s;
	s.gInfo.scid = JString_atoi(oldPathVec[2*i]);
	s.sName = oldPathVec[2*i+1];
	pInfo.physPath.push_back(s);
      }

      pInfo.version = fail_path_ver;
      pInfo.lastHBRecd = (unsigned long)now(path_hb::startTime);
      // this path definitely has been created fully, so note this
      // while adding the path info for this
      pInfo.numCreated = pInfo.physPath.size();
      Paths::setPathInfo(pathID, path_dest_scid, pInfo);
    }

    PathInfo_t &p = Paths::getPathInfo(pathID, path_dest_scid);

    // find position in sub-path where i am
    PathInfo_t subPath;
    int subPosn = -1;
    bool dup = false;
    subPath.physPath.push_back(p.physPath[posn]);
    for(int i = 1; i < numInNewSubPath-1; i++) {
      ServInst_t s;
      s.gInfo.scid = JString_atoi(newSubPathVec[2*i]);
      s.sName = newSubPathVec[2*i+1];
      if(s.sName != "noop") {
	cerr << "peerPatchPathLocal: sub-path has unknown service "
	     << s.sName << endl;
	exit(114);
      }
      subPath.physPath.push_back(s);
      if(s.gInfo.scid == NetworkGraph::mySCID) {
	if(subPosn != -1) { dup = true; }
	subPosn = i;
      }
    }
    subPath.physPath.push_back(p.physPath[posn+1]);

    // sanity check again
    if((subPosn == -1) || dup) {
      if(dbg->use_stream) {
	dbg->print() << "sub-path-is-weird subPosn=" << subPosn
		     << " dup=" << dup << endl;
      } else {
	LogLine_t ll; ll.desc = sub_path_is_weird_desc;
	ll.subPosn = subPosn; ll.utype = ERR_U; ll.u.err.init();
	ll.u.err.dup = dup; dbg->addLL(ll);
      }
      dbg->close(); exit(115);
    }

    unsigned short serv_dest_scid = subPath.physPath[subPosn-1].gInfo.scid;
    const char *this_dest_spec = sync_createServInst
      (pathID, path_dest_scid, fail_path_ver, "noop",
       &subPath.physPath[subPosn], prev_dest_spec, serv_dest_scid);
    subPath.physPath[subPosn].dest_spec = this_dest_spec;

    Paths::setRepairInfo(pathID, path_dest_scid, fail_dest_scid,
			 fail_origin_scid, posn, subPath);
    p.repairExists = true;
    p.fail_dest_scid = fail_dest_scid;
    p.fail_origin_scid = fail_origin_scid;
    p.failPosn = posn;

    /*---- send request upstream ----*/

    // form the packet to send
    ServCompIFPkt_t scPkt;
    UDPLibPkt_t *uPkt = listen_pktBlock->getPkt();
    if(uPkt == NULL) {
      fprintf(stderr, "peerPatchPathLocal: can't alloc for uPkt\n");
      exit(65);
    }
    scPkt.fixed.type = PEER_PATCH_PATH;
    scPkt.fixed.repType = LOCAL_R;
    scPkt.fixed.path_dest_scid = path_dest_scid;
    scPkt.fixed.pathID = pathID;
    scPkt.fixed.fail_origin_scid = fail_origin_scid;
    scPkt.fixed.fail_dest_scid = fail_dest_scid;
    scPkt.addPair(PREV_DEST_SPEC_F, prev_dest_spec);
    scPkt.addPair(OLD_PATH_F, oldPath);
    scPkt.addPair(NEW_SUB_PATH_F, newSubPathStr);
    scPkt.fixed.failPosn = posn;
    scPkt.fixed.version = p.version;
    int uPktLen = scPkt.toStr(uPkt->data, sizeof(uPkt->data));

    // get the address to send
    unsigned short nextSCID = subPath.physPath[subPosn+1].gInfo.scid;
    Peer_t *peer = NetworkGraph::getPeerWithSCID(nextSCID);

    if(dbg->use_stream) {
      dbg->print() << "send-req-upstream nextSCID=" << nextSCID << endl;
    } else {
      LogLine_t ll; ll.desc = send_req_upstream_desc;
      ll.nextSCID = nextSCID; dbg->addLL(ll);
    }

    // send the packet
    UDPLibPkt_t *u_ackPkt = listen_pktBlock->getPkt();
    if(u_ackPkt == NULL) {
      fprintf(stderr, "peerPatchPathLocal: can't alloc for u_ackPkt\n");
      exit(66);
    }
    struct timeval timeout = peer->getServCompTimeout();
    int send_fd = listen_fd; // can use same socket for sending
    bool use_tunnel = true;
    int reqNo = listen_rrBlock->insertRequest
      (send_fd, uPkt, uPktLen, peer->ip_addr, SERV_COMP_PORT,
       ack_listen_fd, u_ackPkt, timeout, use_tunnel, peer->scid,
       NetworkGraph::mySCID);
    if(reqNo == -1) {
      fprintf(stderr, "peerPatchPathLocal: no space for request\n");
      exit(67);
    }

    // add the reqNo to the queue
    *toReply = 0;
    bool local = true;
    Peer::sc_repair_addReq(reqNo, reqPktRaw, from_ip_addr,
			   SERV_COMP_ACK_PORT, pathID, path_dest_scid,
			   fail_dest_scid, fail_origin_scid, posn, local);

  } // this ends the case when i'm NOT the final node in the patched path

} // End ServComp::peerPatchPathLocal()

void ServComp::peerPatchBunch(UDPLibPkt_t *reqPktRaw,
			      ServCompIFPkt_t *reqIFPkt,
			      ServCompIFPkt_t *replyIFPkt,
			      struct in_addr from_ip_addr,
			      int *replyPort, int *toReply) {
  dbg->setCurrFn(PEERPATCHBUNCH_FN);
  *replyPort = SERV_COMP_ACK_PORT;

  unsigned short num_in_bunch = reqIFPkt->fixed.num_in_bunch;
  if(num_in_bunch == 0) {
    if(dbg->use_stream) {
      dbg->print() << "invalid-num-in-bunch num_in_bunch="
		   << num_in_bunch << endl;
    } else {
      LogLine_t ll; ll.desc = invalid_num_in_bunch_desc;
      ll.utype = PATCH_U; ll.u.patch.init();
      ll.u.patch.num_in_bunch = num_in_bunch; dbg->addLL(ll);
    }
    *toReply = 1;
    replyIFPkt->fixed.success = 0;
    return;
  }

  unsigned short fail_origin_scid = reqIFPkt->fixed.fail_origin_scid;
  unsigned short fail_dest_scid = reqIFPkt->fixed.fail_dest_scid;

  Bunch_t b;
  b.fail_origin_scid = fail_origin_scid;
  b.fail_dest_scid = fail_dest_scid;

  for(unsigned int i = 0; i < num_in_bunch; i++) {
    unsigned short path_dest_scid = reqIFPkt->failed[i].path_dest_scid;
    int pathID = reqIFPkt->failed[i].pathID;
    unsigned short fail_path_ver = reqIFPkt->failed[i].version;

    /* SUPRESS
    if(dbg->use_stream) {
      dbg->print() << "e2e-req p_id=" << pathID << ':' << path_dest_scid
		   << " fail_dest_scid=" << fail_dest_scid
		   << " fail_origin_scid=" << fail_origin_scid
		   << " fail_path_ver=" << fail_path_ver << endl;
    } else {
      LogLine_t ll; ll.desc = e2e_req_desc; ll.pathID = pathID;
      ll.path_dest_scid = path_dest_scid;
      ll.fail_dest_scid = fail_dest_scid;
      ll.fail_origin_scid = fail_origin_scid;
      ll.fail_version = fail_path_ver; dbg->addLL(ll);
    }
    */

    Peer_t *peer = NetworkGraph::getPeerWithIP(from_ip_addr);
    if(peer == NULL) {
      if(dbg->use_stream) {
	dbg->print() << "unknown-peer peerIP=" << ip2str(from_ip_addr) << endl;
      } else {
	LogLine_t ll; ll.desc = unknown_peer_desc;
	ll.utype = ERR_U; ll.u.err.init();
	ll.u.err.peerIP.s_addr = from_ip_addr.s_addr; dbg->addLL(ll);
      }
      dbg->close(); exit(124);
      return;
    }
    bool pathExists = Paths::pathExists(pathID, path_dest_scid);
    if(!pathExists) {
      if(dbg->use_stream) {
	dbg->print() << "non-existent-path p_id=" << pathID
		     << ':' << path_dest_scid
		     << " fail_dest_scid=" << fail_dest_scid
		     << " fail_origin_scid=" << fail_origin_scid
		     << " fail_path_ver=" << fail_path_ver << endl;
      } else {
	LogLine_t ll; ll.desc = non_existent_path_desc;
	ll.pathID = pathID; ll.path_dest_scid = path_dest_scid;
	ll.fail_dest_scid = fail_dest_scid;
	ll.fail_origin_scid = fail_origin_scid;
	ll.fail_version = fail_path_ver; dbg->addLL(ll);
      }
      continue;
    }
    PathInfo_t &p = Paths::getPathInfo(pathID, path_dest_scid);
    path_id_t p_id = { pathID, path_dest_scid };
    patchEdgeE2EHelper(peer, &p_id, p, fail_dest_scid, fail_origin_scid,
		       fail_path_ver, &b);
  } // End for(handling each path in the bunch)

  // send out the bunched packets
  sendBunch(&b);

  *toReply = 1;
  replyIFPkt->fixed.success = 1;
} // End ServComp::peerPatchBunch()

void ServComp::peerPatchPathE2E(UDPLibPkt_t *reqPktRaw,
				ServCompIFPkt_t *reqIFPkt,
				ServCompIFPkt_t *replyIFPkt,
				struct in_addr from_ip_addr,
				int *replyPort, int *toReply) {

  dbg->setCurrFn(PEERPATCHPATHE2E_FN);

  *replyPort = SERV_COMP_ACK_PORT;

  unsigned short path_dest_scid = reqIFPkt->fixed.path_dest_scid;
  int pathID = reqIFPkt->fixed.pathID;
  unsigned short fail_origin_scid = reqIFPkt->fixed.fail_origin_scid;
  unsigned short fail_dest_scid = reqIFPkt->fixed.fail_dest_scid;
  unsigned short fail_path_ver = reqIFPkt->fixed.version;

  /* SUPRESS
  if(dbg->use_stream) {
    dbg->print() << "e2e-req p_id=" << pathID << ':' << path_dest_scid
		 << " fail_dest_scid=" << fail_dest_scid
		 << " fail_origin_scid=" << fail_origin_scid
		 << " fail_path_ver=" << fail_path_ver << endl;
  } else {
    LogLine_t ll; ll.desc = e2e_req_desc; ll.pathID = pathID;
    ll.path_dest_scid = path_dest_scid;
    ll.fail_dest_scid = fail_dest_scid;
    ll.fail_origin_scid = fail_origin_scid;
    ll.fail_version = fail_path_ver; dbg->addLL(ll);
  }
  */

  Peer_t *peer = NetworkGraph::getPeerWithIP(from_ip_addr);
  if(peer == NULL) {
    if(dbg->use_stream) {
      dbg->print() << "unknown-peer peerIP=" << ip2str(from_ip_addr) << endl;
    } else {
      LogLine_t ll; ll.desc = unknown_peer_desc;
      ll.utype = ERR_U; ll.u.err.init();
      ll.u.err.peerIP.s_addr = from_ip_addr.s_addr; dbg->addLL(ll);
    }
    dbg->close(); exit(124);
    return;
  }
  bool pathExists = Paths::pathExists(pathID, path_dest_scid);
  if(!pathExists) {
    if(dbg->use_stream) {
      dbg->print() << "non-existent-path p_id=" << pathID
		   << ':' << path_dest_scid
		   << " fail_dest_scid=" << fail_dest_scid
		   << " fail_origin_scid=" << fail_origin_scid
		   << " fail_path_ver=" << fail_path_ver << endl;
    } else {
      LogLine_t ll; ll.desc = non_existent_path_desc; ll.pathID = pathID;
      ll.path_dest_scid = path_dest_scid;
      ll.fail_dest_scid = fail_dest_scid;
      ll.fail_origin_scid = fail_origin_scid;
      ll.fail_version = fail_path_ver; dbg->addLL(ll);
    }
    *toReply = 1;
    replyIFPkt->fixed.success = 0;
    return;
  }
  PathInfo_t &p = Paths::getPathInfo(pathID, path_dest_scid);
  path_id_t p_id = { pathID, path_dest_scid };
  Bunch_t *b = NULL;
  patchEdgeE2EHelper(peer, &p_id, p, fail_dest_scid, fail_origin_scid,
		     fail_path_ver, b);

  *toReply = 1;
  replyIFPkt->fixed.success = 1;

} // End ServComp::peerPatchPathE2E()

void ServComp::createAltPath(Peer_t *peer, path_id_t *path,
			     PathInfo_t &pInfo, unsigned short fail_dest_scid,
			     unsigned short fail_origin_scid) {
  // simply call createPath with the altPath option set to true
  int *replyPort = NULL; // not used
  int *toReply = NULL; // not used
  UDPLibPkt_t *reqPktRaw = NULL; // not used
  ServCompIFPkt_t *replyIFPkt = NULL; // not used
  ServCompIFPkt_t reqIFPkt; // used to pass parameters for path creation
  reqIFPkt.fixed.pathID = path->pathID;
  reqIFPkt.fixed.fail_dest_scid = fail_dest_scid;
  reqIFPkt.fixed.fail_origin_scid = fail_origin_scid;
  bool altPath = true;
  struct in_addr ackIPAddr; // to whom ack must be sent
  ackIPAddr.s_addr = inet_addr("0.0.0.0");
  createPath(reqPktRaw, &reqIFPkt, replyIFPkt, ackIPAddr, replyPort,
	     toReply, altPath);
} // End ServComp::createAltPath()

bool ServComp::checkRepeat(vector<ServInst_t> &oldPath,
			   vector<ServInst_t> &newSubPath) {
  for(unsigned int i = 1; i < newSubPath.size()-1; i++) {
    for(unsigned int j = 0; j < oldPath.size(); j++) {
      if(oldPath[j].gInfo.scid == newSubPath[i].gInfo.scid) {
	return true;
      }
    }
  }
  return false;
} // End ServComp::checkRepeat()

void ServComp::sendBunch(Bunch_t *b) {
  dbg->setCurrFn(SENDBUNCH_FN);

  for(unsigned int i = 0; i < b->peerList.size(); i++) {

    vector<patchInfo_t> &vPatch = b->peerList[i].vPatch;
    unsigned int num_sent = 0;

    // get the address to send
    unsigned short nextSCID = b->peerList[i].scid;
    Peer_t *peer1 = NetworkGraph::getPeerWithSCID(nextSCID);
    if(peer1 == NULL) {
      if(dbg->use_stream) {
	dbg->print() << "unknown-peer peerSCID=" << nextSCID << endl;
      } else {
	LogLine_t ll; ll.desc = unknown_peer_desc;
	ll.peerSCID = nextSCID; dbg->addLL(ll);
      }
      dbg->close(); exit(128);
    }

    while(num_sent < vPatch.size()) {
      unsigned int num2bunch = vPatch.size()-num_sent;
      if(num2bunch > max_num_in_bunch) { num2bunch = max_num_in_bunch; }

      // form the packet to send to peer
      ServCompIFPkt_t scPkt;
      UDPLibPkt_t *uPkt = listen_pktBlock->getPkt();
      if(uPkt == NULL) {
	fprintf(stderr, "sendBunch: can't alloc for uPkt\n");
	exit(77);
      }
      scPkt.fixed.type = PEER_PATCH_BUNCH;
      scPkt.fixed.repType = E2E_R;
      scPkt.fixed.fail_origin_scid = b->fail_origin_scid;
      scPkt.fixed.fail_dest_scid = b->fail_dest_scid;

      scPkt.fixed.num_in_bunch = num2bunch;

      for(unsigned int j = num_sent; j < (num_sent+num2bunch); j++) {
	patchInfo_t &patch = vPatch[j];
	FailedPath_t fp;
	fp.pathID = patch.pathID;
	fp.path_dest_scid = patch.path_dest_scid;
	fp.version = patch.version;
	scPkt.failed.push_back(fp);
      }

      unsigned int totLen = scPkt.totLen();
      if((totLen > (sizeof(uPkt->data)-100)) || (totLen > (1500-100))) {
	// are we reaching close to capacity, or MTU?
	cerr << NetworkGraph::mySCID << ": bunched-pkt-too-big totLen="
	     << totLen << " num_in_bunch=" << vPatch.size() << endl;
	if(dbg->use_stream) {
	  dbg->print() << "bunched-pkt-too-big totLen=" << totLen
		       << " num_in_bunch=" << vPatch.size() << endl;
	} else {
	  LogLine_t ll; ll.desc = bunched_pkt_too_big_desc;
	  ll.utype = PATCH_U; ll.u.patch.init();
	  ll.u.patch.num_in_bunch = vPatch.size();
	  ll.u.patch.totLen = totLen; dbg->addLL(ll);
	}
	dbg->close(); exit(129);
      }

      int uPktLen = scPkt.toStr(uPkt->data, sizeof(uPkt->data));

      if(dbg->use_stream) {
	dbg->print() << "num-in-bunch num_in_bunch=" << num2bunch
		     << " nextSCID=" << nextSCID << endl;
      } else {
	LogLine_t ll; ll.desc = num_in_bunch_desc; ll.nextSCID = nextSCID;
	ll.utype = PATCH_U; ll.u.patch.init();
	ll.u.patch.num_in_bunch = num2bunch; dbg->addLL(ll);
      }

      // send the packet
      UDPLibPkt_t *u_ackPkt = listen_pktBlock->getPkt();
      if(u_ackPkt == NULL) {
	fprintf(stderr, "sendBunch: can't alloc for u_ackPkt\n");
	exit(78);
      }
      struct timeval timeout = peer1->getServCompTimeout();
      int send_fd = listen_fd; // can use same socket for sending too
      bool use_tunnel = true;
      int reqNo = listen_rrBlock->insertRequest
	(send_fd, uPkt, uPktLen, peer1->ip_addr, SERV_COMP_PORT,
	 ack_listen_fd, u_ackPkt, timeout, use_tunnel, peer1->scid,
	 NetworkGraph::mySCID);
      if(reqNo == -1) {
	fprintf(stderr, "sendBunch: no space for request\n");
	exit(79);
      }

      // add the reqNo to the queue, also remember the pkt for the
      // ack, and the peer from whom the original request was recd
      UDPLibPkt_t *reqRaw = NULL; // no request recd from anyone
      struct in_addr ackIPAddr; // to whom ack must be sent
      ackIPAddr.s_addr = inet_addr("0.0.0.0");
      bool local = false;
      int pathID = 0; int failPosn = 0;
      unsigned short path_dest_scid, fail_dest_scid, fail_origin_scid;
      path_dest_scid = fail_dest_scid = fail_origin_scid =
	NetworkGraph::mySCID;
      Peer::sc_repair_addReq(reqNo, reqRaw, ackIPAddr, 0, pathID,
			     path_dest_scid, fail_dest_scid,
			     fail_origin_scid, failPosn, local);

      num_sent += num2bunch;
    } // End while(still some patches to be sent)

  } // End for(handling each peer in the bunch)

} // End ServComp::sendBunch()
