/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Implements class ServComp */

#include <netdb.h>
#include <stdio.h>
#include <unistd.h>
#include <iostream>

#include "ServComp.h"
#include "TreeInfo.h"
#include "Paths.h"
#include "../util/LogLine.h"
#include "../link-state/NetworkGraph.h"
#include "../service-locn/ServiceLocn.h"

queue<path_hb_t, list<path_hb_t> > ServComp::hbQ;
Debug_t *ServComp::dbg;

pthread_t ServComp::listenThr;

UDPPktBlock_t *ServComp::listen_pktBlock;
reqRespBlock_t *ServComp::listen_rrBlock;

UDPPktBlock_t *ServComp::serv_pktBlock;
reqRespBlock_t *ServComp::serv_rrBlock;

int ServComp::ack_listen_fd;
int ServComp::listen_fd;
int ServComp::server_listen_fd;
int ServComp::client_listen_fd;

// This is a HACK
int ServComp::last_app_seq = -1;

bool ServComp::stop_send = false;
bool ServComp::stop_listen = false;
unsigned long ServComp::numSent = 0;
unsigned long ServComp::numRecd = 0;

unsigned int ServComp::max_num_in_bunch = 40;
unsigned int ServComp::max_to_patch_now = 200;
unsigned long ServComp::min_patch_interval = 50; // 50ms

TreeInfo_t *ServComp::buildTree(const vector<JString> &services,
				unsigned short fail_dest_scid,
				unsigned short fail_origin_scid) {
  if(TreeInfo::debug) {
    cerr << "buildtree called on: ";
    for(unsigned int i = 0; i < services.size(); i++) {
      cerr << services[i] << " ";
    }
    if(services.size() == 0) {
      cerr << "<null path>";
    }
    cerr << endl;
  }

  // this is the tree of information for this logical path
  TreeInfo_t *t = TreeInfo::getTreeInfo(services);
  if(t == NULL) {
    t = TreeInfo::makeNewTree(services, NetworkGraph::getNumNodes());
  }

  /* mark each of nodes in the graph white ==> has no services */
  t->numTreeNodes = NetworkGraph::getNumNodes();
  NetworkGraph::getNodes(t->nVec, t->numTreeNodes);
  for(int i = 0; i < t->numTreeNodes; i++) {
    t->nVec[i].initServiceList();
  }

  // build the reverse index (SCID to index into vertex array)
  t->buildReverseIndex();

  /* colour each of nodes in the graph */
  for(unsigned int i = 0; i < services.size(); i++) {
    // colour the locations of services[i] with color i
    const vector<GlobalServInfo_t> *gVec = ServiceLocn::getLocations(services[i]);
    for(unsigned int j = 0; j < gVec->size(); j++) {
      int nIndex = t->getIndexOfSCID((*gVec)[j].scid);
      t->nVec[nIndex].addToServiceList(services[i]);
    }
  }

  t->allocSeenTreeNodes(); // allocate the priority queues

  /* initialize seenTreeNodes[0..numServices] */
  for(unsigned int i = 0; i < services.size()+1; i++) {
    for(int j = 0; j < t->numTreeNodes; j++) {
      // for all vertices v, mark cost of v in seenTreeNodes[i] to be infinity
      t->seenTreeNodes[i].decreaseCost
	(j, PriorityQueue<unsigned short>::INFINITY);
    }
  }

  /* initialize prevTreeNode in path through services */
  for(unsigned int i = 0; i < services.size()+1; i++) {
    for(int j = 0; j < t->numTreeNodes; j++) {
      // foreach vertex v, prevTreeNode in path through services is 'invalid'
      t->nVec[j].prevTreeNode[i] = NetworkGraph::INVALID_NODE;
      t->nVec[j].minCost[i] = PriorityQueue<unsigned short>::INFINITY;
    }
  }

  // mark cost of source in seenTreeNodes[0] to be 0
  int myIndex = t->getIndexOfSCID(NetworkGraph::mySCID);
  t->seenTreeNodes[0].decreaseCost(myIndex, 0);
  // mark prevTreeNode of source in prevTreeNode[0] to be the source itself
  t->nVec[myIndex].prevTreeNode[0] = myIndex;
  t->nVec[myIndex].minCost[0] = 0;

  // keep track of the global min node -- this is the node from which
  // the "constrained" path is the minimum (this will be a node that
  // has services[n-1] -- the last service)
  t->globalMinVertex = NetworkGraph::INVALID_NODE;
  t->globalMinSCID = NetworkGraph::INVALID_SCID;

  for(;;) {
    // find the minimum among all priority queues
    int minExists = 0; 		unsigned short minCost = 0;
    int minVertex = -1;		int minIndex = -1;
    for(unsigned int i = 0; i < services.size()+1; i++) {
      unsigned short thisMinCost;
      int thisMinVertex = t->seenTreeNodes[i].minVertex(&thisMinCost);
      if(thisMinVertex == -1) continue; // no node in the priority queue
      if(!minExists || (thisMinCost <= minCost)) {
	minCost = thisMinCost; 		minVertex = thisMinVertex;
	minIndex = i; 			minExists = 1;
      }
    }

    // this is the termination condition for the algorithm
    if(!minExists) break;

    // remove minVertex from the priority queue
    t->seenTreeNodes[minIndex].removeMin();

    /*
    cerr << "minVertex's scid = " << t->nVec[minVertex].scid
	 << ", minIndex = " << minIndex
	 << ", minCost = " << minCost
	 << ", prevNode = " << t->nVec[minVertex].prevTreeNode[minIndex]
	 << endl;
    */

    // promote node up the chain of services
    unsigned int lastIndex = minIndex;
    for(unsigned int i = minIndex+1; i < services.size()+1; i++) {
      if(t->nVec[minVertex].hasService(services[i-1])) {
	// update prevTreeNode for minVertex, i to minCost
	t->nVec[minVertex].prevTreeNode[i] =
	  t->nVec[minVertex].prevTreeNode[minIndex];
	t->nVec[minVertex].minCost[i] = minCost;
	// remove minVertex from seenTreeNodes[i]
	t->seenTreeNodes[i].removeVertex(minVertex);
	lastIndex = i;
	/*
	cerr << "removing scid " << t->nVec[minVertex].scid
	     << ", lastIndex = " << lastIndex
	     << ", minCost = " << minCost
	     << ", prevNode = " << t->nVec[minVertex].prevTreeNode[lastIndex]
	     << endl;
	*/
      } else {
	break;
      }
    }

    // expand minVertex
    const vector<GraphEdge_t> *eVec =
      NetworkGraph::getIncomingEdges(t->nVec[minVertex].scid);

    unsigned short min_ver_scid = t->nVec[minVertex].scid;

    // for all neighbours u of minVertex
    for(unsigned int i = 0; i < eVec->size(); i++) {
      int srcVertex = (*eVec)[i].srcVertex;
      unsigned short min_nbr_scid = t->nVec[srcVertex].scid;

      unsigned int arcCost = (*eVec)[i].arcCost;

      if((min_ver_scid == fail_dest_scid) &&
	 (min_nbr_scid == fail_origin_scid)) {
	// this is the recently failed edge
	//cerr << NetworkGraph::mySCID << ":0 - failed edge encountered\n";
	arcCost = PriorityQueue<unsigned short>::INFINITY;
      }
      if((min_nbr_scid == fail_dest_scid) &&
	 (min_ver_scid == fail_origin_scid)) {
	// this is the back edge of the recently failed edge
	//cerr << NetworkGraph::mySCID << ":1 - failed edge encountered\n";
	arcCost = PriorityQueue<unsigned short>::INFINITY;
      }

      for(unsigned int j = minIndex; j < lastIndex+1; j++) {
	// update cost of u in seenTreeNodes[minIndex..lastIndex] if necessary
	// update prevTreeNode for u, [minIndex..lastIndex] if cost has
	// been updated
	unsigned short oldCost;
	bool nodeExists = t->seenTreeNodes[j].getCost(srcVertex, &oldCost);
	if(!nodeExists) {
	  // node has already been removed from PQ
	  continue;
	}
	unsigned short newCost = minCost + arcCost;
	if((arcCost == PriorityQueue<unsigned short>::INFINITY) ||
	   (minCost == PriorityQueue<unsigned short>::INFINITY)) {
	  newCost = PriorityQueue<unsigned short>::INFINITY;
	}
	if((newCost < oldCost) ||
	   (minCost == PriorityQueue<unsigned short>::INFINITY)) {
	  if(newCost < oldCost) {
	    t->seenTreeNodes[j].decreaseCost(srcVertex, newCost);
	  }
	  /*
	  cerr << "decreaseCost: "
	       << "srcVertex=" << t->nVec[srcVertex].scid
	       << ", newCost=" << newCost
	       << ", j=" << j
	       << ", prevNode=" << minVertex << endl;
	  */
	  t->nVec[srcVertex].prevTreeNode[j] = minVertex;
	  t->nVec[srcVertex].minCost[j] = newCost;
	}
      }
    } // End for all neighbours u of minVertex

    // if the final service has been reached, mark the globalMinVertex
    if((lastIndex == services.size()) &&
       (t->globalMinVertex == NetworkGraph::INVALID_NODE)) {
      t->globalMinVertex = minVertex;
      t->globalMinSCID = t->nVec[minVertex].scid;
    }

  } // End for(;;)

  // discard the seenTreeNodes priority queues
  t->deallocSeenTreeNodes();

  // note that tree for "services" was built at this time
  gettimeofday(&t->buildTime, NULL);

  return t; // return the tree built

} // End ServComp::buildTree()

int ServComp::findPath(unsigned short origin_scid,
		       const vector<JString> &services,
		       vector<ServInst_t> &path, bool forceRebuild,
		       unsigned short fail_dest_scid,
		       unsigned short fail_origin_scid) {
  // get the tree structure (built earlier) for this set of services
  TreeInfo_t *t = TreeInfo::getTreeInfo(services);
  bool toBuild = forceRebuild;
  if(t == NULL) {
    // this logical path has not been seen earlier
    t = buildTree(services, fail_dest_scid, fail_origin_scid);
    toBuild = false;
  }

  // check the time when buildTree was called for the given vector of services
  float timeElapsed = now(t->buildTime);
  if(timeElapsed > TreeInfo::TREE_CACHE_TIMEOUT) {
    t = buildTree(services, fail_dest_scid, fail_origin_scid);
    toBuild = false;
  }

  // check if tree is being rebuilt forcibly, by caller
  if(toBuild) {
    t = buildTree(services, fail_dest_scid, fail_origin_scid);
  }

  // when caller does not want a specific origin for the path,
  // origin_scid is INVALID_SCID
  unsigned short orig_scid_to_use =
    (origin_scid == NetworkGraph::INVALID_SCID) ?
    t->globalMinSCID : origin_scid;

  // check if there exists a path at all
  int origin_nIndex = t->getIndexOfSCID(orig_scid_to_use);
  if(origin_nIndex == -1) {
    cerr << "findPath: nothing known about orig_scid_to_use "
	 << orig_scid_to_use << endl;
    return -1;
  }
  if(t->nVec[origin_nIndex].minCost[services.size()] ==
     PriorityQueue<unsigned short>::INFINITY) {
    // no valid path exists
    fputc('n', stderr);
    return -1;
  }

  // extract the path from the prevTreeNode information
  int indexOfServToSee = services.size()-1;
  unsigned short curr_scid = orig_scid_to_use;
  unsigned short last_service_scid = NetworkGraph::INVALID_SCID;
  for(;;) {
    int nIndex = t->getIndexOfSCID(curr_scid);

    if(indexOfServToSee < 0) {
      // all services have been seen

      if(curr_scid == NetworkGraph::mySCID) {
	// and we're at the local node
	break; // this is the termination condition
      } else {

	// fill in the current scid and the "noop" service
	if(last_service_scid != curr_scid) {
	  // if the last service was at this scid, then no need for a "noop"
	  ServInst_t s;
	  s.gInfo.scid = curr_scid;
	  s.sName = "noop";
	  path.push_back(s);
	  last_service_scid = curr_scid;
	}

	// update curr_scid
	unsigned short nextIndex = t->nVec[nIndex].prevTreeNode[0];
	curr_scid = t->nVec[nextIndex].scid;
	continue;
      }

    } // end case when (indexOfServToSee < 0)

    if(t->nVec[nIndex].hasService(services[indexOfServToSee])) {

      // the current scid has the service we need to see
      // insert the appropriate entry into the path
      ServInst_t s;
      s.gInfo.scid = curr_scid;
      s.sName = services[indexOfServToSee];
      path.push_back(s);

      // note that you have seen this service and move on (no need to
      // update curr_scid)
      indexOfServToSee--;
      last_service_scid = curr_scid;
      continue;

    } else {

      // the current scid does not have the service we are looking for
      // insert the appropriate entry into the path
      if(last_service_scid != curr_scid) {
	// if the last service was at this scid, then no need for a "noop"
	ServInst_t s;
	s.gInfo.scid = curr_scid;
	s.sName = "noop";
	path.push_back(s);
	last_service_scid = curr_scid;
      }

      // update curr_scid
      unsigned short nextIndex =
	t->nVec[nIndex].prevTreeNode[indexOfServToSee+1];
      curr_scid = t->nVec[nextIndex].scid;
      continue;
    }

  } // End for(;;) loop computing the path

  // insert the null operator in mySCID if the last service seen was
  // somewhere upstream
  if(last_service_scid != NetworkGraph::mySCID) {
    // if the last service was at this scid, then no need for a "noop"
    ServInst_t s;
    s.gInfo.scid = curr_scid;
    s.sName = "noop";
    path.push_back(s);
    last_service_scid = NetworkGraph::mySCID;    
  }

  // have to reverse the path so that path[0] is downstream (at mySCID)
  for(unsigned int i = 0; i < path.size()/2; i++) {
    ServInst_t temp = path[i];
    path[i] = path[path.size()-i-1];
    path[path.size()-i-1] = temp;
  }

  // have to check if failed edge occurs in the path found
  for(unsigned int i = 0; i < (path.size()-1); i++) {
    if((path[i].gInfo.scid == fail_dest_scid) &&
       (path[i+1].gInfo.scid == fail_origin_scid)) {
      //cerr << NetworkGraph::mySCID << ":0 - failed edge in path\n";
      return -1;
    }
    if((path[i+1].gInfo.scid == fail_dest_scid) &&
       (path[i].gInfo.scid == fail_origin_scid)) {
      //cerr << NetworkGraph::mySCID << ":1 - failed edge in path\n";
      return -1;
    }
  }

  return path.size();

} // End ServComp::findPath()

void path_hb::init() {
  // Note down the start time (for sending path HBs)
  gettimeofday(&startTime, NULL);
  // set random seed
  unsigned long seedLong = startTime.tv_usec%211;
  float seedFloat = ((seedLong+1)*1.0)/212;
  //fprintf(stderr, "setting random seed done with %ld(%f)\n",
  //seedLong, seedFloat);
  Uniform::Set(seedFloat);
} // End path_hb::init()

void ServComp::init(Debug_t *arg_dbg) {
  stop_send = stop_listen = false;
  numRecd = numSent = 0;

  // get the request response block for listening
  listen_rrBlock = UDPLib::getReqRespBlock();
  if(listen_rrBlock == NULL) {
    fprintf(stderr, "ServComp: cannot get listen_rrBlock\n");
    exit(68);
  }

  // get a block of packet space for listening
  listen_pktBlock = UDPPktBlock::getUDPPktBlock();
  if(listen_pktBlock == NULL) {
    fprintf(stderr, "ServComp: cannot get listen_pktBlock\n");
    exit(69);
  }

  // get the request response block for comm with local services
  serv_rrBlock = UDPLib::getReqRespBlock();
  if(serv_rrBlock == NULL) {
    fprintf(stderr, "ServComp: cannot get serv_rrBlock\n");
    exit(68);
  }

  // get a block of packet space for comm with local services
  serv_pktBlock = UDPPktBlock::getUDPPktBlock();
  if(serv_pktBlock == NULL) {
    fprintf(stderr, "ServComp: cannot get serv_pktBlock\n");
    exit(69);
  }

  // socket for listening to acks
  ack_listen_fd = sock_and_bind(SERV_COMP_ACK_PORT);
  if(ack_listen_fd < 0) {
    perror("ServComp::init");
    exit(70);
  }

  // socket for listening
  listen_fd = sock_and_bind(SERV_COMP_PORT);
  if(listen_fd < 0) {
    perror("ServComp::init");
    exit(70);
  }

  // socket for interaction with servers
  server_listen_fd = sock_and_bind(SERV_COMP_SERVER_PORT);
  if(server_listen_fd < 0) {
    perror("ServComp::init");
    exit(70);
  }

  // socket for interaction with clients
  client_listen_fd = sock_and_bind(SERV_COMP_CLIENT_PORT);
  if(client_listen_fd < 0) {
    perror("ServComp::init");
    exit(70);
  }

  // call init on Paths
  Paths::init(arg_dbg);

  // call init on path_hb
  path_hb::init();

  // get the debug file
  dbg = arg_dbg;

  if(cfg_read_paths) {
    ifstream fin("paths.cfg");
    if(fin.fail()) {
      cerr << "Can't open paths.cfg for reading" << endl;
      exit(43);
    }

    int line_num = 0;
    for(;;) {
      // read each line
      if(fin.eof()) break;
      char line[2048];
      fin.getline(line, sizeof(line)); // read a line
      if(fin.eof()) break;
      if(fin.fail()) {
	cerr << "Error reading" << endl; exit(99);
      }
      line_num++;
      LogLine_t ll(line);
      if(!ll.succ) {
	cerr << "Format wrong while reading line: " << line_num << endl;
	cerr << line << endl;
	exit(9);
      }

      if(ll.path_dest_scid != NetworkGraph::mySCID) { continue; }

      // extract path info from line
      PathInfo_t p;
      p.path_origin_scid = ll.path_origin_scid;
      p.version = ll.version;
      const JString &serv = ll.strip_getValue("services");
      split(serv, p.services);

      const JString &phys = ll.strip_getValue("physPath");
      vector<JString> pVec;
      int numServ = split(phys, pVec);
      for(int i = 0; i < numServ; i++) {
	ServInst_t sInfo;

	char buf[128];
	JString_cp(buf, pVec[i]);
	char *col_pos = index(buf, ':');
	if(col_pos == NULL) {
	  cerr << "colon not found in physPath buf=" << buf << endl;
	  cerr << ll.toStr() << endl;
	  exit(35);
	}
	*col_pos = 0;
	char *sNamePos = col_pos+1;
	sInfo.gInfo.scid = atoi(buf);
	sInfo.sName = sNamePos;

	p.physPath.push_back(sInfo);
      }

      // add path info to db
      Paths::setReadPathInfo(ll.pathID, ll.path_dest_scid, p);

    } // End for(;;) reading file

    fin.close();
  } // End case cfg_read_paths

  // sanity check
  if(Paths::joint_listen && Peer::use_queue) {
    cerr << NetworkGraph::mySCID << ": joint_listen => (!use_queue)\n";
    exit(23);
  }

} // End ServComp::init()

void ServComp::startListenThread() {
  // Spawn the thread
  pthread_create(&listenThr, NULL, ServComp::listenThread, 0);
} // End ServComp::startListenThread()

void ServComp::killPath(ServCompIFPkt_t *req, UDPLibPkt_t *uReq,
			struct in_addr from_addr) {

  // get the path info for the given path
  int pathID = req->fixed.pathID;
  unsigned short path_dest_scid = req->fixed.path_dest_scid;

  vector<ServInst_t> pVec;
  vector<ServInst_t>::iterator iter = NULL;
  int success = 1;
  int numDestroyed = 0; // keep track of position in path

  if(path_dest_scid != NetworkGraph::mySCID) {
    // path_dest_scid does not match, return failure
    success = 0;
  } else {
    PathInfo_t &p = Paths::getPathInfo(pathID, path_dest_scid);
    pVec = p.physPath;
    numDestroyed = 0; // keep track of position in path

    // call destroyServInst for all portions of path that reside at 'mySCID'
    for(iter = pVec.begin(); iter != pVec.end(); iter++) {
      if(iter->gInfo.scid != NetworkGraph::mySCID) {
	break; // have to pass on the rest of the path information upstream
      }
      int succ1 = destroyServInst(pathID, NetworkGraph::mySCID,
				  p.version, iter->sName, &iter->gInfo);
      success &= succ1;
      numDestroyed++;
    }
  }

  if((path_dest_scid != NetworkGraph::mySCID) || (iter == pVec.end())) {
    // all instances have been killed, send ack back to the app
    ServCompIFPkt_t replyIFPkt;
    replyIFPkt.fixed.success = success;
    replyIFPkt.fixed.pathID = pathID;
    replyIFPkt.fixed.path_dest_scid = NetworkGraph::mySCID;
    int reply_fd = listen_fd; // same socket can be used for sending response
    int replyPort = req->fixed.port;
    int replyPktLen = replyIFPkt.toStr(uReq->data, sizeof(uReq->data));
    bool use_tunnel = false; // should not send pkt via the delaying tunnel
    int replyRes = UDPLib::sendResp
      (reply_fd, uReq, replyPktLen, from_addr, replyPort, use_tunnel, &numSent);
    if(replyRes != 1) {
      fprintf(stderr, "ServComp warning: error sending response to %s\n",
	      inet_ntoa(from_addr));
    }

    // free the packet allocated for receiving
    listen_pktBlock->freePkt(uReq);

    // clean up the space used for the path
    Paths::recyclePathID(pathID, path_dest_scid);

  } else {
    // not all instances have been killed, send request to kill
    // instances upstream

    // form the packet to send
    ServCompIFPkt_t scPkt;
    UDPLibPkt_t *uPkt = listen_pktBlock->getPkt();
    if(uPkt == NULL) {
      fprintf(stderr, "ServComp: can't alloc for uPkt\n");
      exit(85);
    }
    scPkt.fixed.type = PEER_KILL_SERV_INST;
    scPkt.fixed.path_dest_scid = NetworkGraph::mySCID;
    scPkt.fixed.pathID = pathID;
    scPkt.fixed.succSoFar = success;
    scPkt.fixed.numDestroyed = numDestroyed;
    int uPktLen = scPkt.toStr(uPkt->data, sizeof(uPkt->data));

    // get the address to send
    unsigned short nextSCID = iter->gInfo.scid;
    Peer_t *peer = NetworkGraph::getPeerWithSCID(nextSCID);

    // send the packet
    UDPLibPkt_t *u_ackPkt = listen_pktBlock->getPkt();
    if(u_ackPkt == NULL) {
      fprintf(stderr, "ServComp: can't alloc for u_ackPkt\n");
      exit(86);
    }
    struct timeval timeout = peer->getServCompTimeout();
    int send_fd = listen_fd; // can use same socket for sending
    bool use_tunnel = true;
    int reqNo = listen_rrBlock->insertRequest
      (send_fd, uPkt, uPktLen, peer->ip_addr, SERV_COMP_PORT,
       ack_listen_fd, u_ackPkt, timeout, use_tunnel, peer->scid,
       NetworkGraph::mySCID);
    if(reqNo == -1) {
      fprintf(stderr, "ServComp: no space for request\n");
      exit(67);
    }

    // add the reqNo to the queue
    int replyPort = req->fixed.port;
    bool to_app = true; // ack has to be sent to app (bypass tunnel)
    peer->sc_addReq(reqNo, uReq, from_addr, replyPort,
		    pathID, NetworkGraph::mySCID, to_app);
  }

} // End ServComp::killPath()

void ServComp::peerKillServInst(UDPLibPkt_t *reqPktRaw,
				ServCompIFPkt_t *reqIFPkt,
				ServCompIFPkt_t *replyIFPkt,
				struct in_addr from_ip_addr,
				int *replyPort, int *toReply) {

  *replyPort = SERV_COMP_ACK_PORT;
  // packet from peer: request to kill "service" instance (given pathID)
  unsigned short path_dest_scid = reqIFPkt->fixed.path_dest_scid;
  int pathID = reqIFPkt->fixed.pathID;
  unsigned short path_ver = reqIFPkt->fixed.version;
  int succSoFar = reqIFPkt->fixed.succSoFar;
  unsigned int numDestroyed = reqIFPkt->fixed.numDestroyed;

  // get the path information
  PathInfo_t &p = Paths::getPathInfo(pathID, path_dest_scid);
  vector<ServInst_t> &path = p.physPath;

  // destroy the services that have their location locally
  unsigned int i;
  for(i = numDestroyed; i < path.size(); i++) {
    if(path[i].gInfo.scid != NetworkGraph::mySCID) { break; }
    int succ1 = destroyServInst
      (pathID, path_dest_scid, path_ver, path[i].sName, &path[i].gInfo);
    succSoFar &= succ1;
  } // End for(i)
  numDestroyed = i;

  if(numDestroyed == path.size()) {
    // done killing all services for this path, return
    replyIFPkt->fixed.success = succSoFar;
    replyIFPkt->fixed.pathID = pathID;
    replyIFPkt->fixed.path_dest_scid = path_dest_scid;
  } else {
    *toReply = 0; // no need to send reply to downstream peer now
    // there are more services to be killed
    // send kill request upstream so that the rest of the service
    // instances are destroyed

    unsigned short nextSCID = path[numDestroyed].gInfo.scid;

    // form the packet to send
    ServCompIFPkt_t scPkt;
    UDPLibPkt_t *uPkt = listen_pktBlock->getPkt();
    if(uPkt == NULL) {
      fprintf(stderr, "ServComp: can't alloc for uPkt\n");
      exit(89);
    }
    scPkt.fixed.type = PEER_KILL_SERV_INST;
    scPkt.fixed.path_dest_scid = path_dest_scid;
    scPkt.fixed.pathID = pathID;
    scPkt.fixed.succSoFar = succSoFar;
    scPkt.fixed.numDestroyed = numDestroyed;
    int uPktLen = scPkt.toStr(uPkt->data, sizeof(uPkt->data));

    // get the address to send
    Peer_t *peer = NetworkGraph::getPeerWithSCID(nextSCID);

    // send the packet
    UDPLibPkt_t *u_ackPkt = listen_pktBlock->getPkt();
    if(u_ackPkt == NULL) {
      fprintf(stderr, "ServComp: can't alloc for u_ackPkt\n");
      exit(90);
    }
    struct timeval timeout = peer->getServCompTimeout();
    int send_fd = listen_fd; // can use same socket for sending too
    bool use_tunnel = true;
    int reqNo = listen_rrBlock->insertRequest
      (send_fd, uPkt, uPktLen, peer->ip_addr, SERV_COMP_PORT,
       ack_listen_fd, u_ackPkt, timeout, use_tunnel, peer->scid,
       NetworkGraph::mySCID);
    if(reqNo == -1) {
      fprintf(stderr, "ServComp: no space for request\n");
      exit(91);
    }

    // add the reqNo to the queue, also remember the pkt for the
    // ack, and the peer from whom the original request was recd
    peer->sc_addReq(reqNo, reqPktRaw, from_ip_addr,
		    SERV_COMP_ACK_PORT, pathID, path_dest_scid);

    // note down that this path no more has an edge with this peer
    peer->delPath(pathID, path_dest_scid);

  } // end case when a msg has to be sent upstream

} // End ServComp::peerKillServInst()

int ServComp::destroyServInst(int pathID, unsigned short path_dest_scid,
			      unsigned short version,
			      const JString &sName, GlobalServInfo_t *gInfo) {
  // form the packet to send
  UDPLibPkt_t uSendPkt, uRcvPkt;
  int send_len, recv_len;
  ServCompIFPkt_t scSendPkt;
  scSendPkt.fixed.type = APP_KILL_SERV_INST;
  scSendPkt.fixed.pathID = pathID;
  scSendPkt.fixed.path_dest_scid = path_dest_scid;
  scSendPkt.fixed.version = version;
  send_len = scSendPkt.toStr(uSendPkt.data, sizeof(uSendPkt.data));
  struct timeval timeout = { 2, 0 }; // 2sec

  // send the service start request to the application
  int server_send_fd = server_listen_fd; // can use same socket for sending
  bool use_tunnel = false; // should not send pkt through the delaying tunnel
  unsigned long rcv_count = 0;
  unsigned short dest_scid = NetworkGraph::INVALID_SCID;
  int res =
    UDPLib::syncReqResp(server_send_fd, &uSendPkt, send_len,
			gInfo->serviceIP, gInfo->servicePort,
			server_listen_fd, &uRcvPkt, &recv_len, &timeout,
			use_tunnel, &numSent, &rcv_count, dest_scid,
			NetworkGraph::mySCID);
  if(res == -1) {
    cerr << "ServComp: timeout waiting for response for " << sName
	 << " from " << inet_ntoa(gInfo->serviceIP) << endl;
    return 0; // failure
  } else {
    return 1; // success
  }
} // End ServComp::destroyServInst()

bool ServComp::checkPath(const vector<ServInst_t> &path) {
  for(unsigned int i = 0; i < path.size()-1; i++) {

    if(path[i].gInfo.scid != path[i+1].gInfo.scid) {

      unsigned short cost1 = NetworkGraph::getEdgeCost
	(path[i].gInfo.scid, path[i+1].gInfo.scid);
      unsigned short cost2 = NetworkGraph::getEdgeCost
	(path[i+1].gInfo.scid, path[i].gInfo.scid);
      if((cost1 == PriorityQueue<unsigned short>::INFINITY) ||
	 (cost2 == PriorityQueue<unsigned short>::INFINITY)) {
	return false;
      }

    }

  }
  return true;
} // End ServComp::checkPath()

bool ServComp::pathsEqual(const vector<ServInst_t> &p, const vector<JString> &sVec) {

  if((p.size()*2) != sVec.size()) { return false; }

  for(unsigned int i = 0; i < p.size(); i++) {
    unsigned short scid = JString_atoi(sVec[2*i]);
    const JString &sName = sVec[2*i+1];
    if((p[i].gInfo.scid != scid) || (p[i].sName != sName)) {
      return false;
    }
  }

  return true;
} // End ServComp::pathsEqual()

void ServComp::stopSend() {
  stop_send = true;
} // End ServComp::stopSend()

void ServComp::stopListenThread() {
  stop_listen = true;
  pthread_join(listenThr, NULL);
} // End ServComp::stopThreads()

unsigned long ServComp::getNumSent() {
  return numSent + listen_rrBlock->numSent;
} // End ServComp::getNumSent()

unsigned long ServComp::getNumRecd() {
  return numRecd + listen_rrBlock->numRecd;
} // End ServComp::getNumRecd()

void ServComp::printGraph() {
  vector<GraphNode_t> &nodes = NetworkGraph::nodes;
  char valStr[1024]; char *lastPos = valStr;
  *lastPos = '['; lastPos++;
  for(unsigned int i = 0; i < nodes.size(); i++) {
    unsigned short dest_scid = nodes[i].scid;
    vector<GraphEdge_t> &incoming = nodes[i].incoming;
    for(unsigned int j = 0; j < incoming.size(); j++) {
      if(incoming[j].arcCost == PriorityQueue<unsigned short>::INFINITY) {
	unsigned short origin_scid = nodes[incoming[j].srcVertex].scid;
	lastPos = longPrint(lastPos, dest_scid);
	lastPos[0] = '<'; lastPos[1] = '-'; lastPos[2] = '-'; lastPos += 3;
	lastPos = longPrint(lastPos, origin_scid);
	*lastPos = ' '; lastPos++;
      }
    }
  }
  *lastPos = ']'; lastPos++; *lastPos = 0;
  if(dbg->use_stream) {
    dbg->print() << "dead-links dead=" << valStr << endl;
  } else {
    LogLine_t ll; ll.desc = dead_links_desc;
    ll.addPair(DEAD_LK, valStr); dbg->addLL(ll);
  }
} // End ServComp::printGraph()
