/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
#ifndef SERV_COMP_H
#define SERV_COMP_H

#include <pthread.h>

#include "../link-state/Peer.h"
#include "../udp-lib/UDPPktBlock.h"
#include "../util/newran02/include.h"
#include "../util/newran02/newran.h"
#include "../util/Debug.h"
#include "ServCompIFPkt.h"
#include "TreeInfo.h"
#include "Paths.h"
#include "Bunch.h"

#define SERV_COMP_PORT 3600
#define SERV_COMP_ACK_PORT 3602
#define SERV_COMP_SERVER_PORT 3604
#define SERV_COMP_CLIENT_PORT 3608

/** The acknowledgement type */
enum ackType_t { INVALID_A=0, CREATE_A=1, REPAIR_A=2, ALT_A=3, HB_A=4 };

/** This is the main class that interfaces with the
    application/clients for composing services */
class ServComp {

  //protected:
public:
  /** max number in a single recovery msg */
  static unsigned int max_num_in_bunch;

  /** The max number of paths to patch at a time */
  static unsigned int max_to_patch_now;
  /** The min patch interval (rate control), in ms */
  static unsigned long min_patch_interval;

  /** Should I stop sending? */
  static bool stop_send;
  /** Should I stop listening? */
  static bool stop_listen;
  /** The object for handling debug log files */
  static Debug_t *dbg;

  /** Queue of path_hb_t -- recording when to send HBs */
  static queue<path_hb_t, list<path_hb_t> > hbQ;

  /*-------- Interface with the application and peers --------*/
  /** See 'startListenThread' */
  static pthread_t listenThr;
  /** The number of packets sent by this thread in syncReqResp and
      sendResp calls */
  static unsigned long numSent;
  /** The number of packets recd by this thread in blockingRecv calls */
  static unsigned long numRecd;
  

  /** The block used by the listenThread to allocate space for udp-lib
      async request/response */
  static UDPPktBlock_t *listen_pktBlock;
  /** The block used by listenThread for inserting async req/resp */
  static reqRespBlock_t *listen_rrBlock;

  /** The block used by the listenThread to allocate space for udp-lib
      async request/response -- for communication with local services */
  static UDPPktBlock_t *serv_pktBlock;
  /** The block used by listenThread for inserting async req/resp --
      for communication with local services */
  static reqRespBlock_t *serv_rrBlock;

  /** The socket descriptor used for listening to acks for service
      composition messages */
  static int ack_listen_fd;
  /** The socket descriptor used for listening to serv-comp requests,
      and message exchanges with peers */
  static int listen_fd;
  /** The socket for communication with servers */
  static int server_listen_fd;
  /** The socket for communication with clients */
  static int client_listen_fd;

  /** HACK: The last app_seq recd for APP_CREATE_PATH (assume that
      only one host generates createPath requests, with increasing
      integer values for app_seq) */
  static int last_app_seq;

  /** Implements 'listenThr' */
  static void *listenThread(void*);
  /** Helper routine for the listen thread.  Returns true if a packet
      was recd, false otherwise.  If a packet was recd, the pkt type
      is returned in 'type'.  The 'fd' argument gives the fd on which
      pkt has been received. */
  static bool listenHelper(ServCompPktType_t *type, int fd);

  /** Helper routine to check for acks for serv-comp msgs.  The
      'should_listen' argument is the same as in checkServAcks(). */
  static void checkAcks(bool should_listen);
  /** An ack is ready to be received */
  static void recvAck();
  static void checkAckHelper(int ackRecd, ackType_t aType,
			     ackInfo_t *aPtr, repair_ackInfo_t *rep_aPtr,
			     struct in_addr ack_from);

  /** Checking acks from service instances.  The 'should_listen'
      argument tells whether the routine should actually listen for
      packets.  If should_listen is false, it is assumed that a
      separate thread (listenThread) listens for acks one by one. */
  static void checkServAcks(bool should_listen);
  /** A service ack is ready to be received */
  static void recvServAck();
  static void checkSAHelper(int ackRecd, ackInfo_t *aInfo);

  /** Helper routine to check for acks for serv-comp path heart-beat
      msgs (cleaner to separate it from checkAcks).  The
      'should_listen' argument is the same as in checkServAcks(). */
  static void checkHBAcks(bool should_listen);
  static void checkHBHelper(int ackRecd, int reqNo);

  /** Routine to send path heart-beat for all paths */
  static void sendPathHBs();
  /** Routine to send the heart-beat to the next hop in the path.
      'posn' gives the position of mySCID in the path.  Returns
      success value (true/false) */
  static bool sendPathHBHelper(int pathID, unsigned short path_dest_scid,
			       unsigned short path_ver, unsigned short posn,
			       unsigned short subPosn);
  /** Routine to handle the PEER_PATH_HB message from a peer */
  static void peerPathHB(UDPLibPkt_t *reqPktRaw,
			 ServCompIFPkt_t *reqIFPkt,
			 ServCompIFPkt_t *replyIFPkt,
			 struct in_addr from_ip_addr,
			 int *replyPort, int *toReply);

  /** Create the path with the given request structure from the
      application.  The request reponse block to insert asynchronous
      requests and the packet block from which to allocate blocks for
      sending and receiving, are assumed to be those for the
      listenThread.  'uReq' is the udp-lib packet received from
      'from_addr' (both parameters to be later used for sending ack).

      Interface with the application:

      expected from app:
      KEY	VALUE
      type	<int value of APP_CREATE_PATH>
      port	<port to which reply has to be sent>
      app_seq	<a unique int value for each call> -- hack! assume
      		that only one host is sending createPath requests
      path_origin_scid	<if there is a specific scid from which the
      			path should start>
      dest_spec		<a JString that is understood by the application>
      services		<logical path, with services listed downstream
      			to upstream>

      response sent to app:
      KEY	VALUE
      success	1 on success, 0 on failure
      app_seq	for the app to match with the request it sent
      pathID	identifies the path along with path_dest_scid
      path_dest_scid identifies the path along with pathID */
  static int createPath(UDPLibPkt_t *reqPktRaw,
			ServCompIFPkt_t *reqIFPkt,
			ServCompIFPkt_t *replyIFPkt,
			struct in_addr from_ip_addr,
			int *replyPort, int *toReply,
			bool altPath);

  /** Create the local service instance.  The pathID and the
      destination SCID of this path are given (to be passed onto the
      application - for storage).  The name of the service to be
      instantiated, and the destination specification of the previous
      service are also given.

      Interface with app (service):

      sent to service:
      KEY		VALUE
      type		APP_CREATE_SERV_INST
      pathID		pathID -- identifies physical path
      path_dest_scid	path_dest_scid -- identifies physical path
      prev_dest_spec	prev_dest_spec -- where this service should
      			send data (format understood at application layer)

      expected from service:
      KEY		VALUE
      dest_spec		JString specifying where this service expects
      			data (format understoof at application layer) 

  */
  static const char *sync_createServInst
  (int pathID, unsigned short path_dest_scid, unsigned short version,
   const JString &sName, ServInst_t *sInfo, const char *prev_dest_spec,
   unsigned short serv_dest_scid);

  /** Same as sync_createServInst, but the interface with the service
      instance is asynchronous.  'toInst' gives the index into
      'p.physPath'.  The return value is the request number on
      insertion into serv_rrBlock. */
  static int createServInst
  (int pathID, unsigned short path_dest_scid, unsigned short version,
   PathInfo_t &path, int toInst, const char *prev_dest_spec,
   unsigned short serv_dest_scid);

  /** Change the service instance -- ask it to redirect data to some
      other place.

      Interface with app (service):

      sent to service:
      KEY		VALUE
      type		APP_CREATE_SERV_INST
      pathID		pathID -- identifies physical path
      path_dest_scid	path_dest_scid -- identifies physical path
      prev_dest_spec	prev_dest_spec -- where this service should
      			send data (format understood at application layer)

      expected from service:
      KEY		VALUE
      success		0 or 1
  */
  static int changeServInst(int pathID, unsigned short path_dest_scid,
			    unsigned short version,
			    GlobalServInfo_t *gInfo,
			    const char *prev_dest_spec,
			    unsigned short serv_dest_scid);

  /** Kill the path identified by the given request.  The arguments
      are the same as in createPath().  */
  static void killPath(ServCompIFPkt_t *req, UDPLibPkt_t *uReq,
		       struct in_addr from_addr);

  /** Destroy the service instance 'sName' for the given
      path/path_dest_scid.  The machine where the service is running
      is given in gInfo.  Return 0 on failure, 1 on success. */
  static int destroyServInst(int pathID, unsigned short path_dest_scid,
			     unsigned short version, const JString &sName,
			     GlobalServInfo_t *gInfo);

  /*--------- Graph related stuff --------*/
  /** Run the modified Dijkstra's algorithm for the given list of
      services.  Note: this list of services has to be in the reverse
      order.  That is, buildTree will build the data structure for
      composing services[n-1], services[n-2]... with destination
      NetworkGraph::mySCID ('n' being the number of services in the
      given vector).  See 'findPath' for explanation of fail_dest_scid
      and fail_origin_scid */
  static TreeInfo_t *buildTree(const vector<JString> &services,
			       unsigned short fail_dest_scid,
			       unsigned short fail_origin_scid);

  /** Return the path from 'origin_scid' to mySCID through list of
      services.  If origin_scid is INVALID_SCID, then caller just
      wants a path terminating at mySCID.  This list of services is
      also in reverse order -- i.e. downstream to upstream of data
      flow.  The caller can force a rebuild of the tree structure by
      setting 'forceRebuild' to be true.  Caller can specify an edge
      to have failed, by setting values for fail_dest_scid and
      fail_origin_scid (can set them to INVALID_SCID otherwise).
      Returns number of nodes in the path (-1 on failure) */
  static int findPath(unsigned short origin_scid,
		      const vector<JString> &services,
		      vector<ServInst_t> &path, bool forceRebuild,
		      unsigned short fail_dest_scid,
		      unsigned short fail_origin_scid);

  /** Checks if the given physical path is still valid -- checks
      against the graph maintained by the link-state layer.  Returns
      true if path is okay, false otherwise. */
  static bool checkPath(const vector<ServInst_t> &path);

  /** Check if the given path's physical path is the same as the one
      given in the vector of SCIDs and service names */
  static bool pathsEqual(const vector<ServInst_t> &p,
			 const vector<JString> &sVec);

  /** To handle a kill request from a peer */
  static void peerKillServInst(UDPLibPkt_t *reqPktRaw,
			       ServCompIFPkt_t *reqIFPkt,
			       ServCompIFPkt_t *replyIFPkt,
			       struct in_addr from_ip_addr,
			       int *replyPort, int *toReply);

  /** To handle path creation request from peer */
  static void peerCreateServInst(UDPLibPkt_t *reqPktRaw,
				 ServCompIFPkt_t *reqIFPkt,
				 ServCompIFPkt_t *replyIFPkt,
				 struct in_addr from_ip_addr,
				 int *replyPort, int *toReply,
				 bool altPath);

  /** To check for failures of edges */
  static void checkFailures();
  /** To check for node repeat */
  static bool checkRepeat(vector<ServInst_t> &oldPath,
			  vector<ServInst_t> &newSubPath);
  /** Patch paths from given peer */
  static void patchPathsFromPeer(Peer_t *peer);
  /** Patch paths from given peer, with rate control */
  static void patchPathsFromPeerRC(Peer_t *peer);
  /** Patch the given path that passes via the edge from the given peer */
  static void patchPath(Peer_t *peer, path_id_t *path, Bunch_t *b);
  /** Try to do the patch locally.  'posn' is the position of the
      failed edge in the physical path vector */
  static void patchEdgeLocal(Peer_t *peer, path_id_t *path,
			     PathInfo_t &pInfo, unsigned int posn,
			     Bunch_t *b);
  /** Try to do the path end2end */
  static void patchEdgeE2E(Peer_t *peer, path_id_t *path,
			   PathInfo_t &pInfo, Bunch_t *b);
  /** Called by both 'patchEdgeE2E' and 'peerPatchPathE2E' */
  static void patchEdgeE2EHelper(Peer_t *peer, path_id_t *path,
				 PathInfo_t &pInfo,
				 unsigned short fail_dest_scid,
				 unsigned short fail_origin_scid,
				 unsigned short fail_path_ver,
				 Bunch_t *b);
  /** Handle the PEER_PATCH_PATH message */
  static void peerPatchPath(UDPLibPkt_t *reqPktRaw,
			    ServCompIFPkt_t *reqIFPkt,
			    ServCompIFPkt_t *replyIFPkt,
			    struct in_addr from_ip_addr,
			    int *replyPort, int *toReply);
  /** Handle the PEER_PATCH_PATH message when it corresponds to local
      recovery */
  static void peerPatchPathLocal(UDPLibPkt_t *reqPktRaw,
				 ServCompIFPkt_t *reqIFPkt,
				 ServCompIFPkt_t *replyIFPkt,
				 struct in_addr from_ip_addr,
				 int *replyPort, int *toReply);
  /** Handle the PEER_PATCH_PATH message when it corresponds to
      end2end recovery */
  static void peerPatchPathE2E(UDPLibPkt_t *reqPktRaw,
			       ServCompIFPkt_t *reqIFPkt,
			       ServCompIFPkt_t *replyIFPkt,
			       struct in_addr from_ip_addr,
			       int *replyPort, int *toReply);
  /** Handle the PEER_PATCH_BUNCH message */
  static void peerPatchBunch(UDPLibPkt_t *reqPktRaw,
			     ServCompIFPkt_t *reqIFPkt,
			     ServCompIFPkt_t *replyIFPkt,
			     struct in_addr from_ip_addr,
			     int *replyPort, int *toReply);
  /** Send a bunch of patch packets */
  static void sendBunch(Bunch_t *b);
  /** Create an alternate path for the given path */
  static void createAltPath(Peer_t *peer, path_id_t *path, PathInfo_t &pInfo,
			    unsigned short fail_dest_scid,
			    unsigned short fail_origin_scid);

  /** Print the graph known */
  static void printGraph();

public:
  /** Init internal data structures */
  static void init(Debug_t *arg_dbg);
  /** Starts 'listenThr' -- the thread to listen on packets from
      applications and peers */
  static void startListenThread();
  /** Stop sending */
  static void stopSend();
  /** Stop the listen thread */
  static void stopListenThread();
  /** Return the number of pkts sent by all the threads */
  static unsigned long getNumSent();
  /** Return the number of packets received */
  static unsigned long getNumRecd();

  static void checkPatchQueues();
  static void checkPatchQ(Peer_t *peer);

}; // End class ServComp

#endif
