/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Implements the part of ServComp.h that relates to interfacing over
    the network (either with an application, or with a peer) */

#include <netdb.h>
#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <iostream>

#include "ServComp.h"
#include "TreeInfo.h"
#include "Paths.h"
#include "../link-state/NetworkGraph.h"
#include "../service-locn/ServiceLocn.h"

float meas[10000];
int num_meas = 0;
float max_meas = 0;
ServCompPktType_t max_type = INVALID_TYPE;
int max_recd_now_3602 = 0;
int max_recd_now_serv = 0;

void* ServComp::listenThread(void*) {

  struct timeval lastCheck;
  gettimeofday(&lastCheck, NULL);

  for(;;) {
    // do I have to stop now?
    if(stop_listen) {

      // print stats
      if(NetworkGraph::mySCID == 6) {
	float sum = 0, sum_sq = 0;
	for(int i = 0; i < num_meas; i++) {
	  sum += meas[i];
	  sum_sq += meas[i]*meas[i];
	}
	float avg = 0, se = 0;
	if(num_meas > 0) {
	  avg = sum/num_meas; se = sqrt(sum_sq/num_meas - avg*avg);
	}
	cerr << NetworkGraph::mySCID << ": listenHelper.avg=" << avg
	     << '(' << se << ')' << endl
	     << NetworkGraph::mySCID << ": listenHelper.num_meas="
	     << num_meas << endl
	     << NetworkGraph::mySCID << ": listenHelper.max_meas="
	     << max_meas << endl
	     << NetworkGraph::mySCID << ": listenHelper.max_type="
	     << ServCompIFPkt::type2str(max_type) << endl;
	cerr << NetworkGraph::mySCID
	     << ": max_recd_now_3602=" << max_recd_now_3602 << endl
	     << NetworkGraph::mySCID
	     << ": max_recd_now_serv=" << max_recd_now_serv << endl;
	cerr << NetworkGraph::mySCID << ": numRecd3600=" << numRecd << endl
	     << NetworkGraph::mySCID << ": numRecd3602="
	     << listen_rrBlock->numRecd << endl
	     << NetworkGraph::mySCID << ": totalRecd="
	     << (numRecd + listen_rrBlock->numRecd) << endl;
      }

      break;
    }

    float timeSinceLastCheck = now(lastCheck);
    if(timeSinceLastCheck >= 50) {
      // should listen separately if not joint listen
      bool should_listen = !Paths::joint_listen;
      // check for acknowledgements received
      checkAcks(should_listen);

      // check for acks from services
      checkServAcks(should_listen);

      // check for acks from path HBs
      checkHBAcks(should_listen);

      if(!stop_send) {
	// check if heart-beats have to be sent
	if(!HB::run_send_thread) { HB::all_checkSend(); }
	// check if lat-meas pkts have to be sent
	if(!LatMeas::run_send_thread) { LatMeas::all_checkSend(); }
	// check if link-state pkts have to be sent
	if(!LinkState::run_send_thread) {
	  LinkState::sendHelper();
	  // this checks for retransmits
	  LinkState::checkAcks(LS_SEND, LinkState::run_send_thread);
	}
	if(!LinkState::run_listen_thread) {
	  // this checks for retransmits
	  LinkState::checkAcks(LS_FWD, LinkState::run_listen_thread);
	}

	// check for failures of edges
	checkFailures();
	// check for patches waiting in queue
	if(Paths::rate_control) { checkPatchQueues(); }
	// send path heart-beats
	sendPathHBs();
      }

      gettimeofday(&lastCheck, NULL);
    }

    // listen for packet, and deal as appropriate
    fd_set rfds; FD_ZERO(&rfds);
    FD_SET(listen_fd, &rfds); FD_SET(client_listen_fd, &rfds);
    int max_fd = listen_fd;
    if(max_fd < client_listen_fd) { max_fd = client_listen_fd; }

    if(Paths::joint_listen) {
      // sc stuff
      FD_SET(server_listen_fd, &rfds);
      if(max_fd < server_listen_fd) { max_fd = server_listen_fd; }
      FD_SET(ack_listen_fd, &rfds);
      if(max_fd < ack_listen_fd) { max_fd = ack_listen_fd; }
    }

    if(!HB::run_listen_thread) {
      // hb listen
      FD_SET(HB::listen_fd, &rfds);
      if(HB::listen_fd > max_fd) { max_fd = HB::listen_fd; }
    }

    if(!LatMeas::run_listen_thread) {
      // lat-meas listen
      FD_SET(LatMeas::listen_fd, &rfds);
      if(LatMeas::listen_fd > max_fd) { max_fd = LatMeas::listen_fd; }
    }

    if(!LinkState::run_listen_thread) {
      // link-state listen
      FD_SET(LinkState::listen_fd, &rfds);
      if(LinkState::listen_fd > max_fd) { max_fd = LinkState::listen_fd; }
      FD_SET(LinkState::f_ack_listen_fd, &rfds);
      if(LinkState::f_ack_listen_fd > max_fd) { max_fd = LinkState::f_ack_listen_fd; }
    }

    if(!LinkState::run_send_thread || LinkState::stop_send) {
      // link-state listen
      FD_SET(LinkState::s_ack_listen_fd, &rfds);
      if(LinkState::s_ack_listen_fd > max_fd) { max_fd = LinkState::s_ack_listen_fd; }
    }

    struct timeval timeout = { 0, 50*1000 }; // 50ms
    int retval = select(max_fd+1, &rfds, NULL, NULL, &timeout);
    if(retval < 0) { perror("ServComp::listenThread.select"); continue; }

    struct timeval t1; gettimeofday(&t1, NULL);

    ServCompPktType_t type = INVALID_TYPE;

    // call appropriate routines based on which fd recd packet

    if(!HB::run_listen_thread && FD_ISSET(HB::listen_fd, &rfds)) {
      HB::recvHelper();
    }
    if(!LatMeas::run_listen_thread && FD_ISSET(LatMeas::listen_fd, &rfds)) {
      LatMeas::recvHelper();
    }
    if(!LinkState::run_listen_thread &&
       FD_ISSET(LinkState::listen_fd, &rfds)) {
      bool pkt_ready = true;
      LinkState::listenHelper(pkt_ready);
    }
    if(!LinkState::run_listen_thread &&
       FD_ISSET(LinkState::f_ack_listen_fd, &rfds)) {
      LinkState::recvAck(LS_FWD);
    }
    if((!LinkState::run_send_thread || LinkState::stop_send) &&
       FD_ISSET(LinkState::s_ack_listen_fd, &rfds)) {
      LinkState::recvAck(LS_SEND);
    }

    if(FD_ISSET(client_listen_fd, &rfds)) {
      // pkt from client recd
      listenHelper(&type, client_listen_fd);
    }
    if(FD_ISSET(listen_fd, &rfds)) {
      // pkt from peer recd
      listenHelper(&type, listen_fd);
    }

    if(Paths::joint_listen && FD_ISSET(server_listen_fd, &rfds)) {
      recvServAck(); // pkt from server (must be an ack)
    }
    if(Paths::joint_listen && FD_ISSET(ack_listen_fd, &rfds)) {
      recvAck(); // ack pkt
    }

    // measuring time
    if(num_meas < 10000) {
      meas[num_meas] = now(t1);
      if(meas[num_meas] > max_meas) {
	max_meas = meas[num_meas];
	max_type = type;
      }
      num_meas++;
    }

  } // End for(;;)

  return 0;
} // End ServComp::listenThread()

void ServComp::recvAck() {
  int reqNo = listen_rrBlock->check_one(ack_listen_fd);
  if(reqNo < 0) {
    cerr << '[' << NetworkGraph::mySCID << ":r=-1]";
    return;
  }
  struct sockaddr_in ack_from = listen_rrBlock->reqs[reqNo].recv_from;
  // have to see if this is a hb, create, repair, or alt
  ackType_t aType = INVALID_A;
  bool repeat = false;
  // is this an HB ack?
  if((reqNo < Peer::sc_hb_reqNo_vec.size()) &&
     Peer::sc_hb_reqNo_vec[reqNo]) {
    if(aType == INVALID_A) { aType = HB_A; }
    else { repeat = true; }
  }

  // is this a CREATE ack?
  if((reqNo < Peer::sc_reqNo_vec.size()) &&
     Peer::sc_reqNo_vec[reqNo].active) {
    if(aType == INVALID_A) { aType = CREATE_A; }
    else { repeat = true; }
  }

  // is this a REPAIR ack?
  if((reqNo < Peer::sc_repair_reqNo_vec.size()) &&
     Peer::sc_repair_reqNo_vec[reqNo].active) {
    if(aType == INVALID_A) { aType = REPAIR_A; }
    else { repeat = true; }
  }

  // is this an ALT ack?
  if((reqNo < Peer::sc_alt_reqNo_vec.size()) &&
     Peer::sc_alt_reqNo_vec[reqNo].active) {
    if(aType == INVALID_A) { aType = ALT_A; }
    else { repeat = true; }
  }

  // sanity check
  if(repeat) {
    cerr << NetworkGraph::mySCID << ": ack-type-repeat aType="
	 << aType << " reqNo=" << reqNo << endl;
    dbg->close(); exit(45);
  }

  int ackRecd = 1;
  if(aType == HB_A) {
    checkHBHelper(ackRecd, reqNo);
    Peer::sc_hb_reqNo_vec[reqNo] = false;

  } else if(aType == CREATE_A) {
    ackInfo_t *aPtr = &Peer::sc_reqNo_vec[reqNo].aInfo;
    repair_ackInfo_t *rep_aPtr = NULL;
    checkAckHelper(ackRecd, aType, aPtr, rep_aPtr, ack_from.sin_addr);
    Peer::sc_reqNo_vec[reqNo].active = false;

  } else if(aType == REPAIR_A) {
    ackInfo_t *aPtr = NULL;
    repair_ackInfo_t *rep_aPtr =
      &Peer::sc_repair_reqNo_vec[reqNo].rep_aInfo;
    checkAckHelper(ackRecd, aType, aPtr, rep_aPtr, ack_from.sin_addr);
    Peer::sc_repair_reqNo_vec[reqNo].active = false;

  } else if(aType == ALT_A) {
    ackInfo_t *aPtr = &Peer::sc_alt_reqNo_vec[reqNo].aInfo;
    repair_ackInfo_t *rep_aPtr = NULL;
    checkAckHelper(ackRecd, aType, aPtr, rep_aPtr, ack_from.sin_addr);
    Peer::sc_alt_reqNo_vec[reqNo].active = false;

  } else {
    cerr << NetworkGraph::mySCID << ": invalid-ack-type aType="
	 << aType << " reqNo=" << reqNo << endl;
    dbg->close(); exit(45);
  }
} // End ServComp::recvAck()

void ServComp::recvServAck() {
  int reqNo = serv_rrBlock->check_one(server_listen_fd);
  if(reqNo < 0) {
    cerr << '[' << NetworkGraph::mySCID << "s:r=-1]";
    return;
  }
  if(reqNo >= Peer::sc_serv_reqNo_vec.size()) {
    cerr << NetworkGraph::mySCID
	 << ": SC::listenThread.server_listen_fd reqNo=" << reqNo << endl;
    exit(43);
  }
  active_ackInfo_t &a_aInfo = Peer::sc_serv_reqNo_vec[reqNo];
  if(!a_aInfo.active) {
    cerr << NetworkGraph::mySCID << ": inactive reqNo=" << reqNo << endl;
    exit(45);
  }
  int ackRecd = 1;
  checkSAHelper(ackRecd, &(a_aInfo.aInfo));
  Peer::sc_serv_reqNo_vec[reqNo].active = false;
} // End ServComp::recvServAck()

bool ServComp::listenHelper(ServCompPktType_t *type, int fd) {
  // blocking receive
  UDPLibPkt_t *reqPktRaw = listen_pktBlock->getPkt();
  if(reqPktRaw == NULL) {
    cerr << "ServComp: cannot alloc for reqPktRaw" << endl;
    exit(80);
  }
  int reqPktLen;
  struct sockaddr_in from_addr;
  struct timeval timeout = { 0, 50*1000 }; // 50ms
  unsigned long rcv_count = 0; // this is the count for the number of
  // pkts recd on client_listen_fd -- this is just a dummy var
  unsigned long *rcv_count_ptr =
    (fd == client_listen_fd) ? &rcv_count : &numRecd;
  bool pkt_ready = true;
  int res = UDPLib::blockingRecv
    (fd, reqPktRaw, &reqPktLen, &from_addr, &timeout, rcv_count_ptr,
     pkt_ready);

  if(res == -1) {
    //perror("ServComp::listenThread.blockingRecv2");
    listen_pktBlock->freePkt(reqPktRaw);
    return false;
  }

  if(stop_send) {
    listen_pktBlock->freePkt(reqPktRaw);
    return true;
  }

  // parse the packet recd, and determine type
  ServCompIFPkt_t reqIFPkt;
  if(!reqIFPkt.parsePkt(reqPktRaw->data, reqPktLen)) {
    fprintf(stderr, "ServComp: error parsing pkt from %s\n",
	    inet_ntoa(from_addr.sin_addr));
    listen_pktBlock->freePkt(reqPktRaw);
    return false;
  }

  ServCompIFPkt_t replyIFPkt;
  int toReply = 1;
  int replyPort = -1;
  bool use_tunnel = true;

  if(reqIFPkt.fixed.type == APP_CREATE_PATH) {
    // pkt from appln, request to create path
    bool altPath = false;
    createPath(reqPktRaw, &reqIFPkt, &replyIFPkt,
	       from_addr.sin_addr, &replyPort, &toReply, altPath);
    use_tunnel = false;

    // End case APP_CREATE_PATH
  } else if(reqIFPkt.fixed.type == APP_KILL_PATH) {

    use_tunnel = false;
    replyPort = reqIFPkt.fixed.port;
    // pkt from appln, request to kill path
    killPath(&reqIFPkt, reqPktRaw, from_addr.sin_addr); // kill the path
    toReply = 0; // the reply would have been sent by killPath if
    // path was fully created locally, else, have to wait for ack
    // from upstream peer before sending ack to application

    // End case APP_KILL_PATH
  } else if(reqIFPkt.fixed.type == APP_FWD_DATA) {

    use_tunnel = false;
    replyPort = reqIFPkt.fixed.port;
    // pkt from appln, request to send data to another piece of the path
    // return the 'success' in the repsonse
    cerr << "APP_FWD_DATA not yet implemented" << endl;
    exit(94);

    // End case APP_FWD_DATA
  } else if(reqIFPkt.fixed.type == PEER_CREATE_SERV_INST) {

    bool altPath = false;
    peerCreateServInst(reqPktRaw, &reqIFPkt, &replyIFPkt,
		       from_addr.sin_addr, &replyPort, &toReply, altPath);

    // end case PEER_CREATE_SERV_INST
  } else if(reqIFPkt.fixed.type == PEER_KILL_SERV_INST) {

    peerKillServInst(reqPktRaw, &reqIFPkt, &replyIFPkt,
		     from_addr.sin_addr, &replyPort, &toReply);

    // End case PEER_KILL_SERV_INST
  } else if(reqIFPkt.fixed.type == PEER_FWD_DATA) {

    replyPort = SERV_COMP_ACK_PORT;
    // pkt from peer: request to send app_pkt to another piece of the path
    // sendAppPkt(&reqIFPkt, &replyIFPkt);
    cerr << "PEER_FWD_DATA not yet implemented" << endl;
    exit(95);

    // End case PEER_FWD_DATA
  } else if(reqIFPkt.fixed.type == PEER_PATCH_PATH) {

    peerPatchPath(reqPktRaw, &reqIFPkt, &replyIFPkt,
		  from_addr.sin_addr, &replyPort, &toReply);

    // End case PEER_PATCH_PATH
  } else if(reqIFPkt.fixed.type == PEER_PATCH_BUNCH) {

    peerPatchBunch(reqPktRaw, &reqIFPkt, &replyIFPkt,
		   from_addr.sin_addr, &replyPort, &toReply);

    // End case PEER_PATCH_BUNCH
  } else if(reqIFPkt.fixed.type == PEER_ALT_PATH) {

    bool altPath = true;
    peerCreateServInst(reqPktRaw, &reqIFPkt, &replyIFPkt,
		       from_addr.sin_addr, &replyPort, &toReply, altPath);

    // End case PEER_ALT_PATH
  } else if(reqIFPkt.fixed.type == PEER_PATH_HB) {

    peerPathHB(reqPktRaw, &reqIFPkt, &replyIFPkt,
	       from_addr.sin_addr, &replyPort, &toReply);

    // End case PEER_PATH_HB
  } else if(reqIFPkt.fixed.type == INVALID_TYPE) {

    cerr << "Invalid pkt type" << endl;
    toReply = 1;

    // End case INVALID_TYPE
  }

  if(toReply) { 
    int reply_fd = fd; // same socket can be used for sending response
    int replyPktLen =
      replyIFPkt.toStr(reqPktRaw->data, sizeof(reqPktRaw->data));
    int replyRes =
      UDPLib::sendResp(reply_fd, reqPktRaw, replyPktLen,
		       from_addr.sin_addr, replyPort, use_tunnel, &numSent);
    if(replyRes != 1) {
      fprintf(stderr, "ServComp warning: error sending response\n");
    }

    // free the packet allocated for receiving
    listen_pktBlock->freePkt(reqPktRaw);
  }

  *type = (ServCompPktType_t)reqIFPkt.fixed.type;
  return true;
} // End ServComp::listenHelper()

int ServComp::createPath(UDPLibPkt_t *reqPktRaw,
			 ServCompIFPkt_t *reqIFPkt,
			 ServCompIFPkt_t *replyIFPkt,
			 struct in_addr from_ip_addr,
			 int *replyPort, int *toReply,
			 bool altPath) {

  dbg->setCurrFn(CREATEPATH_FN);

  if(!altPath) {
    //cerr << "APP_CREATE_PATH from " << ip2str(from_ip_addr) << endl;
    *replyPort = reqIFPkt->fixed.port;
  }

  int pathID;
  unsigned short path_origin_scid;
  unsigned short path_ver;
  vector<JString> services;
  const char *dest_spec;
  unsigned short fail_dest_scid;
  unsigned short fail_origin_scid;

  // check for duplicate
  int app_seq = altPath ? 0 : reqIFPkt->fixed.app_seq;
  pathID = app_seq;
  if(!altPath && (app_seq < 0)) {
    if(dbg->use_stream) {
      dbg->print() << "app-seq-not-specified" << endl;
    } else {
      LogLine_t ll; ll.desc = app_seq_not_specified_desc; dbg->addLL(ll);
    }
    *toReply = 1;
    replyIFPkt->fixed.success = 0;
    return -1;
  }
  if(!altPath) {
    /** This is a HACK -- assuming that all create-path requests are
	from the same host, with increasing integer values for app_seq.
	Need to fix this later */
    if(app_seq <= last_app_seq) {
      //cerr << "duplicate APP_CREATE_PATH" << endl;

      PathInfo_t &p = Paths::getPathInfo(app_seq, NetworkGraph::mySCID);
      if(p.numCreated == p.physPath.size()) {
	// path has been created fully ==> send success
	if(dbg->use_stream) {
	  dbg->print() << "lost-ack app_seq=" << app_seq << " p_id="
		       << pathID << ':' << NetworkGraph::mySCID << endl;
	} else {
	  LogLine_t ll; ll.desc = lost_ack_desc;
	  ll.pathID = pathID;
	  ll.path_dest_scid = NetworkGraph::mySCID; dbg->addLL(ll);
	}
	replyIFPkt->fixed.success = 1;
	replyIFPkt->fixed.app_seq = app_seq;
	replyIFPkt->fixed.pathID = app_seq;
	replyIFPkt->fixed.path_dest_scid = NetworkGraph::mySCID;
	*toReply = 1;
	return app_seq;
      }
      if(dbg->use_stream) {
	dbg->print() << "dup-req app_seq=" << app_seq << " p_id="
		     << pathID << ':' << NetworkGraph::mySCID << endl;
      } else {
	LogLine_t ll; ll.desc = dup_req_desc;
	ll.pathID = pathID;
	ll.path_dest_scid = NetworkGraph::mySCID; dbg->addLL(ll);
      }
      *toReply = 0;
      listen_pktBlock->freePkt(reqPktRaw);
      return -1;
    }
    last_app_seq = app_seq;
    pathID = app_seq; // hack!
  } else {
    // this is an alternate path creation
    // need to check that there is no alt path creation in progress
    // for this path currently
    pathID = reqIFPkt->fixed.pathID;
    bool altPathExists = Paths::altPathExists(pathID, NetworkGraph::mySCID);
    if(altPathExists) {
      cerr << "(a)";
      if(dbg->use_stream) {
	dbg->print() << "alt-path-in-progress" << " p_id=" << pathID
		     << ':' << NetworkGraph::mySCID << endl;
      } else {
	LogLine_t ll; ll.desc = alt_path_in_progress_desc; ll.pathID = pathID;
	ll.path_dest_scid = NetworkGraph::mySCID; dbg->addLL(ll);
      }
      return -1;
    }

  }
  // checking for duplicates done at this point

  // get the parameters required for path construction from the request
  if(!altPath) {
    path_origin_scid = reqIFPkt->fixed.path_origin_scid;
    dest_spec = reqIFPkt->getValue(PREV_DEST_SPEC_F);
    const char *services_str = reqIFPkt->getValue(SERVICES_F);
    if((dest_spec == NULL) || (services_str == NULL)) {
      cerr << "dest_spec and services_str should be specified" << endl;
      exit(122);
    }
    // the list of services
    int numServices = str_split(services_str, services);
    if(numServices < 0) {
      cerr << "numServices is " << numServices
	   << " in ServComp::createPath" << endl;
      exit(96);
    }

    if(cfg_read_paths) {
      bool readExists = Paths::readPathExists(pathID, NetworkGraph::mySCID);
      if(!readExists) {
	cerr << NetworkGraph::mySCID << ": nothing known about pathID="
	     << pathID << endl;
	exit(35);
      }
      PathInfo_t &p1 = Paths::getReadPathInfo(pathID, NetworkGraph::mySCID);
      services = p1.services;
      path_origin_scid = p1.path_origin_scid;
    }

    // the path version
    path_ver = 100;

    // the failed edge (non-existent)
    fail_dest_scid = fail_origin_scid = NetworkGraph::INVALID_SCID;

  } else {
    PathInfo_t &p = Paths::getPathInfo(pathID, NetworkGraph::mySCID);
    path_origin_scid = p.path_origin_scid;
    fail_dest_scid = reqIFPkt->fixed.fail_dest_scid;
    fail_origin_scid = reqIFPkt->fixed.fail_origin_scid;
    dest_spec = p.dest_spec.cstr();
    services = p.services;
    path_ver = p.version + 100; // increment the version number

  }
  // end getting params for path construction

  char valStr[1024]; char *lastPos = valStr;
  *lastPos = '['; lastPos++;
  for(unsigned int i = 0; i < services.size(); i++) {
    lastPos = my_strcpy(lastPos, services[i].cstr());
    *lastPos = ' '; lastPos++;
  }
  *lastPos = ']'; lastPos++; *lastPos = 0;

  if(dbg->use_stream) {
    dbg->print() << "cp-req p_id=" << pathID << ':' << NetworkGraph::mySCID
		 << " path_origin_scid=" << path_origin_scid
		 << " fail=" << fail_dest_scid << '-' << fail_origin_scid
		 << " path_ver=" << path_ver << " services="
		 << valStr << endl;
  } else {
    LogLine_t ll; ll.desc = cp_req_desc; ll.pathID = pathID;
    ll.path_dest_scid = NetworkGraph::mySCID;
    ll.path_origin_scid = path_origin_scid;
    ll.fail_dest_scid = fail_dest_scid;
    ll.fail_origin_scid = fail_origin_scid; ll.version = path_ver;
    ll.addPair(SERVICES_LK, valStr); dbg->addLL(ll);
  }

  // call findPath
  PathInfo_t p; int pathLen;
  bool forceRebuild = false;
  vector<ServInst_t> &path = p.physPath;
  pathLen = findPath(path_origin_scid, services, path, forceRebuild,
		     fail_dest_scid, fail_origin_scid);
  if(!altPath && cfg_read_paths) {
    PathInfo_t &p1 = Paths::getReadPathInfo(pathID, NetworkGraph::mySCID);
    p.physPath = p1.physPath;
    pathLen = p.physPath.size();
  }

  bool pathOK = true;
  if(pathLen != -1) {
    // check the path for broken edges -- results when after the last
    // tree build time, some network edges have broken
    bool allEdgesExist = checkPath(path);
    if(!allEdgesExist) { pathOK = false; }
  } else {
    pathOK = false;
  }

  if(!pathOK) {
    if(dbg->use_stream) {
      dbg->print() << "graph-state-stale" << endl;
    } else {
      LogLine_t ll; ll.desc = graph_state_stale_desc; dbg->addLL(ll);
    }
    // discard the old physPath
    while(path.size() > 0) { path.pop_back(); }
    // try calling findPath again -- will use the new graph
    forceRebuild = true;
    pathLen = findPath(path_origin_scid, services, path,
		       forceRebuild, fail_dest_scid, fail_origin_scid);
  }

  // see if there exists a path at all
  if(pathLen == -1) {
    printGraph();
    if(dbg->use_stream) { dbg->print() << "no-phys-path" << endl; }
    else { LogLine_t ll; ll.desc = no_phys_path_desc; dbg->addLL(ll); }
    if(!altPath) {
      // path creation failed at this time, send failure as response
      ServCompIFPkt_t replyIFPkt;
      replyIFPkt.fixed.success = 0;
      replyIFPkt.fixed.app_seq = app_seq;
      replyIFPkt.fixed.pathID = pathID;
      replyIFPkt.fixed.path_dest_scid = NetworkGraph::mySCID;
      *toReply = 1;
      return -1;
    } else {
      return -1;
    }
  }

  // print the path
  char valStr1[1024]; char *lastPos1 = valStr1;
  *lastPos1 = '['; lastPos1++;
  for(unsigned int i = 0; i < path.size(); i++) {
    lastPos1 = longPrint(lastPos1, path[i].gInfo.scid);
    *lastPos1 = ':'; lastPos1++;
    lastPos1 = my_strcpy(lastPos1, path[i].sName.cstr());
    *lastPos1 = ' '; lastPos1++;
  }
  *lastPos1 = ']'; lastPos1++; *lastPos1 = 0;
  if(dbg->use_stream) {
    dbg->print() << "phys-path-found physPath=" << valStr1 << endl;
  } else {
    LogLine_t ll; ll.desc = phys_path_found_desc;
    ll.addPair(PHYSPATH_LK, valStr1); dbg->addLL(ll);
  }

  // record the path info
  p.dest_spec = dest_spec;
  p.services = services;
  p.path_origin_scid = path_origin_scid;
  p.lastHBRecd = (unsigned long)now(path_hb::startTime);
  p.version = path_ver;
  if(!altPath) {
    Paths::setPathInfo(pathID, NetworkGraph::mySCID, p);
  } else {
    Paths::setAltPathInfo(pathID, NetworkGraph::mySCID, p);
  }

  PathInfo_t &p1 = altPath ?
    Paths::getAltPathInfo(pathID, NetworkGraph::mySCID) :
    Paths::getPathInfo(pathID, NetworkGraph::mySCID);
  vector<ServInst_t> &path1 = p1.physPath;

  // call createServInst on the first element in the physical path
  int toInst = 0;
  if(path1[toInst].gInfo.scid != NetworkGraph::mySCID) {
    cerr << NetworkGraph::mySCID
	 << ": error in phys-path toInst=" << toInst
	 << " path[toInst].gInfo.scid=" << path1[toInst].gInfo.scid
	 << " mySCID=" << NetworkGraph::mySCID << endl;
    dbg->close();
    exit(68);
  }
  const char *prev_dest_spec = dest_spec;
  unsigned short serv_dest_scid = NetworkGraph::INVALID_SCID;
  int reqNo = createServInst
    (pathID, NetworkGraph::mySCID, path_ver, p1, toInst,
     prev_dest_spec, serv_dest_scid);

  // insert the reqNo in the queue
  bool to_app = true; // ack has to be sent to app (bypass tunnel)
  int replyPort1 = altPath ? 0 : *replyPort;
  Peer::sc_serv_addReq(reqNo, reqPktRaw, from_ip_addr, replyPort1,
		       pathID, NetworkGraph::mySCID, p1.version, to_app);

  if(!altPath) { *toReply = 0; }

  return pathID;
} // End ServComp::createPath()

const char *ServComp::sync_createServInst(int pathID,
					  unsigned short path_dest_scid,
					  unsigned short version,
					  const JString &sName,
					  ServInst_t *sInfo,
					  const char *prev_dest_spec,
					  unsigned short serv_dest_scid) {

  dbg->setCurrFn(CREATESERVINST_FN);

  // get the list of locations of the service
  const vector<GlobalServInfo_t> *gVec = ServiceLocn::getLocations(sName);
  if(gVec == NULL) {
    cerr << NetworkGraph::mySCID << ":sync_createServInst "
	 << "service " << sName << " not found at all" << endl;
    exit(72);
  }

  // find the location within the local cluster
  int localLocnIndex = -1;
  for(unsigned int i = 0; i < gVec->size(); i++) {
    if((*gVec)[i].scid == NetworkGraph::mySCID) {
      localLocnIndex = i; break;
    }
  }
  if(localLocnIndex == -1) {
    cerr << NetworkGraph::mySCID << ":sync_createServInst "
	 << "service " << sName << " not found locally" << endl;
    exit(73);
  }

  // store info in path
  sInfo->gInfo.serviceIP = (*gVec)[localLocnIndex].serviceIP;
  sInfo->gInfo.servicePort = (*gVec)[localLocnIndex].servicePort;

  /* SUPRESS
  if(dbg->use_stream) {
    dbg->print() << "cs-req sIP=" << ip2str(sInfo->gInfo.serviceIP)
		 << " sPort=" << sInfo->gInfo.servicePort << " sName="
		 << sName << endl;
  } else {
    LogLine_t ll; ll.desc = cs_req_desc;
    ll.utype = SERV_U; ll.u.serv.init();
    ll.u.serv.sIP.s_addr = sInfo->gInfo.serviceIP.s_addr;
    ll.u.serv.sPort = sInfo->gInfo.servicePort;
    ll.addPair(SNAME_LK, sName); dbg->addLL(ll);
  }
  */

  // form the packet to send
  UDPLibPkt_t uSendPkt, uRcvPkt;
  int send_len, recv_len;
  ServCompIFPkt_t scSendPkt, scRcvPkt;
  scSendPkt.fixed.type = APP_CREATE_SERV_INST;
  scSendPkt.fixed.pathID = pathID;
  scSendPkt.fixed.path_dest_scid = path_dest_scid;
  scSendPkt.fixed.version = version;
  scSendPkt.fixed.serv_dest_scid = serv_dest_scid;
  scSendPkt.fixed.serv_my_scid = NetworkGraph::mySCID;
  scSendPkt.addPair(PREV_DEST_SPEC_F, prev_dest_spec);
  send_len = scSendPkt.toStr(uSendPkt.data, sizeof(uSendPkt.data));
  struct timeval timeout = { 2, 0 }; // 2sec

  // send the service start request to the application
  int server_send_fd = server_listen_fd;
  bool use_tunnel = false; // should not send pkt through the delaying tunnel
  unsigned long rcv_count = 0;
  unsigned short dest_scid = NetworkGraph::INVALID_SCID;
  int res = UDPLib::syncReqResp
    (server_send_fd, &uSendPkt, send_len, (*gVec)[localLocnIndex].serviceIP,
     (*gVec)[localLocnIndex].servicePort, server_listen_fd, &uRcvPkt,
     &recv_len, &timeout, use_tunnel, &numSent, &rcv_count, dest_scid,
     NetworkGraph::mySCID);
  if(res == -1) {
    cerr << NetworkGraph::mySCID
	 << ": timeout waiting for response from " << sName << endl;
    exit(74);
  }

  // parse the response to get the destination specification
  int parseSuccess = scRcvPkt.parsePkt(uRcvPkt.data, recv_len);
  if(!parseSuccess) {
    cerr << "ServComp: cant parse response from " << sName << endl;
    exit(75);
  }
  sInfo->dest_spec = scRcvPkt.getValue(PREV_DEST_SPEC_F);
  return sInfo->dest_spec.cstr();
} // End ServComp::sync_createServInst()

int ServComp::createServInst(int pathID, unsigned short path_dest_scid,
			     unsigned short version,
			     PathInfo_t &p, int toInst,
			     const char *prev_dest_spec,
			     unsigned short serv_dest_scid) {

  dbg->setCurrFn(CREATESERVINST_FN);

  ServInst_t &sInfo = p.physPath[toInst];

  // get the list of locations of the service
  const vector<GlobalServInfo_t> *gVec =
    ServiceLocn::getLocations(sInfo.sName);
  if(gVec == NULL) {
    cerr << NetworkGraph::mySCID << ":createServInst "
	 << "service " << sInfo.sName << " not found at all" << endl;
    exit(72);
  }

  // find the location within the local cluster
  int localLocnIndex = -1;
  for(unsigned int i = 0; i < gVec->size(); i++) {
    if((*gVec)[i].scid == NetworkGraph::mySCID) {
      localLocnIndex = i; break;
    }
  }
  if(localLocnIndex == -1) {
    cerr << NetworkGraph::mySCID << ":createServInst "
	 << "service " << sInfo.sName << " not found locally" << endl;
    exit(73);
  }

  // store info in path
  sInfo.gInfo.serviceIP = (*gVec)[localLocnIndex].serviceIP;
  sInfo.gInfo.servicePort = (*gVec)[localLocnIndex].servicePort;

  /* SUPRESS
  if(dbg->use_stream) {
    dbg->print() << "cs-req sIP=" << ip2str(sInfo.gInfo.serviceIP)
		 << " sPort=" << sInfo.gInfo.servicePort << " sName="
		 << sInfo.sName << endl;
  } else {
    LogLine_t ll; ll.desc = cs_req_desc;
    ll.utype = SERV_U; ll.u.serv.init();
    ll.u.serv.sIP.s_addr = sInfo.gInfo.serviceIP.s_addr;
    ll.u.serv.sPort = sInfo.gInfo.servicePort;
    ll.addPair(SNAME_LK, sInfo.sName); dbg->addLL(ll);
  }
  */

  // form the packet to send
  UDPLibPkt_t *uSendPkt = serv_pktBlock->getPkt();
  UDPLibPkt_t *uRcvPkt = serv_pktBlock->getPkt();
  if((uSendPkt == NULL) || (uRcvPkt == NULL)) {
    cerr << "Can't alloc for uSendPkt/uRcvPkt\n";
    exit(99);
  }

  ServCompIFPkt_t scSendPkt, scRcvPkt;
  scSendPkt.fixed.type = APP_CREATE_SERV_INST;
  scSendPkt.fixed.pathID = pathID;
  scSendPkt.fixed.path_dest_scid = path_dest_scid;
  scSendPkt.fixed.version = version;
  scSendPkt.fixed.serv_dest_scid = serv_dest_scid;
  scSendPkt.fixed.serv_my_scid = NetworkGraph::mySCID;
  scSendPkt.addPair(PREV_DEST_SPEC_F, prev_dest_spec);
  int send_len = scSendPkt.toStr(uSendPkt->data, sizeof(uSendPkt->data));
  struct timeval timeout = { 2, 0 }; // 2sec

  // send the service start request to the application
  int server_send_fd = server_listen_fd;
  bool use_tunnel = false; // should not send pkt through the delaying tunnel
  unsigned short pkt_dest_scid = NetworkGraph::INVALID_SCID;
  int reqNo = serv_rrBlock->insertRequest
    (server_send_fd, uSendPkt, send_len,
     (*gVec)[localLocnIndex].serviceIP, (*gVec)[localLocnIndex].servicePort,
     server_listen_fd, uRcvPkt, timeout, use_tunnel, pkt_dest_scid,
     NetworkGraph::mySCID);
  if(reqNo == -1) {
    cerr << "createServInst: no space for request\n"; exit(74);
  }

  return reqNo;
} // End ServComp::createServInst()

void ServComp::peerCreateServInst(UDPLibPkt_t *reqPktRaw,
				  ServCompIFPkt_t *reqIFPkt,
				  ServCompIFPkt_t *replyIFPkt,
				  struct in_addr from_ip_addr,
				  int *replyPort, int *toReply,
				  bool altPath) {

  dbg->setCurrFn(PEERCREATESERVINST_FN);

  *replyPort = SERV_COMP_ACK_PORT;
  // packet from peer: create "service" instance with given name,
  // and given output specifications
  unsigned short path_dest_scid = reqIFPkt->fixed.path_dest_scid;
  int pathID = reqIFPkt->fixed.pathID;
  const char *createServStr = reqIFPkt->getValue(CREATE_SERV_F);
  const char *prev_dest_spec = reqIFPkt->getValue(PREV_DEST_SPEC_F);
  const char *downStreamPath = reqIFPkt->getValue(DOWNSTREAM_PATH_F);
  unsigned short path_ver = reqIFPkt->fixed.version;

  if(dbg->use_stream) {
    dbg->print() << "req p_id=" << pathID << ':' << path_dest_scid
		 << " path_ver=" << path_ver << " downStreamPath=["
		 << downStreamPath << "] createServ=[" << createServStr
		 << ']' << endl;
  } else {
    LogLine_t ll; ll.desc = req_desc; ll.pathID = pathID;
    ll.path_dest_scid = path_dest_scid; ll.version = path_ver;
    dbg->addLL(ll);
  }

  // find the downstream path (already instantiated)
  vector<char*> dsServices;
  char dsMutate[1024]; strcpy(dsMutate, downStreamPath);
  int numFields1 = str_split_mutate(dsMutate, dsServices);
  int numDSServ = numFields1/2;
  if(((numFields1 % 2) != 0) || (numDSServ == 0)) {
    cerr << NetworkGraph::mySCID << ": ServComp: numFields1 ("
	 << numFields1 << ") is not even" << endl;
    cerr << "req p_id=" << pathID << ':' << path_dest_scid
	 << " path_ver=" << path_ver << " downStreamPath=["
	 << downStreamPath << "] createServ=[" << createServStr
	 << ']' << endl;
    cerr << "downStreamPath=[" << downStreamPath << "]" << endl;
    dbg->close(); exit(88);
  }

  // check if this is a duplicate request
  bool pathExists = Paths::pathExists(pathID, path_dest_scid);
  bool altPathExists = Paths::altPathExists(pathID, path_dest_scid);
  if(!altPath) {

    if(pathExists) {
      PathInfo_t &p = Paths::getPathInfo(pathID, path_dest_scid);

      if(path_ver < p.version) {
	// must be a lost-ack for a previous path creation
	if(dbg->use_stream) {
	  dbg->print() << "lost-ack-for-past-version curr_path_ver="
		       << p.version << endl;
	} else {
	  LogLine_t ll; ll.desc = lost_ack_for_past_version_desc;
	  ll.utype = ERR_U; ll.u.err.init();
	  ll.u.err.curr_path_ver = p.version; dbg->addLL(ll);
	}
	*toReply = 1;
	replyIFPkt->fixed.success = 0;
	return;

      } else if(path_ver == p.version) {

	if(p.numCreated == p.physPath.size()) {
	  // must be a lost-ack case for current version
	  if(dbg->use_stream) {
	    dbg->print() << "lost-ack-for-curr-version" << endl;
	  } else {
	    LogLine_t ll; ll.desc = lost_ack_for_curr_version_desc;
	    dbg->addLL(ll);
	  }
	  *toReply = 1;
	  replyIFPkt->fixed.success = 1;
	  return;
	}
	if(numDSServ <= p.numCreated) {
	  // must be a dup-req case for current version
	  if(dbg->use_stream) {
	    dbg->print() << "dup-req-for-curr-version" << endl;
	  } else {
	    LogLine_t ll; ll.desc = dup_req_for_curr_version_desc;
	    dbg->addLL(ll);
	  }
	  *toReply = 0;
	  return;
	}

      }

      // End case if(pathExists)
    }

    if(altPathExists) {
      PathInfo_t &altP = Paths::getAltPathInfo(pathID, path_dest_scid);
      // this must be some old packet
      if(dbg->use_stream) {
	dbg->print() << "lost-ack-for-old-version-alt-in-construction"
		     << "curr_alt_ver=" << altP.version << endl;
      } else {
	LogLine_t ll; ll.desc = lost_ack_for_old_version_alt_in_construction_desc;
	ll.utype = ERR_U; ll.u.err.init();
	ll.u.err.curr_alt_ver = altP.version; dbg->addLL(ll);
      }
      *toReply = 1;
      replyIFPkt->fixed.success = 0;
      return;
    }

    // End case if(!altPath)
  } else {
    if(pathExists) {

      PathInfo_t &p = Paths::getPathInfo(pathID, path_dest_scid);
      if(p.version >= path_ver) {
	if(p.numCreated != p.physPath.size()) {
	  if(dbg->use_stream) {
	    dbg->print() << "err-incomplete-alt-path-in-use" << endl;
	  } else {
	    LogLine_t ll; ll.desc = err_incomplete_alt_path_in_use_desc;
	    dbg->addLL(ll);
	  }
	  *toReply = 1;
	  replyIFPkt->fixed.success = 0;
	  return;
	}

	// must be a lost-ack case
	if(dbg->use_stream) { dbg->print() << "lost-ack" << endl; }
	else { LogLine_t ll; ll.desc = lost_ack_desc; dbg->addLL(ll); }
	*toReply = 1;
	replyIFPkt->fixed.success = 1;
	return;
      }

    }

    if(altPathExists) {
      PathInfo_t &altP = Paths::getAltPathInfo(pathID, path_dest_scid);

      if(altP.version > path_ver) {

	if(dbg->use_stream) {
	  dbg->print() << "lost-ack-for-old-version curr_alt_ver="
		       << altP.version << endl;
	} else {
	  LogLine_t ll; ll.desc = lost_ack_for_old_version_desc;
	  ll.utype = ERR_U; ll.u.err.init();
	  ll.u.err.curr_alt_ver = altP.version; dbg->addLL(ll);
	}
	*toReply = 1;
	replyIFPkt->fixed.success = 0;
	return;

      } else if(altP.version == path_ver) {

	if(altP.numCreated == altP.physPath.size()) {
	  if(dbg->use_stream) {
	    dbg->print() << "err-alt-created-but-not-used" << endl;
	  } else {
	    LogLine_t ll; ll.desc = err_alt_created_but_not_used_desc;
	    dbg->addLL(ll);
	  }
	  *toReply = 1;
	  replyIFPkt->fixed.success = 1;
	  return;
	}
	if(numDSServ <= altP.numCreated) {
	  if(dbg->use_stream) { dbg->print() << "dup-req" << endl; }
	  else { LogLine_t ll; ll.desc = dup_req_desc; dbg->addLL(ll); }
	  *toReply = 0;
	  return;
	}

      }

    }

    // End case altPath
  }

  // find the subpath to instantiate
  vector<char*> serviceSpecs;
  char ssMutate[1024]; strcpy(ssMutate, createServStr);
  int numFields = str_split_mutate(ssMutate, serviceSpecs);
  if((numFields % 2) != 0) {
    cerr << NetworkGraph::mySCID << ": ServComp: numFields ("
	 << numFields << ") is not even" << endl;
    cerr << "req p_id=" << pathID << ':' << path_dest_scid
	 << " path_ver=" << path_ver << " downStreamPath=["
	 << downStreamPath << "] createServ=[" << createServStr
	 << ']' << endl;
    cerr << "createServStr=[" << createServStr << "]" << endl;
    exit(76);
  }
  int numServices = numFields/2;

  // store the path; handle case where this is a folded path separately
  bool toStoreNow = false;
  if(!altPath) {
    if(pathExists) {
      // this must be a folded path; i'm in more than one position
      PathInfo_t &p = Paths::getPathInfo(pathID, path_dest_scid);
      p.numCreated = numDSServ;
      p.version = path_ver;
    } else {
      // this is the first time i'm seeing this path
      toStoreNow = true;
    }

    // end case !altPath
  } else {
    if(altPathExists) {
      PathInfo_t &p = Paths::getAltPathInfo(pathID, path_dest_scid);
      if(p.version == path_ver) {
	// this must be a folded path; i'm in more than one position
	p.numCreated = numDSServ;
	p.version = path_ver;
      } else {
	// p.version must be < path_ver
	// the older version never got created; creating newer version now
	toStoreNow = true;
      }
    } else {
      // creating altPath for the first time
      toStoreNow = true;
    }
  }

  if(toStoreNow) {
    PathInfo_t temp_p;
    if(!altPath) {
      Paths::setPathInfo(pathID, path_dest_scid, temp_p);
    } else {
      Paths::setAltPathInfo(pathID, path_dest_scid, temp_p);
    }
    PathInfo_t &p = (altPath) ?
      Paths::getAltPathInfo(pathID, path_dest_scid) :
      Paths::getPathInfo(pathID, path_dest_scid);
    p.numCreated = numDSServ;
    p.version = path_ver;
    for(int i = 0; i < numDSServ; i++) {
      ServInst_t temp_s; p.physPath.push_back(temp_s);
      ServInst_t &s = p.physPath[p.physPath.size()-1];
      s.gInfo.scid = atoi(dsServices[2*i]);
      s.sName = dsServices[2*i+1];
    }
    for(int i = 0; i < numServices; i++) {
      ServInst_t temp_s; p.physPath.push_back(temp_s);
      ServInst_t &s = p.physPath[p.physPath.size()-1];
      s.gInfo.scid = atoi(serviceSpecs[2*i]);
      s.sName = serviceSpecs[2*i+1];
    }
  }

  PathInfo_t &p1 = altPath ?
    Paths::getAltPathInfo(pathID, path_dest_scid) :
    Paths::getPathInfo(pathID, path_dest_scid);
  p1.lastHBRecd = (unsigned long)now(path_hb::startTime);
  vector<ServInst_t> &path1 = p1.physPath;

  // instantiate the next service in the physical path
  int toInst = numDSServ;
  if(path1[toInst].gInfo.scid != NetworkGraph::mySCID) {
    cerr << NetworkGraph::mySCID
	 << ": error in phys-path toInst=" << toInst
	 << " path[toInst].gInfo.scid=" << path1[toInst].gInfo.scid
	 << " pathExists=" << pathExists << " altPathExists=" << altPathExists
	 << " mySCID=" << NetworkGraph::mySCID << endl;
    dbg->close();
    exit(68);
  }
  unsigned short serv_dest_scid = path1[toInst-1].gInfo.scid;
  int reqNo = createServInst(pathID, path_dest_scid, path_ver, p1,
			     toInst, prev_dest_spec, serv_dest_scid);

  // insert the reqNo in the queue
  bool to_app = false; // ack has to be sent to peer (don't bypass tunnel)
  unsigned short version = p1.version;
  Peer::sc_serv_addReq(reqNo, reqPktRaw, from_ip_addr, *replyPort,
		       pathID, path_dest_scid, version, to_app);

  *toReply = 0; // no need to reply now

  return;
} // End ServComp::peerCreateServInst()

int ServComp::changeServInst(int pathID, unsigned short path_dest_scid,
			     unsigned short version,
			     GlobalServInfo_t *gInfo,
			     const char *prev_dest_spec,
			     unsigned short serv_dest_scid) {

  dbg->setCurrFn(CHANGESERVINST_FN);

  if(dbg->use_stream) {
    dbg->print() << "csi-req sIP=" << ip2str(gInfo->serviceIP)
		 << " sPort=" << gInfo->servicePort << endl;
  } else {
    LogLine_t ll; ll.desc = csi_req_desc;
    ll.utype = SERV_U; ll.u.serv.init();
    ll.u.serv.sIP.s_addr = gInfo->serviceIP.s_addr;
    ll.u.serv.sPort = gInfo->servicePort; dbg->addLL(ll);
  }

  // form the packet to send
  UDPLibPkt_t uSendPkt, uRcvPkt;
  int send_len, recv_len;
  ServCompIFPkt_t scSendPkt, scRcvPkt;
  scSendPkt.fixed.type = APP_CHANGE_SERV_INST;
  scSendPkt.fixed.pathID = pathID;
  scSendPkt.fixed.path_dest_scid = path_dest_scid;
  scSendPkt.fixed.version = version;
  scSendPkt.fixed.serv_dest_scid = serv_dest_scid;
  scSendPkt.fixed.serv_my_scid = NetworkGraph::mySCID;
  scSendPkt.addPair(PREV_DEST_SPEC_F, prev_dest_spec);
  send_len = scSendPkt.toStr(uSendPkt.data, sizeof(uSendPkt.data));
  struct timeval timeout = { 2, 0 }; // 2sec

  // send the redirect request to the application
  int server_send_fd = server_listen_fd;
  bool use_tunnel = false; // should not send pkt through the delaying tunnel
  unsigned long rcv_count = 0;
  unsigned short dest_scid = NetworkGraph::INVALID_SCID;
  int res = UDPLib::syncReqResp
    (server_send_fd, &uSendPkt, send_len, gInfo->serviceIP,
     gInfo->servicePort, server_listen_fd, &uRcvPkt, &recv_len,
     &timeout, use_tunnel, &numSent, &rcv_count, dest_scid,
     NetworkGraph::mySCID);
  if(res == -1) {
    cerr << NetworkGraph::mySCID
	 << ": timeout waiting for response from "
	 << ip2str(gInfo->serviceIP) << endl;
    exit(74);
  }

  // parse the response to get the success info
  int parseSuccess = scRcvPkt.parsePkt(uRcvPkt.data, recv_len);
  if(!parseSuccess) {
    cerr << "ServComp: can't parse response from "
	 << ip2str(gInfo->serviceIP) << endl;
    exit(75);
  }

  return scRcvPkt.fixed.success;
} // End ServComp::changeServInst()

void ServComp::checkAcks(bool should_listen) {
  dbg->setCurrFn(CHECKACKS_FN);

  ackType_t aType = CREATE_A;

  if(should_listen) {
    int num_recd_now = listen_rrBlock->check_all(ack_listen_fd);
    if(num_recd_now > 100) { fprintf(stderr, "[s=%d]", num_recd_now); }
    if(num_recd_now > max_recd_now_3602) { max_recd_now_3602 = num_recd_now; }
  }

  vector<reqInfo_t> &reqs = listen_rrBlock->reqs;

  for(;;) {
    // the outer for(;;) loop is a aType loop

    //const int MAX_REQ_NOS = 1024;
    //ackInfo_t aInfo[MAX_REQ_NOS];
    //repair_ackInfo_t rep_aInfo[MAX_REQ_NOS];

    vector<ackInfo_t> &aInfo = (aType == CREATE_A) ?
      Peer::sc_vec : Peer::sc_alt_vec;
    vector<repair_ackInfo_t> &rep_aInfo = Peer::sc_repair_vec;

    // get the list of request numbers to check for
    int numAcks;
    if(aType == CREATE_A) {
      //numAcks = Peer::sc_popReqNos(aInfo, MAX_REQ_NOS);
      numAcks = Peer::sc_vec_pop();
    } else if(aType == REPAIR_A) {
      //numAcks = Peer::sc_repair_popReqNos(rep_aInfo, MAX_REQ_NOS);
      numAcks = Peer::sc_repair_vec_pop();
    } else {
      //numAcks = Peer::sc_alt_popReqNos(aInfo, MAX_REQ_NOS);
      numAcks = Peer::sc_alt_vec_pop();
    }

    for(int j = 0; j < numAcks; j++) {

      int reqNo;
      int pathID; unsigned short path_dest_scid;
      unsigned short fail_dest_scid = NetworkGraph::INVALID_SCID;
      unsigned short fail_origin_scid = NetworkGraph::INVALID_SCID;
      short posn = -1;
      UDPLibPkt_t *uPkt;
      struct in_addr toReplyAddr;
      int toReplyPort;
      bool to_app = false;
      bool local = true;
      bool active = false;

      if(aType == REPAIR_A) {
	reqNo = rep_aInfo[j].reqNo;
	pathID = rep_aInfo[j].pathID;
	path_dest_scid = rep_aInfo[j].path_dest_scid;
	fail_dest_scid = rep_aInfo[j].fail_dest_scid;
	fail_origin_scid = rep_aInfo[j].fail_origin_scid;
	posn = rep_aInfo[j].posn;
	uPkt = rep_aInfo[j].uPkt;
	toReplyPort = rep_aInfo[j].toReplyPort;
	toReplyAddr = rep_aInfo[j].toReplyAddr;
	local = rep_aInfo[j].local;
	if(!should_listen) {
	  active = Peer::sc_repair_reqNo_vec[reqNo].active;
	}
      } else {
	reqNo = aInfo[j].reqNo;
	pathID = aInfo[j].pathID;
	path_dest_scid = aInfo[j].path_dest_scid;
	uPkt = aInfo[j].uPkt;
	toReplyPort = aInfo[j].toReplyPort;
	toReplyAddr = aInfo[j].toReplyAddr;
	to_app = aInfo[j].to_app;
	if(!should_listen) {
	  active = (aType == CREATE_A) ?
	    Peer::sc_reqNo_vec[reqNo].active :
	    Peer::sc_alt_reqNo_vec[reqNo].active;
	}
      }

      int ackRecd = 0;
      if(should_listen && reqs[reqNo].recd_now) { ackRecd = 1; }
      else { ackRecd = listen_rrBlock->check_t_o_resend(reqNo); }
      if(!should_listen && (ackRecd == 1)) {
	cerr << NetworkGraph::mySCID << ": csa-err p_id=" << pathID
	     << ':' << path_dest_scid << " reqNo=" << reqNo
	     << " active=" << active << " aType=" << (int)aType << endl;
	dbg->close(); exit(44);
      }
      struct sockaddr_in ack_from = reqs[reqNo].recv_from;
      //struct sockaddr_in ack_from;
      //int ackRecd = listen_rrBlock->checkResp(reqNo, NULL, &ack_from);

      if(ackRecd == -1) {
	if(dbg->use_stream) {
	  dbg->print() << "ca-timeout host="
		       << ip2str(ack_from.sin_addr) << endl;
	} else {
	  LogLine_t ll; ll.desc = ca_timeout_desc;
	  ll.utype = ERR_U; ll.u.err.init();
	  ll.u.err.peerIP.s_addr = ack_from.sin_addr.s_addr;
	  dbg->addLL(ll);
	}
      }

      if(ackRecd != 0) {
	ackInfo_t *aPtr = (aType != REPAIR_A) ? &aInfo[j] : NULL;
	repair_ackInfo_t *rep_aPtr =
	  (aType == REPAIR_A) ? &rep_aInfo[j] : NULL;
	checkAckHelper(ackRecd, aType, aPtr, rep_aPtr, ack_from.sin_addr);
      } else {

	// no ack recd yet, and no timeout has happened either
	// push back the reqNo onto the queue for later examination
	if(aType == CREATE_A) {
	  Peer::sc_addReq(reqNo, uPkt, toReplyAddr, toReplyPort,
			  pathID, path_dest_scid, to_app);
	} else if(aType == REPAIR_A) {
	  Peer::sc_repair_addReq
	    (reqNo, uPkt, toReplyAddr, toReplyPort, pathID,
	     path_dest_scid, fail_dest_scid, fail_origin_scid, posn, local);
	} else {
	  Peer::sc_alt_addReq(reqNo, uPkt, toReplyAddr, toReplyPort,
			      pathID, path_dest_scid);
	}

      }

    } // End for(j)

    if(aType == ALT_A) { break; }
    else if(aType == REPAIR_A) { aType = ALT_A; }
    else if(aType == CREATE_A) { aType = REPAIR_A; }

  } // End for(;;) loop on aType

} // End ServComp::checkAcks()

void ServComp::checkAckHelper(int ackRecd, ackType_t aType,
			      ackInfo_t *aInfo, repair_ackInfo_t *rep_aInfo,
			      struct in_addr ack_from) {
  dbg->setCurrFn(CHECKACKS_FN);

  int reqNo;
  int pathID; unsigned short path_dest_scid;
  unsigned short fail_dest_scid = NetworkGraph::INVALID_SCID;
  unsigned short fail_origin_scid = NetworkGraph::INVALID_SCID;
  short posn = -1;
  UDPLibPkt_t *uPkt;
  struct in_addr toReplyAddr;
  int toReplyPort;
  bool to_app = false;
  bool local = true;

  if(aType == REPAIR_A) {
    reqNo = rep_aInfo->reqNo;
    pathID = rep_aInfo->pathID;
    path_dest_scid = rep_aInfo->path_dest_scid;
    fail_dest_scid = rep_aInfo->fail_dest_scid;
    fail_origin_scid = rep_aInfo->fail_origin_scid;
    posn = rep_aInfo->posn;
    uPkt = rep_aInfo->uPkt;
    toReplyPort = rep_aInfo->toReplyPort;
    toReplyAddr = rep_aInfo->toReplyAddr;
    local = rep_aInfo->local;
  } else {
    reqNo = aInfo->reqNo;
    pathID = aInfo->pathID;
    path_dest_scid = aInfo->path_dest_scid;
    uPkt = aInfo->uPkt;
    toReplyPort = aInfo->toReplyPort;
    toReplyAddr = aInfo->toReplyAddr;
    to_app = aInfo->to_app;
  }

  int success = 0;
  if(ackRecd == 1) {
    // check for success
    ServCompIFPkt_t recvP;
    int parseRes = recvP.parsePkt
      (listen_rrBlock->reqs[reqNo].recv_pkt->data,
       listen_rrBlock->reqs[reqNo].recv_len);

    if(parseRes) {
      unsigned int recvSucc = recvP.fixed.success;

      if(recvSucc) {

	if(aType == CREATE_A) {

	  // update the state about the path
	  PathInfo_t &p = Paths::getPathInfo(pathID, path_dest_scid);
	  if(p.numCreated != p.physPath.size()) {
	    p.numCreated = p.physPath.size();
	    if(path_dest_scid == NetworkGraph::mySCID) {
	      path_hb_t p_hb;
	      float nowTime = now(path_hb::startTime);
	      p_hb.pathID = pathID;
	      p_hb.version = p.version;
	      p_hb.whenToSend = nowTime + path_hb::hb_period +
		(2*path_hb::U.Next() - 1)*path_hb::hb_period_var;    
	      hbQ.push(p_hb);
	    }

	    if(dbg->use_stream) {
	      dbg->print() << "cp-done p_id=" << pathID << ':'
			   << path_dest_scid << " path_ver="
			   << p.version << endl;
	    } else {
	      LogLine_t ll; ll.desc = cp_done_desc;
	      ll.pathID = pathID; ll.path_dest_scid = path_dest_scid;
	      ll.version = p.version; dbg->addLL(ll);
	    }
	  }

	} else if(aType == REPAIR_A) {

	  if(local) {
	    PathInfo_t &rep =
	      Paths::getRepairInfo(pathID, path_dest_scid,
				   fail_dest_scid,
				   fail_origin_scid, posn);
	    rep.numCreated = rep.physPath.size();

	    PathInfo_t &p = Paths::getPathInfo(pathID, path_dest_scid);
	    // no harm in setting this too -- this would have to
	    // be set when the path-entry was created only
	    // because the sub-path passes through me
	    p.numCreated = p.physPath.size();

	    if(dbg->use_stream) {
	      dbg->print() << "repair-done p_id=" << pathID << ':'
			   << path_dest_scid << " path_ver=" << p.version
			   << " fail_dest_scid=" << fail_dest_scid
			   << " fail_origin_scid=" << fail_origin_scid
			   << " failPosn=" << posn << endl;
	    } else {
	      LogLine_t ll; ll.desc = repair_done_desc;
	      ll.pathID = pathID; ll.path_dest_scid = path_dest_scid;
	      ll.version = p.version;
	      ll.fail_dest_scid = fail_dest_scid;
	      ll.fail_origin_scid = fail_origin_scid;
	      ll.failPosn = posn; dbg->addLL(ll);
	    }

	  } else {

	    if(dbg->use_stream) {
	      dbg->print() << "ack-for-sending-e2e-patch-towards-dest"
			   << endl;
	    } else {
	      LogLine_t ll;
	      ll.desc = ack_for_sending_e2e_patch_towards_dest_desc;
	      dbg->addLL(ll);
	    }

	  }

	} else {

	  bool altExists = Paths::altPathExists(pathID, path_dest_scid);
	  if(altExists) {
	    PathInfo_t &p =
	      Paths::getAltPathInfo(pathID, path_dest_scid);
	    p.numCreated = p.physPath.size();
	    if(path_dest_scid == NetworkGraph::mySCID) {
	      path_hb_t p_hb;
	      float nowTime = now(path_hb::startTime);
	      p_hb.pathID = pathID;
	      p_hb.version = p.version;
	      p_hb.whenToSend = nowTime + path_hb::hb_period +
		(2*path_hb::U.Next() - 1)*path_hb::hb_period_var;    
	      hbQ.push(p_hb);
	    }

	    if(dbg->use_stream) {
	      dbg->print() << "alt-done p_id=" << pathID << ':'
			   << path_dest_scid << " path_ver="
			   << p.version << endl;
	    } else {
	      LogLine_t ll; ll.desc = alt_done_desc;
	      ll.pathID = pathID; ll.path_dest_scid = path_dest_scid;
	      ll.version = p.version; dbg->addLL(ll);
	    }

	    Paths::setPathInfo(pathID, path_dest_scid, p);
	    Paths::recycleAltPath(pathID, path_dest_scid);

	  }

	}

	// note that the path uses this edge (from whom the ack was
	// received)
	if((aType != REPAIR_A) || local) {
	  // this means that this is either a non-repair case,
	  // or a local-repair case (NOT a case where I sent
	  // an e2e patch request towards the path dest)

	  Peer_t *peer = NetworkGraph::getPeerWithIP(ack_from);
	  if(peer == NULL) {
	    if(dbg->use_stream) {
	      dbg->print() << "bug-ack-from-non-peer ip="
			   << ip2str(ack_from) << endl;
	    } else {
	      LogLine_t ll; ll.desc = bug_ack_from_non_peer_desc;
	      ll.utype = ERR_U; ll.u.err.init();
	      ll.u.err.peerIP.s_addr = ack_from.s_addr;
	      dbg->addLL(ll);
	    }
	    dbg->close(); exit(110);
	  } else {
	    peer->addPath(pathID, path_dest_scid);
	    success = 1;
	    if(dbg->use_stream) {
	      dbg->print() << "add-path p_id=" << pathID << ':'
			   << path_dest_scid << " peerSCID="
			   << peer->scid << endl;
	    } else {
	      LogLine_t ll; ll.desc = add_path_desc;
	      ll.pathID = pathID; ll.path_dest_scid = path_dest_scid;
	      ll.peerSCID = peer->scid; dbg->addLL(ll);
	    }
	  }

	}

	// end if(recvSucc)
      } else {
	if(dbg->use_stream) {
	  dbg->print() << "rcv-fail p_id=" << pathID << ':'
		       << path_dest_scid << endl;
	} else {
	  LogLine_t ll; ll.desc = rcv_fail_desc; ll.pathID = pathID;
	  ll.path_dest_scid = path_dest_scid; dbg->addLL(ll);
	}
      }

      // end if(parseRes)
    } else {
      if(dbg->use_stream) {
	dbg->print() << "parse-err p_id=" << pathID << ':'
		     << path_dest_scid << endl;
      } else {
	LogLine_t ll; ll.desc = parse_err_desc; ll.pathID = pathID;
	ll.path_dest_scid = path_dest_scid; dbg->addLL(ll);
      }
      dbg->close(); exit(124);
    }

  } // end case (ackRcd == 1)

  if(!success && (aType == REPAIR_A) && local) {

    // DO NOT discard repair info -- reject future requests
    // for local recovery too (path has a bad history) --
    // actually, detecting duplicates while discarding this
    // bad history is hard to implement

    if(fail_dest_scid == NetworkGraph::mySCID) {
      // local recovery failed -- have to try for e2e recovery
      Peer_t *peer = NetworkGraph::getPeerWithSCID(fail_origin_scid);
      if(peer == NULL) {
	cerr << "checkAcks: can't get peer with fail_origin_scid "
	     << fail_origin_scid << endl;
	exit(119);
      }
      // can discard repair info now
      Paths::recycleRepair
	(pathID, path_dest_scid, fail_dest_scid,
	 fail_origin_scid, posn);
      path_id_t p_id = { pathID, path_dest_scid };
      PathInfo_t &p = Paths::getPathInfo(pathID, path_dest_scid);
      p.repairExists = false;

      if(!stop_send) {
	if(dbg->use_stream) {
	  dbg->print() << "local-fail-try-e2e p_id=" << pathID
		       << ':' << path_dest_scid << " path_ver="
		       << p.version << endl;
	} else {
	  LogLine_t ll; ll.desc = local_fail_try_e2e_desc;
	  ll.pathID = pathID; ll.version = p.version;
	  ll.path_dest_scid = path_dest_scid; dbg->addLL(ll);
	}
	Bunch_t *bun = NULL;
	patchEdgeE2E(peer, &p_id, p, bun);
      }
      else {
	if(dbg->use_stream) {
	  dbg->print() << "sc-stopped-not-starting-recovery" << endl;
	} else {
	  LogLine_t ll; ll.desc = sc_stopped_not_starting_recovery_desc;
	  dbg->addLL(ll);
	}
      }
    }

  }

  // send the ack to the downstream node/peer/appln
  if(uPkt != NULL) {
    if(dbg->use_stream) {
      dbg->print() << "pass-ack succ=" << success
		   << " to_app=" << to_app << endl;
    } else {
      LogLine_t ll; ll.desc = pass_ack_desc;
      ll.success = success; dbg->addLL(ll);
    }

    ServCompIFPkt_t replyIFPkt;
    replyIFPkt.fixed.success = success;
    replyIFPkt.fixed.pathID = pathID;
    replyIFPkt.fixed.path_dest_scid = path_dest_scid;
    int reply_fd = ack_listen_fd; // same sock can be used to send resp
    int replyPort = toReplyPort;
    int replyPktLen =
      replyIFPkt.toStr(uPkt->data, sizeof(uPkt->data));
    bool use_tunnel = !to_app;
    int replyRes =
      UDPLib::sendResp(reply_fd, uPkt, replyPktLen,
		       toReplyAddr, replyPort, use_tunnel, &numSent);
    if(replyRes != 1) {
      cerr << "ServComp warning: error sending response to "
	   << ip2str(toReplyAddr) << endl;
    }

    // free the packet allocated for receiving
    listen_pktBlock->freePkt(uPkt);
  }

  // ack has been received, or timeout has happened
  // free the send and receive packets
  listen_pktBlock->freePkt(listen_rrBlock->reqs[reqNo].send_pkt);
  listen_pktBlock->freePkt(listen_rrBlock->reqs[reqNo].recv_pkt);
  // free the udp-lib entry
  listen_rrBlock->freeEntry(reqNo);
} // End ServComp::checkAckHelper()

void ServComp::sendPathHBs() {
  dbg->setCurrFn(SENDPATHHBS_FN);
  for(;;) {
    if(hbQ.empty()) { break; } // nothing to examine
    path_hb_t p_hb = hbQ.front();
    float nowTime = now(path_hb::startTime);

    if(p_hb.whenToSend <= nowTime) {
      // time to send heart-beat
      unsigned short path_dest_scid = NetworkGraph::mySCID;
      unsigned short posn = 0;
      unsigned short subPosn = 0;
      bool res = sendPathHBHelper
	(p_hb.pathID, path_dest_scid, p_hb.version, posn, subPosn);
      hbQ.pop();
      // have to add the entry back to the queue
      if(res) {
	p_hb.whenToSend = nowTime + path_hb::hb_period +
	  (2*path_hb::U.Next() - 1)*path_hb::hb_period_var;
	hbQ.push(p_hb);
      } else {
	if(dbg->use_stream) {
	  dbg->print() << "removing-old-path-hb-entry p_id=" << p_hb.pathID
		       << ':' << NetworkGraph::mySCID << " path_ver="
		       << p_hb.version << endl;
	} else {
	  LogLine_t ll; ll.desc = removing_old_path_hb_entry_desc;
	  ll.pathID = p_hb.pathID; ll.path_dest_scid = NetworkGraph::mySCID;
	  ll.version = p_hb.version; dbg->addLL(ll);
	}
      }
    } else {
      break;
    }
  } // End for(;;)
} // End ServComp::sendPathHBs()

bool ServComp::sendPathHBHelper(int pathID, unsigned short path_dest_scid,
				unsigned short path_ver, unsigned short posn,
				unsigned short subPosn) {
  dbg->setCurrFn(SENDPATHHBHELPER_FN);

  bool pathExists = Paths::pathExists(pathID, path_dest_scid);
  if(!pathExists) {
    if(dbg->use_stream) {
      dbg->print() << "non-existent-path p_id=" << pathID << ':'
		   << path_dest_scid << " posn=" << posn
		   << " subPosn=" << subPosn << endl;
    } else {
      LogLine_t ll; ll.desc = non_existent_path_desc;
      ll.pathID = pathID; ll.path_dest_scid = path_dest_scid;
      ll.posn = posn; ll.subPosn = subPosn; dbg->addLL(ll);
    }
    return false;
  }

  PathInfo_t &pInfo = Paths::getPathInfo(pathID, path_dest_scid);

  /* SUPRESS
  if(dbg->use_stream) {
    dbg->print() << "hb-req p_id=" << pathID << ':' << path_dest_scid
		 << " path_ver=" << path_ver << " posn=" << posn
		 << " subPosn=" << subPosn << endl;
  } else {
    LogLine_t ll; ll.desc = hb_req_desc; ll.pathID = pathID;
    ll.path_dest_scid = path_dest_scid; ll.version = path_ver;
    ll.posn = posn; ll.subPosn = subPosn; dbg->addLL(ll);
  }
  */

  if(path_ver != pInfo.version) {
    if(dbg->use_stream) {
      dbg->print() << "version-mismatch curr_path_ver="
		   << pInfo.version << endl;
    } else {
      LogLine_t ll; ll.desc = version_mismatch_desc;
      ll.utype = ERR_U; ll.u.err.init();
      ll.u.err.curr_path_ver = pInfo.version; dbg->addLL(ll);
    }
    return false;
  }

  if(pInfo.numCreated != pInfo.physPath.size()) {
    if(dbg->use_stream) { dbg->print() << "incomplete-path" << endl; }
    else { LogLine_t ll; ll.desc = incomplete_path_desc; dbg->addLL(ll); }
    return false;
  }
  if(posn >= pInfo.physPath.size()) {
    if(dbg->use_stream) { dbg->print() << "invalid-posn" << endl; }
    else { LogLine_t ll; ll.desc = invalid_posn_desc; dbg->addLL(ll); }
    return false;
  }

  // do all sanity checks to verify that this is indeed a valid
  // path/sub-path
  if(subPosn == 0) {
    if(pInfo.physPath[posn].gInfo.scid != NetworkGraph::mySCID) {
      if(dbg->use_stream) {
	dbg->print() << "am-not-in-posn" << endl;
      } else {
	LogLine_t ll; ll.desc = am_not_in_posn_desc; dbg->addLL(ll);
      }
      return false;
    }
  } else {
    // I must be along a local recovery sub-path
    if(!pInfo.repairExists) {
      if(dbg->use_stream) {
	dbg->print() << "unknown-repair-path" << endl;
      } else {
	LogLine_t ll; ll.desc = unknown_repair_path_desc; dbg->addLL(ll);
      }
      return false;
    }
    bool repairExists = Paths::repairExists
      (pathID, path_dest_scid, pInfo.fail_dest_scid,
       pInfo.fail_origin_scid, pInfo.failPosn);
    if(!repairExists) {
      if(dbg->use_stream) {
	dbg->print() << "repair-not-in-db" << endl;
      } else {
	LogLine_t ll; ll.desc = repair_not_in_db_desc; dbg->addLL(ll);
      }
      return false;
    }

    PathInfo_t &repInfo = Paths::getRepairInfo
      (pathID, path_dest_scid, pInfo.fail_dest_scid,
       pInfo.fail_origin_scid, pInfo.failPosn);
    if(subPosn >= (repInfo.physPath.size()-1)) {
      if(dbg->use_stream) {
	dbg->print() << "invalid-sub-posn" << endl;
      } else {
	LogLine_t ll; ll.desc = invalid_sub_posn_desc; dbg->addLL(ll);
      }
      return false;
    }

    if(repInfo.physPath[subPosn].gInfo.scid != NetworkGraph::mySCID) {
      if(dbg->use_stream) {
	dbg->print() << "am-not-in-sub-posn" << endl;
      } else {
	LogLine_t ll; ll.desc = am_not_in_sub_posn_desc; dbg->addLL(ll);
      }
      return false;
    }
  }

  // check for possible duplicate -- crude check -- HB recd too soon
  // after previous record time, and at a posn in the path that was
  // already seen
  unsigned long timeNow = (unsigned long)now(path_hb::startTime);
  if((posn <= pInfo.lastHBPosn) &&
     (timeNow <= (pInfo.lastHBRecd + path_hb::hb_period/2))) {
    if(dbg->use_stream) {
      dbg->print() << "dup-hb lastHBPosn=" << pInfo.lastHBPosn << endl;
    } else {
      LogLine_t ll; ll.desc = dup_hb_desc;
      ll.utype = ERR_U; ll.u.err.init();
      ll.u.err.lastHBPosn = pInfo.lastHBPosn; dbg->addLL(ll);
    }
    return true;
  }

  // find nextSCID
  unsigned short nextSCID = NetworkGraph::INVALID_SCID;
  unsigned short nextPosn = 0;
  unsigned short nextSubPosn = 0;

  if(subPosn == 0) {

    if(pInfo.repairExists && (pInfo.fail_dest_scid == NetworkGraph::mySCID)) {
      PathInfo_t &repInfo = Paths::getRepairInfo
	(pathID, path_dest_scid, pInfo.fail_dest_scid,
	 pInfo.fail_origin_scid, pInfo.failPosn);
      nextSCID = repInfo.physPath[1].gInfo.scid;
      nextPosn = posn;
      nextSubPosn = 1;

    } else {

      for(unsigned short i = posn+1; i < pInfo.physPath.size(); i++) {
	if(pInfo.physPath[i].gInfo.scid != NetworkGraph::mySCID) {
	  nextSCID = pInfo.physPath[i].gInfo.scid;
	  nextPosn = i;
	  nextSubPosn = 0;
	  break;
	}
      }

    }

  } else {

    PathInfo_t &repInfo = Paths::getRepairInfo
      (pathID, path_dest_scid, pInfo.fail_dest_scid,
       pInfo.fail_origin_scid, pInfo.failPosn);
    if(subPosn == (repInfo.physPath.size()-2)) {
      // break out of sub-path
      nextSubPosn = 0;
      nextPosn = posn+1;
      nextSCID = pInfo.fail_origin_scid;
    } else {
      nextSubPosn = subPosn+1;
      nextPosn = posn;
      nextSCID = repInfo.physPath[nextSubPosn].gInfo.scid;
    }

  }

  // mark that this path is still alive -- this is true irrespective
  // of whether i'm sending or forwarding path HBs
  pInfo.lastHBRecd = timeNow;
  // note down the last position from where HB was forwarded -- for
  // use in detecting future duplicates
  pInfo.lastHBPosn = nextPosn;

  if(nextSCID == NetworkGraph::INVALID_SCID) {
    // path ends here, no need to forward HB
    /* SUPRESS
    if(dbg->use_stream) { dbg->print() << "path-ends-here" << endl; }
    else { LogLine_t ll; ll.desc = path_ends_here_desc; dbg->addLL(ll); }
    */
    return true;
  }

  /* SUPRESS
  if(dbg->use_stream) {
    dbg->print() << "send-or-fwd-hb nextSCID=" << nextSCID
		 << " nextPosn=" << nextPosn << " nextSubPosn="
		 << nextSubPosn << endl;
  } else {
    LogLine_t ll; ll.desc = send_or_fwd_hb_desc;
    ll.nextSCID = nextSCID; ll.posn = nextPosn;
    ll.subPosn = nextSubPosn; dbg->addLL(ll);
  }
  */

  // form the packet to send
  ServCompIFPkt_t scPkt;
  UDPLibPkt_t *uPkt = listen_pktBlock->getPkt();
  if(uPkt == NULL) {
    fprintf(stderr, "sendPathHBHelper: can't alloc for uPkt\n");
    exit(65);
  }
  scPkt.fixed.type = PEER_PATH_HB;
  scPkt.fixed.path_dest_scid = path_dest_scid;
  scPkt.fixed.pathID = pathID;
  scPkt.fixed.posn = nextPosn;
  scPkt.fixed.subPosn = nextSubPosn;
  scPkt.fixed.version = path_ver;
  int uPktLen = scPkt.toStr(uPkt->data, sizeof(uPkt->data));

  // get the address to send
  Peer_t *peer = NetworkGraph::getPeerWithSCID(nextSCID);
  if(peer == NULL) {
    if(dbg->use_stream) {
      dbg->print() << "send-or-fwd-hb-to-unknown-peer nextSCID="
		   << nextSCID << endl;
    } else {
      LogLine_t ll; ll.desc = send_or_fwd_hb_to_unknown_peer_desc;
      ll.nextSCID = nextSCID; dbg->addLL(ll);
    }
    dbg->close(); exit(129);
  }

  // send the packet
  UDPLibPkt_t *u_ackPkt = listen_pktBlock->getPkt();
  if(u_ackPkt == NULL) {
    fprintf(stderr, "sendPathHBHelper: can't alloc for u_ackPkt\n");
    exit(66);
  }
  struct timeval timeout = peer->getServCompTimeout();
  int send_fd = listen_fd; // can use same socket for sending
  bool use_tunnel = true;
  int reqNo = listen_rrBlock->insertRequest
    (send_fd, uPkt, uPktLen, peer->ip_addr, SERV_COMP_PORT,
     ack_listen_fd, u_ackPkt, timeout, use_tunnel, peer->scid,
     NetworkGraph::mySCID);
  if(reqNo == -1) {
    fprintf(stderr, "sendPathHBHelper: no space for request\n");
    exit(67);
  }

  // add the reqNo to the queue
  Peer::sc_hb_addReq(reqNo);

  return true;

} // End ServComp::sendPathHBHelper()

void ServComp::peerPathHB(UDPLibPkt_t *reqPktRaw,
			  ServCompIFPkt_t *reqIFPkt,
			  ServCompIFPkt_t *replyIFPkt,
			  struct in_addr from_ip_addr,
			  int *replyPort, int *toReply) {

  unsigned short path_dest_scid = reqIFPkt->fixed.path_dest_scid;
  unsigned short pathID = reqIFPkt->fixed.pathID;
  unsigned short posn = reqIFPkt->fixed.posn;
  unsigned short subPosn = reqIFPkt->fixed.subPosn;
  unsigned short path_ver = reqIFPkt->fixed.version;

  *replyPort = SERV_COMP_ACK_PORT;

  bool succ = sendPathHBHelper(pathID, path_dest_scid, path_ver, posn, subPosn);
  if(!succ) {
    //cerr << "[p]";
  }
  *toReply = 1;
  replyIFPkt->fixed.success = succ;

} // End ServComp::peerPathHB()

void ServComp::checkHBAcks(bool should_listen) {
  dbg->setCurrFn(CHECKHBACKS_FN);

  //const int MAX_REQ_NOS = 128;
  //int reqNoList[MAX_REQ_NOS];
  //int numAcks = Peer::sc_hb_popReqNos(reqNoList, MAX_REQ_NOS);

  if(should_listen) {
    int num_recd_now = listen_rrBlock->check_all(ack_listen_fd);
    if(num_recd_now > 100) { fprintf(stderr, "[sa=%d]", num_recd_now); }
    if(num_recd_now > max_recd_now_3602) { max_recd_now_3602 = num_recd_now; }
  }

  vector<reqInfo_t> &reqs = listen_rrBlock->reqs;
  vector<int> &reqNoList = Peer::sc_hb_vec;
  int numAcks = Peer::sc_hb_vec_pop();

  for(int j = 0; j < numAcks; j++) {
    int reqNo = reqNoList[j];

    int ackRecd = 0;
    if(should_listen && reqs[reqNo].recd_now) { ackRecd = 1; }
    else { ackRecd = listen_rrBlock->check_t_o_resend(reqNo); }
    if(!should_listen && (ackRecd == 1)) {
      cerr << NetworkGraph::mySCID << ": chb-ack-err reqNo=" << reqNo
	   << " active=" << Peer::sc_hb_reqNo_vec[reqNo] << endl;
      dbg->close(); exit(44);
    }
    struct sockaddr_in ack_from = reqs[reqNo].recv_from;
    //struct sockaddr_in ack_from;
    //int ackRecd = listen_rrBlock->checkResp(reqNo, NULL, &ack_from);

    if(ackRecd != 0) {
      /* SUPRESS
      if(dbg->use_stream) {
	dbg->print() << "hb-ack ackRecd=" << ackRecd << " host="
		     << ip2str(ack_from.sin_addr) << endl;
      } else {
	LogLine_t ll; ll.desc = hb_ack_desc; ll.success = ackRecd;
	ll.utype = ERR_U; ll.u.err.init();
	ll.u.err.peerIP.s_addr = ack_from.sin_addr.s_addr; dbg->addLL(ll);
      }
      */
      checkHBHelper(ackRecd, reqNo);
    } else {
      Peer::sc_hb_addReq(reqNo);
    }

  } // End for(j)

} // End ServComp::checkHBAcks()

void ServComp::checkHBHelper(int ackRecd, int reqNo) {
  // ack has been received, or timeout has happened
  // free the send and receive packets
  listen_pktBlock->freePkt(listen_rrBlock->reqs[reqNo].send_pkt);
  listen_pktBlock->freePkt(listen_rrBlock->reqs[reqNo].recv_pkt);
  // free the udp-lib entry
  listen_rrBlock->freeEntry(reqNo);
} // End ServComp::checkHBHelper()

void ServComp::checkServAcks(bool should_listen) {
  //const int MAX_REQ_NOS = 128;
  //ackInfo_t aInfo[MAX_REQ_NOS];
  //int numAcks = Peer::sc_serv_popReqNos(aInfo, MAX_REQ_NOS);

  if(should_listen) {
    int num_recd_now = serv_rrBlock->check_all(server_listen_fd);
    if(num_recd_now > 100) { fprintf(stderr, "[sa=%d]", num_recd_now); }
    if(num_recd_now > max_recd_now_serv) { max_recd_now_serv = num_recd_now; }
  }

  vector<reqInfo_t> &reqs = serv_rrBlock->reqs;
  vector<ackInfo_t> &aInfo = Peer::sc_serv_vec;
  int numAcks = Peer::sc_serv_vec_pop();

  for(int j = 0; j < numAcks; j++) {
    int reqNo = aInfo[j].reqNo;
    int pathID = aInfo[j].pathID;
    unsigned short path_dest_scid = aInfo[j].path_dest_scid;
    UDPLibPkt_t *uPkt = aInfo[j].uPkt;
    int toReplyPort = aInfo[j].toReplyPort;
    struct in_addr toReplyAddr = aInfo[j].toReplyAddr;
    bool to_app = aInfo[j].to_app;
    unsigned short version = aInfo[j].version;

    int ackRecd = 0;
    if(should_listen && reqs[reqNo].recd_now) { ackRecd = 1; }
    else { ackRecd = serv_rrBlock->check_t_o_resend(reqNo); }
    if(!should_listen && (ackRecd == 1)) {
      cerr << NetworkGraph::mySCID << ": csa-err p_id=" << pathID
	   << ':' << path_dest_scid << " path_ver=" << version
	   << " reqNo=" << reqNo << " active="
	   << Peer::sc_serv_reqNo_vec[reqNo].active << endl;
      dbg->close(); exit(44);
    }
    struct sockaddr_in ack_from = reqs[reqNo].recv_from;
    //struct sockaddr_in ack_from;
    //int ackRecd = serv_rrBlock->checkResp(reqNo, NULL, &ack_from);

    if(ackRecd != 0) {
      checkSAHelper(ackRecd, &aInfo[j]);
    } else {
      Peer::sc_serv_addReq(reqNo, uPkt, toReplyAddr, toReplyPort,
			   pathID, path_dest_scid, version, to_app);
    }

  } // End for(j)

} // End ServComp::checkServAcks()

void ServComp::checkSAHelper(int ackRecd, ackInfo_t *aInfo) {
  dbg->setCurrFn(CHECKSERVACKS_FN);

  int reqNo = aInfo->reqNo;
  int pathID = aInfo->pathID;
  unsigned short path_dest_scid = aInfo->path_dest_scid;
  UDPLibPkt_t *uPkt = aInfo->uPkt;
  int toReplyPort = aInfo->toReplyPort;
  struct in_addr toReplyAddr = aInfo->toReplyAddr;
  bool to_app = aInfo->to_app;
  unsigned short version = aInfo->version;
  bool altPath = (version != 100);

  // sanity check -- path exists
  bool pathExists = (altPath) ?
    Paths::altPathExists(pathID, path_dest_scid) :
    Paths::pathExists(pathID, path_dest_scid);
  if(!pathExists) {
    if(dbg->use_stream) {
      dbg->print() << "path-not-found p_id=" << pathID << ':'
		   << path_dest_scid << " path_ver=" << version << endl;
    } else {
      LogLine_t ll; ll.desc = path_not_found_desc;
      ll.pathID = pathID; ll.path_dest_scid = path_dest_scid;
      ll.version = version; dbg->addLL(ll);
    }
    dbg->close(); exit(77);
  }

  // get path info
  PathInfo_t &p = (altPath) ?
    Paths::getAltPathInfo(pathID, path_dest_scid) :
    Paths::getPathInfo(pathID, path_dest_scid);

  // sanity check -- numCreated
  if(p.numCreated == p.physPath.size()) {
    // what is this ack?
    if(dbg->use_stream) {
      dbg->print() << "path-complete p_id=" << pathID << ':'
		   << path_dest_scid << " path_ver=" << version
		   << " reqNo=" << reqNo << endl;
    } else {
      LogLine_t ll; ll.desc = path_complete_desc;
      ll.pathID = pathID; ll.path_dest_scid = path_dest_scid;
      ll.version = version; dbg->addLL(ll);
    }
    dbg->close(); exit(78);
  }

  // sanity check -- version
  if(p.version != version) {
    // what is this ack?
    if(dbg->use_stream) {
      dbg->print() << "version-mismatch p_id=" << pathID << ':'
		   << path_dest_scid << " path_ver=" << version << endl;
    } else {
      LogLine_t ll; ll.desc = version_mismatch_desc;
      ll.pathID = pathID; ll.path_dest_scid = path_dest_scid;
      ll.version = version; dbg->addLL(ll);
    }
    dbg->close(); exit(78);
  }

  vector<ServInst_t> &path = p.physPath;
  ServInst_t &sDone = path[p.numCreated];

  // if timeout, die
  if(ackRecd == -1) {
    cerr << NetworkGraph::mySCID << ": timeout waiting for response from "
	 << sDone.sName << endl;
    exit(74);
  }

  // check for success, get dest_spec
  // parse the response to get the destination specification
  UDPLibPkt_t *recv_pkt = serv_rrBlock->reqs[reqNo].recv_pkt;
  int recv_len = serv_rrBlock->reqs[reqNo].recv_len;
  ServCompIFPkt_t scRcvPkt;
  int parseSuccess = scRcvPkt.parsePkt(recv_pkt->data, recv_len);
  if(!parseSuccess) {
    cerr << "ServComp: cant parse response from " << sDone.sName << endl;
    exit(75);
  }
  sDone.dest_spec = scRcvPkt.getValue(PREV_DEST_SPEC_F);

  // note that we're one step ahead in phys path creation
  p.numCreated++;
  int toInst = p.numCreated;
  const char *prev_dest_spec = sDone.dest_spec.cstr();

  // if done, send ack
  // else, create next service, or send request upstream
  if(p.numCreated == p.physPath.size()) {
    if(dbg->use_stream) {
      dbg->print() << "cp-done p_id=" << pathID << ':' << path_dest_scid
		   << " path_ver=" << version << " reqNo=" << reqNo << endl;
    } else {
      LogLine_t ll; ll.desc = cp_done_desc;
      ll.pathID = pathID; ll.path_dest_scid = path_dest_scid;
      ll.version = version; dbg->addLL(ll);
    }
    if(altPath) {
      Paths::setPathInfo(pathID, path_dest_scid, p);
      Paths::recycleAltPath(pathID, path_dest_scid);
    }

    // send ack
    if(uPkt != NULL) {
      ServCompIFPkt_t replyIFPkt;
      replyIFPkt.fixed.success = 1;
      replyIFPkt.fixed.pathID = pathID;
      replyIFPkt.fixed.path_dest_scid = path_dest_scid;
      int reply_fd = ack_listen_fd; // same sock can be used to send resp
      int replyPort = toReplyPort;
      int replyPktLen =
	replyIFPkt.toStr(uPkt->data, sizeof(uPkt->data));
      bool use_tunnel = !to_app;
      int replyRes =
	UDPLib::sendResp(reply_fd, uPkt, replyPktLen,
			 toReplyAddr, replyPort, use_tunnel, &numSent);
      if(replyRes != 1) {
	cerr << "ServComp warning: error sending response to "
	     << ip2str(toReplyAddr) << endl;
      }

      // free the packet allocated for receiving
      serv_pktBlock->freePkt(uPkt);
    }

    // end case when path creation is done
  } else {

    if(path[toInst].gInfo.scid == NetworkGraph::mySCID) {
      // next service instance is also here
      unsigned short serv_dest_scid = path[toInst-1].gInfo.scid;
      int next_reqNo = createServInst
	(pathID, path_dest_scid, version, p, toInst, prev_dest_spec,
	 serv_dest_scid);

      // add new reqNo to queue
      Peer::sc_serv_addReq(next_reqNo, uPkt, toReplyAddr, toReplyPort,
			   pathID, path_dest_scid, version, to_app);

      // end case when next service instance is also local
    } else {
      // form the pkt to send
      char dsBuf[1024]; char *dsLast = dsBuf;
      for(unsigned int i = 0; i < p.numCreated; i++) {
	dsLast = longPrint(dsLast, path[i].gInfo.scid);
	*dsLast = ' '; dsLast++;
	dsLast = my_strcpy(dsLast, path[i].sName.cstr());
	*dsLast = ' '; dsLast++;
      }
      *dsLast = 0;

      char csBuf[1024]; char *csLast = csBuf;
      for(unsigned int i = p.numCreated; i < p.physPath.size(); i++) {
	csLast = longPrint(csLast, path[i].gInfo.scid);
	*csLast = ' '; csLast++;
	csLast = my_strcpy(csLast, path[i].sName.cstr());
	*csLast = ' '; csLast++;
      }
      *csLast = 0;

      unsigned short nextSCID = path[toInst].gInfo.scid;

      if(dbg->use_stream) {
	dbg->print() << "pc-send-upstream p_id=" << pathID << ':'
		     << path_dest_scid << " path_ver=" << version
		     << " nextSCID=" << nextSCID << endl;
      } else {
	LogLine_t ll; ll.desc = pc_send_upstream_desc; ll.nextSCID = nextSCID;
	ll.pathID = pathID; ll.path_dest_scid = path_dest_scid;
	ll.version = version; dbg->addLL(ll);
      }

      ServCompIFPkt_t scPkt;
      UDPLibPkt_t *send_uPkt = listen_pktBlock->getPkt();
      if(send_uPkt == NULL) {
	fprintf(stderr, "checkServAcks: can't alloc for send_uPkt\n");
	exit(65);
      }
      if(!altPath) {
	scPkt.fixed.type = PEER_CREATE_SERV_INST;
      } else {
	scPkt.fixed.type = PEER_ALT_PATH;
      }
      scPkt.fixed.path_dest_scid = path_dest_scid;
      scPkt.fixed.pathID = pathID;
      scPkt.fixed.version = version;
      scPkt.addPair(CREATE_SERV_F, csBuf);
      scPkt.addPair(DOWNSTREAM_PATH_F, dsBuf);
      scPkt.addPair(PREV_DEST_SPEC_F, prev_dest_spec);
      int uPktLen = scPkt.toStr(send_uPkt->data, sizeof(send_uPkt->data));

      // get the address to send
      Peer_t *peer = NetworkGraph::getPeerWithSCID(nextSCID);

      // send the packet
      UDPLibPkt_t *u_ackPkt = listen_pktBlock->getPkt();
      if(u_ackPkt == NULL) {
	fprintf(stderr, "checkServAcks: can't alloc for u_ackPkt\n");
	exit(66);
      }
      struct timeval timeout = peer->getServCompTimeout();
      int send_fd = listen_fd; // can use same socket for sending
      bool use_tunnel = true;
      int next_reqNo = listen_rrBlock->insertRequest
	(send_fd, send_uPkt, uPktLen, peer->ip_addr, SERV_COMP_PORT,
	 ack_listen_fd, u_ackPkt, timeout, use_tunnel, peer->scid,
	 NetworkGraph::mySCID);
      if(reqNo == -1) {
	fprintf(stderr, "checkServAcks: no space for request\n");
	exit(67);
      }

      // add the reqNo to the queue
      if(!altPath) {
	Peer::sc_addReq(next_reqNo, uPkt, toReplyAddr, toReplyPort,
			pathID, path_dest_scid, to_app);
      } else {
	Peer::sc_alt_addReq(next_reqNo, uPkt, toReplyAddr, toReplyPort,
			    pathID, path_dest_scid);
      }

      // end case when next service instance is an upstream peer
    }

    // end case when path creation was continued
  }

  // ack has been received, or timeout has happened
  // free the send and receive packets
  serv_pktBlock->freePkt(serv_rrBlock->reqs[reqNo].send_pkt);
  serv_pktBlock->freePkt(serv_rrBlock->reqs[reqNo].recv_pkt);
  // free the udp-lib entry
  serv_rrBlock->freeEntry(reqNo);
} // End ServComp::checkSAHelper()
