/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Implements ServCompIFPkt.h */

#include <stdio.h>

#include "ServCompIFPkt.h"
#include "../util/Util.h"

FixedPkt::FixedPkt() {
  /** Packet type */
  type = INVALID_TYPE;

  /** Path info */
  pathID = app_seq = port = -1;
  path_dest_scid = path_origin_scid = LogLine::INVALID_SCID;
  version = 0;

  /** Path traversal info */
  posn = subPosn = -1;

  /** Repair info */
  repType = INVALID_F;
  fail_dest_scid = fail_origin_scid = LogLine::INVALID_SCID;
  fail_version = 0;
  failPosn = -1;

  /** Failure notification */
  failure_scid = LogLine::INVALID_SCID;
  succSoFar = 0;
  numDestroyed = -1;

  /** Communication with services */
  serv_dest_scid = serv_my_scid = serv_upstream_scid = LogLine::INVALID_SCID;

  /** Success indication */
  success = 0;

  /** The number of bunched failure info */
  num_in_bunch = 0;
  /** The number of string fields */
  numStr = 0;

} // End FixedPkt::FixedPkt()

StrField::StrField() {
  key = INVALID_F;
  len = 0;
} // End StrField::StrField()

StrField::StrField(const StrField &sf) {
  key = sf.key; len = sf.len;
  strncpy(value, sf.value, sf.len);
  value[len] = 0;
} // End StrField::StrField(StrField&)

const int num_pkt_types = 20;
pkt_type_name_t ServCompIFPkt::pkt_type_name_map[num_pkt_types] = {
  { INVALID_TYPE, "INVALID_TYPE" },

  { APP_CREATE_PATH, "APP_CREATE_PATH" },
  { APP_KILL_PATH, "APP_KILL_PATH" },
  { APP_FWD_DATA, "APP_FWD_DATA" },

  { PEER_CREATE_SERV_INST, "PEER_CREATE_SERV_INST" },
  { PEER_KILL_SERV_INST, "PEER_KILL_SERV_INST" },
  { PEER_FWD_DATA, "PEER_FWD_DATA" },
  { PEER_PATCH_PATH, "PEER_PATCH_PATH" },
  { PEER_ALT_PATH, "PEER_ALT_PATH" },
  { PEER_PATH_HB, "PEER_PATH_HB" },
  { PEER_PATCH_BUNCH, "PEER_PATCH_BUNCH" },

  { APP_CREATE_SERV_INST, "APP_CREATE_SERV_INST" },
  { APP_KILL_SERV_INST, "APP_KILL_SERV_INST" },
  { APP_CHANGE_SERV_INST, "APP_CHANGE_SERV_INST" },

  { APP_UPSTREAM_INFO, "APP_UPSTREAM_INFO" },
  { APP_PATH_STATE, "APP_PATH_STATE" },
  { APP_HB, "APP_HB" },
  { APP_KILL, "APP_KILL" },

  { TTS_REQ, "TTS_REQ" },
  { TTS_RESP, "TTS_RESP" },
};

const char *ServCompIFPkt::type2str(ServCompPktType_t type) {
  for(int i = 0; i < num_pkt_types; i++) {
    if(pkt_type_name_map[i].type == type) {
      return pkt_type_name_map[i].name;
    }
  }
  return NULL;
} // End ServCompIFPkt::type2str()

int ServCompIFPkt::parsePkt(char *data, int len) {
  if(len < sizeof(FixedPkt_t)) { fprintf(stderr, "1\n"); return 0; }
  FixedPkt_t *fp = (FixedPkt_t*)data;

  /** Packet type */
  fixed.type = ntohs(fp->type);

  /** Path info */
  fixed.pathID = ntohl(fp->pathID); fixed.app_seq = ntohl(fp->app_seq);
  fixed.port = ntohl(fp->port);
  fixed.path_dest_scid = ntohs(fp->path_dest_scid);
  fixed.path_origin_scid = ntohs(fp->path_origin_scid);
  fixed.version = ntohs(fp->version);

  /** Path traversal info */
  fixed.posn = ntohs(fp->posn); fixed.subPosn = ntohs(fp->subPosn);

  /** Repair info */
  fixed.repType = ntohs(fp->repType);
  fixed.fail_dest_scid = ntohs(fp->fail_dest_scid);
  fixed.fail_origin_scid = ntohs(fp->fail_origin_scid);
  fixed.fail_version = ntohs(fp->fail_version);
  fixed.failPosn = ntohs(fp->failPosn);

  /** Failure notification */
  fixed.failure_scid = ntohs(fp->failure_scid);
  fixed.succSoFar = ntohs(fp->succSoFar);
  fixed.numDestroyed = ntohs(fp->numDestroyed);

  /** Communication with services */
  fixed.serv_dest_scid = ntohs(fp->serv_dest_scid);
  fixed.serv_my_scid = ntohs(fp->serv_my_scid);
  fixed.serv_upstream_scid = ntohs(fp->serv_upstream_scid);

  /** Success indication */
  fixed.success = ntohs(fp->success);

  /** The number of bunched failure info */
  fixed.num_in_bunch = ntohs(fp->num_in_bunch);
  /** The number of string fields */
  fixed.numStr = ntohs(fp->numStr);

  if(len < (sizeof(FixedPkt_t) + fixed.num_in_bunch*sizeof(FailedPath_t))) {
    fprintf(stderr, "2\n"); return 0;
  }

  FailedPath_t *failed_p = (FailedPath_t*)(data + sizeof(FixedPkt_t));
  for(unsigned int i = 0; i < fixed.num_in_bunch; i++) {
    FailedPath_t fp1;
    fp1.pathID = ntohl(failed_p[i].pathID);
    fp1.path_dest_scid = ntohs(failed_p[i].path_dest_scid);
    fp1.version = ntohs(failed_p[i].version);
    failed.push_back(fp1);
  }

  char *s = (char*)(data + sizeof(FixedPkt_t) + fixed.num_in_bunch*sizeof(FailedPath_t));
  StrField_t *sf_p = (StrField_t*)s;
  for(unsigned int i = 0; i < fixed.numStr; i++) {
    StrField_t sf;
    sf.len = ntohs(sf_p->len);
    int this_sf_len = (sizeof(sf) - sizeof(sf.value)) + sf.len;
    if(len < (sizeof(FixedPkt_t) +
	      fixed.num_in_bunch*sizeof(FailedPath_t) + this_sf_len)) {
      fprintf(stderr, "3\n"); return 0;
    }
    sf.key = ntohs(sf_p->key);
    if(sizeof(sf.value) < (sf.len+1)) {
      fprintf(stderr, "size=%d, len+1=%d\n", (int)(sizeof(sf.value)),
	      (int)(sf.len+1));
      return 0;
    }
    strncpy(sf.value, sf_p->value, sf.len);
    sf.value[sf.len] = 0;
    strFields.push_back(sf);
    s += this_sf_len;
    sf_p = (StrField_t*)s;
  }

  if(s != (data+len)) { fprintf(stderr, "5\n"); return 0; }
  return 1;
} // End ServCompIFPkt::parsePkt()

int ServCompIFPkt::toStr(char *data, int maxLen) {
  fixed.num_in_bunch = failed.size();
  fixed.numStr = strFields.size();

  int len = sizeof(FixedPkt_t);
  if(maxLen < len) { return 0; }
  FixedPkt_t *fp = (FixedPkt_t*)data;

  /** Packet type */
  fp->type = htons(fixed.type);

  /** Path info */
  fp->pathID = htonl(fixed.pathID); fp->app_seq = htonl(fixed.app_seq);
  fp->port = htonl(fixed.port);
  fp->path_dest_scid = htons(fixed.path_dest_scid);
  fp->path_origin_scid = htons(fixed.path_origin_scid);
  fp->version = htons(fixed.version);

  /** Path traversal info */
  fp->posn = htons(fixed.posn); fp->subPosn = htons(fixed.subPosn);

  /** Repair info */
  fp->repType = htons(fixed.repType);
  fp->fail_dest_scid = htons(fixed.fail_dest_scid);
  fp->fail_origin_scid = htons(fixed.fail_origin_scid);
  fp->fail_version = htons(fixed.fail_version);
  fp->failPosn = htons(fixed.failPosn);

  /** Failure notification */
  fp->failure_scid = htons(fixed.failure_scid);
  fp->succSoFar = htons(fixed.succSoFar);
  fp->numDestroyed = htons(fixed.numDestroyed);

  /** Communication with services */
  fp->serv_dest_scid = htons(fixed.serv_dest_scid);
  fp->serv_my_scid = htons(fixed.serv_my_scid);
  fp->serv_upstream_scid = htons(fixed.serv_upstream_scid);

  /** Success indication */
  fp->success = htons(fixed.success);

  /** The number of bunched failure info */
  fp->num_in_bunch = htons(fixed.num_in_bunch);
  /** The number of string fields */
  fp->numStr = htons(fixed.numStr);

  len += fixed.num_in_bunch*sizeof(FailedPath_t);
  if(maxLen < len) { return 0; }

  FailedPath_t *failed_p = (FailedPath_t*)(data + sizeof(FixedPkt_t));
  for(unsigned int i = 0; i < fixed.num_in_bunch; i++) {
    failed_p[i].pathID = htonl(failed[i].pathID);
    failed_p[i].path_dest_scid = htons(failed[i].path_dest_scid);
    failed_p[i].version = htons(failed[i].version);
  }

  char *s = (char*)(data + len);
  StrField_t *sf_p = (StrField_t*)s;
  for(unsigned int i = 0; i < fixed.numStr; i++) {
    StrField_t &sf = strFields[i];
    int this_sf_len = (sizeof(sf) - sizeof(sf.value)) + sf.len;
    len += this_sf_len;
    if(maxLen < len) { return 0; }
    sf_p->len = htons(sf.len);
    sf_p->key = htons(sf.key);
    strncpy(sf_p->value, sf.value, sf.len);
    s += this_sf_len;
    sf_p = (StrField_t*)s;
  }

  if(s != (data+len)) { return 0; }
  return len;
} // End ServCompIFPkt::toStr()

void ServCompIFPkt::addPair(const StrKey_t key, const char *value) {
  vector<StrField_t>::iterator iter;
  int found = 0;
  for(iter = strFields.begin(); iter != strFields.end(); iter++) {
    if(iter->key == key) {
      found = 1; break;
    }
  }

  if(found) {
    strcpy(iter->value, value);
    iter->len = strlen(value);
  } else {
    StrField_t sf;
    sf.key = key; sf.len = strlen(value);
    strcpy(sf.value, value);
    strFields.push_back(sf);
  }
} // End ServCompIFPkt::addPair()

const char *ServCompIFPkt::getValue(const StrKey_t key) {
  vector<StrField_t>::iterator iter;
  int found = 0;
  for(iter = strFields.begin(); iter != strFields.end(); iter++) {
    if(iter->key == key) {
      found = 1; break;
    }
  }
  if(found) {
    return iter->value;
  } else {
    return NULL;
  }
} // End ServCompIFPkt::getValue()

int ServCompIFPkt::totLen() {
  fixed.num_in_bunch = failed.size();
  fixed.numStr = strFields.size();

  int len = sizeof(FixedPkt_t);
  len += fixed.num_in_bunch*sizeof(FailedPath_t);

  for(unsigned int i = 0; i < fixed.numStr; i++) {
    StrField_t &sf = strFields[i];
    int this_sf_len = (sizeof(sf) - sizeof(sf.value)) + sf.len;
    len += this_sf_len;
  }

  return len;
} // End ServCompIFPkt::totLen()
