/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
#ifndef TREE_INFO_H
#define TREE_INFO_H

#include <vector>

#include "../util/JHashtable.h"
#include "../util/JString.h"
#include "../util/Util.h"
#include "../util/PriorityQueue.h"

typedef class TreeInfo TreeInfo_t;
typedef class TreeNode TreeNode_t;

class TreeNode {
protected:
  vector<JString> services;
public:
  /** The vector of previous node info.  prevTreeNode[i] has the
      information for stage 'i' of the modified Dijkstra's algorithm
      (there are numServices+1 stages). */
  vector<int> prevTreeNode;
  /** The vector of min costs.  minCost[i] has the information for
      stage 'i' of the modified Dijkstra's algorithm (there are
      numServices+1 stages). */
  vector<unsigned short> minCost;
  /** The SCID of this node */
  unsigned short scid;

  /** The constructor: has to be given the number of services in the
      logical path */
  TreeNode(unsigned int numServices) {
    prevTreeNode.reserve(numServices+1);
    minCost.reserve(numServices+1);
  }

  /** Initialize the list of services at this node to be a null list */
  void initServiceList() {
    services.erase(services.begin(), services.end());
  }

  /** Add a service to the list of services at this node */
  void addToServiceList(const JString &sName) {
    services.push_back(sName);
  }

  /** See if the given service is there locally */
  int hasService(const JString &sName) {
    for(unsigned int i = 0; i < services.size(); i++) {
      if(services[i] == sName) return 1;
    }
    return 0;
  } // End TreeNode::hasService()

}; // End class TreeNode

/** This class defines the information stored to figure out the
    "constrained" shortest path for a particular logical path pf
    services. */
class TreeInfo {
protected:
  static JHashtable<JString, TreeInfo_t*> tHash;

  /** The JString of services */
  vector<JString> services;
  /** The hashtable to get from SCID to the index in the nVec array */
  JHashtable<unsigned short, int> scidHash;

public:
  /** for printing debug messages */
  static bool debug;

  /** The vector of nodes */
  TreeNode_t *nVec;
  /** The number of nodes */
  int numTreeNodes;
  /** The array of seenTreeNodes */
  class PriorityQueue<unsigned short> *seenTreeNodes;
  /** The index of the global minimum vertex */
  int globalMinVertex;
  /** The SCID of the global minimum vertex */
  unsigned short globalMinSCID;

  /** The last time this tree was built */
  struct timeval buildTime;

  /** The timeout value for path caches (in ms) */
  static float TREE_CACHE_TIMEOUT;

  /** Build a reverse index of SCIDs to node indices */
  void buildReverseIndex();
  /** Get the index (into the 'nVec' vector), given the SCID.  Return
      -1 on failure */
  int getIndexOfSCID(unsigned short scid);

  /** Allocate the 'seenTreeNodes' set of priority queues */
  void allocSeenTreeNodes();
  /** De-allocate the 'seenTreeNodes' set of priority queues */
  void deallocSeenTreeNodes();

  /** The constructor */
  TreeInfo(const vector<JString> &arg_services, int numTreeNodes) {
    scidHash.setHashFunction(ushortHash); // setting hash fn
    nVec = new TreeNode[numTreeNodes](arg_services.size()); // allocating
    // copy the list of services
    for(unsigned int i = 0; i < arg_services.size(); i++) {
      services.push_back(arg_services[i]);
    }
  } // End TreeInfo::TreeInfo()

  /** The destructor */
  ~TreeInfo() {
    delete[] nVec;
  }

  /** Return the pointer to the tree-info associated with this logical
      path.  Returns NULL if no such tree exists. */
  static TreeInfo_t *getTreeInfo(const vector<JString> &services);
  /** Make a new tree with the given set of services as the index
      (into the local hash table), and return a pointer to the new
      tree.  If there was any old tree with the same set of services,
      it is destroyed.  The number of nodes in the graph is given in
      'numTreeNodes'. */
  static TreeInfo_t *makeNewTree(const vector<JString> &services,
				 int numTreeNodes);

}; // End class TreeInfo

#endif
