/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Program to start the serv-comp threads */

#include <vector>

#include <unistd.h>
#include <math.h>

#include "ServComp.h"
#include "../util/Control.h"
#include "../service-locn/ServiceLocn.h"
#include "../link-state/NetworkGraph.h"
#include "../link-state/LinkState.h"
#include "../perf-meas/HB.h"
#include "../perf-meas/LatMeas.h"
#include "../udp-lib/timeSync.h"
#include "../util/Debug.h"

extern int print_ctl;

void usage(const char *progname) {
  cerr << "Usage: " << progname
       << " <scid-mapping-file> <graph-file> <mySCID>"
       << " <service-locn-file>"
       << endl;
} // End usage()

int main(int argc, char *argv[]) {
  util_init();
  if(argc != 5) { usage(argv[0]); exit(1); }

  char *scid_mapping, *graph, *scid_str, *service_locn;
  scid_mapping = argv[1];
  graph = argv[2];
  scid_str = argv[3];
  service_locn = argv[4];
  int mySCID = atoi(scid_str);
  print_ctl = mySCID;

  // Init the debug object
  struct timeval startTime;
  struct in_addr ipAddr;
  ipAddr.s_addr = inet_addr(timeServer);
  bool res1 = getStartTimeAtHost(&startTime, ipAddr);
  if(!res1) {
    cerr << "Can't get time from server " << timeServer << endl;
    exit(11);
  }

  char logFile[128];
  sprintf(logFile, "log.%d", mySCID);
  Debug_t dbg(logFile, mySCID, startTime, cfg_to_file, cfg_use_stream);

  char ls_send_logFile[128];
  sprintf(ls_send_logFile, "ls-send-log.%d", mySCID);
  Debug_t ls_send_dbg(ls_send_logFile, mySCID, startTime, cfg_to_file,
		      cfg_use_stream);

  char ls_listen_logFile[128];
  sprintf(ls_listen_logFile, "ls-listen-log.%d", mySCID);
  Debug_t ls_listen_dbg(ls_listen_logFile, mySCID, startTime,
			cfg_to_file, cfg_use_stream);

  // Init the udp-lib
  UDPLib::init();

  // Init the NetworkGraph
  bool read_all = false;
  NetworkGraph::init(scid_mapping, graph, (unsigned short)mySCID, read_all);

  bool run_listen_thread = false, run_send_thread = false;
  bool run_ls_listen_thread = false, run_ls_send_thread = false;

  // Start the listen threads for HB and LatMeas
  HB::startListenThread(run_listen_thread);
  LatMeas::startListenThread(run_listen_thread);

  // Allocate the HB and LatMeas threads for each peer
  // Start the send threads on HB/LatMeas for each peer
  unsigned short numPeers;
  vector<Peer_t> &peer_list = NetworkGraph::getPeerList(&numPeers);
  for(int i = 0; i < numPeers; i++) {
    peer_list[i].hb = new HB(peer_list[i].ip_addr, peer_list[i].scid);
    peer_list[i].latMeas =
      new LatMeas(peer_list[i].ip_addr, peer_list[i].scid);
  }

  // Init stuff
  if(!LinkState::init(&ls_send_dbg, &ls_listen_dbg)) {
    cerr << "LinkState::init() failed" << endl; exit(100);
  }
  if(!ServiceLocn::init()) {
    cerr << "ServiceLocn::init() failed" << endl; exit(102);
  }
  ServiceLocn::readFile(service_locn);
  ServComp::init(&dbg);

  // Start the listen threads for LS
  LinkState::startListenThread(run_ls_listen_thread);

  // Start SC thread -- this might actually be listening for other
  // stuff too
  ServComp::startListenThread();

  // connect and register with central server
  int cl_fd = sock_and_connect(controlServer, CONTROL_PORT);
  if(cl_fd < 0) {
    perror("sock_and_connect");
    exit(24);
  }
  struct cPkt c1, c2;
  c1.cmd = htons(REGISTER);
  int res = send(cl_fd, (void *)(&c1), sizeof(c1), 0);
  if(res != sizeof(c1)) { perror("send"); exit(4); }
  res = recv(cl_fd, &c2, sizeof(c2), 0);
  if(res != sizeof(c2)) { perror("recv"); exit(5); }
  if(ntohs(c2.cmd) != REGOK) {
    cerr << mySCID << ": recv from control server " << ntohs(c2.cmd)
	 << endl;
    exit(6);
  }

  // wait for start cmd from control server
  res = recv(cl_fd, &c2, sizeof(c2), 0);
  if(res != sizeof(c2)) { perror("recv"); exit(5); }
  if(ntohs(c2.cmd) != START) {
    cerr << mySCID << ": recv from control server " << ntohs(c2.cmd)
	 << endl;
    exit(6);
  }

  // Start the send threads on HB/LatMeas for each peer
  HB::start_all_sends(run_send_thread, mySCID);
  LatMeas::start_all_sends(run_send_thread, mySCID);

  // sleep so that i get a good picture of my peers
  struct timeval randT;
  gettimeofday(&randT, NULL);
  long randPeriod = 6*1000 + randT.tv_usec%(3367);
  fprintf(stderr, "~~~~~~~~~~~ %d-%d-%d-%d: SLEEP %ld ms ~~~~~~~~~~~\n",
	  mySCID, mySCID, mySCID, mySCID, randPeriod);
  usleep(randPeriod*1000);

  fprintf(stderr, "~~~~~~~~~~~ %d-%d-%d-%d: Starting LS Send ~~~~~~~~~~~\n",
	  mySCID, mySCID, mySCID, mySCID);
  LinkState::startSendThread(run_ls_send_thread);

  // sleep some more -- let link-state stuff propagate
  sleep(6);

  fprintf(stderr, "----------- %d-%d-%d-%d: SC started -----------\a\n",
	  mySCID, mySCID, mySCID, mySCID);

  // wait for stop cmd from control server
  res = recv(cl_fd, &c2, sizeof(c2), 0);
  if(res != sizeof(c2)) { perror("recv"); exit(55); }
  if(ntohs(c2.cmd) != STOP) {
    cerr << mySCID << ": recv from control server " << ntohs(c2.cmd)
	 << endl;
    exit(6);
  }

  fprintf(stderr, "=========== %d-%d-%d-%d: Stopping SC ===========\a\n",
	  mySCID, mySCID, mySCID, mySCID);

  ServComp::stopSend();
  LinkState::stopSendThread();
  LatMeas::stopSendThreads();
  HB::stopSendThreads();
  sleep(10);
  ServComp::stopListenThread();
  LinkState::stopListenThread();
  LatMeas::stopListenThread();
  HB::stopListenThread();

  unsigned long numSent = ServComp::getNumSent() +
    LinkState::getNumSent() + LatMeas::getNumSent() + HB::getNumSent();
  unsigned long numRecd = ServComp::getNumRecd() +
    LinkState::getNumRecd() + LatMeas::getNumRecd() + HB::getNumRecd();

  dbg.setCurrFn(MAIN_FN);
  if(dbg.use_stream) {
    dbg.print() << "total-sent total_sent=" << numSent << endl;
  } else {
    LogLine_t ll;
    ll.desc = total_sent_desc; ll.total_sent = numSent; dbg.addLL(ll);
  }
  if(dbg.use_stream) {
    dbg.print() << "total-recd total_recd=" << numRecd << endl;
  } else {
    LogLine_t ll;
    ll.desc = total_recd_desc; ll.total_recd = numRecd; dbg.addLL(ll);
  }

  dbg.close();
  ls_listen_dbg.close();
  ls_send_dbg.close();

} // End main()
