/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Program to test/measure Service Composition */

#include <unistd.h>
#include <vector>

#include <math.h>

#include "ServComp.h"
#include "../service-locn/ServiceLocn.h"
#include "../link-state/NetworkGraph.h"
#include "../link-state/LinkState.h"
#include "../perf-meas/HB.h"
#include "../perf-meas/LatMeas.h"
#include "../udp-lib/timeSync.h"
#include "../util/Debug.h"
#include "../util/JString.h"
#include "../util/Util.h"

void usage(const char *progname) {
  cerr << "Usage: " << progname
       << " <scid-mapping-file> <graph-file> <mySCID> <service-locn-file>"
       << endl;
} // End usage()

int main(int argc, char *argv[]) {
  util_init();
  if(argc != 5) { usage(argv[0]); exit(1); }

  char *scid_mapping, *graph, *scid_str, *service_locn;
  scid_mapping = argv[1];
  graph = argv[2];
  scid_str = argv[3];
  service_locn = argv[4];
  int mySCID = atoi(scid_str);

  // Init the debug object
  struct timeval startTime;
  struct in_addr ipAddr;
  ipAddr.s_addr = inet_addr(timeServer);
  bool res = getStartTimeAtHost(&startTime, ipAddr);
  if(!res) {
    cerr << "Can't get time from server " << timeServer << endl;
    exit(11);
  }
  char logFile[128];
  sprintf(logFile, "log.%d", mySCID);
  Debug_t dbg(logFile, mySCID, startTime, cfg_to_file, cfg_use_stream);

  char ls_send_logFile[128];
  sprintf(ls_send_logFile, "ls-send-log.%d", mySCID);
  Debug_t ls_send_dbg(ls_send_logFile, mySCID, startTime, cfg_to_file,
		      cfg_use_stream);

  char ls_listen_logFile[128];
  sprintf(ls_listen_logFile, "ls-listen-log.%d", mySCID);
  Debug_t ls_listen_dbg(ls_listen_logFile, mySCID, startTime,
			cfg_to_file, cfg_use_stream);

  // Init the udp-lib
  UDPLib::init();

  // Init the NetworkGraph
  bool read_all = true;
  NetworkGraph::init(scid_mapping, graph, (unsigned short)mySCID, read_all);

  // Start the listen threads for HB and LatMeas
  //bool run_listen_thread = true, run_send_thread = true;
  //HB::startListenThread(run_listen_thread);
  //LatMeas::startListenThread(run_send_thread);

  // Allocate the HB and LatMeas threads for each peer
  // Also start the send threads on each of those
  unsigned short numPeers;
  vector<Peer_t> &peer_list = NetworkGraph::getPeerList(&numPeers);
  for(int i = 0; i < numPeers; i++) {
    peer_list[i].hb = new HB(peer_list[i].ip_addr, peer_list[i].scid);
    peer_list[i].latMeas =
      new LatMeas(peer_list[i].ip_addr, peer_list[i].scid);
  }
  //HB::start_all_sends(run_send_thread);
  //LatMeas::start_all_sends(run_send_thread);

  // LinkState stuff
  bool run_ls_listen_thread = true, run_ls_send_thread = true;
  if(!LinkState::init(&ls_send_dbg, &ls_listen_dbg)) {
    cerr << "LinkState::init() failed" << endl;
    exit(100);
  }
  //LinkState::startListenThread(run_ls_listen_thread);
  /** Wait for peers to start up before starting to send LS updates */
  //fprintf(stderr, "Sleeping for 30sec before LS send\n");
  //sleep(30);
  //LinkState::startSendThread(run_ls_send_thread);

  // Service location stuff
  if(!ServiceLocn::init()) {
    cerr << "ServiceLocn::init() failed" << endl;
    exit(102);
  }
  ServiceLocn::readFile(service_locn);

  // ServComp stuff
  ServComp::init(&dbg);
  //ServComp::startListenThread();

  /*---- Test the path stuff ----*/

  // the required logical path (specified downstream to upstream)
  vector<JString> services;
  services.push_back("s0");
  services.push_back("s3");
  services.push_back("s1");

  // set random seed
  struct timeval tv;
  gettimeofday(&tv, NULL);
  srandom(tv.tv_usec);
  fprintf(stderr, "set random seed done with %ld\n", tv.tv_usec);

  // the array of measurements
  const int numMeas = 1000;
  float M[numMeas];
  short numNodes = NetworkGraph::getNumNodes();

  for(int i = 0; i < numMeas; i++) {
    int randSCID = random()%(numNodes+1);
    vector<ServInst_t> path;

    struct timeval start;

    gettimeofday(&start, NULL);

    int numNodesInPath;
    bool forceRebuild = true;
    if(randSCID == 0) {
      randSCID = NetworkGraph::INVALID_SCID;
    }
    unsigned short fail_dest_scid, fail_origin_scid;
    fail_dest_scid = fail_origin_scid = NetworkGraph::INVALID_SCID;
    numNodesInPath =
      ServComp::findPath(randSCID, services, path, forceRebuild,
			 fail_dest_scid, fail_origin_scid);
    if(numNodesInPath == -1) {
      cerr << "Error finding path: ";
      for(unsigned int i = 0; i < services.size(); i++) {
	cerr << services[i] << " ";
      }
      if(services.size() == 0) {
	cerr << "<null path>";
      }
      cerr << endl;
    }

    M[i] = now(start);

    if(false) {
      // NOTE: path is printed downstream to upstream
      cerr << "----------------- from " << randSCID << " to " << mySCID
	   << endl;
      for(unsigned int i = 0; i < path.size(); i++) {
	cerr << path[i].sName << " at " << path[i].gInfo.scid << endl;
      }

      fprintf(stderr, "Enter to continue: ");
      char readIn[64];
      cin.getline(readIn, sizeof(readIn));
    }

  } // End for(;;) test cases

  float sum, sumsq, avg, se, max, min;
  sum = sumsq = avg = se = 0;
  max = min = M[0];
  int minIndex = 0, maxIndex = 0;

  for(int i = 0; i < numMeas; i++) {
    sum += M[i];
    sumsq += M[i]*M[i];
    if(M[i] < min) {
      min = M[i]; minIndex = i;
    }
    if(M[i] > max) {
      max = M[i]; maxIndex = i;
    }
  }

  avg = sum/numMeas;
  se = sqrt(sumsq/numMeas - avg*avg);

  cerr << "avg = " << avg << ", se = " << se << endl;
  cerr << "min = " << min << ", minIndex = " << minIndex << endl;
  cerr << "max = " << max << ", maxIndex = " << maxIndex << endl;

} // End main()
