/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Implements the routines in ServiceLocn.h */

#include <netdb.h>
#include <stdio.h>
#include <unistd.h>
#include <iostream>
#include <fstream>
#include <strstream>

#include "ServiceLocn.h"
#include "../link-state/NetworkGraph.h"
#include "../util/Util.h"

reqRespBlock_t *ServiceLocn::send_rrBlock;
UDPPktBlock_t *ServiceLocn::send_pktBlock;
reqRespBlock_t *ServiceLocn::fwd_rrBlock;
UDPPktBlock_t *ServiceLocn::fwd_pktBlock;

vector<LocalServInfo_t> ServiceLocn::lServInfoVec;

pthread_t ServiceLocn::appIFThr;
pthread_t ServiceLocn::fwdThr;
pthread_t ServiceLocn::sendThr;

int ServiceLocn::ack_listen_fd;

struct timeval ServiceLocn::lastChange;
struct timeval ServiceLocn::lastSend;
const long ServiceLocn::sendPeriod = 10 * 60 * 1000; // 10min in ms

ServHash_t ServiceLocn::servHash;

int ServiceLocn::init() {
  // socket for listening for acks
  ack_listen_fd =
    socket(PF_INET, SOCK_DGRAM, getprotobyname("udp")->p_proto);
  if(ack_listen_fd < 0) {
    perror("ServiceLocn::init.socket");
    return 0;
  }

  // bind to SERV_INFO_ACK_PORT
  struct sockaddr_in listen_addr;
  bzero((char *)&listen_addr, sizeof(listen_addr));
  listen_addr.sin_family =  AF_INET;
  listen_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  listen_addr.sin_port = htons(SERV_INFO_ACK_PORT);
  if(bind(ack_listen_fd, (struct sockaddr *)&listen_addr,
	  sizeof(listen_addr)) < 0) {
    fprintf(stderr, "Can't bind to port %d\n", SERV_INFO_ACK_PORT);
    perror("bind failed");
    close(ack_listen_fd);
    return 0;
  }

  // get the request response block for "sending" packets
  send_rrBlock = UDPLib::getReqRespBlock();
  if(send_rrBlock == NULL) {
    fprintf(stderr, "Can't alloc for send_rrBlock in ServiceLocn::init\n");
    return 0;
  }

  // get a block of packet space to receive acks for packets "sent"
  send_pktBlock = UDPPktBlock::getUDPPktBlock();
  if(send_pktBlock == NULL) {
    fprintf(stderr, "Can't alloc for send_pktBlock in ServiceLocn::init\n");
    return 0;
  }

  // get the request response block for "forwarding" packets
  fwd_rrBlock = UDPLib::getReqRespBlock();
  if(fwd_rrBlock == NULL) {
    fprintf(stderr, "Can't alloc for fwd_rrBlock in ServiceLocn::init\n");
    return 0;
  }

  // get a block of packet space to receive acks for packets "forwarded"
  fwd_pktBlock = UDPPktBlock::getUDPPktBlock();
  if(fwd_pktBlock == NULL) {
    fprintf(stderr, "Can't alloc for fwd_pktBlock in ServiceLocn::init\n");
    return 0;
  }

  // init the last time of modification of service information (NOTE:
  // this may have to be read off a file to initialize)
  gettimeofday(&lastChange, NULL);

  // set the hash function for JString hashes
  servHash.setHashFunction(JStringHash);

  return 1;
} // End ServiceLocn::init()

int ServiceLocn::readFile(const char *fName) {
  //cerr << "reading service config file " << fName << endl;

  // read the file for the list of peers
  ifstream fin(fName);
  for(;;) {
    // read each line
    if(fin.eof()) break;
    char line[256];
    fin.getline(line, sizeof(line)); // read a line
    if(fin.eof()) break;
    if(fin.fail()) {
      cerr << "Error reading " << fName << endl;
      exit(99);
    }

    // line has the format: "serviceName scid serviceIP servicePort"
    vector<JString> sVec;
    int numFields = split(line, sVec);
    if(numFields != 4) {
      cerr << "Format of " << fName << " is wrong" << endl;
      exit(97);
    }

    // read the fields in the line
    LocalServInfo_t lServInfo;
    ostrstream sout(lServInfo.serviceName, sizeof(lServInfo.serviceName));
    sout << sVec[0];
    lServInfo.serviceName[sVec[0].length()] = 0;

    int scidInt = JString_atoi(sVec[1]);
    if(scidInt <= 0) {
      cerr << "scid is " << scidInt << " (has to be > 0)" << endl;
      exit(98);
    }
    unsigned short scid = (unsigned short)scidInt;
    lServInfo.serviceIP = nslookup(sVec[2].cstr()); // DNS lookup
    lServInfo.servicePort = JString_atoi(sVec[3]);

    // insert the service info in the database
    addServInfo(scid, &lServInfo);

  } // End for(;;) reading all lines in config file

  return 1;
} // End ServiceLocn::readFile()

void ServiceLocn::startAppIFThread() {
  // Spawn the thread
  pthread_create(&appIFThr, NULL, ServiceLocn::appIFThread, 0);
} // End ServiceLocn::startAppIFThread()

void *ServiceLocn::appIFThread(void*) {
  // socket for listening
  int app_fd =
    socket(PF_INET, SOCK_DGRAM, getprotobyname("udp")->p_proto);
  if(app_fd < 0) {
    perror("ServiceLocn::appIFThread.socket");
    return 0;
  }

  // bind to APP_IF_PORT
  struct sockaddr_in listen_addr;
  bzero((char *)&listen_addr, sizeof(listen_addr));
  listen_addr.sin_family =  AF_INET;
  listen_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  listen_addr.sin_port = htons(APP_IF_PORT);
  if(bind(app_fd, (struct sockaddr *)&listen_addr,
	  sizeof(listen_addr)) < 0) {
    fprintf(stderr, "Can't bind to port %d\n", APP_IF_PORT);
    perror("bind failed");
    close(app_fd);
    return 0;
  }

  UDPLibPkt_t p;
  struct sockaddr_in from_addr;
  int recv_len;

  // forever, listen for packets, and deal as appropriate
  for(;;) {
    // receive packet (blocking)
    unsigned long rcv_count = 0;
    bool pkt_ready = false;
    int res = UDPLib::blockingRecv(app_fd, &p, &recv_len,
				   &from_addr, NULL, &rcv_count, pkt_ready);
    if(res == -1) {
      perror("ServiceLocn::appIFThread.blockingRecv");
      continue;
    }

    // parse the packet into an appPkt structure
    appPkt_t appPkt;
    int parseRes = appPkt.parsePkt(p.data, recv_len);
    if(parseRes != 1) {
      fprintf(stderr, "ServiceLocn::appIFThread.parsePkt failed\n");
      continue;
    }

    // acknowledge the receipt of the packet -- send UDPLib response
    // to the port specified by the application (service)
    int send_len = 0; // this is just an ack
    bool use_tunnel = true;
    unsigned long count = 0;
    res =
      UDPLib::sendResp(app_fd, &p, send_len, from_addr.sin_addr,
		       appPkt.lServInfo.servicePort, use_tunnel, &count);
    if(res != 1) {
      fprintf(stderr, "Warning: error sending response\n");
    }

    // deal with the packet as appropriate
    if(appPkt.type == AppPkt::REGISTER) {
      registerService(&appPkt);
    } else if(appPkt.type == AppPkt::DEREGISTER) {
      deregisterService(&appPkt);
    } else {
      fprintf(stderr, "Warning: wrong packet type\n");
      continue;
    }

  } // End for(;;)

} // End ServiceLocn::appIFThread()

void ServiceLocn::registerService(appPkt_t *ap) {
  // see if element already exists.  if so, ignore request, else add
  // to end of vector
  vector<LocalServInfo_t>::iterator iter;
  int found = 0;
  for(iter = lServInfoVec.begin(); iter != lServInfoVec.end(); iter++) {
    if((strcmp(iter->serviceName, ap->lServInfo.serviceName) == 0) &&
       (iter->serviceIP.s_addr == ap->lServInfo.serviceIP.s_addr) &&
       (iter->servicePort == ap->lServInfo.servicePort)) {
      found = 1; break;
    }
  }

  if(!found) {
    lServInfoVec.push_back(ap->lServInfo);
    gettimeofday(&lastChange, NULL);
    addServInfo(NetworkGraph::mySCID, &ap->lServInfo);
  }
} // End ServiceLocn::registerService()

void ServiceLocn::deregisterService(appPkt_t *ap) {
  // if element exists, delete it, else do nothing
  vector<LocalServInfo_t>::iterator iter;
  for(iter = lServInfoVec.begin(); iter != lServInfoVec.end(); iter++) {
    if((strcmp(iter->serviceName, ap->lServInfo.serviceName) == 0) &&
       (iter->serviceIP.s_addr == ap->lServInfo.serviceIP.s_addr) &&
       (iter->servicePort == ap->lServInfo.servicePort)) {
      lServInfoVec.erase(iter);
      gettimeofday(&lastChange, NULL);
      break;
    }
  }
} // End ServiceLocn::deregisterService()

void ServiceLocn::startFwdThread() {
  // Spawn the thread
  pthread_create(&fwdThr, NULL, ServiceLocn::fwdThread, 0);
} // End ServiceLocn::startFwdThread()

void *ServiceLocn::fwdThread(void*) {
  // socket for listening
  int listen_fd =
    socket(PF_INET, SOCK_DGRAM, getprotobyname("udp")->p_proto);
  if(listen_fd < 0) {
    perror("ServiceLocn::fwdThread.socket");
    return 0;
  }

  // bind to SERV_INFO_PORT
  struct sockaddr_in listen_addr;
  bzero((char *)&listen_addr, sizeof(listen_addr));
  listen_addr.sin_family =  AF_INET;
  listen_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  listen_addr.sin_port = htons(SERV_INFO_PORT);
  if(bind(listen_fd, (struct sockaddr *)&listen_addr,
	  sizeof(listen_addr)) < 0) {
    fprintf(stderr, "Can't bind to port %d\n", SERV_INFO_PORT);
    perror("bind failed");
    close(listen_fd);
    return 0;
  }

  // listen for packets forever, deal as appropriate
  for(;;) {
    struct sockaddr_in from_addr;
    UDPLibPkt_t *p = fwd_pktBlock->getPkt();
    if(p == NULL) {
      fprintf(stderr, "no space for receiving service info\n");
      continue;
    }
    int recv_len;

    // receive packet (blocking)
    unsigned long rcv_count1 = 0;
    bool pkt_ready = false;
    int res = UDPLib::blockingRecv(listen_fd, p, &recv_len,
				   &from_addr, NULL, &rcv_count1, pkt_ready);
    if(res == -1) {
      perror("ServiceLocn::listenThread.blockingRecv");
      fwd_pktBlock->freePkt(p);
      continue;
    }
    p->data[recv_len] = 0; // end the JString with a '\0'

    // send UDPLib response
    int send_fd = listen_fd; // can use the same socket for sending
    int send_len = 0; // this is just an ack
    bool use_tunnel = true;
    unsigned long count = 0;
    res =
      UDPLib::sendResp(send_fd, p, send_len,
		       from_addr.sin_addr, SERV_INFO_ACK_PORT,
		       use_tunnel, &count);
    if(res != 1) {
      fprintf(stderr, "Warning: error sending service info ack\n");
    }

    // get the origin_scid and the recv_seqNo; also get the pointer to
    // the rest of service information data
    unsigned short origin_scid; unsigned short recv_seqNo;
    char *servData = parseServInfoPkt(p->data, &origin_scid, &recv_seqNo);
    if(servData == NULL) {
      fprintf(stderr, "serv info parse error\n");
      fwd_pktBlock->freePkt(p);
      continue;
    }

    // check if duplicate
    if(NetworkGraph::isDuplicateServInfo(origin_scid, recv_seqNo)) {
      fprintf(stderr, "Duplicate pkt from %s\n",
	      inet_ntoa(from_addr.sin_addr));
      fwd_pktBlock->freePkt(p);
      continue;
    }

    // store the service information in the packet
    int parseError = 0;
    while(servData[0] != 0) {
      char *nlPos = index(servData, '\n');
      if(nlPos == NULL) {
	fprintf(stderr, "ServiceLocn:parseError -- no newline\n");
	parseError = 1; break;
      }
      *nlPos = 0;
      LocalServInfo_t servInfo;
      int res = servInfo.parsePkt(servData);
      if(!res) {
	parseError = 1; break;
      }
      addServInfo(origin_scid, &servInfo);
      servData = nlPos+1;
    }
    if(parseError) {
      fwd_pktBlock->freePkt(p);
      continue;
    }

    // forward message along all directions (using UDPLib)
    unsigned short numPeers;
    vector<Peer_t> &peer_list = NetworkGraph::getPeerList(&numPeers);

    int checkNumForwards = 0; // just a check
    unsigned short flags[MAX_PEERS];
    for(int i = 0; i < numPeers; i++) {
      if(peer_list[i].ip_addr.s_addr == from_addr.sin_addr.s_addr) {
	// no need to flood along link on which pkt was received
	flags[i] = 0;
      } else {
	checkNumForwards++;
	flags[i] = 1;
      }
    } // End for(i)

    if(checkNumForwards != (numPeers-1)) {
      fprintf(stderr, "ServiceLocn:ERR: checkNumForwards=%d, numPeers-1=%d\n",
	      checkNumForwards, numPeers-1);
      exit(32);
    }
    // need to wait for 'numPeers-1' acks before freeing up 'p'
    fwd_pktBlock->setNumUsers(p, numPeers-1);

    // forward to peers
    sendToPeers(send_fd, p, recv_len, peer_list, numPeers, flags, FWD);

  } // End for(;;)

} // End ServiceLocn::fwdThread()

void ServiceLocn::startSendThread() {
  // Spawn the thread
  pthread_create(&sendThr, NULL, ServiceLocn::sendThread, 0);
} // End ServiceLocn::startSendThread()

void *ServiceLocn::sendThread(void*) {
  /*---- periodically broadcast all local service information ----*/

  // can use same socket for sending
  int send_fd = ack_listen_fd;

  // the sequence number of the send
  unsigned int seqNo = 0;
  int first_send = 1;

  // forever, form packet, send to all peers, check for acks
  for(;;) {
    // check if need to send
    int toSend = 0;
    float timeSinceLastChange = now(lastChange);
    float timeSinceLastSend = now(lastSend);

    // is this the first time we're sending?
    if(first_send) {
      toSend = 1;
      first_send = 0;
    } else {
      // has the last update been sent?
      if(timeSinceLastSend > timeSinceLastChange) { toSend = 1; }
      // too much time since last send?
      if(timeSinceLastSend > sendPeriod) { toSend = 1; }
    }

    /* ---- ----*/
    if(toSend) {
      UDPLibPkt_t *p = send_pktBlock->getPkt();
      if(p == NULL) {
	fprintf(stderr, "Can't alloc pkt for sending\n");
	continue;
      }

      // form the packet.  format is:
      // scid:<scid>\n
      // seqNo:<seqNo>\n;
      // <serviceName1>:<serviceIP1>:<servicePort1>\n
      // <serviceName2>:<serviceIP2>:<servicePort2>\n
      // ...
      int data_len = 0;
      sprintf(p->data, "scid:%d\nseqNo:%d\n", NetworkGraph::mySCID, seqNo);
      data_len = strlen(p->data);
      vector<LocalServInfo_t>::iterator iter;
      for(iter = lServInfoVec.begin(); iter != lServInfoVec.end(); iter++) {
	int this_len = iter->toStr(p->data + data_len);
	data_len += this_len;
      }

      // send it to all peers
      unsigned short numPeers;
      vector<Peer_t> &peer_list = NetworkGraph::getPeerList(&numPeers);
      unsigned short flags[MAX_PEERS];
      for(int i = 0; i < numPeers; i++) {
	flags[i] = 1;
      }
      send_pktBlock->setNumUsers(p, numPeers); // expect numPeers acks
      sendToPeers(send_fd, p, data_len, peer_list, numPeers, flags, SEND);

      // update last send time
      gettimeofday(&lastSend, NULL);

    } // End if(toSend)

    /* ---- ----*/
    /* check if any space in the send_pktBlock or fwd_pktBlock has to
       be freed */
    enum send_or_fwd_t send_or_fwd;
    for(send_or_fwd = SEND; send_or_fwd == SEND; send_or_fwd = FWD) {
      // NOTE: this outer loop on 'send_or_fwd' is a weird one that
      // executes once for SEND and then for FWD

      // get list of peers
      unsigned short numPeers;
      vector<Peer_t> &peer_list = NetworkGraph::getPeerList(&numPeers);

      for(int i = 0; i < numPeers; i++) {
	const int MAX_REQ_NOS = 32;
	int reqNos[MAX_REQ_NOS];
	int numReqs; UDPPktBlock_t *pktBlock; reqRespBlock_t *rrBlock;
	if(send_or_fwd == SEND) {
	  numReqs = peer_list[i].send_popReqNosServInfo(reqNos, MAX_REQ_NOS);
	  pktBlock = send_pktBlock;
	  rrBlock = send_rrBlock;
	} else {
	  numReqs = peer_list[i].fwd_popReqNosServInfo(reqNos, MAX_REQ_NOS);
	  pktBlock = fwd_pktBlock;
	  rrBlock = fwd_rrBlock;
	}

	for(int j = 0; i < numReqs; j++) {
	  struct sockaddr_in ack_from;
	  int ackRecd = rrBlock->checkResp(reqNos[j], NULL, &ack_from);
	  if(ackRecd == -1) {
	    fprintf(stderr, "timeout occurred waiting for %s\n",
		    inet_ntoa(ack_from.sin_addr));
	  }
	  if(ackRecd != 0) {
	    pktBlock->freePkt(rrBlock->reqs[reqNos[j]].send_pkt);
	    pktBlock->freePkt(rrBlock->reqs[reqNos[j]].recv_pkt);
	    rrBlock->freeEntry(reqNos[j]);
	  }
	} // End for(j)

      } // End for(i)

    } // End for(send_or_fwd)

    // sleep for a second
    sleep(1);

  } // End for(;;)

} // End ServiceLocn::sendThread()

void ServiceLocn::sendToPeers(int send_fd, UDPLibPkt_t *p,
			      int data_len, vector<Peer_t> &peer_list,
			      int numPeers, unsigned short *flags,
			      enum send_or_fwd_t send_or_fwd) {
  UDPPktBlock_t *pktBlock;
  reqRespBlock_t *rrBlock;
  if(send_or_fwd == SEND) {
    pktBlock = send_pktBlock;
    rrBlock = send_rrBlock;
  } else {
    pktBlock = fwd_pktBlock;
    rrBlock = fwd_rrBlock;
  }

  for(int i = 0; i < numPeers; i++) {
    if(!flags[i]) { continue; }

    UDPLibPkt_t *ack_pkt = pktBlock->getPkt();
    if(ack_pkt == NULL) {
      fprintf(stderr, "ServiceLocn::sendToPeers: No space for ack pkt\n");
      exit(37);
    }

    struct timeval timeout = peer_list[i].getServInfoTimeout();
    bool use_tunnel = true;
    int reqNo =
      rrBlock->insertRequest(send_fd, p, data_len,
			     peer_list[i].ip_addr, SERV_INFO_PORT,
			     ack_listen_fd, ack_pkt, timeout,
			     use_tunnel, peer_list[i].scid,
			     NetworkGraph::mySCID);
    if(reqNo == -1) {
      fprintf(stderr, "no space for request\n");
      continue;
    } else {
      if(send_or_fwd == SEND) { peer_list[i].send_addServInfoReq(reqNo); }
      else { peer_list[i].fwd_addServInfoReq(reqNo); }
    }
  }

} // End ServiceLocn::sendToPeers()

char *ServiceLocn::parseServInfoPkt(char *data, unsigned short *origin_scid,
				    unsigned short *recv_seqNo) {

  // extract 'origin_scid' and 'recv_seqNo' from packet.  format is:
  // scid:<scid>\n
  // seqNo:<seqNo>\n;
  // <serviceName1>:<serviceIP1>:<servicePort1>\n
  // <serviceName2>:<serviceIP2>:<servicePort2>\n
  // ...
  char *nlPos1 = index(data, '\n');
  if(nlPos1 == NULL) {
    fprintf(stderr, "serv info pkt parse error -- no newline after SCID\n");
    return NULL;
  }
  char *nlPos2 = index(nlPos1+1, '\n');
  if(nlPos2 == NULL) {
    fprintf(stderr, "serv info pkt parse error -- no newline after seqNo\n");
    return NULL;
  }

  *nlPos1 = *nlPos2 = 0;
  char *scidStr = data; char *seqNoStr = nlPos1+1;

  if(strncmp("scid:", scidStr, strlen("scid:")) != 0) {
    fprintf(stderr, "line1 does not begin with \"scid:\"\n");
    return NULL;
  }
  if(strncmp("seqNo:", seqNoStr, strlen("seqNo:")) != 0) {
    fprintf(stderr, "line2 does not begin with \"seqNo:\"\n");
    return NULL;
  }

  *origin_scid = atoi(scidStr+strlen("scid:"));
  *recv_seqNo = atoi(seqNoStr+strlen("seqNo:"));

  return nlPos2+1;
} // End ServiceLocn::parseServInfoPkt()

void ServiceLocn::addServInfo(unsigned short origin_scid,
			      LocalServInfo_t *servInfo) {
  JString sName = servInfo->serviceName;
  if(!servHash.containsKey(sName)) {
    vector<GlobalServInfo_t> gVec;
    servHash.put(sName, gVec);
  }
  vector<GlobalServInfo_t> &gVec = servHash.get(sName);
  GlobalServInfo_t g;
  g.scid = origin_scid;
  g.serviceIP = servInfo->serviceIP;
  g.servicePort = servInfo->servicePort;
  gVec.push_back(g);
} // End ServiceLocn::addServInfo()

const vector<GlobalServInfo_t> *ServiceLocn::getLocations(const JString &sName) {
  if(servHash.containsKey(sName)) {
    const vector<GlobalServInfo_t> &gVec = servHash.get(sName);
    return &gVec;
  } else {
    return NULL;
  }
} // End ServiceLocn::getLocations()
