/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
#ifndef SERVICE_LOCN_H
#define SERVICE_LOCN_H

#include <pthread.h>
#include <vector>
#include <sys/time.h>

#include "../udp-lib/UDPLib.h"
#include "../udp-lib/UDPPktBlock.h"
#include "../link-state/Peer.h"
#include "../util/JString.h"
#include "AppPkt.h"

#define APP_IF_PORT 3600
#define SERV_INFO_PORT 3650
#define SERV_INFO_ACK_PORT 3652

typedef JHashtable<JString, vector<GlobalServInfo_t> > ServHash_t;

/** The class that exchanges information about services, and handles
    service regn/deregn. */
class ServiceLocn {
protected:
  /** The block of space for reliable UDP -- for "sending" packets */
  static reqRespBlock_t *send_rrBlock;
  /** To allocate space for ack packets for packets "sent" */
  static UDPPktBlock_t *send_pktBlock;

  /** The block of space for reliable UDP -- for "forwarding" packets */
  static reqRespBlock_t *fwd_rrBlock;
  /** To allocate space for ack packets for packets "forwarded" */
  static UDPPktBlock_t *fwd_pktBlock;

  /** The vector of local services */
  static vector<LocalServInfo_t> lServInfoVec;

  /** See 'startAppIFThread' */
  static pthread_t appIFThr;
  /** See 'startFwdThread' */
  static pthread_t fwdThr;
  /** See 'startSendThread' */
  static pthread_t sendThr;

  /** Socket to listen on acks for packets sent/forwarded */
  static int ack_listen_fd;

  /** Time of last change */
  static struct timeval lastChange;
  /** Time of last send */
  static struct timeval lastSend;
  /** The send period in ms */
  static const long sendPeriod;

  /** The hashtable for storing the service information */
  static ServHash_t servHash;

  /** Implements 'appIFThr' */
  static void *appIFThread(void*);
  /** Implements 'fwdThr' */
  static void *fwdThread(void*);
  /** Implements 'sendThr' */
  static void *sendThread(void*);

  /** Register the given service */
  static void registerService(appPkt_t *ap);
  /** Deregister the given service */
  static void deregisterService(appPkt_t *ap);

  enum send_or_fwd_t { SEND=1, FWD=2 };

  /** Help routine -- to send a given packet to all peers */
  static void sendToPeers(int send_fd, UDPLibPkt_t *send_pkt,
			  int send_len, vector<Peer_t> &peer_list,
			  int numPeers, unsigned short *flags,
			  enum send_or_fwd_t send_or_fwd);

  /** Parse the given service information packet to get the
      'origin_scid' and the 'recv_seqNo'.  Return the beginning of the
      rest of the service information.  Returns NULL on parse error. */
  static char *parseServInfoPkt(char *data, unsigned short *origin_scid,
				unsigned short *recv_seqNo);

  /** Add the given service information to the database */
  static void addServInfo(unsigned short origin_scid,
			  LocalServInfo_t *servInfo);

public:
  /** Init the class variables ReqRespBlock for UDPLib, etc.  Return 1
      on success, 0 on failure */
  static int init();

  /** Init the set of services in the network, with the given file.
      The file format is, in each line:
      "<serviceName> <scid> <serviceIP> <servicePort>" */
  static int readFile(const char *fName);

  /** Starts 'appIFThr' -- the thread that manages regn/deregn
      requests from the application */
  static void startAppIFThread();

  /** Starts 'fwdThr' -- the thread for listening on SERV_INFO_PORT
      and forwards packets on it to neighbours */
  static void startFwdThread();

  /** Starts 'sendThr' -- the thread that broadcasts local service
      information periodically */
  static void startSendThread();

  /** Return the list of 'service' locations in the given vector.
      Return value is NULL on failure. */
  static const vector<GlobalServInfo_t> *getLocations(const JString &sName);

}; // End class ServiceLocn

#endif
