/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** The client program to invoke the text-to-speech composed service.
    This writes the audio output to a file/STDOUT */

#include <iostream>
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "../app/generic_client_info.h"
#include "../udp-lib/UDPLib.h"
#include "../serv-comp/ServCompIFPkt.h"
#include "../serv-comp/ServComp.h"
#include "../serv-comp/Paths.h"
#include "../util/LogLine.h"
#include "../util/Debug.h"
#include "../util/Util.h"
#include "../util/rtp.h"
#include "../udp-lib/timeSync.h"
#include "gapStats.h"

#define CLIENT_PORT 14000

char name[128] = "tts_sink";

class tts_sink_cl: public generic_client_info {
protected:
  FILE *fout;

  bool first_time;
  unsigned int seqno;

  int fmt_int;
  int expected_len;
  int numpkts;
  char out_file[128];

  unsigned int startSentence;
  long startBytePosn;

public:
  gapStats_t gStats;
  struct timeval startRecvTime;
  struct timeval lastRecvTime;

  int terminate1;

  tts_sink_cl() :generic_client_info() {
    fout = NULL;
    first_time = true;
    terminate1 = 0;
    fmt_int = RTP_PT_PCMU;
    expected_len = 160;
    numpkts = 0;
    startSentence = 0; startBytePosn = 0;
  }
  /** Handle incoming data pkt */
  void handleData();
  /** Handle incoming ctl pkt */
  void handleCtl();
  /** This should not recd */
  void handleDataAck();
  /** This should not be recd */
  void handleCtlAck() { fprintf(stderr, "#"); }

  /** Should also get path info from upstream */
  void parseUpstreamInfo(ServCompIFPkt_t *sPkt, ServCompIFPkt_t *resp_sPkt);
  /** Nothing to do for sendUpstreamInfo since I'm the sink */
  void sendUpstreamInfo() {}

  void sendDataReq();

  /** Handle incoming soft-state */
  void handleSoftState();
  bool recordSoftState(ServCompIFPkt_t *sPkt, JString &err_str);

  /** No need to timeout path */
  void checkPathTimeout() {}

  /** Init the file to which audio pkts are to be written */
  void init_file(const char *fName);
  /** Handle a Ctrl-C */
  void intr_catcher();
  /** End the session */
  void endSession() {
    gStats.end();
    generic_client_info::endSession();
    fclose(fout);
    if(dbg != NULL) { dbg->close(); dbg = NULL; }
  }

  virtual ~tts_sink_cl() {}
}; // End class tts_sink_cl
typedef class tts_sink_cl tts_sink_cl_t;

void tts_sink_cl::init_file(const char *fName) {
  strcpy(out_file, fName);
  if(strcmp(fName, "-") == 0) { fout = stdout; return; }
  fout = fopen(fName, "w");
  if(fout == NULL) { perror("fopen"); exit(87); }
} // End tts_sink_cl::init_file()

void tts_sink_cl::handleData() {
  struct sockaddr_in from_addr;
  UDPLibPkt_t upkt;

  // receive packet (blocking)
  // NOTE: called only when something is ready to be read
  int recv_len = lib_recvfrom(data_sock_fd, &upkt, sizeof(upkt), &from_addr);
  if(recv_len == -1) { perror("tts_sink_cl.handleData.recvfrom"); return; }

  // check sender
  bool correct_sender = true; int mismatch_code = 1;
  if(ntohs(from_addr.sin_port) != (upstreamPort+DATA_OFFSET)) {
    correct_sender = false; mismatch_code *= 2;
  }
  if(from_addr.sin_addr.s_addr != upstreamIP.s_addr) {
    correct_sender = false; mismatch_code *= 3;
  }
#ifdef IP_TUNNEL
  if(ntohs(upkt.my_scid) != serv_upstream_scid) {
    correct_sender = false; mismatch_code *= 5;
  }
  if(ntohs(upkt.dest_scid) != serv_my_scid) {
    correct_sender = false; mismatch_code *= 7;
  }
#endif
  if(!correct_sender) { fprintf(stderr, "o"); return; }

  struct myRTPpacket *rp = (struct myRTPpacket*)(upkt.data);
  int rtp_size =
    recv_len - sizeof(struct rtphdr) - (sizeof(upkt) - MAXDATASZ);

  // sanity check
  if(rtp_size != expected_len) {
    fprintf(stderr, "[r=%d,e=%d]", rtp_size, expected_len);
    return;
  }

  // check seqno
  if(first_time) {
    first_time = false;
    seqno = ntohs(rp->rh.rh_seqno);
  } else {
    seqno++;
    if(seqno != ntohs(rp->rh.rh_seqno)) {
      //fprintf(stderr, "[l:e=%d,r=%d]", (int)seqno,
      //(int)ntohs(rp->rh.rh_seqno));
      fprintf(stderr, "L");
    }
    seqno = ntohs(rp->rh.rh_seqno);

    // note stats
    float time_since_start = now(dbg->startTime);
    float time_since_last_recv = now(lastRecvTime);
    if(time_since_last_recv > gapStats::GAP_THRESHOLD) {
      gStats.add(time_since_last_recv, time_since_start);
    }
  }
  gettimeofday(&lastRecvTime, NULL);

  // check flags
  int flags = ntohs(rp->rh.rh_flags);
  if((flags & (RTP_VERSION << 14 | fmt_int)) !=
     (RTP_VERSION << 14 | fmt_int)) {
    fprintf(stderr, "[f]");
  }
  fprintf(stderr, "-");

  // write onto file
  fwrite((void*)(rp->data), 1, rtp_size, fout);
  fflush(fout);

  numpkts++;

} // End tts_sink_cl::handleData()

void tts_sink_cl::handleCtl() {
  sendHB(APP_KILL); // kill the old path
  generic_client_info::handleCtl();
  sendDataReq();
} // End tts_sink_cl::handleCtl()

void tts_sink_cl::parseUpstreamInfo(ServCompIFPkt_t *sPkt,
				    ServCompIFPkt_t *resp_sPkt) {
  generic_client_info::parseUpstreamInfo(sPkt, resp_sPkt);
  pathID = sPkt->fixed.pathID;
  path_dest_scid = sPkt->fixed.path_dest_scid;
  version = sPkt->fixed.version;
} // End tts_sink_cl::parseUpstreamInfo()

void tts_sink_cl::sendDataReq() {
  ServCompIFPkt_t req_sPkt;
  req_sPkt.fixed.type = TTS_REQ;
  req_sPkt.fixed.pathID = pathID;
  req_sPkt.fixed.path_dest_scid = path_dest_scid;
  req_sPkt.fixed.version = version;
  JString state_str =
    long2str(startSentence) + ":" + long2str(startBytePosn);
  req_sPkt.addPair(REQ_ARGS_F, state_str.cstr());
  int send_len = req_sPkt.toStr(reqPkt.data, sizeof(reqPkt.data));
  struct timeval timeout = { 2, 0 }; // 2sec

  // send the request upstream to get text
  int send_fd = data_sock_fd; // use same socket for sending
  // send pkt through the delaying tunnel
  bool use_tunnel = false;
  reqNo = rrBlock->insertRequest
    (send_fd, &reqPkt, send_len, upstreamIP, upstreamPort+DATA_OFFSET,
     data_ack_sock_fd, &ackPkt, timeout, use_tunnel,
     serv_upstream_scid, serv_my_scid);
  if(reqNo == -1) {
    cerr << "tts.sendDataReq: no space for request\n";
    exit(74);
  }

  dbg->print() << name << " tts-req p_id=" << pathID << ':'
	       << path_dest_scid << " path_ver=" << version
	       << " serv_my_scid=" << serv_my_scid
	       << " serv_upstream_scid=" << serv_upstream_scid << endl;
} // End tts_sink_cl::sendDataReq()

void tts_sink_cl::handleDataAck() {
  int reqNo_recd = rrBlock->check_one(data_ack_sock_fd);
  if(reqNo_recd < 0) { cerr << "[ts:r=-1]"; return; }
  if(reqNo != reqNo_recd) {
    cerr << "tts_sink.handleDataAck: reqNo mismatch, reqNo=" << reqNo
	 << " reqNo_recd=" << reqNo_recd << endl;
    exit(99);
  }
  rrBlock->freeEntry(reqNo); reqNo = -1;
  dbg->print() << name << " data-ack p_id=" << pathID << ':'
	       << path_dest_scid << " path_ver=" << version << endl;
} // End tts_sink_cl::handleDataAck()

void tts_sink_cl::handleSoftState() {
  // recv the pkt
  UDPLibPkt_t uPkt; ServCompIFPkt_t sPkt;
  int recv_len; struct sockaddr_in from_addr;
  unsigned long rcv_count = 0; bool pkt_ready = true;
  int res = UDPLib::blockingRecv(state_sock_fd, &uPkt, &recv_len,
				 &from_addr, NULL, &rcv_count, pkt_ready);
  if(res != 1) { cerr << "res = " << res << endl; return; }

  // check sender
  if(ntohs(from_addr.sin_port) != (upstreamPort+STATE_OFFSET)) {
    fprintf(stderr, "S"); return;
  }

  // parse it
  int parseRes = sPkt.parsePkt(uPkt.data, recv_len);
  if(parseRes != 1) { cerr << "parseRes = " << parseRes << endl; return; }

  JString err_str;
  if(!recordSoftState(&sPkt, err_str)) {
    cerr << "record soft state error" << endl;
    dbg->print() << name << ' ' << err_str << " p_id=" << pathID << ':'
		 << path_dest_scid << " path_ver=" << version << " type="
		 << ServCompIFPkt::type2str((ServCompPktType_t)sPkt.fixed.type)
		 << endl;
    dbg->close();
    exit(85);
  }
  fprintf(stderr, "[s:%d:%ld]", startSentence, startBytePosn);
} // End tts_sink_cl::handleSoftState()

bool tts_sink_cl::recordSoftState(ServCompIFPkt_t *sPkt, JString &err_str) {
  // sanity checks
  if(sPkt->fixed.type != APP_PATH_STATE) {
    err_str = "unknown-pkt-type"; return false;
  }
  const char *state_str = sPkt->getValue(PATH_STATE_F);
  if(state_str == NULL) { err_str = "no-state-str"; return false; }

  // get state info
  char *col_pos = index(state_str, ':');
  if(col_pos == NULL) { err_str = "no-col"; return false; }
  *col_pos = 0;
  unsigned int nextSentence = atoi(state_str);
  long nextStartBytePosn = atoi(col_pos+1);

  // sanity check, for reordering
  if(nextSentence > startSentence) {
    startSentence = nextSentence;
    startBytePosn = nextStartBytePosn;
    err_str = "ok"; return true;
  }
  if(nextSentence < startSentence) {
    cerr << "[reorder]"; err_str = "ok-reorder"; return true;
  }
  if(nextStartBytePosn < startBytePosn) {
    cerr << "[reorder]"; err_str = "ok-reorder"; return true;
  }
  startBytePosn = nextStartBytePosn;
  err_str = "ok"; return true;

} // End tts_sink_cl::recordSoftState()

void tts_sink_cl::intr_catcher() {
  if(terminate1 == 1) {
    fprintf(stderr, "%d pkts written onto file %s\n", numpkts, out_file);
    endSession(); exit(0);
  }
  fprintf(stderr, "Caught Ctrl-C; Going to terminate...\n");
  fprintf(stderr, "Hit Ctrl-C again if I don't terminate...\n");
  terminate1 = 1;
} // End tts_sink_cl::intr_catcher()

tts_sink_cl_t sink_cl;

void intr_catcher(int) {
  sink_cl.intr_catcher();
} // End intr_catcher

void usage(const char *progname) {
  if(cfg_read_paths) {
    fprintf(stderr, "Usage: %s <scid-mapping-file> <audio-output>\n",
	    progname);
  } else {
    fprintf(stderr, "Usage: %s <scid-mapping-file> <audio-output> <sc-host-scid>\n",
	    progname);
  }
  fprintf(stderr, "       (give '-' to write audio to STDOUT)\n");
  fprintf(stderr, "       cfg_read_paths=%d\n", (int)cfg_read_paths);
} // End usage()

unsigned short readPaths() {
  ifstream fin("../serv-comp/paths.cfg");
  if(fin.fail()) {
    cerr << "Can't open ../serv-comp/paths.cfg for reading" << endl;
    return LogLine::INVALID_SCID;
  }

  int line_num = 0;
  unsigned short to_ret = LogLine::INVALID_SCID;
  for(;;) {
    if(line_num != 0) {
      cerr << "Expecting only one line in ../serv-comp/paths.cfg\n";
      break;
    }
    // read the line
    if(fin.eof()) { break; }
    char line[2048];
    fin.getline(line, sizeof(line)); // read a line
    if(fin.eof()) break;
    if(fin.fail()) { cerr << "Error reading" << endl; break; }
    line_num++;
    LogLine_t ll(line);
    if(!ll.succ) {
      cerr << "Format wrong while reading line: " << line_num << endl;
      cerr << line << endl;
      break;
    }

    if((ll.pathID < 0) || (ll.path_dest_scid == LogLine::INVALID_SCID)) {
      cerr << "pathID=" << ll.pathID << ", path_dest_scid="
	   << ll.path_dest_scid << endl;
      break;
    }

    to_ret = ll.path_dest_scid;

  } // End for(;;) reading file

  fin.close();
  return to_ret;
} // End readPaths()

int main(int argc, char *argv[]) {
  util_init();

  int arg_expected = 4;
  if(cfg_read_paths) { arg_expected = 3; }

  // check command line args
  if(argc != arg_expected) { usage(argv[0]); exit(1); }
  const char *scid_mapping = argv[1];
  const char *outFile = argv[2];
  unsigned short serv_scid = cfg_read_paths ? readPaths() : atoi(argv[3]);
  if(serv_scid == LogLine::INVALID_SCID) {
    cerr << "serv_scid is invalid\n"; exit(55);
  }

  vector<sc_node_t> sc1;
  util_readSCIDMapping(scid_mapping, sc1);
  int serv_index = -1;
  for(unsigned int i = 0; i < sc1.size(); i++) {
    if(sc1[i].scid == serv_scid) {
      serv_index = i; break;
    }
  }
  if(serv_index == -1) {
    cerr << "scid " << serv_scid << " not found in " << scid_mapping
	 << endl;
    exit(7);
  }
  struct in_addr scIPAddr = sc1[serv_index].ipAddr;

  // prepare the UDP socket for listening
  int udp_sock = sock_and_bind(CLIENT_PORT);
  if(udp_sock < 0) { perror("sock_and_bind"); exit(3); }

  // init the udp library
  UDPLib::init();

  // Init the debug object
  struct timeval startTime;
  struct in_addr ipAddr;
  ipAddr.s_addr = inet_addr(timeServer);
  bool res1 = getStartTimeAtHost(&startTime, ipAddr);
  if(!res1) {
    cerr << "Can't get time from server " << timeServer << endl;
    exit(11);
  }
  char logFile[128];
  sprintf(logFile, "tts-sink-%s.log", getenv("HOSTNAME"));
  bool to_file = true, use_stream = true;
  Debug_t sink_dbg(logFile, 0, startTime, to_file, use_stream);
  Debug_t *dbg = &sink_dbg;

  // init the sink client
  generic_client_info::myIP = nslookup(getenv("HOSTNAME"));
  generic_client_info::dbg = dbg;
  strcpy(generic_client_info::name, name);
  // get the request response block for sends
  generic_client_info::rrBlock = UDPLib::getReqRespBlock();
  if(generic_client_info::rrBlock == NULL) {
    cerr << "rrBlock is NULL" << endl; exit(23);
  }
  int data_port_base = CLIENT_PORT+100;
  sink_cl.init_sock(data_port_base);
  sink_cl.init_file(outFile);
  sink_cl.state = PROCESSING_CL;

  // init the stats structure
  char statsFile[256];
  sprintf(statsFile, "tts-sink-%s.stat", getenv("HOSTNAME"));
  sink_cl.gStats.init(statsFile);

  // prepare the packet to send
  UDPLibPkt_t uPkt;
  ServCompIFPkt_t sPkt;
  sPkt.fixed.type = APP_CREATE_PATH;
  sPkt.fixed.app_seq = 1;
  sPkt.fixed.port = CLIENT_PORT;
  JString my_dest_spec =
    ip2str(generic_client_info::myIP) + ":" + long2str(data_port_base);
  sPkt.addPair(PREV_DEST_SPEC_F, my_dest_spec.cstr());
  sPkt.addPair(SERVICES_F, "tts email");
  //sPkt.fixed.path_origin_scid = 8;
  int uPktLen = sPkt.toStr(uPkt.data, sizeof(uPkt.data));

  // do a synchronous request response
  UDPLibPkt_t resp_uPkt;
  ServCompIFPkt_t resp_sPkt;
  int resp_len;
  bool use_tunnel = false; // should not send pkt through the delaying tunnel
  unsigned long count = 0, rcv_count = 0;
  int send_fd = udp_sock; // can use same socket for sending
  unsigned short dest_scid = LogLine::INVALID_SCID;
  unsigned short my_scid = LogLine::INVALID_SCID;
  int res = UDPLib::syncReqResp(send_fd, &uPkt, uPktLen, scIPAddr,
				SERV_COMP_CLIENT_PORT, udp_sock, &resp_uPkt,
				&resp_len, NULL, use_tunnel, &count,
				&rcv_count, dest_scid, my_scid);
  if(res != 1) { cerr << "res = " << res << endl; exit(3); }

  int parseRes = resp_sPkt.parsePkt(resp_uPkt.data, resp_len);
  if(!parseRes) { cerr << "parseRes = " << parseRes << endl; exit(3); }

  int pathID = resp_sPkt.fixed.pathID;
  unsigned short path_dest_scid = resp_sPkt.fixed.path_dest_scid;
  int success = resp_sPkt.fixed.success;

  dbg->print() << name << " sc-reply p_id=" << pathID << ':'
	       << path_dest_scid << " success=" << success << endl;

  if(!success) { cerr << "success=" << success; dbg->close(); exit(3); }

  // init the times
  gettimeofday(&sink_cl.startRecvTime, NULL);
  gettimeofday(&sink_cl.lastRecvTime, NULL);

  // Activate interrupt catcher
  signal(SIGINT, intr_catcher);
  signal(SIGTERM, intr_catcher);

  // wait on data/ctl ports, and handle when packets arrive
  while(sink_cl.terminate1 == 0) {
    /*---- Check to receive packets ----*/
    fd_set rfds; FD_ZERO(&rfds);
    FD_SET(udp_sock, &rfds);
    int max_fd = -1;
    FD_SET(sink_cl.data_sock_fd, &rfds);
    if(max_fd < sink_cl.data_sock_fd) { max_fd = sink_cl.data_sock_fd; }
    FD_SET(sink_cl.ctl_sock_fd, &rfds);
    if(max_fd < sink_cl.ctl_sock_fd) { max_fd = sink_cl.ctl_sock_fd; }
    FD_SET(sink_cl.data_ack_sock_fd, &rfds);
    if(max_fd < sink_cl.data_ack_sock_fd) { max_fd = sink_cl.data_ack_sock_fd; }
    FD_SET(sink_cl.ctl_ack_sock_fd, &rfds);
    if(max_fd < sink_cl.ctl_ack_sock_fd) { max_fd = sink_cl.ctl_ack_sock_fd; }
    FD_SET(sink_cl.state_sock_fd, &rfds);
    if(max_fd < sink_cl.state_sock_fd) { max_fd = sink_cl.state_sock_fd; }

    struct timeval timeout = { 0, 20*1000 }; // 20ms
    int retval = select(max_fd+1, &rfds, NULL, NULL, &timeout);
    if(retval < 0) { perror("tts_sink.select"); continue; }

    if(FD_ISSET(sink_cl.data_sock_fd, &rfds)) { sink_cl.handleData(); }
    if(FD_ISSET(sink_cl.ctl_sock_fd, &rfds)) { sink_cl.handleCtl(); }
    if(FD_ISSET(sink_cl.data_ack_sock_fd, &rfds)) { sink_cl.handleDataAck(); }
    if(FD_ISSET(sink_cl.ctl_ack_sock_fd, &rfds)) { sink_cl.handleCtlAck(); }
    if(FD_ISSET(sink_cl.state_sock_fd, &rfds)) { sink_cl.handleSoftState(); }

    sink_cl.checkDataSend();
    sink_cl.checkCtlSend();
    sink_cl.checkSoftStateSend();
    sink_cl.checkHBSend();
    sink_cl.checkPathTimeout();
  } // End while()

  sink_cl.endSession();
  dbg->close();
  cerr << "\a";

} // End main()
