/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** This file implements the routines related to passing requests and
    responses between the UDPLib thread and the application thread */

#include "ReqRespBlock.h"
#include "../util/Util.h"
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <stdio.h>

#include <iostream>

extern int print_ctl;

/** The times at which a UDP packet must be resent for response */
const long ReqRespBlock::retryTimes[5] = { 1000, 1500, 2000, 2000, 2000 }; // in ms

int ReqRespBlock::insertRequest(int send_fd, UDPLibPkt_t* send_pkt,
				int send_len, struct in_addr dest_ip,
				int dest_port, int recv_fd,
				UDPLibPkt_t* recv_pkt,
				struct timeval timeout,
				bool use_tunnel,
				unsigned short dest_scid,
				unsigned short my_scid) {
  int index = -1;
  for(unsigned int i = 0; i < reqs.size(); i++) {
    if(!reqs[i].taken) {
      index = i;
      break;
    }
  }
  if(index == -1) {
    reqInfo_t r;
    r.taken = 0;
    index = reqs.size();
    reqs.push_back(r);
  }

  reqs[index].send_fd = send_fd;
  reqs[index].send_pkt = send_pkt;
  reqs[index].send_len = send_len;
  reqs[index].use_tunnel = use_tunnel;
  reqs[index].dest_scid = dest_scid;
  reqs[index].my_scid = my_scid;
  // set the timestamp and nonce for the packet to be sent
  send_pkt->ts = htonl(timestamp());
  send_pkt->nonce = htonl(random());
  send_pkt->reqNo = htonl(index);

  //fprintf(stderr, "--> ts=%u nonce=%u\n", ntohl(send_pkt->ts),
  //ntohl(send_pkt->nonce));

  bzero((char *)&reqs[index].send_dest, sizeof(reqs[index].send_dest));
  reqs[index].send_dest.sin_family = PF_INET;
  reqs[index].send_dest.sin_addr = dest_ip;
  reqs[index].send_dest.sin_port = htons(dest_port);

  reqs[index].recv_fd = recv_fd;
  reqs[index].recv_pkt = recv_pkt;
  reqs[index].timeout = timeout;
  reqs[index].recd = 0;

  gettimeofday(&reqs[index].startTime, NULL);
  reqs[index].lastSendTime = reqs[index].startTime;
  reqs[index].retryIndex = 0;

  reqs[index].sending = 1;
  reqs[index].taken = 1;

  // send packet
  int res = lib_sendto(send_fd, send_pkt,
		       send_len+(sizeof(UDPLibPkt_t)-MAXDATASZ), 0,
		       &reqs[index].send_dest, use_tunnel, &numSent,
		       reqs[index].dest_scid, reqs[index].my_scid);
  if(res == -1) {
    perror("ReqRespBlock::insertRequest.lib_sendto");
  }
  return index;
} // End ReqRespBlock::insertRequest()

int ReqRespBlock::checkResp(int reqNo, int *recv_len_p,
			    struct sockaddr_in *recv_from_p) {
  if(reqs[reqNo].recd == 0) {
    check_req_helper(&reqs[reqNo]);
    if(reqs[reqNo].sending) {
      check_resend(&reqs[reqNo]);
    }
  }
  if(recv_len_p != NULL) *recv_len_p = reqs[reqNo].recv_len;
  if(recv_from_p != NULL) *recv_from_p = reqs[reqNo].recv_from;
  return reqs[reqNo].recd;
} // End ReqRespBlock::checkResp()

int ReqRespBlock::nonBlockingRecv(int recv_fd, UDPLibPkt_t* recv_pkt,
				  struct timeval timeout) {
  int index = -1;
  for(unsigned int i = 0; i < reqs.size(); i++) {
    if(!reqs[i].taken) {
      index = i;
      break;
    }
  }
  if(index == -1) {
    reqInfo_t r;
    r.taken = 0;
    index = reqs.size();
    reqs.push_back(r);
  }

  reqs[index].recv_fd = recv_fd;
  reqs[index].recv_pkt = recv_pkt;
  reqs[index].timeout = timeout;
  gettimeofday(&reqs[index].startTime, NULL);
  reqs[index].recd = 0;

  reqs[index].sending = 0;
  reqs[index].taken = 1;
  return index;
} // End ReqRespBlock::nonBlockingRecv()

int ReqRespBlock::freeEntry(int reqNo) {
  if(reqs[reqNo].taken) {
    reqs[reqNo].taken = 0;
    return 1;
  } else {
    return 0;
  }
} // End ReqRespBlock::freeEntry()

void ReqRespBlock::check_req_helper(reqInfo_t *req) {
  // check for timeout on receive
  float timeoutFloat = time2float(req->timeout);
  float timeNow = now(req->startTime);
  if(timeNow > timeoutFloat) {
    req->recd = -1;
    return;
  }

  // now check if packet has been received
  fd_set rfds;
  FD_ZERO(&rfds); FD_SET(req->recv_fd, &rfds);
  struct timeval t = { 0, 0 };
  int retval = select(req->recv_fd+1, &rfds, NULL, NULL, &t);
  if(retval < 0) {
    perror("UDPLib::thread.select");
  }

  if(!FD_ISSET(req->recv_fd, &rfds)) { return; }

  // read packet
  int res = lib_recvfrom(req->recv_fd, req->recv_pkt,
			 sizeof(UDPLibPkt_t), &req->recv_from);
  numRecd++;

  if(req->sending) {

    // this is an ack packet in response to a pkt we sent
    if((req->recv_pkt->ts == req->send_pkt->ts) &&
       (req->recv_pkt->nonce == req->send_pkt->nonce)) {
      // ts and nonce match -- mark entry
      req->recv_len = res - (sizeof(UDPLibPkt_t) - MAXDATASZ);
      req->recd = 1;

    } else {

      if(gdbg) { fputc('?', stderr); }
      //ntohl(req->recv_pkt->ts), ntohl(req->recv_pkt->nonce));
      // check the rest of the request blocks if they have the
      // same recv_fd
      bool efficient = false;
      if(efficient) {
	unsigned long reqNo = ntohl(req->recv_pkt->reqNo);
	if(reqNo >= reqs.size()) {
	  fprintf(stderr, "%d: reqNo=%ld, reqs.size()=%ld\n", print_ctl,
		  reqNo, (unsigned long)reqs.size());
	  exit(34);
	}
	reqInfo_t *req1 = &reqs[reqNo];
	if((req1->taken == 1) && (req1->recd == 0) &&
	   (req1->recv_fd == req->recv_fd) &&
	   (req->recv_pkt->ts == req1->send_pkt->ts) &&
	   (req->recv_pkt->nonce == req1->send_pkt->nonce)) {
	  // ts and nonce match -- now copy the data
	  memcpy((void*)req1->recv_pkt, (void*)req->recv_pkt, res);
	  req1->recv_len = res - (sizeof(UDPLibPkt_t) - MAXDATASZ);
	  req1->recd = 1;
	  req1->recv_from = req->recv_from;
	  if(gdbg) { fputc('!', stderr); }
	}

	// end case if(efficient)
      } else {

	for(unsigned int j1 = 0; j1 < reqs.size(); j1++) {
	  reqInfo_t *req1 = &reqs[j1];
	  if((req == req1) || (req1->recd != 0) || !req1->taken) {
	    continue;
	  }
	  if((req1->recv_fd == req->recv_fd) && !req1->sending) {
	    fprintf(stderr, "recv_fd's match, not sending\n");
	  }
	  if((req1->recv_fd == req->recv_fd) &&
	     (req->recv_pkt->ts == req1->send_pkt->ts) &&
	     (req->recv_pkt->nonce == req1->send_pkt->nonce)) {
	    // ts and nonce match -- now copy the data
	    memcpy((void*)req1->recv_pkt, (void*)req->recv_pkt, res);
	    req1->recv_len = res - (sizeof(UDPLibPkt_t) - MAXDATASZ);
	    req1->recd = 1;
	    req1->recv_from = req->recv_from;
	    if(gdbg) { fputc('!', stderr); }
	    break; // out of the j1 loop
	  }
	} // End for(j1)

	// end else case of if(efficient)
      }

    } // End else of if("original req's ts, nonce match")

    // End case if(req->sending)
  } else {
    // this is an unsolicited receive -- no ts/nonce checking reqd
    // mark entry as 'recd'
    req->recv_len = res - (sizeof(UDPLibPkt_t) - MAXDATASZ);
    req->recd = 1;
  }
} // End ReqRespBlock::check_req_helper()

void ReqRespBlock::check_resend(reqInfo_t *req) {
  float timeSinceLastSend = now(req->lastSendTime);
  float randRetryTime = retryTimes[req->retryIndex] +
    (2*U.Next()-1)*100; // plus or minus a 100ms
  if(timeSinceLastSend > randRetryTime) {
    // time to send again
    int res = lib_sendto(req->send_fd, req->send_pkt,
			 req->send_len+(sizeof(UDPLibPkt_t)-MAXDATASZ),
			 0, &req->send_dest, req->use_tunnel,
			 &numSent, req->dest_scid, req->my_scid);
    if(res == -1) {
      perror("ReqRespBlock::insertRequest.lib_sendto");
    }
    gettimeofday(&req->lastSendTime, NULL);
    if(req->retryIndex < 4) { req->retryIndex++; }
  }
} // End ReqRespBlock::check_resend()

int ReqRespBlock::check_one(int fd) {
  struct sockaddr_in recv_from;
  UDPLibPkt_t recv_pkt;
  // read packet
  int res = lib_recvfrom(fd, &recv_pkt, sizeof(UDPLibPkt_t), &recv_from);
  numRecd++;

  unsigned long reqNo = ntohl(recv_pkt.reqNo);
  if(reqNo >= reqs.size()) {
    fprintf(stderr, "%d: reqNo=%ld, reqs.size()=%ld\n", print_ctl,
	    reqNo, (unsigned long)reqs.size());
    exit(34);
  }
  reqInfo_t *req1 = &reqs[reqNo];
  if((req1->taken == 1) && (req1->recd == 0)) {
    if((req1->recv_fd == fd) &&
       (recv_pkt.ts == req1->send_pkt->ts) &&
       (recv_pkt.nonce == req1->send_pkt->nonce)) {
      // ts and nonce match -- now copy the data
      memcpy((void*)req1->recv_pkt, (void*)&recv_pkt, res);
      req1->recv_len = res - (sizeof(UDPLibPkt_t) - MAXDATASZ);
      req1->recd = 1;
      req1->recv_from = recv_from;
      req1->recd_now = true;
      return reqNo;
    } else {
      fputc('#', stderr);
      return -1;
      /*
      cerr << "[E] reqNo=" << reqNo << " ts=" << ntohl(recv_pkt.ts)
	   << " nonce=" << ntohl(recv_pkt.nonce)
	   << " exp_ts=" << ntohl(req1->send_pkt->ts)
	   << " exp_nonce=" << ntohl(req1->send_pkt->nonce)
	   << " peer=" << ip2str(recv_from.sin_addr) << endl;
      */
    }
  }
  return -1;
} // End ReqRespBlock::check_one()

int ReqRespBlock::check_all(int fd) {
  for(unsigned int i = 0; i < reqs.size(); i++) {
    reqs[i].recd_now = false;
  }

  // now check if packet has been received
  // check for a max of reqs.size() packets
  int num_recd_now = 0;
  for(unsigned int i = 0; i < reqs.size(); i++) {
    if((reqs[i].taken != 1) || (reqs[i].recd != 0)) { continue; }
    if(reqs[i].recv_fd != fd) {
      cerr << "Unknown fd " << fd << endl;
      exit(80);
    }

    fd_set rfds;
    FD_ZERO(&rfds); FD_SET(fd, &rfds);
    struct timeval t = { 0, 0 };
    int retval = select(fd+1, &rfds, NULL, NULL, &t);
    if(retval < 0) {
      perror("UDPLib::thread.select");
    }

    if(!FD_ISSET(fd, &rfds)) { return num_recd_now; }
    int reqNo = check_one(fd);
    if(reqNo >= 0) { num_recd_now++; }
  } // End for(i) receiving packets

  return num_recd_now;
} // End ReqRespBlock::check_all()

int ReqRespBlock::check_t_o_resend(int reqNo) {
  reqInfo_t *req = &reqs[reqNo];
  if(req->recd != 0) return req->recd;
  // check for timeout on receive
  float timeoutFloat = time2float(req->timeout);
  float timeNow = now(req->startTime);
  if(timeNow > timeoutFloat) {
    req->recd = -1;
    return -1;
  }
  check_resend(req);
  return req->recd;
} // End ReqRespBlock::check_t_o_resend()
