/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Library to implement reliable UDP request/response mechanisms.
    This file implements the "class" used in the library.  This only
    implements request retry -- the same request may appear twice at
    the other end if there is retransmit due to loss of the
    response/ack. */

#include <iostream>
#include "UDPLib.h"
#include "UDPPktBlock.h"
#include "../util/Util.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/** Should I stop running? */
bool UDPLib::stop = false;

/** The array of ReqRespBlock's */
reqRespBlock_t UDPLib::blocks[MAXBLOCKS];
/** The array of markers for keeping track of which ReqRespBlock's
    are allocated */
int UDPLib::blockAllocated[MAXBLOCKS];
/** The array of markers to indicate if the 'thread' is currently
    checking on the status of packets received for this block.  This
    array of vars is used only by the 'thread' */
int UDPLib::checking[MAXBLOCKS];

/** The mutex to control allocation of ReqRespBlock's */
pthread_mutex_t UDPLib::allocMutex;
/** The thread that implements the library */
pthread_t UDPLib::libThr;

void UDPLib::init() {
  stop = false;

  // Init the mutex
  pthread_mutex_init(&allocMutex, NULL);

  // Init the internal data structures
  for(int i = 0; i < MAXBLOCKS; i++) {
    blockAllocated[i] = 0;
  }

  // set random seed
  struct timeval tv;
  gettimeofday(&tv, NULL);
  unsigned long seedLong = tv.tv_usec%211;
  float seedFloat = ((seedLong+1)*1.0)/212;
  Uniform::Set(seedFloat);
  srandom(tv.tv_usec);

  // Finally, spawn the thread
  //pthread_create(&libThr, NULL, UDPLib::thread, 0);

} // End UDPLib::init()

reqRespBlock_t* UDPLib::getReqRespBlock() {
  // Lock the mutex, grab an available location, and return a pointer
  // to it
  pthread_mutex_lock(&allocMutex);
  int index = -1;
  for(int i = 0; i < MAXBLOCKS; i++) {
    if(!blockAllocated[i]) {
      index = i;
      blockAllocated[i] = 1;
      blocks[i].numSent = 0;
      blocks[i].numRecd = 0;
      break;
    }
  }
  pthread_mutex_unlock(&allocMutex);
  if(index == -1) return NULL;
  else return &blocks[index];
} // End UDPLib::getReqRespBlock()

void* UDPLib::thread(void*) {

  bool stopped = false;
  struct timeval stopTime;
  for(;;) {
    if(stop && !stopped) {
      stopped = true;
      gettimeofday(&stopTime, NULL);
    }

    if(stopped && (now(stopTime) > 15000)) { break; }

    // For all pending replies (and unsolicited recv's), check if any
    // packet has been received.  If received, check with nonce and then
    // mark as received.  If timeout, signal error.
    for(int i = 0; i < MAXBLOCKS; i++) {

      if(!blockAllocated[i]) { continue; }

      for(unsigned int j = 0; j < blocks[i].reqs.size(); j++) {
	reqInfo_t *req = &blocks[i].reqs[j];
	if((req->recd != 0) || !req->taken) {
	  continue;
	}
	blocks[i].check_req_helper(req);
      } // End for(j)
    } // End for(i)

    if(stopped) { continue; } // don't send any more packets

    /*---- For all pending sends, send (or resend) a request ----*/
    for(int i = 0; i < MAXBLOCKS; i++) {
      if(!blockAllocated[i]) { continue; }

      for(unsigned int j = 0; j < blocks[i].reqs.size(); j++) {
	reqInfo_t *req = &blocks[i].reqs[j];
	if((req->recd != 0) || !req->taken || !req->sending) {
	  continue;
	}
	blocks[i].check_resend(req);
      } // End for(j)

    } // End for(i)

  } // End for(;;)

  return 0;
} // End UDPLib::thread()

int UDPLib::blockingRecv(int recv_fd, UDPLibPkt_t* recv_pkt, int* recv_len,
			 struct sockaddr_in *from_addr,
			 struct timeval *timeout, unsigned long *rcv_count,
			 bool pkt_ready) {
  bool can_recv = pkt_ready;
  if(!pkt_ready) {
    // select on the socket for the timeout period.  return appropriate value.
    fd_set rfds;
    FD_ZERO(&rfds); FD_SET(recv_fd, &rfds);
    int retval = select(recv_fd+1, &rfds, NULL, NULL, timeout);
    if(retval < 0) {
      perror("UDPLib::thread.select");
      return -1;
    }
    if(FD_ISSET(recv_fd, &rfds)) { can_recv = true; }
  }

  if(can_recv) {
    int res = lib_recvfrom(recv_fd, recv_pkt, MAXDATASZ, from_addr);
    if(res == -1) {
      perror("UDPLib::blockingRecv.recvfrom"); return -1;
    }
    *recv_len = res - (sizeof(UDPLibPkt_t) - MAXDATASZ);
    (*rcv_count)++;
    return 1;
  } else {
    return -1; // timeout occurred
  }
} // End UDPLib::blockingRecv()

int UDPLib::blockingRecv2(int fd1, int fd2, UDPLibPkt_t* recv_pkt,
			  int* recv_len, struct sockaddr_in *from_addr,
			  struct timeval *timeout,
			  unsigned long *rcv_count1,
			  unsigned long *rcv_count2) {
  // select on the socket for the timeout period.  return appropriate value.
  fd_set rfds;
  FD_ZERO(&rfds); FD_SET(fd1, &rfds); FD_SET(fd2, &rfds);
  int max_fd = (fd1 > fd2) ? fd1 : fd2;
  int retval = select(max_fd+1, &rfds, NULL, NULL, timeout);
  if(retval < 0) {
    perror("UDPLib::thread.select");
    return -1;
  }

  if(FD_ISSET(fd1, &rfds)) {
    int res = lib_recvfrom(fd1, recv_pkt, MAXDATASZ, from_addr);
    *recv_len = res - (sizeof(UDPLibPkt_t) - MAXDATASZ);
    (*rcv_count1)++;
    return fd1;
  } else if(FD_ISSET(fd2, &rfds)) {
    int res = lib_recvfrom(fd2, recv_pkt, MAXDATASZ, from_addr);
    *recv_len = res - (sizeof(UDPLibPkt_t) - MAXDATASZ);
    (*rcv_count2)++;
    return fd2;
  } else {
    return -1; // timeout occurred
  }
} // End UDPLib::blockingRecv2()

int UDPLib::syncReqResp(int send_fd, UDPLibPkt_t* send_pkt, int send_len,
			struct in_addr dest_ip, int dest_port, int recv_fd,
			UDPLibPkt_t* recv_pkt, int *recv_len,
			struct timeval *timeout, bool use_tunnel,
			unsigned long *count, unsigned long *rcv_count,
			unsigned short dest_scid, unsigned short my_scid) {
  // send request, retry as many times as is allowed by 'timeout', and
  // then 'select' for the response.

  // get the destination address
  struct sockaddr_in dest_addr;
  bzero((void *)&dest_addr, sizeof(dest_addr));
  dest_addr.sin_family = PF_INET;
  dest_addr.sin_addr = dest_ip;
  dest_addr.sin_port = ntohs(dest_port);

  // set the timestamp and nonce of the packet
  send_pkt->ts = htonl(timestamp());
  send_pkt->nonce = htonl(random());

  float timeoutFloat = -1;
  if(timeout != NULL) timeoutFloat = time2float(*timeout);
  int retryIndex = 0; // Index into the retryTimes array -- will
  // increase upto 4, and stay there
  struct timeval startTime;
  // send packet
  int res = lib_sendto(send_fd, send_pkt,
		       send_len+(sizeof(UDPLibPkt_t)-MAXDATASZ), 0,
		       &dest_addr, use_tunnel, count, dest_scid, my_scid);
  if(res == -1) {
    perror("UDPLib::syncReqResp.lib_sendto");
  }
  gettimeofday(&startTime, NULL);
  float lastSendTime = 0; // this is the time when the last pkt was
  // sent (w.r.t. the startTime)

  for(;;) {
    // figure out next time when pkt should be resent
    int lastTry = 0; // is this going to be the last try?
    unsigned long toWait;
    if((timeoutFloat == -1) ||
       ((timeoutFloat-lastSendTime) > ReqRespBlock::retryTimes[retryIndex])) {
      toWait = ReqRespBlock::retryTimes[retryIndex];
      lastTry = 0;
    } else {
      toWait = (unsigned long)(timeoutFloat - lastSendTime);
      lastTry = 1;
    }

    while(toWait > 1) {
      // at least one more milli-second to wait

      struct timeval t;
      t.tv_sec = toWait/1000;
      t.tv_usec = (toWait%1000)*1000;

      struct timeval waitStart;
      gettimeofday(&waitStart, NULL);

      // select
      fd_set rfds;
      FD_ZERO(&rfds); FD_SET(recv_fd, &rfds);
      int retval = select(recv_fd+1, &rfds, NULL, NULL, &t);
      if(retval < 0) {
	perror("UDPLib::thread.select");
      }

      float timeWaited = now(waitStart);
      toWait -= (unsigned long)timeWaited;

      if(FD_ISSET(recv_fd, &rfds)) {
	struct sockaddr_in from_addr;
	int res = lib_recvfrom(recv_fd, recv_pkt, MAXDATASZ, &from_addr);
	*recv_len = res - (sizeof(UDPLibPkt_t) - MAXDATASZ);
	(*rcv_count)++;
	// check for timestamp and nonce
	if((recv_pkt->ts == send_pkt->ts) &&
	   (recv_pkt->nonce == send_pkt->nonce)) {
	  return 1;
	} else {
	  if(gdbg) { fprintf(stderr, "[?]"); }
	}
      } else {
	break; // out of while loop
      }

    } // End while(toWait > 1)

    if(lastTry) {
      return -1; // timeout occurred
    } else {
      // should retransmit request
    }

    // send packet
    int res = lib_sendto(send_fd, send_pkt,
			 send_len+(sizeof(UDPLibPkt_t)-MAXDATASZ), 0,
			 &dest_addr, use_tunnel, count, dest_scid, my_scid);
    if(res == -1) {
      perror("UDPLib::syncReqResp.lib_sendto");
    }
    lastSendTime = now(startTime);
    if(retryIndex < 4) retryIndex++;
  } // End for(;;)

} // End UDPLib::syncReqResp()

int UDPLib::sendResp(int send_fd, UDPLibPkt_t *send_pkt, int send_len,
		     struct in_addr dest_ip, int dest_port,
		     bool use_tunnel, unsigned long *count) {

  // get the destination address
  struct sockaddr_in dest_addr;
  bzero((void *)&dest_addr, sizeof(dest_addr));
  dest_addr.sin_family = PF_INET;
  dest_addr.sin_addr = dest_ip;
  dest_addr.sin_port = ntohs(dest_port);

  // send packet
  unsigned short dest_scid = 0, my_scid = 0;
#ifdef IP_TUNNEL
  dest_scid = ntohs(send_pkt->my_scid);
  my_scid = ntohs(send_pkt->dest_scid);
#endif
  int res = lib_sendto(send_fd, send_pkt,
		       send_len+(sizeof(UDPLibPkt_t)-MAXDATASZ), 0,
		       &dest_addr, use_tunnel, count, dest_scid, my_scid);
  if(res < 0) { perror("UDPLib::sendResp.lib_sendto"); return -1; }
  else { return 1; }

} // End UDPLib::sendResp()

void UDPLib::stopThreads() {
  stop = true;
  //pthread_join(libThr, NULL);
  //thread(0);
} // End UDPLib::stopThreads()
