/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
#ifndef UDPLIB_H
#define UDPLIB_H

/** The class that implements the library of routines for reliable UDP
    request/response */

#include <pthread.h>
#include "ReqRespBlock.h"

#define MAXBLOCKS 6

class UDPLib {
protected:
  /** Should I stop running? */
  static bool stop;
  /** The array of ReqRespBlock's */
  static reqRespBlock_t blocks[MAXBLOCKS];
  /** The array of markers for keeping track of which ReqRespBlock's
      are allocated */
  static int blockAllocated[MAXBLOCKS];
  /** The array of markers to indicate if the 'thread' is currently
      checking on the status of packets received for this block.  This
      array of vars is used only by the 'thread' */
  static int checking[MAXBLOCKS];

  /** The mutex to control allocation of ReqRespBlock's */
  static pthread_mutex_t allocMutex;
  /** The thread that implements the library */
  static pthread_t libThr;
  /** The thread that implements the non-blocking, acked UDP messages
      and request/response pairs. */
  static void* thread(void*);

public:
  /** The init function -- to be called by the application before
      using the thread library.  This call is NOT thread safe and NOT
      idempotent.  This is supposed to be called only once by the main
      application thread. */
  static void init();

  /** Return a block of memory (encapsulated as an object) to be used
      for communication between the UDPLib thread and the application
      thread.  Returns NULL on failure. */
  static reqRespBlock_t* getReqRespBlock();

  /** Block for a message on the given socket.  Return the packet in
      'recv_pkt'.  Return value is 1 on success, -1 on timeout.  (This
      convention is in sync with the non blocking version).  The
      'timeout' parameter has the same semantics as in the 'select'
      call.  The length of the packet received is in *recv_len.  This
      length excludes the length of the header.  If the 'pkt_ready'
      flag is set, a 'select' has already been done */
  static int blockingRecv(int recv_fd, UDPLibPkt_t* recv_pkt,
			  int *recv_len, struct sockaddr_in *from_addr,
			  struct timeval *timeout, unsigned long *rcv_count,
			  bool pkt_ready);

  /** Same as blockingRecv, but listens on two fds.  Return value is
      -1 on failure, the 'fd' read on success.  NOTE: The 'pkt_ready'
      argument in blockingRecv is absent in this routine. */
  static int blockingRecv2(int fd1, int fd2, UDPLibPkt_t *recv_pkt,
			   int *recv_len, struct sockaddr_in *from_addr,
			   struct timeval *timeout,
			   unsigned long *rcv_count1, unsigned long *rcv_count2);

  /** Send request and wait for a response.  This is a blocking call.
      Return the packet in 'recv_pkt'.  Return value is 1 on success,
      -1 on timeout.  (This convention is in sync with the non
      blocking version).  The 'timeout' parameter has the same
      semantics as in the 'select' call.  The length of the packet
      received is in *recv_len.  This length excludes the length of
      the header.  */
  static int syncReqResp(int send_fd, UDPLibPkt_t* send_pkt, int send_len,
			 struct in_addr dest_ip, int dest_port, int recv_fd,
			 UDPLibPkt_t* recv_pkt, int *recv_len,
			 struct timeval *timeout, bool use_tunnel,
			 unsigned long *count, unsigned long *rcv_count,
			 unsigned short dest_scid, unsigned short my_scid);

  /** Send a response/ack.  Return 1 on success, -1 on failure.  The
      timestamp and nonce are supposed to the same as the ones
      received in the original request packet.  */
  static int sendResp(int send_fd, UDPLibPkt_t *send_pkt, int send_len,
		      struct in_addr dest_ip, int dest_port,
		      bool use_tunnel, unsigned long *count);

  /** Stop the thread */
  static void stopThreads();

}; // End class UDPLib

#endif
